#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"

#if defined(WIN32)
#include <direct.h>
#include <windows.h>

static char mexWndName[32];
HANDLE hMexDpWnd = 0;
#define WM_ML_DISPATCH        (WM_USER+201)

#ifndef HWND_MESSAGE
#define HWND_MESSAGE     ((HWND)-3)
#endif

#elif defined(UNIX)

#include <signal.h>
#define WM_ML_DISPATCH  SIGUSR2

pthread_t mlabMainThread = 0;
int mexlog(char *tag,char *text,...);
#endif

#define	PROPMAX 1024
#define	DEVMAX 4096
#define	MAX_TINE_BUFFER 512000 //256000

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

static int mexStartDispatchHndlrDone = FALSE;

static mxArray *dpParams[3];
static char cwd[256] = "";
static char mxCalledPrp[PROPERTY_NAME_SIZE];
static char mxCalledDev[DEVICE_NAME_SIZE];

typedef struct PrpHndlrStruct
{
  char prpName[PROPERTY_NAME_SIZE];
  char fcnName[128];
  struct PrpHndlrStruct *nxt;
} PrpHndlrList;
PrpHndlrList *prpHndlrLst = NULL;
void addPrpHndlrTolist(char *prp, char *fcn)
{
  PrpHndlrList *phndlr;
  if (fcn == NULL || prp == NULL) return;
  for (phndlr=prpHndlrLst; phndlr != NULL; phndlr=phndlr->nxt)
  {
    if (!strnicmp(prp,phndlr->prpName,PROPERTY_NAME_SIZE))
    {
      strncpy(phndlr->fcnName,fcn,128);
      return;
    }
  }
  if ((phndlr=(PrpHndlrList *)calloc(1,sizeof(PrpHndlrList))) != NULL)
  {
    strncpy(phndlr->fcnName,fcn,128);
    strncpy(phndlr->prpName,prp,PROPERTY_NAME_SIZE);
    phndlr->nxt = prpHndlrLst;
    prpHndlrLst = phndlr;
  }
  return;
}
PrpHndlrList *getPrpHndlr(char *prp)
{
  PrpHndlrList *phndlr;
mexlog("getPrpHndlr","look for %s\n",prp);
  if (prp == NULL) return NULL;
  for (phndlr=prpHndlrLst; phndlr != NULL; phndlr=phndlr->nxt)
  {
mexlog("getPrpHndlr","loop over %s\n",phndlr->prpName);
    if (!strnicmp(prp,phndlr->prpName,PROPERTY_NAME_SIZE))
    {
      return phndlr;
    }
  }
mexlog("getPrpHndlr","nothing found\n");
  return NULL;
}
int prpSizIn=0;
int prpFmtIn=CF_NULL;
char Buffer[MAX_TINE_BUFFER];

int mexlog(char *tag,char *text,...)
{
  char str[512];
  FILE *fp = NULL;
  va_list args;
  time_t t = time(NULL);
  static char *fmod = "a+";

  sprintf(str,"%.24s[%s] ",ctime(&t),tag);

  va_start(args,text);
  vsnprintf(&str[strlen(str)],256,text,args); str[256] = 0;
  va_end(args);

  strcat(str,"\n");
  if ((fp=fopen("mex.log","a+")) == NULL) return -1;
  fwrite(str,strlen(str),1,fp);

  fclose(fp);
  return 0;
}

int handleDispatchMessage(int msg)
{
  mxArray *lhs[1];
  char prp[72], dev[72], fn[512], dfcn[128], *fcn=NULL;
  int cc=not_posted, len, i;
  PrpHndlrList *hndlr;
  double *lp;
  struct stat sbuf;

mexlog("dispatch","thread id : %d, msg %d\n",(int)pthread_self(),msg);
  switch (msg)
  {
    case WM_ML_DISPATCH:
mexlog("dispatch","mxGetN on dpParams\n");
strncpy(prp,mxCalledPrp,PROPERTY_NAME_SIZE);
strncpy(dev,mxCalledDev,DEVICE_NAME_SIZE);
      dpParams[0] = mxCreateString(prp);
      dpParams[1] = mxCreateString(dev);
      len = mxGetN(dpParams[1]) + 1;
      mxGetString(dpParams[1],dev,len);
mexlog("dispatch","mxGetN on dpParams\n");
      len = mxGetN(dpParams[0]) + 1;
      mxGetString(dpParams[0],prp,len);
mexlog("dispatch","got %s %s as prp and dev\n",prp,dev);

      hndlr = getPrpHndlr(prp);
mexlog("dispatch","got hndlr @ %lx\n",hndlr);
      if (hndlr != NULL) 
      {
        fcn = hndlr->fcnName;
      }
      else
      {
        sprintf(dfcn,"tine_dispatch_%.64s",prp);
        fcn = dfcn;
      }
      sprintf(fn,"%.256s%c%.128s.m",cwd,FS_DELIMITER,fcn);
mexlog("dispatch","look for function %s\n",fn);
      if (stat(fn,&sbuf) == 0)
      { // MatLab dispatch function seems to exist
mexlog("dispatch","found function %s\n",fn);
        cc = pullBufferedData(prp,dev,(BYTE *)Buffer,prpSizIn);
mexlog("dispatch","got input data %d\n",cc);
        switch (prpFmtIn)
        {
          case CF_TEXT: 
            dpParams[2] = mxCreateString(Buffer);
            break;
          case CF_NAME8:
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME48:
          case CF_NAME64:
            dpParams[2] = mxCreateString(Buffer);
            prpSizIn = 1;
            break;
          case CF_BYTE:
            dpParams[2] = mxCreateDoubleMatrix(prpSizIn,1,mxREAL);
            lp = (double *)mxGetPr(dpParams[2]);
            for(i=0; i<prpSizIn; i++) *lp++ = (double)((BYTE *)Buffer)[i];
            break;
          case CF_INT16:
            dpParams[2] = mxCreateDoubleMatrix(prpSizIn,1,mxREAL);
            lp = (double *)mxGetPr(dpParams[2]);
            for(i=0; i<prpSizIn; i++) *lp++ = (double)((short *)Buffer)[i];
            break;
          case CF_INT32:
            dpParams[2] = mxCreateDoubleMatrix(prpSizIn,1,mxREAL);
            lp = (double *)mxGetPr(dpParams[2]);
            for(i=0; i<prpSizIn; i++) *lp++ = (double)((int *)Buffer)[i];
            break;
      #   ifdef SINT64
          case CF_INT64:
            dpParams[2] = mxCreateDoubleMatrix(prpSizIn,1,mxREAL);
            lp = (double *)mxGetPr(dpParams[2]);
            for(i=0; i<prpSizIn; i++) *lp++ = (double)((SINT64 *)Buffer)[i];
            break;
      #   endif
          case CF_FLOAT:
            dpParams[2] = mxCreateDoubleMatrix(prpSizIn,1,mxREAL);
            lp = (double *)mxGetPr(dpParams[2]);
            for(i=0; i<prpSizIn; i++) *lp++ = (double)((float *)Buffer)[i];
            break;
          case CF_DOUBLE:
            dpParams[2] = mxCreateDoubleMatrix(prpSizIn,1,mxREAL);
            lp = (double *)mxGetPr(dpParams[2]);
            for(i=0; i<prpSizIn; i++) *lp++ = ((double *)Buffer)[i];
            break;
          default:
            cc = illegal_format; 
            dpParams[2] = mxCreateString("oops!");
        }
mexlog("dispatch","calling into MLAB: %s\n",fcn);
  lhs[0] = mxCreateDoubleMatrix(1,1,mxREAL);
        if (mexCallMATLAB(1,lhs,3,dpParams,fcn) != 0)
        {
mexlog("dispatch","call to MLAB choked!\n");
          cc = code_failure;
        }
        else
        {
mexlog("dispatch","call to MLAB returned!\n");
          if (mxIsDouble(lhs[0]))
          {
            cc = (int)(*(double *)mxGetPr(lhs[0]));
          }
        }
      }
mexlog("dispatch","setting completion to %d\n",cc);
      SetBufferedPropertyCompletion(prp,cc);
  mxDestroyArray(dpParams[0]);
  mxDestroyArray(dpParams[1]);
      return cc;
    default:
mexlog("dispatch","default case !!!!!!\n");
      break;
  }
  return code_failure;
}
#if defined(WIN32)
__declspec(dllexport) LRESULT CALLBACK mexDispatchProc(HWND hWnd, WORD iMessage, WORD wParam, LONG lParam)
{
  switch (iMessage)
  {
    case WM_ML_DISPATCH:
      return (LRESULT)handleDispatchMessage(iMessage);
    default: 
      return DefWindowProc (hWnd, iMessage, wParam, lParam);
  }
  return 0;
}
int CreateMexWindow(void)
{
  static WNDCLASS wndclass;
  int err, pid = GetCurrentProcessId();

  if (mexStartDispatchHndlrDone) return 0;
  sprintf(mexWndName,"mxEqmWnd-%x",pid);
  mexWndName[32] = 0;

  mexStartDispatchHndlrDone = TRUE;
  if (hMexDpWnd != 0 && FindWindow(mexWndName,mexWndName) != 0) 
  { /* still active */
    return 0; 
  }
  
  wndclass.style         = CS_GLOBALCLASS|CS_SAVEBITS; 
  wndclass.lpfnWndProc   = (WNDPROC) mexDispatchProc;
  wndclass.cbClsExtra    = 0;
  wndclass.cbWndExtra    = 0;
  wndclass.hInstance     = (HINSTANCE)HWND_MESSAGE;
  wndclass.hIcon         = LoadIcon (0, IDI_APPLICATION);
  wndclass.hCursor       = LoadCursor (0, IDC_ARROW);
  wndclass.hbrBackground = (HBRUSH)COLOR_BACKGROUND; //GetStockObject (WHITE_BRUSH);
  wndclass.lpszMenuName  = 0;
  wndclass.lpszClassName = mexWndName;

  if (!RegisterClass(&wndclass)) 
  {
    if ((err=GetLastError()) != ERROR_CLASS_ALREADY_EXISTS)
    {
      MessageBeep(-1);
      return -1;
    }
  }

  if (!(hMexDpWnd = CreateWindowEx (
      WS_EX_NOPARENTNOTIFY,
			mexWndName, mexWndName,               
			WS_BORDER|WS_CAPTION,         
			CW_USEDEFAULT, CW_USEDEFAULT, 
			CW_USEDEFAULT, CW_USEDEFAULT, 
			HWND_MESSAGE, 0,                 
			(HINSTANCE)HWND_MESSAGE,               
			(LPSTR)""                     
      ) ) ) 
  {
    err = GetLastError();
    MessageBeep(-1);
    MessageBeep(-1);
    return -1;
  }

  return 0;
}
void mexStartDispatchMsgHndlr(void)
{
  CreateMexWindow();
}
#elif defined(UNIX)
siginfo_t mexDispSigInfo;
void mexDispSigFcn(int sig, siginfo_t *inf,void *ctx)
{
  handleDispatchMessage(sig);
}
void mexStartDispatchMsgHndlr(void)
{
  sigset_t set;
  static struct sigaction act;
  sigemptyset(&set);
  act.sa_handler = NULL;
  act.sa_sigaction = &mexDispSigFcn;
  act.sa_mask = set;
  act.sa_flags = SA_SIGINFO;
  sigaction(SIGUSR2,&act,NULL);
  mlabMainThread = pthread_self();
mexlog("startup","thread id : %d\n",mlabMainThread);
}
#endif

int sendDispMessage(void)
{
  int rc = 0;
# if defined(WIN32)
  int err;
# if defined(OS_64BIT)
  err = (int)SendMessageTimeout((HWND)hMexDpWnd,WM_ML_DISPATCH,0,0,SMTO_NORMAL,100,(PDWORD_PTR)&rc);
# else
  err = (int)SendMessageTimeout((HWND)hMexDpWnd,WM_ML_DISPATCH,0,0,SMTO_NORMAL,100,(LPDWORD)&rc);
# endif
  if (err == 0) rc = code_failure;
# elif defined(UNIX)
  //rc = raise(SIGUSR2);
  rc = pthread_kill(mlabMainThread,SIGUSR2);
# endif
  return rc;
}

void _mxDpCb(int id)
{
  int rc=0;
  char prpName[72],devName[72];
  rc = getNotifiedPropertyAndDevice(prpName,devName);
  prpName[PROPERTY_NAME_SIZE] = 0;
  devName[DEVICE_NAME_SIZE] = 0;
  dpParams[0] = mxCreateString(prpName);
  dpParams[1] = mxCreateString(devName);
  strncpy(mxCalledPrp,prpName,PROPERTY_NAME_SIZE);
  strncpy(mxCalledDev,devName,DEVICE_NAME_SIZE);
mexlog("cb","enter %s %s as prp and dev\n",prpName,devName);
  rc = sendDispMessage();
  if (rc != 0) dbglog("MatLab dispatch error %d",rc);
  mxDestroyArray(dpParams[0]);
  mxDestroyArray(dpParams[1]);
  dpParams[0] = dpParams[1] = NULL;
}
#define usageStr "tine_attach_handler usage:\n ret=tine_attach_handler('<property>'[,'<handler name>')"
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[]) 
{
  mxArray	*data[PROPMAX];
  const char *field[PROPMAX];
  char *prp=NULL, *dev=NULL,*mxerrStr=NULL,*fldStr,*fcn=NULL;
  double *lp;
  char buf[128],stsbuf[64],fmtOut[64],fmtIn[64];
  int	i,len,err,n;
  int ndevs=0,isScheduled=FALSE;
    
  if (plhs == NULL || prhs == NULL) return;
  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 2: // handler function name
      if (!mxIsChar(prhs[1])) mxerr(usageStr);
      if (mxGetM(prhs[1]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[1]) + 1;
      fcn = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[1],fcn,len);
    case 1: // handler property
      if (!mxIsChar(prhs[0])) mxerr(usageStr);
      if (mxGetM(prhs[0]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[0]) + 1;
      prp = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[0],prp,len);
      err = GetRegisteredPropertyInformation(NULL,prp,NULL,NULL,NULL,
              NULL,&prpSizIn,NULL,fmtIn,NULL);
      if (err != 0) goto inpt_err;
      prpFmtIn = GetFormatFromString(fmtIn);
      err = RegisterServerNotifier(prp,_mxDpCb);
      addPrpHndlrTolist(prp,fcn);
      break;
    default:
      mxerr(usageStr);
  }   
  if (!mexStartDispatchHndlrDone) 
  {
    getcwd(cwd,256);
    mexStartDispatchMsgHndlr();
  }
inpt_err:
  if (err != 0)
  {
    GetLastLinkError(err,stsbuf);
    sprintf(buf,"tine_attach_handler : <%d> %.64s",err,stsbuf);
    mexErrMsgTxt(buf);
  }
  data[0] = mxCreateDoubleMatrix(1,1,mxREAL);
  lp = (double *)mxGetPr(data[0]);
  *lp = (double)err;
  plhs[0] = data[0];  
err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
  if (prp != NULL) mxFree(prp);
}
