#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"

#if defined(WIN32)
#include <direct.h>
#include <windows.h>

static char mexWndName[32];
HANDLE hMexDpWnd = 0;
#define WM_ML_DISPATCH        (WM_USER+201)

#ifndef HWND_MESSAGE
#define HWND_MESSAGE     ((HWND)-3)
#endif

#elif defined(UNIX)

#include <signal.h>
#include <pthread.h>
#define WM_ML_DISPATCH  SIGUSR2

pthread_t mlabMainThrd = 0;

#endif

#define	PROPMAX 1024
#define	DEVMAX 4096
#define	MAX_TINE_BUFFER 512000 //256000

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

static int mexStartDispatchHndlrDone = FALSE;

static mxArray *dpParams[3];
static char cwd[256] = "";
static char mxCalledPrp[PROPERTY_NAME_SIZE];
static char mxCalledDev[DEVICE_NAME_SIZE];

typedef struct PrpHndlrStruct
{
  char prpName[PROPERTY_NAME_SIZE];
  char fcnName[128];
  int prpSizIn;
  int prpFmtIn;
  int bufSizIn;
  char prpTagIn[TAG_NAME_SIZE];
  BYTE *dataIn;
  struct PrpHndlrStruct *nxt;
} PrpHndlrList;
PrpHndlrList *prpHndlrLst = NULL;
void addPrpHndlrTolist(char *prp, int prpSizIn, int prpFmtIn,char *prpTagIn, char *fcn)
{
  PrpHndlrList *phndlr;
  char dfcn[128];
  BYTE *tmp;
  int sizeInBytes, fsiz = GetFormatSize(prpFmtIn);
  if (prp == NULL) return;
  if (fcn == NULL)
  {
    sprintf(dfcn,"tine_dispatch_%.64s",prp);
    fcn = dfcn;
  }
  if (strlen(prpTagIn) > 0) fsiz = GetStructSize(prpTagIn);
  sizeInBytes = prpSizIn * fsiz;

  for (phndlr=prpHndlrLst; phndlr != NULL; phndlr=phndlr->nxt)
  {
    if (!strnicmp(prp,phndlr->prpName,PROPERTY_NAME_SIZE))
    {
      strncpy(phndlr->fcnName,fcn,128);
      phndlr->prpSizIn = prpSizIn;
      phndlr->prpFmtIn = prpFmtIn;
      strncpy(phndlr->prpTagIn,prpTagIn,TAG_NAME_SIZE);
      if (sizeInBytes > phndlr->bufSizIn)
      {
        if ((tmp=(BYTE *)realloc(phndlr->dataIn,sizeInBytes*fsiz)) != NULL)
        {
          phndlr->dataIn = tmp;
          phndlr->bufSizIn = phndlr->dataIn == NULL ? 0 : sizeInBytes;
        }
      }
      return;
    }
  }
  if ((phndlr=(PrpHndlrList *)calloc(1,sizeof(PrpHndlrList))) != NULL)
  {
    strncpy(phndlr->fcnName,fcn,128);
    strncpy(phndlr->prpName,prp,PROPERTY_NAME_SIZE);
    phndlr->prpSizIn = prpSizIn;
    phndlr->prpFmtIn = prpFmtIn;
    strncpy(phndlr->prpTagIn,prpTagIn,TAG_NAME_SIZE);
    phndlr->dataIn = (BYTE *)calloc(sizeInBytes,fsiz);
    phndlr->bufSizIn = phndlr->dataIn == NULL ? 0 : sizeInBytes;
    phndlr->nxt = prpHndlrLst;
    prpHndlrLst = phndlr;
  }
  return;
}
PrpHndlrList *getPrpHndlr(char *prp)
{
  PrpHndlrList *phndlr;
  if (prp == NULL) return NULL;
  for (phndlr=prpHndlrLst; phndlr != NULL; phndlr=phndlr->nxt)
  {
    if (!strnicmp(prp,phndlr->prpName,PROPERTY_NAME_SIZE))
    {
      return phndlr;
    }
  }
  return NULL;
}

int isTaggable(short fmt)
{
  switch (fmt)
  {
    case CF_STRUCT:
    case CF_BITFIELD8:
    case CF_BITFIELD16:
    case CF_BITFIELD32:
    case CF_BITFIELD64:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}

#define getDblVal(i,b,f) (\
  (double)((f) == CF_BYTE ? (b)[i] :\
           (f) == CF_INT16 ? ((short *)(b))[i] :\
           (f) == CF_INT32 ? ((int *)(b))[i] :\
           (f) == CF_INT64 ? ((SINT64 *)(b))[i] :\
           (f) == CF_FLOAT ? ((float *)(b))[i] :\
           (f) == CF_DOUBLE ? ((double *)(b))[i] : 0))

#define get1stDblVal(i,b,f) (\
  (double)((f) == CF_INTINT ? ((INTINT *)(b))[i].i1val :\
           (f) == CF_FLTINT ? ((FLTINT *)(b))[i].fval :\
           (f) == CF_DBLDBL ? ((DBLDBL *)(b))[i].d1val :\
           (f) == CF_FLTFLT ? ((FLTFLT *)(b))[i].f1val : 0))

#define get2ndDblVal(i,b,f) (\
  (double)((f) == CF_INTINT ? ((INTINT *)(b))[i].i2val :\
           (f) == CF_FLTINT ? ((FLTINT *)(b))[i].ival :\
           (f) == CF_DBLDBL ? ((DBLDBL *)(b))[i].d2val :\
           (f) == CF_FLTFLT ? ((FLTFLT *)(b))[i].f2val : 0))

#define get1stQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i1val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].ival :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].i1val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].t1 :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].f1val : 0))

#define get2ndQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i2val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].f1val :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].i2val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].t2 :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].i1val : 0))

#define get3rdQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i3val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].f2val :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].f1val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].i1val :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].f2val : 0))

#define get4thQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i4val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].f3val :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].f2val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].i2val :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].i2val : 0))

#define get1stTriVal(i,b,f) (\
  (double)((f) == CF_INTINTINT ? ((INTINTINT *)(b))[i].i1val :\
           (f) == CF_FLTFLTINT ? ((FLTFLTINT *)(b))[i].f1val :\
           (f) == CF_INTFLTINT ? ((INTFLTINT *)(b))[i].i1val :\
           (f) == CF_FLTINTINT ? ((FLTINTINT *)(b))[i].fval :\
           (f) == CF_DBLDBLDBL ? ((DBLDBLDBL *)(b))[i].d1val : 0))

#define get2ndTriVal(i,b,f) (\
  (double)((f) == CF_INTINTINT ? ((INTINTINT *)(b))[i].i2val :\
           (f) == CF_FLTFLTINT ? ((FLTFLTINT *)(b))[i].f2val :\
           (f) == CF_INTFLTINT ? ((INTFLTINT *)(b))[i].fval :\
           (f) == CF_FLTINTINT ? ((FLTINTINT *)(b))[i].i1val :\
           (f) == CF_DBLDBLDBL ? ((DBLDBLDBL *)(b))[i].d2val : 0))

#define get3rdTriVal(i,b,f) (\
  (double)((f) == CF_INTINTINT ? ((INTINTINT *)(b))[i].i3val :\
           (f) == CF_FLTFLTINT ? ((FLTFLTINT *)(b))[i].ival :\
           (f) == CF_INTFLTINT ? ((INTFLTINT *)(b))[i].i2val :\
           (f) == CF_FLTINTINT ? ((FLTINTINT *)(b))[i].i2val :\
           (f) == CF_DBLDBLDBL ? ((DBLDBLDBL *)(b))[i].d3val : 0))

mxArray *getMxData(int fmt,int siz,char *tag,BYTE *buf)
{
  mxArray *da, *fdata[256];
  mxChar *mxc;
  int fmtsiz = fmtsizeof(fmt);
  int i, k, n, nsiz, nf;
  mwSize dims[2];
  const char *nifield[2], *sfield[256];
  double *lp;
  struct structStruct *ss;
  struct structFormat *sf;

  switch(fmt)
  {	
    case CF_TEXT:
      da = mxCreateString((const char *)buf);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      dims[0] = siz;
      dims[1] = fmtsiz;
      if (siz > 1)
      {
        da = mxCreateCharArray(2,dims);
        mxc = (mxChar *)mxGetData(da);
        for (i=0; i<siz; i++)
        { /* string arrays transposed in MatLab ? */
          for (k=0; k<fmtsiz; k++) mxc[siz*k+i] = (mxChar)buf[fmtsiz*i+k];
        }
      }
      else
      {
        da = mxCreateString((const char *)buf);
      }
      break;
    case CF_STRING:
      if (siz > 0)
      {
        da = mxCreateCharMatrixFromStrings(siz,(const char **)buf);
      }
      break;
    case CF_NAME8I:
    case CF_NAME16I:
    case CF_NAME32I:
    case CF_NAME48I:
    case CF_NAME64I:
      nifield[0] = (char *)mxCalloc(5, sizeof(char));
      nifield[1] = (char *)mxCalloc(6, sizeof(char));
      strcpy((char *)nifield[0],"name");
      strcpy((char *)nifield[1],"value");
      nsiz = fmtsiz - 4;
      dims[0] = 1;
      dims[1] = nsiz;
      if (siz > 0)
      {
        da = mxCreateStructMatrix(siz,1,2,&nifield[0]);
        for (i=0; i<siz; i++)
        {
          fdata[0] = mxCreateCharArray(2,dims);
          fdata[1] = mxCreateDoubleMatrix(1,1,mxREAL);
          mxc = (mxChar *)mxGetData(fdata[0]);
          lp = (double *)mxGetPr(fdata[1]);
          for (k=0; k<nsiz; k++) mxc[k] = (mxChar)buf[fmtsiz*i+k];
          *lp = double(*(int *)&buf[fmtsiz*i+nsiz]);
          mxSetFieldByNumber(da,i,0,fdata[0]);
          mxSetFieldByNumber(da,i,1,fdata[1]);
        }
      }
      break;
    case CF_STRUCT:
      if ((ss=GetStructFromTag(tag)) != NULL && siz > 0)
      {
        fmtsiz = GetStructSize(tag);
        for (k=0, sf=ss->s; sf != NULL && k < 256; k++, sf=sf->nxt) if (sf->fmt == CF_NULL) k--;
        nf = k; // number of fields (n.b. legacy registrations might have a 'CF_NULL' in termination)
        for (n=nf-1, sf=ss->s; sf != NULL && n >= 0; n--, sf=sf->nxt)
        {
          if (sf->fmt == CF_NULL) { n++; continue; }
          nsiz = MIN(strlen(sf->field),16);
          sfield[n] = (char *)mxCalloc(nsiz+1,sizeof(char));
          strncpy((char *)sfield[n],sf->field,nsiz);
        }
        da = mxCreateStructMatrix(siz,1,nf,&sfield[0]);
        for (i=0; i<siz; i++)
        {
          for (n=nf-1, sf=ss->s; sf != NULL && n >=0; n--, sf=sf->nxt)
          { 
            nsiz = sf->off;
            switch (sf->fmt)
            {
              case CF_INT16:
              case CF_INT32:
              case CF_FLOAT:
              case CF_DOUBLE:
                fdata[n] = mxCreateDoubleMatrix(sf->siz,1,mxREAL);
                lp = (double *)mxGetPr(fdata[n]);
                for (k=0; k<sf->siz; k++) *lp++ = getDblVal(k,&buf[fmtsiz*i+nsiz],sf->fmt);
                mxSetFieldByNumber(da,i,n,fdata[n]);
                break;
              case CF_TEXT:
                dims[0] = 1;
                dims[1] = sf->siz+1;
                fdata[n] = mxCreateCharArray(2,dims);
                mxc = (mxChar *)mxGetData(fdata[n]);
                for (k=0; k<dims[1]; k++) mxc[k] = (mxChar)buf[fmtsiz*i+nsiz+k];
                mxSetFieldByNumber(da,i,n,fdata[n]);
                break;
              case CF_NULL:
                n++; 
                continue;
            }
          }
        }
      }
      break;
    case CF_USTRING:
      buf[96] = '\0';
      sprintf((char *)buf,"%d %g %g %u %s",*(int *)&buf[0],*(float *)&buf[4],*(float *)&buf[8],*(UINT32 *)&buf[12],(char *)&buf[16]);
      da = mxCreateString((const char *)buf);
      break;					
    case CF_BYTE:
    case CF_INT16:
    case CF_INT32:
    case CF_FLOAT:
    case CF_DOUBLE:
      da = mxCreateDoubleMatrix(siz,1,mxREAL);
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) *lp++ = getDblVal(i,buf,fmt);
      break;
    case CF_BOOLEAN:
      da = mxCreateDoubleMatrix(siz,1,mxREAL);
      lp = (double *)mxGetPr(da);
      for(n = 0; n < siz; n++) *lp++ = (double)(((int *)buf)[i] ? 1 : 0);
      break;
    case CF_INTINT:
    case CF_FLTINT:
    case CF_FLTFLT:
    case CF_DBLDBL:
      da = mxCreateDoubleMatrix(siz,2,mxREAL);
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) 
      {
        *lp       = get1stDblVal(i,buf,fmt);
        *(lp+siz) = get2ndDblVal(i,buf,fmt);
        lp++;
      }
      break;
    case CF_IFFF:
    case CF_FIFI:
    case CF_IIII:
    case CF_IIFF:
    case CF_TTII:
      da = mxCreateDoubleMatrix(siz,4,mxREAL),
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) 
      {
        *lp         = get1stQuadVal(i,buf,fmt);
        *(lp + siz) = get2ndQuadVal(i,buf,fmt);
        *(lp+2*siz) = get3rdQuadVal(i,buf,fmt);
        *(lp+3*siz) = get4thQuadVal(i,buf,fmt);
        lp++;
      }
      break;
    case CF_SPECTRUM:
      da = mxCreateDoubleMatrix(siz-96,1,mxREAL),
      lp = (double *)mxGetPr(da);
      for(n = 0; n < siz-96; n++) *lp++ = double(*(float *)&buf[n]);
      break;
    case CF_INTINTINT:
    case CF_DBLDBLDBL:
    case CF_FLTFLTINT:
    case CF_INTFLTINT:
    case CF_FLTINTINT:
      da = mxCreateDoubleMatrix(siz,3,mxREAL),
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) 
      {
        *lp         = get1stTriVal(i,buf,fmt);
        *(lp + siz) = get2ndTriVal(i,buf,fmt);
        *(lp+2*siz) = get3rdTriVal(i,buf,fmt);
        lp++;
      }
      break;
    default:
      sprintf((char *)buf,"data type %d not handled\n",fmt);
      da  = mxCreateString((const char *)buf);
      break;
  }
  return da;
}
/* these cannot be global! all instances use this same address space !
int prpSizIn=0;
int prpFmtIn=CF_NULL;
char prpTagIn[32];
char Buffer[MAX_TINE_BUFFER];
*/
int handleDispatchMessage(int msg)
{
  mxArray *lhs[1];
  char prp[72], dev[72], fn[512], dfcn[128], *fcn=NULL;
  int cc=not_posted, len, i;
  PrpHndlrList *hndlr;
	double *lp;
  struct stat sbuf;

  switch (msg)
  {
    case WM_ML_DISPATCH:
      strncpy(prp,mxCalledPrp,PROPERTY_NAME_SIZE);
      strncpy(dev,mxCalledDev,DEVICE_NAME_SIZE);
      dpParams[0] = mxCreateString(prp);
      dpParams[1] = mxCreateString(dev);
      hndlr = getPrpHndlr(prp);
      if (hndlr != NULL) 
      {
        fcn = hndlr->fcnName;
        sprintf(fn,"%.256s%c%.128s.m",cwd,FS_DELIMITER,fcn);
        if (stat(fn,&sbuf) == 0)
        { // MatLab dispatch function seems to exist
          cc = pullBufferedData(prp,dev,hndlr->dataIn,hndlr->prpSizIn);
          dpParams[2] = getMxData(hndlr->prpFmtIn,hndlr->prpSizIn,hndlr->prpTagIn,hndlr->dataIn);
          if (mexCallMATLAB(1,lhs,3,dpParams,fcn) != 0)
          {
            cc = code_failure;
          }
          else
          {
            if (mxIsDouble(lhs[0]))
            {
              cc = (int)(*(double *)mxGetPr(lhs[0]));
            }
          }
        }
      }
      SetBufferedPropertyCompletion(prp,cc);
      mxDestroyArray(dpParams[0]);
      mxDestroyArray(dpParams[1]);
      mxCalledPrp[0] = mxCalledDev[0] = 0;
      return cc;
    default:
      break;
  }
  return code_failure;
}
#if defined(WIN32)
__declspec(dllexport) LRESULT CALLBACK mexDispatchProc(HWND hWnd, WORD iMessage, WORD wParam, LONG lParam)
{
  switch (iMessage)
  {
    case WM_ML_DISPATCH:
      return (LRESULT)handleDispatchMessage(iMessage);
    default: 
      return DefWindowProc (hWnd, iMessage, wParam, lParam);
  }
  return 0;
}
int CreateMexWindow(void)
{
  static WNDCLASS wndclass;
  int err, pid = GetCurrentProcessId();

  if (mexStartDispatchHndlrDone) return 0;
  sprintf(mexWndName,"mxEqmWnd-%x",pid);
  mexWndName[32] = 0;

  mexStartDispatchHndlrDone = TRUE;
  if (hMexDpWnd != 0 && FindWindow(mexWndName,mexWndName) != 0) 
  { /* still active */
    return 0; 
  }
  
  wndclass.style         = CS_GLOBALCLASS|CS_SAVEBITS; 
  wndclass.lpfnWndProc   = (WNDPROC) mexDispatchProc;
  wndclass.cbClsExtra    = 0;
  wndclass.cbWndExtra    = 0;
  wndclass.hInstance     = (HINSTANCE)HWND_MESSAGE;
  wndclass.hIcon         = LoadIcon (0, IDI_APPLICATION);
  wndclass.hCursor       = LoadCursor (0, IDC_ARROW);
  wndclass.hbrBackground = (HBRUSH)COLOR_BACKGROUND; //GetStockObject (WHITE_BRUSH);
  wndclass.lpszMenuName  = 0;
  wndclass.lpszClassName = mexWndName;

  if (!RegisterClass(&wndclass)) 
  {
    if ((err=GetLastError()) != ERROR_CLASS_ALREADY_EXISTS)
    {
      MessageBeep(-1);
      return -1;
    }
  }

  if (!(hMexDpWnd = CreateWindowEx (
      WS_EX_NOPARENTNOTIFY,
			mexWndName, mexWndName,               
			WS_BORDER|WS_CAPTION,         
			CW_USEDEFAULT, CW_USEDEFAULT, 
			CW_USEDEFAULT, CW_USEDEFAULT, 
			HWND_MESSAGE, 0,                 
			(HINSTANCE)HWND_MESSAGE,               
			(LPSTR)""                     
      ) ) ) 
  {
    err = GetLastError();
    MessageBeep(-1);
    MessageBeep(-1);
    return -1;
  }

  return 0;
}
void mexStartDispatchMsgHndlr(void)
{
  CreateMexWindow();
}
#elif defined(UNIX)

siginfo_t mexDispSigInfo;
void mexDispSigFcn(int sig, siginfo_t *inf,void *ctx)
{
  handleDispatchMessage(sig);
}
void mexStartDispatchMsgHndlr(void)
{
  sigset_t set;
  static struct sigaction act;
  sigemptyset(&set);
  act.sa_handler = NULL;
  act.sa_sigaction = &mexDispSigFcn;
  act.sa_mask = set;
  act.sa_flags = SA_SIGINFO;
  sigaction(SIGUSR2,&act,NULL);
  mlabMainThrd = pthread_self();
}
#endif

int sendDispMessage(void)
{
  int rc = 0;
# if defined(WIN32)
  int err;
# if defined(OS_64BIT)
  err = (int)SendMessageTimeout((HWND)hMexDpWnd,WM_ML_DISPATCH,0,0,SMTO_NORMAL,100,(PDWORD_PTR)&rc);
# else
  err = (int)SendMessageTimeout((HWND)hMexDpWnd,WM_ML_DISPATCH,0,0,SMTO_NORMAL,100,(LPDWORD)&rc);
# endif
  if (err == 0) rc = code_failure;
# elif defined(UNIX)
  rc = pthread_kill(mlabMainThrd,SIGUSR2);
# endif
  return rc;
}

void _mxDpCb(int id)
{
  int rc=0;
  char prpName[72],devName[72];
  if (strlen(mxCalledPrp) > 0) 
  {
    dbglog("MatLab dispatch error : still dispatching property %.64s device %.64s",mxCalledPrp,mxCalledDev);
    return;
  }
  rc = getNotifiedPropertyAndDevice(prpName,devName);
  prpName[PROPERTY_NAME_SIZE] = 0;
  devName[DEVICE_NAME_SIZE] = 0;
  strncpy(mxCalledPrp,prpName,PROPERTY_NAME_SIZE);
  strncpy(mxCalledDev,devName,DEVICE_NAME_SIZE);
  rc = sendDispMessage();
err:
  if (rc != 0) dbglog("MatLab dispatch error %d",rc);
}
#define usageStr "tine_attach_handler usage:\n ret=tine_attach_handler('<property>'[,'<handler name>')"
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[]) 
{
  mxArray	*data[PROPMAX];
  const char *field[PROPMAX];
  char *prp=NULL, *dev=NULL,*mxerrStr=NULL,*fldStr,*fcn=NULL;
  double *lp;
  char *c,buf[128],stsbuf[64],fmtOut[64],fmtIn[64];
  int	i,len,err,n;
  int ndevs=0,isScheduled=FALSE,isLocked=FALSE;
  int prpSizIn=0;
  int prpFmtIn=CF_NULL;
  char prpTagIn[32];
    
  if (plhs == NULL || prhs == NULL) return;
  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 2: // handler function name
      if (!mxIsChar(prhs[1])) mxerr(usageStr);
      if (mxGetM(prhs[1]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[1]) + 1;
      fcn = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[1],fcn,len);
    case 1: // handler property
      if (!mxIsChar(prhs[0])) mxerr(usageStr);
      if (mxGetM(prhs[0]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[0]) + 1;
      prp = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[0],prp,len);
      err = GetRegisteredPropertyInformation(NULL,prp,NULL,NULL,NULL,
              NULL,&prpSizIn,NULL,fmtIn,NULL);
      if (err != 0) goto inpt_err;
      prpFmtIn = GetFormatFromString(fmtIn);
      if ((c=strchr(fmtIn,'.')) != NULL)
      {
        if (isTaggable(prpFmtIn))
        {
          c++;
          strncpy(prpTagIn,c,TAG_NAME_SIZE);
        }
      }
      else
      {
        prpTagIn[0] = 0;
      }
      err = RegisterServerNotifier(prp,_mxDpCb);
      addPrpHndlrTolist(prp,prpSizIn,prpFmtIn,prpTagIn,fcn);
      mexLock();
      isLocked = TRUE;
      break;
    default:
      mxerr(usageStr);
  }   
  if (!mexStartDispatchHndlrDone) 
  {
    getcwd(cwd,256);
    mexStartDispatchMsgHndlr();
  }
inpt_err:
  if (err != 0)
  {
    GetLastLinkError(err,stsbuf);
    sprintf(buf,"tine_attach_handler : <%d> %.64s",err,stsbuf);
    mexErrMsgTxt(buf);
  }
  data[0] = mxCreateDoubleMatrix(1,1,mxREAL);
  lp = (double *)mxGetPr(data[0]);
  *lp = (double)err;
  plhs[0] = data[0];  
err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
    if (isLocked) mexUnlock();
  }
  if (prp != NULL) mxFree(prp);
}
