/*
 * tine_eventdata.cpp : wrap GetArchivedEventData() from eventtools.dll
 * 
 */

#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"
#include "eventtools.h"

double blob[64000];

#define mxerr(s) { mxerrStr = (char *)s; goto err; }
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) 
{
  char *parameter1=NULL, *parameter2=NULL, *c, *prp, *mxerrStr=NULL;
  char *ctx=NULL,*trg=NULL;
  char *dctx=NULL,*dsrv=NULL,*dnam=NULL,*dprp=NULL;
  PrpQueryStruct pqs;
  DTYPE dout;
  mxArray *temp;
  double *ptr;
  int buflen, status;
  char str[128],evnstr[64],dprpbuf[PROPERTY_NAME_SIZE];
  int i, k, ndevices, evn, cc, idx=0, fmtsiz=0, qrydone=FALSE;

  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 4:
      if (!mxIsDouble(prhs[3]) || mxIsComplex(prhs[3]) || mxGetN(prhs[3])*mxGetM(prhs[3]) != 1) 
          mexErrMsgTxt("Input <data_index> must be a scalar.");
      idx = (int)mxGetScalar(prhs[3]);
    case 3:
      mxGetString(prhs[2],evnstr,64);
      evn = (int)strtol(evnstr,NULL,0);
      if (mxGetM(prhs[0]) != 1) mxerr("Input must be a row vector.");
      ndevices = mxGetM(prhs[1]); // num requested devices (consistency check: property must be the same !)
      if (ndevices != 1) mxerr("multiple calls not supported at this time!");
      break;
    default:
      mxerr("tine_eventdata usage:\ntine_eventdata('/<context>/<trigger>','/<context>/<server>/<device>[<property>]','<event_number>'[, <data_index>])\n");
  }
  if (nlhs > 1) mxerr("Too many output arguments");

  buflen = mxGetM(prhs[0])*mxGetN(prhs[0]) + 1;
  parameter1 = (char *) mxCalloc(buflen, sizeof(char));
  status = mxGetString(prhs[0], parameter1, buflen);
  if(status != 0) mexWarnMsgTxt("Not enough space. String is truncated.");
  // parse out the event server context and the event trigger
  if ((ctx=strtok(parameter1,"/")) == NULL) mxerr("1st argument not in the form: /<context>/<trigger>");
  if ((trg=strtok(NULL,"/")) == NULL) mxerr("1st argument not in the form: /<context>/<trigger>");

  buflen = mxGetN(prhs[1]) + 1; 
  parameter2 = (char *) mxCalloc(buflen, sizeof(char));
  status = mxGetString(prhs[1], parameter2, buflen);
  if(status != 0) mexWarnMsgTxt("Not enough space. String is truncated.");
  // parse out (in loop over ndevices) the desired /context/server/device[property]

  for (i=0; i<ndevices; i++)
  { // ensure consistency of requested property
    if (!qrydone)
    {
      if ((cc=GetArchivedPropertyInformation(ctx,parameter2,&pqs)) != 0) mxerr(GetSystemErrorString(cc,str));
      qrydone = TRUE;
    }
    if ((prp=strchr(parameter2,'[')) == NULL) mxerr("2nd argument not in the form: /<context>/<server>/<device>[<property>]");
    *prp++ = 0;
    if ((c=strchr(prp,']')) == NULL) mxerr("2nd argument not in the form: /<context>/<server>/<device>[<property>]");
    *c = 0;
    if (dprp == NULL) 
    {
      strncpy(dprpbuf,prp,PROPERTY_NAME_SIZE);
      dprp = dprpbuf;
      memset(&dout,0,sizeof(DTYPE));
      if ((fmtsiz=fmtsizeof(pqs.prpFormat)) <= 0) mxerr("illegal format in save property");
      if (pqs.prpSize * fmtsiz > sizeof(blob)) 
      {
        mexPrintf("truncate data size from %d to ",pqs.prpSize);
        pqs.prpSize -= (sizeof(blob)/fmtsiz - pqs.prpSize);
        mexPrintf("%d\n",pqs.prpSize);
      }
      dout.dFormat = LFMT(pqs.prpFormat);
      dout.dArrayLength = pqs.prpSize;
      strncpy(dout.dTag,pqs.prpTag,TAG_NAME_SIZE);
      dout.data.vptr = blob;
      /* create matrix for data */
      temp = mxCreateDoubleMatrix(pqs.prpSize, ndevices, mxREAL);
    }
    else 
    { // ensure consistency of requested property
      if (strnicmp(prp,dprp,PROPERTY_NAME_SIZE) != 0) mxerr("inconsistent property names in list!");
    }
    if ((dctx=strtok(parameter2,"/")) == NULL) mxerr("2nd argument not in the form: /<context>/<server>/<device>[<property>]");
    if ((dsrv=strtok(NULL,"/")) == NULL) mxerr("2nd argument not in the form: /<context>/<server>/<device>[<property>]");
    if ((dnam=strtok(NULL,"/")) == NULL) mxerr("2nd argument not in the form: /<context>/<server>/<device>[<property>]");
    if ((cc=GetArchivedEventData(ctx,trg,dctx,dsrv,dprp,dnam,idx,evn,&dout)) != 0) mxerr(GetSystemErrorString(cc,str));
    ptr = mxGetPr(temp);
    switch (dout.dFormat)
    {
      case CF_BYTE:
        for (k=0; k<(int)dout.dArrayLength; k++) ptr[k+i*ndevices] = (double)dout.data.bptr[k];
        break;
      case CF_INT16:
        for (k=0; k<(int)dout.dArrayLength; k++) ptr[k+i*ndevices] = (double)dout.data.sptr[k];
        break;
      case CF_INT32:
        for (k=0; k<(int)dout.dArrayLength; k++) ptr[k+i*ndevices] = (double)dout.data.lptr[k];
        break;
      case CF_FLOAT:
        for (k=0; k<(int)dout.dArrayLength; k++) ptr[k+i*ndevices] = (double)dout.data.fptr[k];
        break;
      case CF_DOUBLE:
        for (k=0; k<(int)dout.dArrayLength; k++) ptr[k+i*ndevices] = dout.data.dptr[k];
        break;
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
      default:
        mxerr("stored format not currently supported");
    }
  }
  plhs[0] = temp; 

err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
  if (parameter1 != NULL) mxFree(parameter1); 
  if (parameter2 != NULL) mxFree(parameter2);
}