/*
 * tine_evenlist.cpp :
 */

#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"
#include "errors.h"
#include "eventtools.h"

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

time_t parseTimeString(char *ts)
{
  time_t t=time(NULL);
  int i, len;
  char *c,day[16],mon[64],year[16];
  char hr[64],mn[32],sec[16];
  struct tm tmts, *tnow;

  hr[0] = 0;
  tnow = localtime(&t);
  len = strlen(ts);
  strupr(ts);
  for (i=0; i<len && isdigit(ts[i]); i++);  
  if (i == len) t = atoi(ts);
  else if (!strcmp(ts,"NOW")) t = time(NULL);
  else if ((c=strstr(ts,".")) != NULL) /* day.month.year format */
  {
    *c++ = 0;
    strncpy(day,ts,16);
    strncpy(mon,c,64);
    if ((c=strstr(mon,".")) != NULL)
    { 
      *c++ = 0;
      strncpy(year,c,16);
      if ((c=strstr(year,"_")) != NULL || (c=strstr(year," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    else
    {
      sprintf(year,"%d",1900+tnow->tm_year);
      if ((c=strstr(mon,"_")) != NULL || (c=strstr(mon," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    if (strlen(hr) > 0)
    {
      mn[0] = 0; sec[0] = 0;
      if ((c=strstr(hr,":")) != NULL)
      {     
        *c++ = 0;
        strncpy(mn,c,32);
        if ((c=strstr(mn,":")) != NULL)
        {
          *c++ = 0;
          strncpy(sec,c,16);
        }
      }
      tmts.tm_hour = atoi(hr);
      tmts.tm_min =  atoi(mn);
      tmts.tm_sec =  atoi(sec);
    }
    else
    {
      memset(&tmts,0,sizeof(struct tm));
    }
    tmts.tm_mon = (atoi(mon)+11)%12;
    tmts.tm_mday = atoi(day);
    tmts.tm_year = atoi(year);
    if (tmts.tm_year > 1900) tmts.tm_year -= 1900;
    tmts.tm_isdst = -1;
    t = mktime(&tmts);
  }
  return t;
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) 
{
  char *parameter1=NULL, *parameter2=NULL, *c, *prp, *mxerrStr=NULL;
  char *ctx=NULL,*trg=NULL;
  time_t stop, start, interim;
  mxArray *temp;
  mxChar *mxch;
  double *ptr;
  int buflen, status;
  char str[128],evnstr[64],dprpbuf[PROPERTY_NAME_SIZE];
  int i, k, cc, depth=86400, lstsiz = 1000, evnlst[1000], dims[2];

  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 3:
      mxGetString(prhs[2],str,128);
      depth = strtol(str,&c,10);
      switch (tolower(*c))
      {
        case 'd': depth *= 86400; 
          break;
        case 'h': depth *= 3600; 
          break;
        case 'm':
          if (strnicmp(c,"min",3) == 0)
          {
            depth *= 60;
          }
          else if (strnicmp(c,"mon",3) == 0)
          {
            depth *= 86400 * 31;
          }
          break;
        case 'w':
          depth *= 86400 * 7;
          break;
        default:
        case 's': 
          break;
      }
    case 2:
      mxGetString(prhs[1],str,128);
      stop = parseTimeString(str);
      start = stop - depth;
      if (mxGetM(prhs[0]) != 1) mxerr("Input must be a row vector.");
      break;
    default:
      mxerr("tine_eventlist usage:\ntine_evenlist('/<context>/<trigger>','<stop time>'[,'<depth>'])\n");
  }
  if (nlhs > 1) mxerr("Too many output arguments");

  buflen = mxGetM(prhs[0])*mxGetN(prhs[0]) + 1;
  parameter1 = (char *) mxCalloc(buflen, sizeof(char));
  status = mxGetString(prhs[0], parameter1, buflen);
  if(status != 0) mexWarnMsgTxt("Not enough space. String is truncated.");
  // parse out the event server context and the event trigger
  if ((ctx=strtok(parameter1,"/")) == NULL) mxerr("1st argument not in the form: /<context>/<trigger>");
  if ((trg=strtok(NULL,"/")) == NULL) mxerr("1st argument not in the form: /<context>/<trigger>");

  if ((cc=GetArchivedEventList(ctx,trg,start,stop,evnlst,&lstsiz)) != 0)  mxerr(GetSystemErrorString(cc,str));
  dims[0] = lstsiz;
  dims[1] = 64;
  temp = mxCreateCharArray(2,dims);
  mxch = (mxChar *)mxGetData(temp);
  for(k=0; k < lstsiz; k++)
  {
    interim = (time_t)evnlst[k];
    _snprintf(evnstr,64,"%d (%.24s)",evnlst[k],ctime(&interim));
    for (i=0; i<64 && evnstr[i] != '\0'; i++)
    {
      mxch[lstsiz*i+k] = (mxChar)evnstr[i];
    }
  }
  plhs[0] = temp; 

err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
  if (parameter1 != NULL) mxFree(parameter1); 
}