/*
 * tine_eventservers.cpp :
 */

#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"
#include "errors.h"
#include "eventtools.h"

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) 
{
  char *parameter1=NULL, *c, *prp, *mxerrStr=NULL;
  char *ctx=NULL,*trg=NULL,*srv=NULL;
  time_t stop, start, interim;
  mxArray *temp;
  mxChar *mxch;
  double *ptr;
  int buflen, status;
  char str[128];
  int i, k, cc, lstsiz = 100, dims[2];
  NAME64 srvlst[100];
  
  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 1:
      if (mxGetM(prhs[0]) != 1) mxerr("Input must be a row vector.");
      break;
    default:
      mxerr("tine_eventservers usage:\ntine_eventservers('/<context>/<trigger>')\n");
  }
  if (nlhs > 1) mxerr("Too many output arguments");

  buflen = mxGetM(prhs[0])*mxGetN(prhs[0]) + 1;
  parameter1 = (char *) mxCalloc(buflen, sizeof(char));
  status = mxGetString(prhs[0], parameter1, buflen);
  if(status != 0) mexWarnMsgTxt("Not enough space. String is truncated.");
  // parse out the event server context and the event trigger
  if ((ctx=strtok(parameter1,"/")) == NULL) mxerr("1st argument not in the form: /<context>/<trigger>");
  if ((trg=strtok(NULL,"/")) == NULL) mxerr("1st argument not in the form: /<context>/<trigger>");

  if ((cc=GetArchivedServers(ctx,trg,srvlst,&lstsiz)) != 0) mxerr(GetSystemErrorString(cc,str));
  dims[0] = lstsiz;
  dims[1] = 64;
  temp = mxCreateCharArray(2,dims);
  mxch = (mxChar *)mxGetData(temp);
  for(k=0; k < lstsiz; k++)
  {
    for (i=0; i<64 && srvlst[k].name[i] != '\0'; i++)
    {
      mxch[lstsiz*i+k] = (mxChar)srvlst[k].name[i];
    }
  }
  plhs[0] = temp; 

err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
  if (parameter1 != NULL) mxFree(parameter1); 
}