#include <math.h>
#include "mex.h"
#include <windows.h>

#define	PROPMAX 10
#define	DEVMAX 4096
#define	MAX_TINE_BUFFER 512000 //256000

#include "tine.h"
#include "tbufsrv.h"

#define mxerr(s) { mxerrStr = s; goto err; }

char Buffer[MAX_TINE_BUFFER];
char *usageStr = "tine_getdevicenumber usage:\n number=tine_getdevicenumber('<device>'[,'<property>'])";
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[]) 
{
	mxArray	*data[PROPMAX];
	const char *field[PROPMAX];
	char prp[PROPERTY_NAME_SIZE], dev[DEVICE_NAME_SIZE], *mxerrStr=NULL;
	double *lp;
	char buf[128],stsbuf[64],fmtIn[64];
	int	i,siz=0,len,fmt,err,n;
  int sizin=0;
  int ndevs=0,isScheduled=FALSE;
    
  if (plhs == NULL || prhs == NULL) return;
  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 2: // device and property and data reference
      if (!mxIsChar(prhs[1])) mxerr(usageStr);
      if (mxGetM(prhs[1]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[1]) + 1;
      mxGetString(prhs[1],prp,len);
    case 1: // device and property and data reference
      if (!mxIsChar(prhs[0])) mxerr(usageStr);
      if (mxGetM(prhs[0]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[0]) + 1;
      mxGetString(prhs[0],dev,len);

      err = GetDeviceNumberEx(char *eqm,char *devname,char *prpname)
      err = GetRegisteredPropertyInformation(NULL,prp,NULL,NULL,NULL,
              NULL,&sizin,NULL,fmtIn,NULL);
      if (err != 0) goto inpt_err;
      fmt = GetFormatFromString(fmtIn);
      if (mxIsChar(prhs[2]))
      { // string type being pulled
        siz = mxGetM(prhs(2));
        len = mxGetN(prhs[2]);
      }
      else if (mxIsDouble(prhs[2])
      { // number type being pulled
        siz = mxGetM(prhs[2]);
      }
      else 
      { // command (no data involved)
        siz = 0;
      }
      if (siz > sizin) siz = sizin;
      break;
    default:
      mxerr(usageStr);
  }   

  err = pullBufferedData(prp,dev,(BYTE *)Buffer,siz);

  switch (fmt)
  {
    case CF_TEXT: 
      data[0] = mxCreateString(Buffer);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      data[0] = mxCreateString(Buffer);
      siz = 1;
      break;
    case CF_BYTE:
      data[0] = mxCreateDoubleMatrix(siz,1,mxREAL);
      lp = (double *)mxGetPr(data[0]);
      for(i=0; i<siz; i++) *lp++ = (double)((BYTE *)Buffer)[i];
      break;
    case CF_INT16:
      data[0] = mxCreateDoubleMatrix(siz,1,mxREAL);
      lp = (double *)mxGetPr(data[0]);
      for(i=0; i<siz; i++) *lp++ = (double)((short *)Buffer)[i];
      break;
    case CF_INT32:
      data[0] = mxCreateDoubleMatrix(siz,1,mxREAL);
      lp = (double *)mxGetPr(data[0]);
      for(i=0; i<siz; i++) *lp++ = (double)((int *)Buffer)[i];
      break;
#   ifdef SINT64
    case CF_INT64:
      data[0] = mxCreateDoubleMatrix(siz,1,mxREAL);
      lp = (double *)mxGetPr(data[0]);
      for(i=0; i<siz; i++) *lp++ = (double)((SINT64 *)Buffer)[i];
      break;
#   endif
    case CF_FLOAT:
      data[0] = mxCreateDoubleMatrix(siz,1,mxREAL);
      lp = (double *)mxGetPr(data[0]);
      for(i=0; i<siz; i++) *lp++ = (double)((float *)Buffer)[i];
      break;
    case CF_DOUBLE:
      data[0] = mxCreateDoubleMatrix(siz,1,mxREAL);
      lp = (double *)mxGetPr(data[0]);
      for(i=0; i<siz; i++) *lp++ = ((double *)Buffer)[i];
      break;
    default:
      err = illegal_format; 
      data[0] = mxCreateString("oops!");
      goto inpt_err; 
  }

inpt_err:
  if (err != 0)
  {
    GetLastLinkError(err,stsbuf);
    sprintf(buf,"tine_pulldata : <%d> %.64s",err,stsbuf);
    mexErrMsgTxt(buf);
  }
  plhs[0] = data[0];
err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
  //if (dev != NULL) mxFree(dev);
  //if (prp != NULL) mxFree(prp);
}
