/*
 * tine_history.cpp : 
 *
 */

#include <math.h>
#include "mex.h"
#include "tine.h"

#define mxerr(s) { mxerrStr = (char *)s; goto err; }
#define ARCHIVE_CAPACITY 8000
DBLDBLDBL dddDataArray[ARCHIVE_CAPACITY];
DBLDBL *ddDataArray = (DBLDBL *)dddDataArray;

time_t parseTimeString(char *ts)
{
  time_t t=time(NULL);
  int i, len;
  char *c,day[16],mon[64],year[16];
  char hr[64],mn[32],sec[16];
  struct tm tmts, *tnow;

  hr[0] = 0;
  tnow = localtime(&t);
  len = strlen(ts);
  strupr(ts);
  for (i=0; i<len && isdigit(ts[i]); i++);  
  if (i == len) t = atoi(ts);
  else if (!strcmp(ts,"NOW")) t = time(NULL);
  else if ((c=strstr(ts,".")) != NULL) /* day.month.year format */
  {
    *c++ = 0;
    strncpy(day,ts,16);
    strncpy(mon,c,64);
    if ((c=strstr(mon,".")) != NULL)
    { 
      *c++ = 0;
      strncpy(year,c,16);
      if ((c=strstr(year,"_")) != NULL || (c=strstr(year," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    else
    {
      sprintf(year,"%d",1900+tnow->tm_year);
      if ((c=strstr(mon,"_")) != NULL || (c=strstr(mon," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    if (strlen(hr) > 0)
    {
      mn[0] = 0; sec[0] = 0;
      if ((c=strstr(hr,":")) != NULL)
      {     
        *c++ = 0;
        strncpy(mn,c,32);
        if ((c=strstr(mn,":")) != NULL)
        {
          *c++ = 0;
          strncpy(sec,c,16);
        }
      }
      tmts.tm_hour = atoi(hr);
      tmts.tm_min =  atoi(mn);
      tmts.tm_sec =  atoi(sec);
    }
    else
    {
      memset(&tmts,0,sizeof(struct tm));
    }
    tmts.tm_mon = (atoi(mon)+11)%12;
    tmts.tm_mday = atoi(day);
    tmts.tm_year = atoi(year);
    if (tmts.tm_year > 1900) tmts.tm_year -= 1900;
    tmts.tm_isdst = -1;
    t = mktime(&tmts);
  }
  return t;
}
int mlParseFullDeviceName(char *fullname,char *dev,char *prp,int isHistoryCall)
{
  char *c,tmp[STATUS_SIZE+8];
  int isBracketed = FALSE;

  strncpy(tmp,fullname,STATUS_SIZE+8);
  if ((c=strchr(tmp,'[')) != 0)
  { /* fullname given as "/<context>/<server>/<device>[<property>]" */
    *c++ = 0;
    strncpy(prp,c,PROPERTY_NAME_SIZE);
    if ((c=strchr(prp,']')) != 0) *c = 0;
    isBracketed = TRUE;
  }
  if (!isBracketed)
  { /* fullname given as "/<context>/<server>/<device>/<property>" */
    strrev(tmp);
    if ((c=strchr(tmp,'/')) == 0) return illegal_delimiter;
    *c++ = 0;
    strncpy(prp,tmp,PROPERTY_NAME_SIZE); strrev(prp);
    strncpy(dev,c,STATUS_SIZE); strrev(dev);
  }
  else
  {
    strncpy(dev,tmp,STATUS_SIZE);
  }
  if (isHistoryCall)
  {
    if (strstr(dev,"HISTORY") == NULL && /* not a central archive call */
      strstr(prp,".HIST") == NULL && strstr(prp,".HST") == NULL) /* meta property not yet appended */
    {
      if (strlen(prp) < PROPERTY_NAME_SIZE-4) strcat(prp,".HST");
    }
  }
  return 0;
}
int mlGetArchivedDataAsDoublet(char *devsrv,time_t start,time_t stop,int sample,int index,DBLDBL *ddDataArray,int *num)
{
  DTYPE din,dout;
  UINT32 startstopArray[4];
  char aServer[STATUS_SIZE],prp[PROPERTY_NAME_SIZE];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;
  if ((cc=mlParseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  startstopArray[2] = (UINT32)index;
  startstopArray[3] = (UINT32)sample;
  dout.dArrayLength = *num;
  dout.dFormat = CF_DBLDBL;
  dout.data.vptr = ddDataArray;
  din.dArrayLength = 4;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY|CA_MUTABLE,200)) != 0) *num = 0;
  else *num = (int)dout.dArrayLength;

  return cc;
}
int mlGetArchivedDataAsTriplet(char *devsrv,time_t start,time_t stop,int sample,int index,DBLDBLDBL *dddDataArray,int *num)
{
  DTYPE din,dout;
  UINT32 startstopArray[4];
  char aServer[STATUS_SIZE],prp[PROPERTY_NAME_SIZE];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;
  if ((cc=mlParseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  startstopArray[2] = (UINT32)index;
  startstopArray[3] = (UINT32)sample;
  dout.dArrayLength = *num;
  dout.dFormat = CF_DBLDBLDBL;
  dout.data.vptr = ddDataArray;
  din.dArrayLength = 4;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY|CA_MUTABLE,200)) != 0) *num = 0;
  else *num = (int)dout.dArrayLength;

  return cc;
}
#define usageStr "tine_history usage:\ntine_history('/<context>/<server>/<name>[<property>]','<stop time>','<depth>'[,sampleInterval,index,'acquireSystemStamps'])\n'ret' contains array of (timestamp,data) or (timestamp,system stamp,data)"
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) 
{
  char *parameter, *mxerrStr=NULL;
  time_t start, stop;
	int num, idx;
  mxArray *temp;
  double *ptr;
  int buflen, status, cc;
  char *c, c0, str[128];
  int i,len,depth=86400,sample=0,acqSysStmp=FALSE;
 
  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 6:
      if (mxIsChar(prhs[5])) 
      {
        len = mxGetN(prhs[5]) + 1;
        c = (char *)mxCalloc(len, sizeof(char));
        mxGetString(prhs[5],c,len);
        c0 = toupper(*c);
        if (c0 == 'Y' || c0 == 'T') acqSysStmp = TRUE;
      }
      else if (mxIsDouble(prhs[5])) 
      {
        acqSysStmp = (int)mxGetScalar(prhs[5]);
      }
    case 5:
      if (!mxIsDouble(prhs[4])) mxerr(usageStr);
      idx = (int)mxGetScalar(prhs[4]);
    case 4:
      if (!mxIsDouble(prhs[3])) mxerr(usageStr);
      sample = (int)mxGetScalar(prhs[3]);
    case 3:
      if (mxIsChar(prhs[2]) != 1) mxerr("depth input must be a string.");
      mxGetString(prhs[2],str,128);
      depth = strtol(str,&c,10);
      switch (tolower(*c))
      {
        case 'd': depth *= 86400; 
          break;
        case 'h': depth *= 3600; 
          break;
        case 'm':
          if (strnicmp(c,"min",3) == 0)
          {
            depth *= 60;
          }
          else if (strnicmp(c,"mon",3) == 0)
          {
            depth *= 86400 * 31;
          }
          break;
        case 'w':
          depth *= 86400 * 7;
          break;
        default:
        case 's': 
          break;
      }
    case 2:
      if (mxIsChar(prhs[1]) != 1) mxerr("stop time input must be a string.");
      mxGetString(prhs[1],str,128);
      stop = parseTimeString(str);
      start = stop - depth;
      if (mxGetM(prhs[0]) != 1) mxerr("Input must be a row vector.");
      if (mxIsChar(prhs[0]) != 1) mxerr("target input must be a string.");
      break;
    default:
      mxerr(usageStr);
  }
  if (nlhs > 1) mxerr("Too many output arguments");

  /* Get the length of the input string. */
  buflen = mxGetM(prhs[0])*mxGetN(prhs[0]) + 1;
  
  /* Allocate memory for input and output strings. */
  parameter = (char *) mxCalloc(buflen, sizeof(char));
  
  /* copy string into parameter */
  status = mxGetString(prhs[0], parameter, buflen);
  if(status != 0) mexWarnMsgTxt("Not enough space. String is truncated.");

  num = ARCHIVE_CAPACITY;

  if (acqSysStmp)
  {
    cc = mlGetArchivedDataAsTriplet(parameter, start, stop, sample, idx, dddDataArray, &num);
  }
  else
  {
    cc = mlGetArchivedDataAsDoublet(parameter, start, stop, sample, idx, ddDataArray, &num);
  }
  if (cc != 0) 
  {
  	mexPrintf("Error reading data, status = %d.\n", cc);
    mexPrintf("Error code is: %s\n", GetRPCLastError(cc));
    num = 0;
  }
  /* create matrix for data */
  if (acqSysStmp)
  {
    temp = mxCreateDoubleMatrix(num, 3, mxREAL);
    ptr = mxGetPr(temp);
    for (i = 0; i < num; i++) {
        ptr[i]     = dddDataArray[i].d2val/86400.0 + 719529.0;
        ptr[i+num] = dddDataArray[i].d3val;
        ptr[i+2*num] = dddDataArray[i].d1val;
    }
  }
  else
  {
    temp = mxCreateDoubleMatrix(num, 2, mxREAL);
    ptr = mxGetPr(temp);
    for (i = 0; i < num; i++) {
        ptr[i]     = ddDataArray[i].d2val/86400.0 + 719529.0;
        ptr[i+num] = ddDataArray[i].d1val;
    }
  }
  plhs[0] = temp; 
err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
  if (parameter != NULL) mxFree(parameter);
}