#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"

#define	PROPMAX 10

int isTaggable(short fmt)
{
  switch (fmt)
  {
    case CF_STRUCT:
    case CF_BITFIELD8:
    case CF_BITFIELD16:
    case CF_BITFIELD32:
    case CF_BITFIELD64:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

#define usageStr "tine_register_property usage:\n ret=tine_register_property('property_name',in_size,'in_type',out_size,'out_type',max,min,'egu',access,'description'[,'array_type',row_length])"
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[]) 
{
  mxArray	*data[PROPMAX];
  char *prp=NULL, *inFmtStr=NULL, *outFmtStr=NULL, *accStr=NULL, *egu=NULL, *dsc=NULL, *atypstr=NULL, *mxerrStr=NULL;
  char *c,buf[128],stsbuf[64],inTag[32],outTag[32];
  double *lp;
  int	i,k,len,err,n;
  int acc=CA_READ, inFmt=CF_NULL, outFmt=CF_NULL, inSiz=0, outSiz=0, atyp=AT_UNKNOWN, rowlen=0;
  float min=0, max=0;
    
  if (plhs == NULL || prhs == NULL) return;
  outTag[0] = inTag[0] = 0;
  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 12:
      if (!mxIsDouble(prhs[11])) mxerr(usageStr);
      if (mxGetM(prhs[11]) != 1) mxerr(usageStr);
      rowlen = (int)mxGetScalar(prhs[11]);
    case 11:
      if (!mxIsChar(prhs[10])) mxerr(usageStr);
      if (mxGetM(prhs[10]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[10]) + 1;
      atypstr = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[10],atypstr,len);
      atyp = (int)GetArrayTypeFromString(atypstr);
    case 10: 
      if (!mxIsChar(prhs[9])) mxerr(usageStr);
      if (mxGetM(prhs[9]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[9]) + 1;
      dsc = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[9],dsc,len);
      if (!mxIsChar(prhs[8])) mxerr(usageStr);
      if (mxGetM(prhs[8]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[8]) + 1;
      accStr = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[8],accStr,len);
      acc = GetAccessMode(accStr);

      if (!mxIsChar(prhs[7])) mxerr(usageStr);
      if (mxGetM(prhs[7]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[7]) + 1;
      egu = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[7],egu,len);
      if (!mxIsDouble(prhs[6])) mxerr(usageStr);
      if (mxGetM(prhs[6]) != 1) mxerr(usageStr);
      min = (float)mxGetScalar(prhs[6]);
      if (!mxIsDouble(prhs[5])) mxerr(usageStr);
      if (mxGetM(prhs[5]) != 1) mxerr(usageStr);
      max = (float)mxGetScalar(prhs[5]);
      if (!mxIsChar(prhs[4])) mxerr(usageStr);
      if (mxGetM(prhs[4]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[4]) + 1;
      outFmtStr = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[4],outFmtStr,len);
      outFmt = GetFormatFromString(outFmtStr);
      if ((c=strchr(outFmtStr,'.')) != NULL)
      {
        if (isTaggable(outFmt))
        {
          c++;
          strncpy(outTag,c,TAG_NAME_SIZE);
        }
      }
      if (!mxIsDouble(prhs[3])) mxerr(usageStr);
      if (mxGetM(prhs[3]) != 1) mxerr(usageStr);
      outSiz = (int)mxGetScalar(prhs[3]);
      if (!mxIsChar(prhs[2])) mxerr(usageStr);
      if (mxGetM(prhs[2]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[2]) + 1;
      inFmtStr = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[2],inFmtStr,len);
      inFmt = GetFormatFromString(inFmtStr);
      if ((c=strchr(inFmtStr,'.')) != NULL)
      {
        if (isTaggable(inFmt))
        {
          c++;
          strncpy(inTag,c,TAG_NAME_SIZE);
        }
      }
      if (!mxIsDouble(prhs[1])) mxerr(usageStr);
      if (mxGetM(prhs[1]) != 1) mxerr(usageStr);
      inSiz = (int)mxGetScalar(prhs[1]);
      if (!mxIsChar(prhs[0])) mxerr(usageStr);
      if (mxGetM(prhs[0]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[0]) + 1;
      prp = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[0],prp,len);
      break;
    default:
      mxerr(usageStr);
  }   

  err = RegisterBufferedPropertyEx2(prp,inSiz,inFmt,inTag,outSiz,outFmt,outTag,max,min,egu,acc,dsc,-1,atyp,rowlen); 

  if (err < 0)
  {
    GetLastLinkError(-err,stsbuf);
    sprintf(buf,"tine_register_property : <%d> %.64s",-err,stsbuf);
    mexErrMsgTxt(buf);
  }
  data[0] = mxCreateDoubleMatrix(1,1,mxREAL);
  lp = (double *)mxGetPr(data[0]);
  *lp = (double)err;
  plhs[0] = data[0];
  
err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
  if (prp != NULL) mxFree(prp);
  if (inFmtStr != NULL) mxFree(inFmtStr);
  if (outFmtStr != NULL) mxFree(outFmtStr);
  if (accStr != NULL) mxFree(accStr);
  if (egu != NULL) mxFree(egu);
  if (dsc != NULL) mxFree(dsc);
  if (atypstr != NULL) mxFree(atypstr);
}
