#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"
#if defined(UNIX)
#include "threader.h"
#endif

#define	PROPMAX 10

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

#define usageStr "tine_register_server usage:\n ret=tine_register_server('export_name','equipment_module_name',device_capacity)"
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[]) 
{
  mxArray	*data[PROPMAX];
  char *exp=NULL, *eqm=NULL, *mxerrStr=NULL;
  char buf[128],stsbuf[64];
  double *lp;
  int	i,k,len,err,n;
  int ndevs=0;
    
  if (plhs == NULL || prhs == NULL) return;
  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 3:
      if (!mxIsDouble(prhs[2])) mxerr(usageStr);
      if (mxGetM(prhs[2]) != 1) mxerr(usageStr);
      ndevs = (int)mxGetScalar(prhs[2]);
      if (!mxIsChar(prhs[1])) mxerr(usageStr);
      if (mxGetM(prhs[1]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[1]) + 1;
      eqm = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[1],eqm,len);
      if (!mxIsChar(prhs[0])) mxerr(usageStr);
      if (mxGetM(prhs[0]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[0]) + 1;
      exp = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[0],exp,len);
      break;
    default:
      mxerr(usageStr);
  }   

  sprintf(buf,"%.32s/INIT=SRV|KEEP",exp);
  err = AttachServerEx(buf,eqm,ndevs,NULL,0);

  if (err != 0)
  {
    GetLastLinkError(err,stsbuf);
    sprintf(buf,"tine_register_server : <%d> %.64s",err,stsbuf);
    mexErrMsgTxt(buf);
  }
# if defined(UNIX)
  else
  {
    SystemSetCycleTimer();
  }
# endif
  data[0] = mxCreateDoubleMatrix(1,1,mxREAL);
  lp = (double *)mxGetPr(data[0]);
  *lp = (double)err;
  plhs[0] = data[0];
  
err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
  if (exp != NULL) mxFree(exp);
  if (eqm != NULL) mxFree(eqm);
}
