/* tinewrite.cpp   Interface zur Tine-Library          */
/* last modified   MH 1.10.2002                      */
/*                 MH, JK 18.6.2003                  */

#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"

#define DEBUG 0
#define	PROPMAX	1024
#define	PROPMAX1 512
#define	MAX_TINE_BUFFER 65000

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

short ftoi(char *fmt)
{
  char *c,fmtstr[64];
  short ifmt;
  strncpy(fmtstr,fmt,64);
  if ((c=strchr(fmtstr,'.')) != NULL) *c = 0;
  strupr(fmtstr);
  if ((ifmt=getSimpleFormatType(fmtstr)) != CF_NULL) return ifmt;
  return getCompoundFormatType(fmtstr);
}
int read_prop(char *dev, char *prop, int *size, char *tag)
{ 
  PropertyQueryStruct	svrProp; // legacy
  PrpQueryStruct pqs;
  DTYPE dout, din;
  int cc, fmt, isLegacy=FALSE;
  char *prp = "PROPS";

  if ((cc=getPropertyInfoFromListenerEx(dev,prop,size,&fmt,tag)) == 0)
  {
    return fmt;
  }

  memset(&din,0,sizeof(DTYPE));
  din.dFormat = CF_TEXT;
  din.dArrayLength = (int)strlen(prop);
  din.data.vptr = prop;

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  dout.data.vptr = &pqs;
  strcpy(dout.dTag,"PRPQSr4");

legacyRetry:
  if ((cc=alsnExecLink(dev,prp,&dout,&din,CA_READ,500)) != 0)
  {
    if (!isLegacy && (cc == illegal_format || cc == invalid_structure_tag))
    {
      isLegacy = TRUE;
      dout.dFormat = CF_STRUCT;
      dout.dArrayLength = PROPERTYQUERYSTRUCT_SIZE;
      dout.data.vptr = &svrProp;
      dout.dTag[0] = '\0';
      goto legacyRetry;
    }
    *size = 0;
    return -cc;
  }

  if (isLegacy)
  {
    *size = svrProp.prpSize;
    fmt = LFMT(svrProp.prpFormat);
  }
  else
  {
    *size = pqs.prpSize;
    fmt = LFMT(pqs.prpFormat);
    if (tag != NULL) strncpy(tag,pqs.prpTag,TAG_NAME_SIZE);
  }
  return fmt;
}

int	write_data(char *dev, char *prop, int type, int size, char *buf,int timeout,char *dtag)
{
  DTYPE	din;
  int	cc;

  memset(&din,0,sizeof(DTYPE));
  din.dFormat = type;
  din.dArrayLength = size;
  din.data.vptr = buf != NULL ? (void *)&buf[0] : NULL;
  strncpy(din.dTag,dtag,TAG_NAME_SIZE);
  cc = alsnExecLink(dev,prop,NULL, &din,CA_WRITE|CA_RETRY,timeout);
  if(cc && cc != -141)
  {
    GetLastLinkError(cc,buf);
    return 1;
  }
  return 0;
}

#define setDblVal(i,n,b,f,v) {\
  for (i=0; i<(n); i++) {\
  if ((f) == CF_BYTE) (b)[i] = (BYTE)((v)[i]);\
    else if ((f) == CF_INT16) ((short *)(b))[i] = (short)((v)[i]);\
    else if ((f) == CF_INT32) ((int *)(b))[i] = (int)((v)[i]);\
    else if ((f) == CF_FLOAT) ((float *)(b))[i] = (float)((v)[i]);\
    else ((double *)(b))[i] = ((v)[i]);\
  }\
}

#define setDbltVal(n,b,f,v) {\
  for (i=0; i<(n); i++) {\
  if ((f) == CF_INTINT) {\
  ((INTINT *)(b))[i].i1val = (int)((v)[i*2]);\
  ((INTINT *)(b))[i].i2val = (int)((v)[i*2+1]);\
  }\
    else if ((f) == CF_FLTINT) {\
    ((FLTINT *)(b))[i].fval = (float)((v)[i*2]);\
    ((FLTINT *)(b))[i].ival = (int)((v)[i*2+1]);\
}\
    else if ((f) == CF_FLTFLT) {\
    ((FLTFLT *)(b))[i].f1val = (float)((v)[i*2]);\
    ((FLTFLT *)(b))[i].f2val = (float)((v)[i*2+1]);\
}\
    else if ((f) == CF_DBLDBL) {\
    ((DBLDBL *)(b))[i].d1val = ((v)[i*2]);\
    ((DBLDBL *)(b))[i].d2val = ((v)[i*2+1]);\
}\
  }\
}

#define setTpltVal(n,b,f,v) {\
  for (i=0; i<(n); i++) {\
  if ((f) == CF_INTINTINT) {\
  ((INTINTINT *)(b))[i].i1val = (int)((v)[i*3]);\
  ((INTINTINT *)(b))[i].i2val = (int)((v)[i*3+1]);\
  ((INTINTINT *)(b))[i].i3val = (int)((v)[i*3+2]);\
  }\
    else if ((f) == CF_FLTINTINT) {\
    ((FLTINTINT *)(b))[i].fval = (float)((v)[i*3]);\
    ((FLTINTINT *)(b))[i].i1val = (int)((v)[i*3+1]);\
    ((FLTINTINT *)(b))[i].i2val = (int)((v)[i*3+2]);\
}\
    else if ((f) == CF_INTFLTINT) {\
    ((INTFLTINT *)(b))[i].i1val = (int)((v)[i*3]);\
    ((INTFLTINT *)(b))[i].fval = (float)((v)[i*3+1]);\
    ((INTFLTINT *)(b))[i].i2val = (int)((v)[i*3+2]);\
}\
    else if ((f) == CF_FLTFLTINT) {\
    ((FLTFLTINT *)(b))[i].f1val = (float)((v)[i*3]);\
    ((FLTFLTINT *)(b))[i].f2val = (float)((v)[i*3+1]);\
    ((FLTFLTINT *)(b))[i].ival = (int)((v)[i*3+2]);\
}\
    else if ((f) == CF_DBLDBLDBL) {\
    ((DBLDBLDBL *)(b))[i].d1val = ((v)[i*3]);\
    ((DBLDBLDBL *)(b))[i].d2val = ((v)[i*3+1]);\
    ((DBLDBLDBL *)(b))[i].d3val = ((v)[i*3+2]);\
}\
  }\
}

#define setQuadVal(n,b,f,v) {\
  for (i=0; i<(n); i++) {\
  if ((f) == CF_INTINTINTINT) {\
  ((DADDRESS *)(b))[i].i1val = (int)((v)[i*4]);\
  ((DADDRESS *)(b))[i].i2val = (int)((v)[i*4+1]);\
  ((DADDRESS *)(b))[i].i3val = (int)((v)[i*4+2]);\
  ((DADDRESS *)(b))[i].i4val = (int)((v)[i*4+3]);\
  }\
    else if ((f) == CF_INTFLTFLTFLT) {\
    ((DFILTER *)(b))[i].ival = (int)((v)[i*4]);\
    ((DFILTER *)(b))[i].f1val = (float)((v)[i*4+1]);\
    ((DFILTER *)(b))[i].f2val = (float)((v)[i*4+2]);\
    ((DFILTER *)(b))[i].f3val = (float)((v)[i*4+3]);\
}\
    else if ((f) == CF_INTINTFLTFLT) {\
    ((FWINDOW *)(b))[i].i1val = (int)((v)[i*4]);\
    ((FWINDOW *)(b))[i].i2val = (int)((v)[i*4+1]);\
    ((FWINDOW *)(b))[i].f1val = (float)((v)[i*4+2]);\
    ((FWINDOW *)(b))[i].f2val = (float)((v)[i*4+3]);\
}\
    else if ((f) == CF_TTII) {\
    ((DWINDOW *)(b))[i].t1 = (int)((v)[i*4]);\
    ((DWINDOW *)(b))[i].t2 = (int)((v)[i*4+1]);\
    ((DWINDOW *)(b))[i].i1val = (int)((v)[i*4+2]);\
    ((DWINDOW *)(b))[i].i2val = (int)((v)[i*4+3]);\
}\
    else if ((f) == CF_FLTINTFLTINT) {\
    ((FLTINTFLTINT *)(b))[i].f1val = (float)((v)[i*4]);\
    ((FLTINTFLTINT *)(b))[i].i1val = (int)((v)[i*4+1]);\
    ((FLTINTFLTINT *)(b))[i].f2val = (float)((v)[i*4+2]);\
    ((FLTINTFLTINT *)(b))[i].i2val = (int)((v)[i*4+3]);\
}\
  }\
}
int setMxData(const mxArray *mxa,int fmt,int siz,char *tag,BYTE *buf)
{
  mxChar *mxc;
  mxArray *sa;
  double *lp;
  int fmtsiz = GetFormatSize(fmt);
  int i, k, n, nf, len;
  char c, *cptr;
  const char *fld;
  struct structStruct *ss;
  struct structFormat *sf;

  switch (fmt)
  {
  case CF_NULL:
    break;    
  case CF_TEXT:
    if (!mxIsChar(mxa)) return argument_list_error;
    len = mxGetN(mxa);
    siz = len;
    mxGetString(mxa,(char *)buf,len+1);
    break;
  case CF_NAME8:
  case CF_NAME16:
  case CF_NAME32:
  case CF_NAME48:
  case CF_NAME64:
    if (!mxIsChar(mxa)) return argument_list_error;
    if (siz > 1)
    {
      const mwSize *dims = mxGetDimensions(mxa);
      for (i=0; i<siz; i++)
      {
        mxc = mxGetChars(mxa);
        for (i=0; i<siz; i++)
        { /* string arrays transposed in MatLab ? */
          memset(&buf[fmtsiz*i],0,fmtsiz);
          for (k=0; k<fmtsiz && k<dims[1]; k++) 
          {
            c = (char)mxc[siz*k+i];
            ((char *)buf)[fmtsiz*i+k] = c;
            if (c == 0) break;
          }
        }
      }
    }
    else
    {
      len = mxGetN(mxa) + 1;
      if (len > fmtsiz) len = fmtsiz;
      mxGetString(mxa,(char *)buf,len);
    }
    break;
  case CF_STRING:
    if (mxIsChar(mxa)) 
    {
      len = 0;
      const mwSize *dims = mxGetDimensions(mxa);
      for (i=0; i<siz; i++)
      {
        mxc = (mxChar *)mxGetData(mxa);
        for (i=0; i<siz; i++)
        { /* string arrays transposed in MatLab ? */
          cptr = (char *)&buf[sizeof(size_t)*siz + len];
          ((size_t *)buf)[i] = (size_t)cptr;
          for (k=0; k<fmtsiz && k<dims[1]; k++) 
          {
            c = (char)mxc[siz*k+i];
            cptr[len++] = c;
            if (c == 0) break;
          }
        }
      }
    }
    else
    {
      return argument_list_error;
    }
    break;
  case CF_STRUCT:
    if(!mxIsStruct(mxa)) return argument_list_error;
    if ((ss=GetStructFromTag(tag)) != NULL && siz > 0)
    {
      fmtsiz = GetStructSize(tag);
      for (k=0, sf=ss->s; sf != NULL && k < 256; k++, sf=sf->nxt) if (sf->fmt == CF_NULL) k--;
      nf = k; // number of fields (n.b. legacy registrations might have a 'CF_NULL' in termination)s
      if (nf != mxGetNumberOfFields(mxa)) return invalid_structure_size;
      for (i=0; i<siz; i++)
      {
        for (n=nf-1, sf=ss->s; sf != NULL && n >=0; n--, sf=sf->nxt)
        { 
          fld = mxGetFieldNameByNumber(mxa,n);
          sa = mxGetField(mxa,i,fld);
          switch (sf->fmt)
          {
          case CF_INT16:
          case CF_INT32:
          case CF_FLOAT:
          case CF_DOUBLE:
            if (!mxIsDouble(sa)) return invalid_structure_tag;
            lp = (double *)mxGetPr(sa);
            setDblVal(k,sf->siz,&buf[fmtsiz*i+sf->off],sf->fmt,lp);
            break;
          case CF_TEXT:
            if (!mxIsChar(sa)) return invalid_structure_tag;
            len = mxGetN(sa);
            mxGetString(sa,(char *)&buf[fmtsiz*i+sf->off],len+1);
            break;
          case CF_NULL:
            n++; 
            continue;
          }
        }
      }
    }
    break;
  case CF_SPECTRUM:
    if (mxIsStruct(mxa)) 
    {
      fld = mxGetFieldNameByNumber(mxa,0); // comment
      sa = mxGetField(mxa,0,fld);
      if (!mxIsChar(sa)) return argument_list_error;
      len = mxGetN(sa); if (len > 79) len = 79;
      mxGetString(sa,(char *)&buf[0],len+1);
      fld = mxGetFieldNameByNumber(mxa,1); // tm
      sa = mxGetField(mxa,0,fld);
      if (!mxIsDouble(sa)) return argument_list_error;
      lp = (double *)mxGetPr(sa);
      setDblVal(i,1,&buf[80],CF_INT32,lp);
      fld = mxGetFieldNameByNumber(mxa,2); // s_start
      sa = mxGetField(mxa,0,fld);
      if (!mxIsDouble(sa)) return argument_list_error;
      lp = (double *)mxGetPr(sa);
      setDblVal(i,1,&buf[84],CF_FLOAT,lp);
      fld = mxGetFieldNameByNumber(mxa,3); // s_inc
      sa = mxGetField(mxa,0,fld);
      if (!mxIsDouble(sa)) return argument_list_error;
      lp = (double *)mxGetPr(sa);
      setDblVal(i,1,&buf[88],CF_FLOAT,lp);
      fld = mxGetFieldNameByNumber(mxa,4); // status
      sa = mxGetField(mxa,0,fld);
      if (!mxIsDouble(sa)) return argument_list_error;
      lp = (double *)mxGetPr(sa);
      setDblVal(i,1,&buf[92],CF_INT32,lp);
      fld = mxGetFieldNameByNumber(mxa,5); // d_spect_array
      sa = mxGetField(mxa,0,fld);
      if (!mxIsDouble(sa)) return argument_list_error;
      lp = (double *)mxGetPr(sa);
      setDblVal(i,siz,&buf[96],CF_FLOAT,lp);
      break;
    }
    if (!mxIsDouble(mxa)) return argument_list_error;
    memset(buf,0,96);
    lp = (double *)mxGetPr(mxa);
    setDblVal(i,siz,&buf[96],CF_FLOAT,lp);
    break;
  case CF_BYTE:
  case CF_SHORT:
  case CF_INT32:
  case CF_FLOAT:
  case CF_DOUBLE:
    if (!mxIsDouble(mxa)) return argument_list_error;
    lp = (double *)mxGetPr(mxa);
    setDblVal(i,siz,buf,fmt,lp);
    break;
  case CF_BOOLEAN:
    if (!mxIsDouble(mxa)) return argument_list_error;
    lp = (double *)mxGetPr(mxa);
    for (i=0; i<siz; i++) ((int *)buf)[i] = ((int)lp[i] == 0 ? 0 : 1);
    break;		
  case CF_INTINT:
  case CF_FLTINT:
  case CF_FLTFLT:
  case CF_DBLDBL:
    if (!mxIsDouble(mxa)) return argument_list_error;
    lp = (double *)mxGetPr(mxa);
    siz /= 2;
    setDbltVal(siz,buf,fmt,lp);
    break;
  case CF_INTINTINT:
  case CF_FLTINTINT:
  case CF_INTFLTINT:
  case CF_FLTFLTINT:
  case CF_DBLDBLDBL:
    if (!mxIsDouble(mxa)) return argument_list_error;
    lp = (double *)mxGetPr(mxa);
    siz /= 3;
    setTpltVal(siz,buf,fmt,lp);
    break;
  case CF_IFFF:
  case CF_IIFF:
  case CF_IIII:
  case CF_TTII:
  case CF_FIFI:
    if (!mxIsDouble(mxa)) return argument_list_error;
    lp = (double *)mxGetPr(mxa);
    siz /= 4;
    setQuadVal(siz,buf,fmt,lp);
    break;		
    /*
    case CF_SPECTRUM:
    need 3 call
    break;
    case CF_USTRING:
    if(!mxIsChar(mxa))
    {
    plhs[0] = mxCreateString("Input argument error !!!");
    return;
    }
    len = mxGetN(mxa) + 1;
    mxGetString(mxa,buf,len);
    siz = 1; // kluge this by forcing only 1 of these to be sent
    break;
    */
  default:
    if (!mxIsChar(mxa)) return illegal_format;
  }
  return 0;
}

#define usageStr "tine_write usage:\n ret=tine_write(values,'/<context>/<server>/<device>[<property>]'[,'<data_type>',timeout]);\n where values = string or double Nx1 array (or omitted to send a COMMAND without input)\n'ret' contains the status/error message"
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) 
{
  char Buffer[MAX_TINE_BUFFER];
  double *lp;
  DTYPE	dout,din;
  char *name, *typestr, *mxerrStr=NULL;
  char *p;
  char buf[1024];
  char bf1[1024],*c=NULL, *c1=NULL, *c2=NULL, *c3=NULL;
  char dtstr[32],dtag[32];
  char dev[80];
  int	len,type=CF_NULL,size=0,fmtsiz,readsize = -1,ln,tmr=500,datatype = -1;
  int	i,j,k,err,count,n,tgt_idx=1,isCmd=0;
  float fl;
  char *pf;


  if (plhs == NULL || prhs == NULL) return;
  dtag[0] = 0;

  /* Check for proper number of arguments. */
  switch (nrhs)
  {
  case 4: // timer interval (int)
    if (!mxIsDouble(prhs[3])) mxerr(usageStr);
    if (mxGetM(prhs[3]) != 1) mxerr(usageStr);
    tmr = (int)mxGetScalar(prhs[3]);
  case 3: // input data type
    if (!mxIsChar(prhs[2])) mxerr(usageStr);
    if (mxGetM(prhs[2]) != 1) mxerr(usageStr);
    len = mxGetN(prhs[2]) + 1;
    typestr = (char *)mxCalloc(len, sizeof(char));
    mxGetString(prhs[2],typestr,len);
    strncpy(dtstr,typestr,32);
    if ((c=strchr(dtstr,'/')) != NULL || 
      (c=strchr(dtstr,'\\')) != NULL ||
      (c=strchr(dtstr,'.')) != NULL)
    {
      *c++ = 0;
      strncpy(dtag,c,TAG_NAME_SIZE);
    }
    strupr(dtstr);
    if (dtstr[strlen(dtstr)-1] == 'S') dtstr[strlen(dtstr)-1] = 0; // in case "floats" instead of "float"
    datatype = ftoi(dtstr);
    mxFree(typestr);
  case 2:
  case 1:
    if (nrhs == 1)
    {
      if (!mxIsChar(prhs[0])) mxerr(usageStr);
      tgt_idx = 0;
      isCmd = -1;
    }
    break;
  default:
    mxerr(usageStr);
  }  

  len = mxGetN(prhs[tgt_idx]) + 1;
  name = (char *)mxCalloc(len, sizeof(char));
  mxGetString(prhs[tgt_idx],name,len);
  strcpy(buf,name);
  if (buf[0] == '?')
  {
    strcpy(buf,usageStr);
    plhs[0] = mxCreateString(buf);
    mxFree(name);
    return;
  }
  if((p=strchr(name,'[')) == NULL || (p=strchr(name,']')) == NULL) 
  {
    mxFree(name);
    mxerr(usageStr);
  }
  if ((c1=strchr(buf,'<')) != NULL)
  {
    c = c1; c++;
    strncpy(bf1,c,sizeof(bf1));
    for (c2=bf1; *c2 == ' '; c2++);  // any leading blanks?
    readsize = atoi(c2);
    if ((c2=strchr(c2,' ')) != NULL) // is the datatype left to parse out ?
    {
      while (*c2 == ' ') c2++;       // remove all leading blanks
      if ((c3=strchr(c2,'>')) != NULL) *c3 = 0;
      strncpy(dtstr,c2,32);
      strupr(dtstr);
      if (dtstr[strlen(dtstr)-1] == 'S') dtstr[strlen(dtstr)-1] = 0;
      datatype = ftoi(dtstr);
    }
  }
  if ((c=strchr(buf,'@')) != NULL)
  {
    *c++ = 0; // peel it off before we start tokenizing
    tmr = atoi(c);
  }
  if (c1 != NULL) *c1 = 0; // peel this one off too, if it's there

  strcpy(dev,strtok(buf,"["));
  strncpy(bf1,"",sizeof(bf1));
  strcpy(bf1,strtok(NULL,"]"));
  if (!isCmd && datatype < 0)
  {
    type = read_prop(dev, bf1, &size, dtag);
  }

  if (datatype > 0) type = datatype;
  size = mxGetM(prhs[0]); // set the size to what was input (num rows) !
  if (type == CF_STRUCT && strlen(dtag) > 0)
  {
    err = AcquireAndRegisterStruct(dev,dtag,size);
  }

  if (!isCmd)
  {
    err = setMxData(prhs[0],type,size,dtag,(BYTE *)Buffer);
    if (err != 0)
    {
      mxerr("could not set input data !");
    }
  }
  err = write_data(dev,bf1,type,size,Buffer,tmr,dtag);
  if(err)	plhs[0] = mxCreateString(Buffer);
  else	plhs[0] = mxCreateString("");

err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
}