#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"

#define	PROPMAX 1024
#define	DEVMAX 4096
#define	MAX_TINE_BUFFER 512000 //256000

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

short ftoi(char *fmt)
{
  char *c,fmtstr[64];
  short ifmt;
  strncpy(fmtstr,fmt,64);
  if ((c=strchr(fmtstr,'.')) != NULL) *c = 0;
  strupr(fmtstr);
  if ((ifmt=getSimpleFormatType(fmtstr)) != CF_NULL) return ifmt;
  return getCompoundFormatType(fmtstr);
}
#ifdef WIN32
int mexlog(char *tag,char *text,...)
{
  char str[512];
  FILE *fp = NULL;
  va_list args;
  time_t t = time(NULL);
  static char *fmod = "a+";

  sprintf(str,"%.24s[%s] ",ctime(&t),tag);

  va_start(args,text);
  _vsnprintf(&str[strlen(str)],256,text,args); str[256] = 0;
  va_end(args);

  strcat(str,"\n");
  if ((fp=fopen("l:\\log\\mex.log","a+")) == NULL) return -1;
  fwrite(str,strlen(str),1,fp);

  fclose(fp);
  return 0;
}
#endif
int mexIsStockProperty(char *prop)
{
  if (!stricmp(prop,"DEVICES")) return TRUE;
  if (!stricmp(prop,"NDEVICES")) return TRUE;
  if (!stricmp(prop,"NPROPS")) return TRUE;
  if (!stricmp(prop,"NPROPERTIES")) return TRUE;
  if (!stricmp(prop,"PROPERTIES")) return TRUE;
  return FALSE;
}
int read_prop(char *dev, char *prop, int *size, char *tag)
{ 
  PropertyQueryStruct	svrProp; // legacy
  PrpQueryStruct pqs;
  DTYPE dout, din;
  int cc, fmt, isLegacy=FALSE;
  char *prp = "PROPS";

  if ((cc=getPropertyInfoFromListenerEx(dev,prop,size,&fmt,tag)) == 0)
  {
    return fmt;
  }
  if (mexIsStockProperty(prop)) prp = "STOCKPROPS";

  memset(&din,0,sizeof(DTYPE));
  din.dFormat = CF_TEXT;
  din.dArrayLength = (int)strlen(prop);
  din.data.vptr = prop;

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  dout.data.vptr = &pqs;
  strcpy(dout.dTag,"PRPQSr4");

legacyRetry:
  if ((cc=alsnExecLink(dev,prp,&dout,&din,CA_READ,500)) != 0)
  {
    if (!isLegacy && (cc == illegal_format || cc == invalid_structure_tag))
    {
      isLegacy = TRUE;
      dout.dFormat = CF_STRUCT;
      dout.dArrayLength = PROPERTYQUERYSTRUCT_SIZE;
      dout.data.vptr = &svrProp;
      dout.dTag[0] = '\0';
      goto legacyRetry;
    }
    *size = 0;
    return -cc;
 	}

  if (isLegacy)
  {
    *size = svrProp.prpSize;
    fmt = LFMT(svrProp.prpFormat);
  }
  else
  {
    *size = pqs.prpSize;
    fmt = LFMT(pqs.prpFormat);
    if (tag != NULL) strncpy(tag,pqs.prpTag,TAG_NAME_SIZE);
  }
  return fmt;
}
int write_read_data(char *dev, char *prop, int writetype, int writesize, int readtype, int readsize, 
                    char *buf,int access_rate,double *pDataTime,char *dtag,char *dtagw)
{
	DTYPE dout, din;
	int cc;

  if (readsize < 0 || writesize < 0) return dimension_error;
  if (writetype < 0 || readtype < 0) return illegal_format;
  if (buf == NULL) return code_failure;
  memset(&din,0,sizeof(DTYPE)); memset(&dout,0,sizeof(DTYPE));
  din.dFormat = writetype;
  din.dArrayLength = writesize;
  din.data.vptr = buf;
  strncpy(din.dTag,dtagw,TAG_NAME_SIZE);
	dout.dFormat = readtype;
	dout.dArrayLength = readsize;
	dout.data.vptr = buf;
  strncpy(dout.dTag,dtag,TAG_NAME_SIZE);
  if ((cc=ExecLinkEx(dev,prop,&dout,&din,CA_READ|CA_WRITE|CA_MUTABLE,access_rate)) != 0)
  {
    GetLastLinkError(cc,buf);
    if (pDataTime != NULL) *pDataTime = (double)time(NULL);
		return 1;
	}
  if (pDataTime != NULL) *pDataTime = dout.dTimeStamp;
	return 0;
}
char *name_corr(char *name)
{
  int	i, len;
  static char cname[256];

  strncpy(cname,name,256);
  len = strlen(cname);
  for (i = 0; i < len; i++) 
  {
    if (cname[i] == '.') 
    {
      cname[i] = '_';
    } 
    else if (cname[i] == '-') 
    {
      cname[i] = '_';
    }
	}
  return cname;
}

#define getDblVal(i,b,f) (\
  (double)((f) == CF_BYTE ? (b)[i] :\
           (f) == CF_INT16 ? ((short *)(b))[i] :\
           (f) == CF_INT32 ? ((int *)(b))[i] :\
           (f) == CF_FLOAT ? ((float *)(b))[i] :\
           (f) == CF_DOUBLE ? ((double *)(b))[i] : 0))

#define get1stDblVal(i,b,f) (\
  (double)((f) == CF_INTINT ? ((INTINT *)(b))[i].i1val :\
           (f) == CF_FLTINT ? ((FLTINT *)(b))[i].fval :\
           (f) == CF_DBLDBL ? ((DBLDBL *)(b))[i].d1val :\
           (f) == CF_FLTFLT ? ((FLTFLT *)(b))[i].f1val : 0))

#define get2ndDblVal(i,b,f) (\
  (double)((f) == CF_INTINT ? ((INTINT *)(b))[i].i2val :\
           (f) == CF_FLTINT ? ((FLTINT *)(b))[i].ival :\
           (f) == CF_DBLDBL ? ((DBLDBL *)(b))[i].d2val :\
           (f) == CF_FLTFLT ? ((FLTFLT *)(b))[i].f2val : 0))

#define get1stQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i1val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].ival :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].i1val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].t1 :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].f1val : 0))

#define get2ndQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i2val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].f1val :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].i2val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].t2 :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].i1val : 0))

#define get3rdQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i3val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].f2val :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].f1val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].i1val :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].f2val : 0))

#define get4thQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i4val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].f3val :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].f2val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].i2val :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].i2val : 0))

#define get1stTriVal(i,b,f) (\
  (double)((f) == CF_INTINTINT ? ((INTINTINT *)(b))[i].i1val :\
           (f) == CF_FLTFLTINT ? ((FLTFLTINT *)(b))[i].f1val :\
           (f) == CF_INTFLTINT ? ((INTFLTINT *)(b))[i].i1val :\
           (f) == CF_FLTINTINT ? ((FLTINTINT *)(b))[i].fval :\
           (f) == CF_DBLDBLDBL ? ((DBLDBLDBL *)(b))[i].d1val : 0))

#define get2ndTriVal(i,b,f) (\
  (double)((f) == CF_INTINTINT ? ((INTINTINT *)(b))[i].i2val :\
           (f) == CF_FLTFLTINT ? ((FLTFLTINT *)(b))[i].f2val :\
           (f) == CF_INTFLTINT ? ((INTFLTINT *)(b))[i].fval :\
           (f) == CF_FLTINTINT ? ((FLTINTINT *)(b))[i].i1val :\
           (f) == CF_DBLDBLDBL ? ((DBLDBLDBL *)(b))[i].d2val : 0))

#define get3rdTriVal(i,b,f) (\
  (double)((f) == CF_INTINTINT ? ((INTINTINT *)(b))[i].i3val :\
           (f) == CF_FLTFLTINT ? ((FLTFLTINT *)(b))[i].ival :\
           (f) == CF_INTFLTINT ? ((INTFLTINT *)(b))[i].i2val :\
           (f) == CF_FLTINTINT ? ((FLTINTINT *)(b))[i].i2val :\
           (f) == CF_DBLDBLDBL ? ((DBLDBLDBL *)(b))[i].d3val : 0))

mxArray *getMxData(int fmt,int siz,char *tag,BYTE *buf)
{
  mxArray *da, *fdata[256];
  mxChar *mxc;
  int fmtsiz = fmtsizeof(fmt);
  int i, k, n, nsiz, nf;
  mwSize dims[2];
  const char *nifield[2], *sfield[256];
  double *lp;
  struct structStruct *ss;
  struct structFormat *sf;

  switch(fmt)
  {	
    case CF_TEXT:
      da = mxCreateString((const char *)buf);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      dims[0] = siz;
      dims[1] = fmtsiz;
      if (siz > 1)
      {
        da = mxCreateCharArray(2,dims);
        mxc = (mxChar *)mxGetData(da);
        for (i=0; i<siz; i++)
        { /* string arrays transposed in MatLab ? */
          for (k=0; k<fmtsiz; k++) mxc[siz*k+i] = (mxChar)buf[fmtsiz*i+k];
        }
      }
      else
      {
        da = mxCreateString((const char *)buf);
      }
      break;
    case CF_STRING:
      if (siz > 0)
      {
        da = mxCreateCharMatrixFromStrings(siz,(const char **)buf);
      }
      break;
    case CF_NAME8I:
    case CF_NAME16I:
    case CF_NAME32I:
    case CF_NAME48I:
    case CF_NAME64I:
      nifield[0] = (char *)mxCalloc(5, sizeof(char));
      nifield[1] = (char *)mxCalloc(6, sizeof(char));
      strcpy((char *)nifield[0],"name");
      strcpy((char *)nifield[1],"value");
      nsiz = fmtsiz - 4;
      dims[0] = 1;
      dims[1] = nsiz;
      if (siz > 0)
      {
        da = mxCreateStructMatrix(siz,1,2,&nifield[0]);
        for (i=0; i<siz; i++)
        {
          fdata[0] = mxCreateCharArray(2,dims);
          fdata[1] = mxCreateDoubleMatrix(1,1,mxREAL);
          mxc = (mxChar *)mxGetData(fdata[0]);
          lp = (double *)mxGetPr(fdata[1]);
          for (k=0; k<nsiz; k++) mxc[k] = (mxChar)buf[fmtsiz*i+k];
          *lp = double(*(int *)&buf[fmtsiz*i+nsiz]);
          mxSetFieldByNumber(da,i,0,fdata[0]);
          mxSetFieldByNumber(da,i,1,fdata[1]);
        }
      }
      break;
    case CF_STRUCT:
      if ((ss=GetStructFromTag(tag)) != NULL && siz > 0)
      {
        fmtsiz = GetStructSize(tag);
        for (k=0, sf=ss->s; sf != NULL && k < 256; k++, sf=sf->nxt) if (sf->fmt == CF_NULL) k--;
        nf = k; // number of fields (n.b. legacy registrations might have a 'CF_NULL' in termination)
        for (n=nf-1, sf=ss->s; sf != NULL && n >= 0; n--, sf=sf->nxt)
        {
          if (sf->fmt == CF_NULL) { n++; continue; }
          nsiz = MIN(strlen(sf->field),16);
          sfield[n] = (char *)mxCalloc(nsiz+1,sizeof(char));
          strncpy((char *)sfield[n],sf->field,nsiz);
        }
        da = mxCreateStructMatrix(siz,1,nf,&sfield[0]);
        for (i=0; i<siz; i++)
        {
          for (n=nf-1, sf=ss->s; sf != NULL && n >=0; n--, sf=sf->nxt)
          { 
            nsiz = sf->off;
            switch (sf->fmt)
            {
              case CF_INT16:
              case CF_INT32:
              case CF_FLOAT:
              case CF_DOUBLE:
                fdata[n] = mxCreateDoubleMatrix(sf->siz,1,mxREAL);
                lp = (double *)mxGetPr(fdata[n]);
                for (k=0; k<sf->siz; k++) *lp++ = getDblVal(k,&buf[fmtsiz*i+nsiz],sf->fmt);
                mxSetFieldByNumber(da,i,n,fdata[n]);
                break;
              case CF_TEXT:
                dims[0] = 1;
                dims[1] = sf->siz+1;
                fdata[n] = mxCreateCharArray(2,dims);
                mxc = (mxChar *)mxGetData(fdata[n]);
                for (k=0; k<dims[1]; k++) mxc[k] = (mxChar)buf[fmtsiz*i+nsiz+k];
                mxSetFieldByNumber(da,i,n,fdata[n]);
                break;
              case CF_NULL:
                n++; 
                continue;
            }
          }
        }
      }
      break;
    case CF_USTRING:
      buf[96] = '\0';
      sprintf((char *)buf,"%d %g %g %u %s",*(int *)&buf[0],*(float *)&buf[4],*(float *)&buf[8],*(UINT32 *)&buf[12],(char *)&buf[16]);
      da = mxCreateString((const char *)buf);
      break;					
    case CF_BYTE:
    case CF_INT16:
    case CF_INT32:
    case CF_FLOAT:
    case CF_DOUBLE:
      da = mxCreateDoubleMatrix(siz,1,mxREAL);
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) *lp++ = getDblVal(i,buf,fmt);
      break;
    case CF_BOOLEAN:
      da = mxCreateDoubleMatrix(siz,1,mxREAL);
      lp = (double *)mxGetPr(da);
      for(n = 0; n < siz; n++) *lp++ = (double)(((int *)buf)[i] ? 1 : 0);
      break;
    case CF_INTINT:
    case CF_FLTINT:
    case CF_FLTFLT:
    case CF_DBLDBL:
      da = mxCreateDoubleMatrix(siz,2,mxREAL);
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) 
      {
        *lp       = get1stDblVal(i,buf,fmt);
        *(lp+siz) = get2ndDblVal(i,buf,fmt);
        lp++;
      }
      break;
    case CF_IFFF:
    case CF_FIFI:
    case CF_IIII:
    case CF_IIFF:
    case CF_TTII:
      da = mxCreateDoubleMatrix(siz,4,mxREAL),
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) 
      {
        *lp         = get1stQuadVal(i,buf,fmt);
        *(lp + siz) = get2ndQuadVal(i,buf,fmt);
        *(lp+2*siz) = get3rdQuadVal(i,buf,fmt);
        *(lp+3*siz) = get4thQuadVal(i,buf,fmt);
        lp++;
      }
      break;
    case CF_SPECTRUM:
      da = mxCreateDoubleMatrix(siz-96,1,mxREAL),
      lp = (double *)mxGetPr(da);
      for(n = 0; n < siz-96; n++) *lp++ = double(*(float *)&buf[n]);
      break;
    case CF_INTINTINT:
    case CF_DBLDBLDBL:
    case CF_FLTFLTINT:
    case CF_INTFLTINT:
    case CF_FLTINTINT:
      da = mxCreateDoubleMatrix(siz,3,mxREAL),
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) 
      {
        *lp         = get1stTriVal(i,buf,fmt);
        *(lp + siz) = get2ndTriVal(i,buf,fmt);
        *(lp+2*siz) = get3rdTriVal(i,buf,fmt);
        lp++;
      }
      break;
    default:
      sprintf((char *)buf,"data type %d not handled\n",fmt);
      da  = mxCreateString((const char *)buf);
      break;
  }
  return da;
}

#define setDblVal(i,n,b,f,v) {\
  for (i=0; i<(n); i++) {\
    if ((f) == CF_BYTE) (b)[i] = (BYTE)((v)[i]);\
    else if ((f) == CF_INT16) ((short *)(b))[i] = (short)((v)[i]);\
    else if ((f) == CF_INT32) ((int *)(b))[i] = (int)((v)[i]);\
    else if ((f) == CF_FLOAT) ((float *)(b))[i] = (float)((v)[i]);\
    else ((double *)(b))[i] = ((v)[i]);\
  }\
}

#define setDbltVal(n,b,f,v) {\
	for (i=0; i<(n); i++) {\
    if ((f) == CF_INTINT) {\
      ((INTINT *)(b))[i].i1val = (int)((v)[i*2]);\
      ((INTINT *)(b))[i].i2val = (int)((v)[i*2+1]);\
    }\
    else if ((f) == CF_FLTINT) {\
      ((FLTINT *)(b))[i].fval = (float)((v)[i*2]);\
      ((FLTINT *)(b))[i].ival = (int)((v)[i*2+1]);\
    }\
    else if ((f) == CF_FLTFLT) {\
      ((FLTFLT *)(b))[i].f1val = (float)((v)[i*2]);\
      ((FLTFLT *)(b))[i].f2val = (float)((v)[i*2+1]);\
    }\
    else if ((f) == CF_DBLDBL) {\
      ((DBLDBL *)(b))[i].d1val = ((v)[i*2]);\
      ((DBLDBL *)(b))[i].d2val = ((v)[i*2+1]);\
    }\
  }\
}

#define setTpltVal(n,b,f,v) {\
	for (i=0; i<(n); i++) {\
    if ((f) == CF_INTINTINT) {\
      ((INTINTINT *)(b))[i].i1val = (int)((v)[i*3]);\
      ((INTINTINT *)(b))[i].i2val = (int)((v)[i*3+1]);\
      ((INTINTINT *)(b))[i].i3val = (int)((v)[i*3+2]);\
    }\
    else if ((f) == CF_FLTINTINT) {\
      ((FLTINTINT *)(b))[i].fval = (float)((v)[i*3]);\
      ((FLTINTINT *)(b))[i].i1val = (int)((v)[i*3+1]);\
      ((FLTINTINT *)(b))[i].i2val = (int)((v)[i*3+2]);\
    }\
    else if ((f) == CF_INTFLTINT) {\
      ((INTFLTINT *)(b))[i].i1val = (int)((v)[i*3]);\
      ((INTFLTINT *)(b))[i].fval = (float)((v)[i*3+1]);\
      ((INTFLTINT *)(b))[i].i2val = (int)((v)[i*3+2]);\
    }\
    else if ((f) == CF_FLTFLTINT) {\
      ((FLTFLTINT *)(b))[i].f1val = (float)((v)[i*3]);\
      ((FLTFLTINT *)(b))[i].f2val = (float)((v)[i*3+1]);\
      ((FLTFLTINT *)(b))[i].ival = (int)((v)[i*3+2]);\
    }\
    else if ((f) == CF_DBLDBLDBL) {\
      ((DBLDBLDBL *)(b))[i].d1val = ((v)[i*3]);\
      ((DBLDBLDBL *)(b))[i].d2val = ((v)[i*3+1]);\
      ((DBLDBLDBL *)(b))[i].d3val = ((v)[i*3+2]);\
    }\
  }\
}

#define setQuadVal(n,b,f,v) {\
	for (i=0; i<(n); i++) {\
    if ((f) == CF_INTINTINTINT) {\
      ((DADDRESS *)(b))[i].i1val = (int)((v)[i*4]);\
      ((DADDRESS *)(b))[i].i2val = (int)((v)[i*4+1]);\
      ((DADDRESS *)(b))[i].i3val = (int)((v)[i*4+2]);\
      ((DADDRESS *)(b))[i].i4val = (int)((v)[i*4+3]);\
    }\
    else if ((f) == CF_INTFLTFLTFLT) {\
      ((DFILTER *)(b))[i].ival = (int)((v)[i*4]);\
      ((DFILTER *)(b))[i].f1val = (float)((v)[i*4+1]);\
      ((DFILTER *)(b))[i].f2val = (float)((v)[i*4+2]);\
      ((DFILTER *)(b))[i].f3val = (float)((v)[i*4+3]);\
    }\
    else if ((f) == CF_INTINTFLTFLT) {\
      ((FWINDOW *)(b))[i].i1val = (int)((v)[i*4]);\
      ((FWINDOW *)(b))[i].i2val = (int)((v)[i*4+1]);\
      ((FWINDOW *)(b))[i].f1val = (float)((v)[i*4+2]);\
      ((FWINDOW *)(b))[i].f2val = (float)((v)[i*4+3]);\
    }\
    else if ((f) == CF_TTII) {\
      ((DWINDOW *)(b))[i].t1 = (int)((v)[i*4]);\
      ((DWINDOW *)(b))[i].t2 = (int)((v)[i*4+1]);\
      ((DWINDOW *)(b))[i].i1val = (int)((v)[i*4+2]);\
      ((DWINDOW *)(b))[i].i2val = (int)((v)[i*4+3]);\
    }\
    else if ((f) == CF_FLTINTFLTINT) {\
      ((FLTINTFLTINT *)(b))[i].f1val = (float)((v)[i*4]);\
      ((FLTINTFLTINT *)(b))[i].i1val = (int)((v)[i*4+1]);\
      ((FLTINTFLTINT *)(b))[i].f2val = (float)((v)[i*4+2]);\
      ((FLTINTFLTINT *)(b))[i].i2val = (int)((v)[i*4+3]);\
    }\
  }\
}
int setMxData(const mxArray *mxa,int fmt,int siz,char *tag,BYTE *buf)
{
  mxChar *mxc;
  mxArray *sa;
  double *lp;
  int fmtsiz = GetFormatSize(fmt);
  int i, k, n, nf, len;
  char c, *cptr;
  const char *fld;
  struct structStruct *ss;
  struct structFormat *sf;

  switch (fmt)
  {
    case CF_NULL:
    	break;    
    case CF_TEXT:
      if (!mxIsChar(mxa)) return argument_list_error;
      len = mxGetN(mxa);
      siz = len;
      mxGetString(mxa,(char *)buf,len+1);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      if (!mxIsChar(mxa)) return argument_list_error;
      if (siz > 1)
      {
        const mwSize *dims = mxGetDimensions(mxa);
        for (i=0; i<siz; i++)
        {
          mxc = mxGetChars(mxa);
          for (i=0; i<siz; i++)
          { /* string arrays transposed in MatLab ? */
            memset(&buf[fmtsiz*i],0,fmtsiz);
            for (k=0; k<fmtsiz && k<dims[1]; k++) 
            {
              c = (char)mxc[siz*k+i];
              ((char *)buf)[fmtsiz*i+k] = c;
              if (c == 0) break;
            }
          }
        }
      }
      else
      {
			  len = mxGetN(mxa) + 1;
        if (len > fmtsiz) len = fmtsiz;
			  mxGetString(mxa,(char *)buf,len);
      }
      break;
    case CF_STRING:
      if (mxIsChar(mxa)) 
      {
        len = 0;
        const mwSize *dims = mxGetDimensions(mxa);
        for (i=0; i<siz; i++)
        {
          mxc = (mxChar *)mxGetData(mxa);
          for (i=0; i<siz; i++)
          { /* string arrays transposed in MatLab ? */
            cptr = (char *)&buf[sizeof(size_t)*siz + len];
            ((size_t *)buf)[i] = (size_t)cptr;
            for (k=0; k<fmtsiz && k<dims[1]; k++) 
            {
              c = (char)mxc[siz*k+i];
              cptr[len++] = c;
              if (c == 0) break;
            }
          }
        }
      }
      else
      {
        return argument_list_error;
      }
      break;
    case CF_STRUCT:
			if(!mxIsStruct(mxa)) return argument_list_error;
      if ((ss=GetStructFromTag(tag)) != NULL && siz > 0)
      {
        fmtsiz = GetStructSize(tag);
        for (k=0, sf=ss->s; sf != NULL && k < 256; k++, sf=sf->nxt) if (sf->fmt == CF_NULL) k--;
        nf = k; // number of fields (n.b. legacy registrations might have a 'CF_NULL' in termination)s
        if (nf != mxGetNumberOfFields(mxa)) return invalid_structure_size;
        for (i=0; i<siz; i++)
        {
          for (n=nf-1, sf=ss->s; sf != NULL && n >=0; n--, sf=sf->nxt)
          { 
            fld = mxGetFieldNameByNumber(mxa,n);
            sa = mxGetField(mxa,i,fld);
            switch (sf->fmt)
            {
              case CF_INT16:
              case CF_INT32:
              case CF_FLOAT:
              case CF_DOUBLE:
           			if(!mxIsDouble(sa)) return invalid_structure_tag;
			          lp = (double *)mxGetPr(sa);
                setDblVal(k,sf->siz,&buf[fmtsiz*i+sf->off],sf->fmt,lp);
                break;
              case CF_TEXT:
           			if(!mxIsChar(sa)) return invalid_structure_tag;
                len = mxGetN(sa);
                mxGetString(sa,(char *)&buf[fmtsiz*i+sf->off],len+1);
                break;
              case CF_NULL:
                n++; 
                continue;
            }
          }
        }
      }
      break;
		case CF_BYTE:
		case CF_SHORT:
		case CF_INT32:
		case CF_FLOAT:
		case CF_DOUBLE:
			if(!mxIsDouble(mxa)) return argument_list_error;
			lp = (double *)mxGetPr(mxa);
      setDblVal(i,siz,buf,fmt,lp)
		  break;
		case CF_BOOLEAN:
			if(!mxIsDouble(mxa)) return argument_list_error;
			lp = (double *)mxGetPr(mxa);
      for (i=0; i<siz; i++) ((int *)buf)[i] = ((int)lp[i] == 0 ? 0 : 1);
		  break;		
		case CF_INTINT:
		case CF_FLTINT:
		case CF_FLTFLT:
		case CF_DBLDBL:
			if(!mxIsDouble(mxa)) return argument_list_error;
			lp = (double *)mxGetPr(mxa);
      siz /= 2;
      setDbltVal(siz,buf,fmt,lp);
		  break;
    case CF_INTINTINT:
    case CF_FLTINTINT:
    case CF_INTFLTINT:
    case CF_FLTFLTINT:
    case CF_DBLDBLDBL:
			if(!mxIsDouble(mxa)) return argument_list_error;
			lp = (double *)mxGetPr(mxa);
      siz /= 3;
      setTpltVal(siz,buf,fmt,lp);
      break;
		case CF_IFFF:
		case CF_IIFF:
		case CF_IIII:
		case CF_TTII:
		case CF_FIFI:
			if(!mxIsDouble(mxa)) return argument_list_error;
			lp = (double *)mxGetPr(mxa);
      siz /= 4;
      setQuadVal(siz,buf,fmt,lp);
      break;		
/*
		case CF_SPECTRUM:
		need 3 call
		break;
    case CF_USTRING:
      if(!mxIsChar(mxa))
      {
				plhs[0] = mxCreateString("Input argument error !!!");
				return;
			}
			len = mxGetN(mxa) + 1;
			mxGetString(mxa,buf,len);
      siz = 1; // kluge this by forcing only 1 of these to be sent
  		break;
*/
    default:
      if (!mxIsChar(mxa)) return illegal_format;
	}
  return 0;
}

char Buffer[MAX_TINE_BUFFER];
#define usageStr "tine_writeread usage:\n ret=tine_writeread(values,'/<context>/<server>/<device>[<property>]'[,'<data_type>',data_size,timeout]);\n where values = string or double Nx1 array (or omitted to send a COMMAND without input)\n'ret' contains the status/error message,timestamps,and returned values"
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[]) 
{
  mxArray	*data[PROPMAX];
  const char *field[PROPMAX];
  double *lp, dt = (double)time(NULL);
  int	error[PROPMAX];
  char *name, *mxerrStr=NULL;
  char *p;
  char buf[1024];
  char *c=NULL, *c1=NULL, *c2=NULL, *c3=NULL, *cf=NULL;
  char dtstr[64], dtag[32], dtagw[32], *typestr;
  char prp[80];
  char dev[132];
  int	plen,len,writetype=CF_NULL,readtype=-1,size,fmtsiz,writesize = 0,readsize = -1,ln,tmr=1000,datatype = -1;
  mwSize dims[2];
  int	i,j,k,err,count,n,tgt_idx=1,isCmd=0,startindex = 3;
  int	lim,nmb;
  mxChar *CharData;
  char tsstr[64];
  
  if (plhs == NULL || prhs == NULL) return;

  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 5: // timer interval (int)
      if (!mxIsDouble(prhs[4])) mxerr(usageStr);
      if (mxGetM(prhs[4]) != 1) mxerr(usageStr);
      tmr = (int)mxGetScalar(prhs[4]);
    case 4: // timer interval (int)
      if (!mxIsDouble(prhs[3])) mxerr(usageStr);
      if (mxGetM(prhs[3]) != 1) mxerr(usageStr);
      readsize = (int)mxGetScalar(prhs[3]);
    case 3: // input/output data type
      if (!mxIsChar(prhs[2])) mxerr(usageStr);
      if (mxGetM(prhs[2]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[2]) + 1;
      typestr = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[2],typestr,len);
      if ((p=strchr(typestr,':')) != NULL)
      { // both write and read types given explicitly as '<write_type>:<read_type>'
        *p++ = 0;
        strncpy(dtstr,typestr,32);
        if ((c=strchr(dtstr,'/')) != NULL || 
            (c=strchr(dtstr,'\\')) != NULL ||
            (c=strchr(dtstr,'.')) != NULL)
        {
          *c++ = 0;
          strncpy(dtagw,c,TAG_NAME_SIZE);
        }
        strupr(dtstr);
        if (dtstr[strlen(dtstr)-1] == 'S') dtstr[strlen(dtstr)-1] = 0; // in case "floats" instead of "float"
        writetype = ftoi(dtstr);
        strncpy(dtstr,p,32);
      }
      else
      {
        strncpy(dtstr,typestr,32);
      }
      if ((c=strchr(dtstr,'/')) != NULL || 
          (c=strchr(dtstr,'\\')) != NULL ||
          (c=strchr(dtstr,'.')) != NULL)
      {
        *c++ = 0;
        strncpy(dtag,c,TAG_NAME_SIZE);
      }
      strupr(dtstr);
      if (dtstr[strlen(dtstr)-1] == 'S') dtstr[strlen(dtstr)-1] = 0; // in case "floats" instead of "float"
      datatype = ftoi(dtstr);
      mxFree(typestr);
    case 2:
      if (datatype < 0)
      { // no type specified, but there are input data showing ...
        if (mxIsChar(prhs[0])) writetype = CF_TEXT; else writetype = CF_INT32;
      }
      writesize = mxGetN(prhs[0]);
    case 1:
      if (nrhs == 1)
      {
        if (!mxIsChar(prhs[0])) mxerr(usageStr);
        tgt_idx = 0;
        isCmd = -1;
      }
      break;
    default:
      mxerr(usageStr);
  }  
  // the requested data endpoint ...
  len = mxGetN(prhs[tgt_idx]) + 1;
  name = (char *)mxCalloc(len, sizeof(char));
  mxGetString(prhs[tgt_idx],name,len);
  strncpy(buf,name,1024);
  mxFree(name);
  if (buf[0] == '?')
  {
    strcpy(buf,usageStr);
    plhs[0] = mxCreateString(buf);
    return;
  }
  if((p=strchr(buf,'[')) == NULL || strchr(buf,']') == NULL) 
  {
    mxerr(usageStr);
  }
  *p++ = 0;
  strncpy(dev,buf,132);
  while (*p == ' ') p++;
  strncpy(prp,p,64);
  if ((p=strchr(prp,']')) != NULL) *p = 0;

  field[0] = (char *)mxCalloc(6, sizeof(char));
  field[1] = (char *)mxCalloc(10, sizeof(char));
  field[2] = (char *)mxCalloc(4, sizeof(char));
  memset((void *)field[0],0,sizeof(field[0]));
  memset((void *)field[1],0,sizeof(field[1]));
  memset((void *)field[2],0,sizeof(field[2]));

  strcpy((char *)field[0],"error");
  strcpy((char *)field[1],"timestamp");
  strcpy((char *)field[2],"utc");
	
  if (!isCmd && (readsize < 0 || datatype < 0))
  { // write type = read type for attributes ... 
    readtype = read_prop(dev, prp, &readsize,dtag);
    if (readtype < 0) readtype = CF_INT32; // try this ...
  }
  else
  {
    readtype = datatype;
  }
  if (writesize > 0)
  {
    if (writetype == CF_NULL) 
    {
      writetype = readtype;
      strncpy(dtagw,dtag,TAG_NAME_SIZE);
    }
    if (writetype == CF_STRUCT && strlen(dtagw) > 0)
    {
      err = AcquireAndRegisterStruct(dev,dtagw,writesize);
    }
    err = setMxData(prhs[0],writetype,writesize,dtagw,(BYTE *)Buffer);
    if (err != 0)
    {
      mxerr("could not set input data !");
    }
  }
  fmtsiz = GetFormatSize(readtype);
  if (readsize * fmtsiz > MAX_TINE_BUFFER) readsize = MAX_TINE_BUFFER / fmtsiz;
  if (readtype == CF_STRUCT && strlen(dtag) > 0)
  {
    err = AcquireAndRegisterStruct(dev,dtag,readsize);
  }

  size = mxGetM(prhs[0]); // set the size to what was input (num rows) !
  plen = strlen(prp);  // length of property string
  if (plen == 0) return;
  field[3] = (char *)mxCalloc(plen+1, sizeof(char));
  strncpy((char *)field[3],name_corr(prp),plen+1);

  err = write_read_data(dev, prp, writetype, writesize, readtype, readsize, Buffer, tmr, &dt, dtag, dtagw);

  if (err) 
  {
    data[3]  = mxCreateString(Buffer);
    data[0]  = mxCreateString(Buffer);
  } 
  else 
  {
    data[3] = getMxData(readtype,readsize,dtag,(BYTE *)Buffer);
  }

  strcpy((char *)field[0],"error");
  plhs[0] = mxCreateStructMatrix(1,1,4,&field[0]);
  strcpy((char *)field[1],"timestamp");
  strcpy((char *)field[2],"utc");
  data[0] = mxCreateString("");
  strncpy(tsstr,getDataTimeString(dt,0),64);
  sprintf(dtstr,"%10.3f",dt);
  data[1] = mxCreateString(tsstr);
  data[2] = mxCreateString(dtstr);

  mxSetFieldByNumber(plhs[0],0,0,data[0]);
  mxSetFieldByNumber(plhs[0],0,1,data[1]);
  mxSetFieldByNumber(plhs[0],0,2,data[2]);
  mxSetFieldByNumber(plhs[0],0,3,data[3]);
err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
}
