#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"
#if defined(UNIX)
#include "threader.h"
#endif

#define	PROPMAX 1024
#define	DEVMAX 4096
#define	MAX_TINE_BUFFER 512000 //256000

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

#define usageStr "tine_attach_server usage:\n ret=tine_attach_server('<equipment module (local) name>'[,'<server export name>',<device list capacity>])"
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[]) 
{
  mxArray	*data[PROPMAX];
  const char *field[PROPMAX];
  char *srv=NULL, *eqm=NULL,*mxerrStr=NULL;
  char buf[128],stsbuf[64];
  double *lp;
  int	i,k,len,err,n;
  int ndevs=0;
    
  if (plhs == NULL || prhs == NULL) return;
  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 3: // device list capacity
      if (!mxIsDouble(prhs[2])) mxerr(usageStr);
      if (mxGetM(prhs[2]) != 1) mxerr(usageStr);
      ndevs = (int)mxGetScalar(prhs[2]);
    case 2: // export name given
      if (!mxIsChar(prhs[1])) mxerr(usageStr);
      if (mxGetM(prhs[1]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[1]) + 1;
      srv = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[1],srv,len);
      if (nrhs == 2) break;
    case 1: // local name
      if (!mxIsChar(prhs[0])) mxerr(usageStr);
      if (mxGetM(prhs[0]) != 1) mxerr(usageStr);
      // the requested data endpoint ...
      len = mxGetN(prhs[0]) + 1;
      eqm = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[0],eqm,len);
      break;
    default:
      mxerr(usageStr);
  }   

  err = AttachServer(srv,eqm,ndevs);

  if (err != 0)
  {
    GetLastLinkError(err,stsbuf);
    sprintf(buf,"tine_attach_server : <%d> %.64s",err,stsbuf);
    mexErrMsgTxt(buf);
  }
# if defined(UNIX)
  else
  {
    SystemSetCycleTimer();
  }
# endif
  data[0] = mxCreateDoubleMatrix(1,1,mxREAL);
  lp = (double *)mxGetPr(data[0]);
  *lp = (double)err;
  plhs[0] = data[0];
  
err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
  if (srv != NULL) mxFree(srv);
  if (eqm != NULL) mxFree(eqm);
}
