#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"

#define	PROPMAX 1024
#define	DEVMAX 4096
#define	MAX_TINE_BUFFER 512000 //256000

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

short ftoi(char *fmt)
{
  char *c,fmtstr[64];
  short ifmt;
  strncpy(fmtstr,fmt,64);
  if ((c=strchr(fmtstr,'.')) != NULL) *c = 0;
  strupr(fmtstr);
  if ((ifmt=getSimpleFormatType(fmtstr)) != CF_NULL) return ifmt;
  return getCompoundFormatType(fmtstr);
}
int isStrFormat(short fmt)
{
  switch (fmt)
  {
    case CF_TEXT:
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
#ifdef WIN32
int mexlog(char *tag,char *text,...)
{
  char str[512];
  FILE *fp = NULL;
  va_list args;
  time_t t = time(NULL);
  static char *fmod = "a+";

  sprintf(str,"%.24s[%s] ",ctime(&t),tag);

  va_start(args,text);
  _vsnprintf(&str[strlen(str)],256,text,args); str[256] = 0;
  va_end(args);

  strcat(str,"\n");
  if ((fp=fopen("l:\\log\\mex.log","a+")) == NULL) return -1;
  fwrite(str,strlen(str),1,fp);

  fclose(fp);
  return 0;
}
#endif
int mexIsStockProperty(char *prop)
{
  if (!stricmp(prop,"DEVICES")) return TRUE;
  if (!stricmp(prop,"NDEVICES")) return TRUE;
  if (!stricmp(prop,"NPROPS")) return TRUE;
  if (!stricmp(prop,"NPROPERTIES")) return TRUE;
  if (!stricmp(prop,"PROPERTIES")) return TRUE;
  return FALSE;
}
typedef struct
{
  char dev[192];
  char prp[64];
  int  fmt;
  int  siz;
} PrpInfoCache;
#define pInfoTblCapacity 100 // just keep a short list
PrpInfoCache pInfoTbl[pInfoTblCapacity];
int pInfoTblSize = 0;
int getPrpInfoTblIndex(char *dev,char *prp)
{
  int i;
  for (i=0; i<pInfoTblSize; i++)
  {
    if (strnicmp(dev,pInfoTbl[i].dev,192)) continue;
    if (strnicmp(prp,pInfoTbl[i].prp,64)) continue;
    return i;
  }
  return -1;
}
int addPrpInfo(char *dev,char *prp,int fmt,int siz)
{
  if (pInfoTblSize < pInfoTblCapacity)
  { // this will stop adding when we're at capacity ....
    strncpy(pInfoTbl[pInfoTblSize].dev,dev,192);
    strncpy(pInfoTbl[pInfoTblSize].prp,prp,64);
    pInfoTbl[pInfoTblSize].fmt = fmt;
    pInfoTbl[pInfoTblSize].siz = siz;
    pInfoTblSize++;
  }
  return 0;
}
int read_prop(char *dev, char *prop, int *size, char *tag)
{ 
  PropertyQueryStruct	svrProp; // legacy
  PrpQueryStruct pqs;
  DTYPE dout, din;
  int cc, fmt, isLegacy=FALSE;
  char *prp = "PROPS";

  if ((cc=getPropertyInfoFromListenerEx(dev,prop,size,&fmt,tag)) == 0)
  {
    return fmt;
  }
  if (mexIsStockProperty(prop)) prp = "STOCKPROPS";

  memset(&din,0,sizeof(DTYPE));
  din.dFormat = CF_TEXT;
  din.dArrayLength = (int)strlen(prop);
  din.data.vptr = prop;

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  dout.data.vptr = &pqs;
  strcpy(dout.dTag,"PRPQSr4");

legacyRetry:
  if ((cc=alsnExecLink(dev,prp,&dout,&din,CA_READ,500)) != 0)
  {
    if (!isLegacy && (cc == illegal_format || cc == invalid_structure_tag))
    {
      isLegacy = TRUE;
      dout.dFormat = CF_STRUCT;
      dout.dArrayLength = PROPERTYQUERYSTRUCT_SIZE;
      dout.data.vptr = &svrProp;
      dout.dTag[0] = '\0';
      goto legacyRetry;
    }
    *size = 0;
    return -cc;
 	}

  if (isLegacy)
  {
    *size = svrProp.prpSize;
    fmt = LFMT(svrProp.prpFormat);
  }
  else
  {
    *size = pqs.prpSize;
    fmt = LFMT(pqs.prpFormat);
    if (tag != NULL) strncpy(tag,pqs.prpTag,TAG_NAME_SIZE);
  }
  return fmt;
}
int mex_tine_call(char *dev, char *prop, DTYPE *dout, DTYPE *din,int access,int interval,double *pDataTime)
{
	int cc;

  if (dout == NULL || din == NULL) return argument_list_error;
  if (dout->dFormat < 0 || din->dFormat < 0) return illegal_format;
  if ((cc=alsnExecLinkEx(dev,prop,dout,din,access,interval,pDataTime)) != 0)
  {
    GetLastLinkError(cc,dout->data.cptr);
    if (pDataTime != NULL) *pDataTime = (double)time(NULL);
		return cc;
	}
	return 0;

}
#define getDblVal(i,b,f) (\
  (double)((f) == CF_BYTE ? (b)[i] :\
           (f) == CF_INT16 ? ((short *)(b))[i] :\
           (f) == CF_INT32 ? ((int *)(b))[i] :\
           (f) == CF_FLOAT ? ((float *)(b))[i] :\
           (f) == CF_DOUBLE ? ((double *)(b))[i] : 0))

#define get1stDblVal(i,b,f) (\
  (double)((f) == CF_INTINT ? ((INTINT *)(b))[i].i1val :\
           (f) == CF_FLTINT ? ((FLTINT *)(b))[i].fval :\
           (f) == CF_DBLDBL ? ((DBLDBL *)(b))[i].d1val :\
           (f) == CF_FLTFLT ? ((FLTFLT *)(b))[i].f1val : 0))

#define get2ndDblVal(i,b,f) (\
  (double)((f) == CF_INTINT ? ((INTINT *)(b))[i].i2val :\
           (f) == CF_FLTINT ? ((FLTINT *)(b))[i].ival :\
           (f) == CF_DBLDBL ? ((DBLDBL *)(b))[i].d2val :\
           (f) == CF_FLTFLT ? ((FLTFLT *)(b))[i].f2val : 0))

#define get1stQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i1val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].ival :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].i1val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].t1 :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].f1val : 0))

#define get2ndQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i2val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].f1val :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].i2val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].t2 :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].i1val : 0))

#define get3rdQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i3val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].f2val :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].f1val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].i1val :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].f2val : 0))

#define get4thQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i4val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].f3val :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].f2val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].i2val :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].i2val : 0))

#define get1stTriVal(i,b,f) (\
  (double)((f) == CF_INTINTINT ? ((INTINTINT *)(b))[i].i1val :\
           (f) == CF_FLTFLTINT ? ((FLTFLTINT *)(b))[i].f1val :\
           (f) == CF_INTFLTINT ? ((INTFLTINT *)(b))[i].i1val :\
           (f) == CF_FLTINTINT ? ((FLTINTINT *)(b))[i].fval :\
           (f) == CF_DBLDBLDBL ? ((DBLDBLDBL *)(b))[i].d1val : 0))

#define get2ndTriVal(i,b,f) (\
  (double)((f) == CF_INTINTINT ? ((INTINTINT *)(b))[i].i2val :\
           (f) == CF_FLTFLTINT ? ((FLTFLTINT *)(b))[i].f2val :\
           (f) == CF_INTFLTINT ? ((INTFLTINT *)(b))[i].fval :\
           (f) == CF_FLTINTINT ? ((FLTINTINT *)(b))[i].i1val :\
           (f) == CF_DBLDBLDBL ? ((DBLDBLDBL *)(b))[i].d2val : 0))

#define get3rdTriVal(i,b,f) (\
  (double)((f) == CF_INTINTINT ? ((INTINTINT *)(b))[i].i3val :\
           (f) == CF_FLTFLTINT ? ((FLTFLTINT *)(b))[i].ival :\
           (f) == CF_INTFLTINT ? ((INTFLTINT *)(b))[i].i2val :\
           (f) == CF_FLTINTINT ? ((FLTINTINT *)(b))[i].i2val :\
           (f) == CF_DBLDBLDBL ? ((DBLDBLDBL *)(b))[i].d3val : 0))

mxArray *getMxData(int fmt,int siz,char *tag,BYTE *buf)
{
  mxArray *da, *fdata[256];
  mxChar *mxc;
  int fmtsiz = fmtsizeof(fmt);
  int i, k, n, nsiz, nf;
  mwSize dims[2];
  const char *nifield[2], *sfield[256];
  double *lp;
  struct structStruct *ss;
  struct structFormat *sf;

  switch(fmt)
  {	
    case CF_TEXT:
      da = mxCreateString((const char *)buf);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      dims[0] = siz;
      dims[1] = fmtsiz;
      if (siz > 1)
      {
        da = mxCreateCharArray(2,dims);
        mxc = (mxChar *)mxGetData(da);
        for (i=0; i<siz; i++)
        { /* string arrays transposed in MatLab ? */
          for (k=0; k<fmtsiz; k++) mxc[siz*k+i] = (mxChar)buf[fmtsiz*i+k];
        }
      }
      else
      {
        da = mxCreateString((const char *)buf);
      }
      break;
    case CF_STRING:
      if (siz > 0)
      {
        da = mxCreateCharMatrixFromStrings(siz,(const char **)buf);
      }
      break;
    case CF_NAME8I:
    case CF_NAME16I:
    case CF_NAME32I:
    case CF_NAME48I:
    case CF_NAME64I:
      nifield[0] = (char *)mxCalloc(5, sizeof(char));
      nifield[1] = (char *)mxCalloc(6, sizeof(char));
      strcpy((char *)nifield[0],"name");
      strcpy((char *)nifield[1],"value");
      nsiz = fmtsiz - 4;
      dims[0] = 1;
      dims[1] = nsiz;
      if (siz > 0)
      {
        da = mxCreateStructMatrix(siz,1,2,&nifield[0]);
        for (i=0; i<siz; i++)
        {
          fdata[0] = mxCreateCharArray(2,dims);
          fdata[1] = mxCreateDoubleMatrix(1,1,mxREAL);
          mxc = (mxChar *)mxGetData(fdata[0]);
          lp = (double *)mxGetPr(fdata[1]);
          for (k=0; k<nsiz; k++) mxc[k] = (mxChar)buf[fmtsiz*i+k];
          *lp = double(*(int *)&buf[fmtsiz*i+nsiz]);
          mxSetFieldByNumber(da,i,0,fdata[0]);
          mxSetFieldByNumber(da,i,1,fdata[1]);
        }
      }
      break;
    case CF_STRUCT:
      if ((ss=GetStructFromTag(tag)) != NULL && siz > 0)
      {
        fmtsiz = GetStructSize(tag);
        for (k=0, sf=ss->s; sf != NULL && k < 256; k++, sf=sf->nxt) if (sf->fmt == CF_NULL) k--;
        nf = k; // number of fields (n.b. legacy registrations might have a 'CF_NULL' in termination)
        for (n=nf-1, sf=ss->s; sf != NULL && n >= 0; n--, sf=sf->nxt)
        {
          if (sf->fmt == CF_NULL) { n++; continue; }
          nsiz = MIN(strlen(sf->field),16);
          sfield[n] = (char *)mxCalloc(nsiz+1,sizeof(char));
          strncpy((char *)sfield[n],sf->field,nsiz);
        }
        da = mxCreateStructMatrix(siz,1,nf,&sfield[0]);
        for (i=0; i<siz; i++)
        {
          for (n=nf-1, sf=ss->s; sf != NULL && n >=0; n--, sf=sf->nxt)
          { 
            nsiz = sf->off;
            switch (sf->fmt)
            {
              case CF_INT16:
              case CF_INT32:
              case CF_FLOAT:
              case CF_DOUBLE:
                fdata[n] = mxCreateDoubleMatrix(sf->siz,1,mxREAL);
                lp = (double *)mxGetPr(fdata[n]);
                for (k=0; k<sf->siz; k++) *lp++ = getDblVal(k,&buf[fmtsiz*i+nsiz],sf->fmt);
                mxSetFieldByNumber(da,i,n,fdata[n]);
                break;
              case CF_TEXT:
                dims[0] = 1;
                dims[1] = sf->siz+1;
                fdata[n] = mxCreateCharArray(2,dims);
                mxc = (mxChar *)mxGetData(fdata[n]);
                for (k=0; k<dims[1]; k++) mxc[k] = (mxChar)buf[fmtsiz*i+nsiz+k];
                mxSetFieldByNumber(da,i,n,fdata[n]);
                break;
              case CF_NULL:
                n++; 
                continue;
            }
          }
        }
      }
      break;
    case CF_USTRING:
      buf[96] = '\0';
      sprintf((char *)buf,"%d %g %g %u %s",*(int *)&buf[0],*(float *)&buf[4],*(float *)&buf[8],*(UINT32 *)&buf[12],(char *)&buf[16]);
      da = mxCreateString((const char *)buf);
      break;					
    case CF_BYTE:
    case CF_INT16:
    case CF_INT32:
    case CF_FLOAT:
    case CF_DOUBLE:
      da = mxCreateDoubleMatrix(siz,1,mxREAL);
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) *lp++ = getDblVal(i,buf,fmt);
      break;
    case CF_BOOLEAN:
      da = mxCreateDoubleMatrix(siz,1,mxREAL);
      lp = (double *)mxGetPr(da);
      for(n = 0; n < siz; n++) *lp++ = (double)(((int *)buf)[i] ? 1 : 0);
      break;
    case CF_INTINT:
    case CF_FLTINT:
    case CF_FLTFLT:
    case CF_DBLDBL:
      da = mxCreateDoubleMatrix(siz,2,mxREAL);
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) 
      {
        *lp       = get1stDblVal(i,buf,fmt);
        *(lp+siz) = get2ndDblVal(i,buf,fmt);
        lp++;
      }
      break;
    case CF_IFFF:
    case CF_FIFI:
    case CF_IIII:
    case CF_IIFF:
    case CF_TTII:
      da = mxCreateDoubleMatrix(siz,4,mxREAL),
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) 
      {
        *lp         = get1stQuadVal(i,buf,fmt);
        *(lp + siz) = get2ndQuadVal(i,buf,fmt);
        *(lp+2*siz) = get3rdQuadVal(i,buf,fmt);
        *(lp+3*siz) = get4thQuadVal(i,buf,fmt);
        lp++;
      }
      break;
    case CF_SPECTRUM:
      da = mxCreateDoubleMatrix(siz-96,1,mxREAL),
      lp = (double *)mxGetPr(da);
      for(n = 0; n < siz-96; n++) *lp++ = double(*(float *)&buf[n]);
      break;
    case CF_INTINTINT:
    case CF_DBLDBLDBL:
    case CF_FLTFLTINT:
    case CF_INTFLTINT:
    case CF_FLTINTINT:
      da = mxCreateDoubleMatrix(siz,3,mxREAL),
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) 
      {
        *lp         = get1stTriVal(i,buf,fmt);
        *(lp + siz) = get2ndTriVal(i,buf,fmt);
        *(lp+2*siz) = get3rdTriVal(i,buf,fmt);
        lp++;
      }
      break;
    default:
      sprintf((char *)buf,"data type %d not handled\n",fmt);
      da  = mxCreateString((const char *)buf);
      break;
  }
  return da;
}

#define setDblVal(i,n,b,f,v) {\
  for (i=0; i<(n); i++) {\
    if ((f) == CF_BYTE) (b)[i] = (BYTE)((v)[i]);\
    else if ((f) == CF_INT16) ((short *)(b))[i] = (short)((v)[i]);\
    else if ((f) == CF_INT32) ((int *)(b))[i] = (int)((v)[i]);\
    else if ((f) == CF_FLOAT) ((float *)(b))[i] = (float)((v)[i]);\
    else ((double *)(b))[i] = ((v)[i]);\
  }\
}

#define setDbltVal(n,b,f,v) {\
	for (i=0; i<(n); i++) {\
    if ((f) == CF_INTINT) {\
      ((INTINT *)(b))[i].i1val = (int)((v)[i*2]);\
      ((INTINT *)(b))[i].i2val = (int)((v)[i*2+1]);\
    }\
    else if ((f) == CF_FLTINT) {\
      ((FLTINT *)(b))[i].fval = (float)((v)[i*2]);\
      ((FLTINT *)(b))[i].ival = (int)((v)[i*2+1]);\
    }\
    else if ((f) == CF_FLTFLT) {\
      ((FLTFLT *)(b))[i].f1val = (float)((v)[i*2]);\
      ((FLTFLT *)(b))[i].f2val = (float)((v)[i*2+1]);\
    }\
    else if ((f) == CF_DBLDBL) {\
      ((DBLDBL *)(b))[i].d1val = ((v)[i*2]);\
      ((DBLDBL *)(b))[i].d2val = ((v)[i*2+1]);\
    }\
  }\
}

#define setTpltVal(n,b,f,v) {\
	for (i=0; i<(n); i++) {\
    if ((f) == CF_INTINTINT) {\
      ((INTINTINT *)(b))[i].i1val = (int)((v)[i*3]);\
      ((INTINTINT *)(b))[i].i2val = (int)((v)[i*3+1]);\
      ((INTINTINT *)(b))[i].i3val = (int)((v)[i*3+2]);\
    }\
    else if ((f) == CF_FLTINTINT) {\
      ((FLTINTINT *)(b))[i].fval = (float)((v)[i*3]);\
      ((FLTINTINT *)(b))[i].i1val = (int)((v)[i*3+1]);\
      ((FLTINTINT *)(b))[i].i2val = (int)((v)[i*3+2]);\
    }\
    else if ((f) == CF_INTFLTINT) {\
      ((INTFLTINT *)(b))[i].i1val = (int)((v)[i*3]);\
      ((INTFLTINT *)(b))[i].fval = (float)((v)[i*3+1]);\
      ((INTFLTINT *)(b))[i].i2val = (int)((v)[i*3+2]);\
    }\
    else if ((f) == CF_FLTFLTINT) {\
      ((FLTFLTINT *)(b))[i].f1val = (float)((v)[i*3]);\
      ((FLTFLTINT *)(b))[i].f2val = (float)((v)[i*3+1]);\
      ((FLTFLTINT *)(b))[i].ival = (int)((v)[i*3+2]);\
    }\
    else if ((f) == CF_DBLDBLDBL) {\
      ((DBLDBLDBL *)(b))[i].d1val = ((v)[i*3]);\
      ((DBLDBLDBL *)(b))[i].d2val = ((v)[i*3+1]);\
      ((DBLDBLDBL *)(b))[i].d3val = ((v)[i*3+2]);\
    }\
  }\
}

#define setQuadVal(n,b,f,v) {\
	for (i=0; i<(n); i++) {\
    if ((f) == CF_INTINTINTINT) {\
      ((DADDRESS *)(b))[i].i1val = (int)((v)[i*4]);\
      ((DADDRESS *)(b))[i].i2val = (int)((v)[i*4+1]);\
      ((DADDRESS *)(b))[i].i3val = (int)((v)[i*4+2]);\
      ((DADDRESS *)(b))[i].i4val = (int)((v)[i*4+3]);\
    }\
    else if ((f) == CF_INTFLTFLTFLT) {\
      ((DFILTER *)(b))[i].ival = (int)((v)[i*4]);\
      ((DFILTER *)(b))[i].f1val = (float)((v)[i*4+1]);\
      ((DFILTER *)(b))[i].f2val = (float)((v)[i*4+2]);\
      ((DFILTER *)(b))[i].f3val = (float)((v)[i*4+3]);\
    }\
    else if ((f) == CF_INTINTFLTFLT) {\
      ((FWINDOW *)(b))[i].i1val = (int)((v)[i*4]);\
      ((FWINDOW *)(b))[i].i2val = (int)((v)[i*4+1]);\
      ((FWINDOW *)(b))[i].f1val = (float)((v)[i*4+2]);\
      ((FWINDOW *)(b))[i].f2val = (float)((v)[i*4+3]);\
    }\
    else if ((f) == CF_TTII) {\
      ((DWINDOW *)(b))[i].t1 = (int)((v)[i*4]);\
      ((DWINDOW *)(b))[i].t2 = (int)((v)[i*4+1]);\
      ((DWINDOW *)(b))[i].i1val = (int)((v)[i*4+2]);\
      ((DWINDOW *)(b))[i].i2val = (int)((v)[i*4+3]);\
    }\
    else if ((f) == CF_FLTINTFLTINT) {\
      ((FLTINTFLTINT *)(b))[i].f1val = (float)((v)[i*4]);\
      ((FLTINTFLTINT *)(b))[i].i1val = (int)((v)[i*4+1]);\
      ((FLTINTFLTINT *)(b))[i].f2val = (float)((v)[i*4+2]);\
      ((FLTINTFLTINT *)(b))[i].i2val = (int)((v)[i*4+3]);\
    }\
  }\
}
int setMxData(const mxArray *mxa,int fmt,int siz,char *tag,BYTE *buf)
{
  mxChar *mxc;
  mxArray *sa;
  double *lp;
  int fmtsiz = GetFormatSize(fmt);
  int i, k, n, nf, len;
  char c, *cptr;
  const char *fld;
  struct structStruct *ss;
  struct structFormat *sf;

  switch (fmt)
  {
    case CF_NULL:
    	break;    
    case CF_TEXT:
      if (!mxIsChar(mxa)) return argument_list_error;
      len = mxGetN(mxa);
      siz = len;
      mxGetString(mxa,(char *)buf,len+1);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      if (!mxIsChar(mxa)) return argument_list_error;
      if (siz > 1)
      {
        const mwSize *dims = mxGetDimensions(mxa);
        for (i=0; i<siz; i++)
        {
          mxc = mxGetChars(mxa);
          for (i=0; i<siz; i++)
          { /* string arrays transposed in MatLab ? */
            memset(&buf[fmtsiz*i],0,fmtsiz);
            for (k=0; k<fmtsiz && k<dims[1]; k++) 
            {
              c = (char)mxc[siz*k+i];
              ((char *)buf)[fmtsiz*i+k] = c;
              if (c == 0) break;
            }
          }
        }
      }
      else
      {
			  len = mxGetN(mxa) + 1;
        if (len > fmtsiz) len = fmtsiz;
			  mxGetString(mxa,(char *)buf,len);
      }
      break;
    case CF_STRING:
      if (mxIsChar(mxa)) 
      {
        len = 0;
        const mwSize *dims = mxGetDimensions(mxa);
        for (i=0; i<siz; i++)
        {
          mxc = (mxChar *)mxGetData(mxa);
          for (i=0; i<siz; i++)
          { /* string arrays transposed in MatLab ? */
            cptr = (char *)&buf[sizeof(size_t)*siz + len];
            ((size_t *)buf)[i] = (size_t)cptr;
            for (k=0; k<fmtsiz && k<dims[1]; k++) 
            {
              c = (char)mxc[siz*k+i];
              cptr[len++] = c;
              if (c == 0) break;
            }
          }
        }
      }
      else
      {
        return argument_list_error;
      }
      break;
    case CF_STRUCT:
			if (!mxIsStruct(mxa)) return argument_list_error;
      if ((ss=GetStructFromTag(tag)) != NULL && siz > 0)
      {
        fmtsiz = GetStructSize(tag);
        for (k=0, sf=ss->s; sf != NULL && k < 256; k++, sf=sf->nxt) if (sf->fmt == CF_NULL) k--;
        nf = k; // number of fields (n.b. legacy registrations might have a 'CF_NULL' in termination)s
        if (nf != mxGetNumberOfFields(mxa)) return invalid_structure_size;
        for (i=0; i<siz; i++)
        {
          for (n=nf-1, sf=ss->s; sf != NULL && n >=0; n--, sf=sf->nxt)
          { 
            fld = mxGetFieldNameByNumber(mxa,n);
            sa = mxGetField(mxa,i,fld);
            switch (sf->fmt)
            {
              case CF_INT16:
              case CF_INT32:
              case CF_FLOAT:
              case CF_DOUBLE:
           			if(!mxIsDouble(sa)) return invalid_structure_tag;
			          lp = (double *)mxGetPr(sa);
                setDblVal(k,sf->siz,&buf[fmtsiz*i+sf->off],sf->fmt,lp);
                break;
              case CF_TEXT:
           			if(!mxIsChar(sa)) return invalid_structure_tag;
                len = mxGetN(sa);
                mxGetString(sa,(char *)&buf[fmtsiz*i+sf->off],len+1);
                break;
              case CF_NULL:
                n++; 
                continue;
            }
          }
        }
      }
      break;
		case CF_BYTE:
		case CF_SHORT:
		case CF_INT32:
		case CF_FLOAT:
		case CF_DOUBLE:
			if(!mxIsDouble(mxa)) return argument_list_error;
			lp = (double *)mxGetPr(mxa);
      setDblVal(i,siz,buf,fmt,lp)
		  break;
		case CF_BOOLEAN:
			if(!mxIsDouble(mxa)) return argument_list_error;
			lp = (double *)mxGetPr(mxa);
      for (i=0; i<siz; i++) ((int *)buf)[i] = ((int)lp[i] == 0 ? 0 : 1);
		  break;		
		case CF_INTINT:
		case CF_FLTINT:
		case CF_FLTFLT:
		case CF_DBLDBL:
			if(!mxIsDouble(mxa)) return argument_list_error;
			lp = (double *)mxGetPr(mxa);
      siz /= 2;
      setDbltVal(siz,buf,fmt,lp);
		  break;
    case CF_INTINTINT:
    case CF_FLTINTINT:
    case CF_INTFLTINT:
    case CF_FLTFLTINT:
    case CF_DBLDBLDBL:
			if(!mxIsDouble(mxa)) return argument_list_error;
			lp = (double *)mxGetPr(mxa);
      siz /= 3;
      setTpltVal(siz,buf,fmt,lp);
      break;
		case CF_IFFF:
		case CF_IIFF:
		case CF_IIII:
		case CF_TTII:
		case CF_FIFI:
			if(!mxIsDouble(mxa)) return argument_list_error;
			lp = (double *)mxGetPr(mxa);
      siz /= 4;
      setQuadVal(siz,buf,fmt,lp);
      break;		
/*
		case CF_SPECTRUM:
		need 3 call
		break;
    case CF_USTRING:
      if(!mxIsChar(mxa))
      {
				plhs[0] = mxCreateString("Input argument error !!!");
				return;
			}
			len = mxGetN(mxa) + 1;
			mxGetString(mxa,buf,len);
      siz = 1; // kluge this by forcing only 1 of these to be sent
  		break;
*/
    default:
      if (!mxIsChar(mxa)) return illegal_format;
	}
  return 0;
}

char Buffer[MAX_TINE_BUFFER];
#define usageStr "tine_call usage:\n ret=tine_call(values,'/<context>/<server>/<device>[<property>]'[,'<data_type>',data_size,timer_interval,'access_mode']);\n where values = string or double Nx1 array\n'ret' contains the status/error message,timestamps,and returned values"
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) 
{
  mxArray	*data[PROPMAX];
  const char *field[PROPMAX];
  double *lp, dt = (double)time(NULL);
  int	error[PROPMAX];
  char *name, *typestr, *modestr, *mxerrStr=NULL;
  char *p;
  char buf[1024];
  char bf1[1024], *c=NULL, *c1=NULL, *c2=NULL, *c3=NULL, *cf=NULL;
  char dtstr[32], dtag[32], dtagw[32];
  char prop[80];
  char dev[80];
  int	len,size,fmtsiz;
  int writesize = 0,readsize = -1,datatype = -1;
  int writetype = CF_NULL, readtype = CF_INT32;
  mwSize dims[2];
  int	i,j,k,err,count=0,n,startindex = 3,cbid=-1;
  int	lim,nmb;
  int tmr = 1000;
  int acc = CA_READ;
  mxChar *CharData;
  char tsstr[64];
  DTYPE din, dout;
  
  if (plhs == NULL || prhs == NULL) return;
  dtagw[0] = dtag[0] = 0;
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 6: // access mode (string)
      if (!mxIsChar(prhs[5])) mxerr(usageStr);
      if (mxGetM(prhs[5]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[5]) + 1;
      modestr = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[5],modestr,len);
      acc = GetAccessMode(modestr);
      mxFree(modestr);
    case 5: // timer interval (int)
      if (!mxIsDouble(prhs[4])) mxerr(usageStr);
      if (mxGetM(prhs[4]) != 1) mxerr(usageStr);
      tmr = (int)mxGetScalar(prhs[4]);
    case 4: // output data size (int)
      if (!mxIsDouble(prhs[3])) mxerr(usageStr);
      if (mxGetM(prhs[3]) != 1) mxerr(usageStr);
      readsize = (int)mxGetScalar(prhs[3]);
    case 3: // input/output data type
      if (!mxIsChar(prhs[2])) mxerr(usageStr);
      if (mxGetM(prhs[2]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[2]) + 1;
      typestr = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[2],typestr,len);
      if ((p=strchr(typestr,':')) != NULL)
      { // both write and read types given explicitly as '<write_type>:<read_type>'
        *p++ = 0;
        strncpy(dtstr,typestr,32);
        if ((c=strchr(dtstr,'/')) != NULL || 
            (c=strchr(dtstr,'\\')) != NULL ||
            (c=strchr(dtstr,'.')) != NULL)
        {
          *c++ = 0;
          strncpy(dtagw,c,TAG_NAME_SIZE);
        }
        strupr(dtstr);
        if (dtstr[strlen(dtstr)-1] == 'S') dtstr[strlen(dtstr)-1] = 0; // in case "floats" instead of "float"
        writetype = ftoi(dtstr);
        strncpy(dtstr,p,32);
      }
      else
      {
        strncpy(dtstr,typestr,32);
      }
      if ((c=strchr(dtstr,'/')) != NULL || 
          (c=strchr(dtstr,'\\')) != NULL ||
          (c=strchr(dtstr,'.')) != NULL)
      {
        *c++ = 0;
        strncpy(dtag,c,TAG_NAME_SIZE);
      }
      strupr(dtstr);
      if (dtstr[strlen(dtstr)-1] == 'S') dtstr[strlen(dtstr)-1] = 0; // in case "floats" instead of "float"
      datatype = ftoi(dtstr);
      mxFree(typestr);
    case 2: // target
      if (!mxIsChar(prhs[1])) mxerr(usageStr);
      if (mxGetM(prhs[1]) != 1) mxerr(usageStr);
      // the requested data endpoint ...
      len = mxGetN(prhs[1]) + 1;
      name = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[1],name,len);
      field[0] = (char *)mxCalloc(6, sizeof(char));
      field[1] = (char *)mxCalloc(10, sizeof(char));
      field[2] = (char *)mxCalloc(4, sizeof(char));
      memset((void *)field[0],0,sizeof(field[0]));
      memset((void *)field[1],0,sizeof(field[1]));
      memset((void *)field[2],0,sizeof(field[2]));
      if((p=strchr(name,'?')) != NULL)
      {
        strcpy((char *)field[0],"usage");
        plhs[0] = mxCreateStructMatrix(1,1,1,&field[0]);
        data[0] = mxCreateString(usageStr);
        mxSetFieldByNumber(plhs[0],0,0,data[0]);
        mxFree(name);
        return;
 	    }
      strcpy((char *)field[0],"error");
      strcpy((char *)field[1],"timestamp");
      strcpy((char *)field[2],"utc");
	    
      strcpy(buf,name);
      if(strchr(name,'[') == NULL || strchr(name,']') == NULL) mxerr(usageStr);
      mxFree(name);

      if ((c1=strchr(buf,'<')) != NULL)
      { // allow this for backward compatibility ?
        c = c1; c++;
        strncpy(bf1,c,sizeof(bf1));
        for (c2=bf1; *c2 == ' '; c2++);  // any leading blanks?
        if (isalpha(*c2))
        { // just the format given
          cf = c2;
        }
        else
        {
          readsize = atoi(c2);
          if ((c2=strchr(c2,' ')) != NULL) // is the datatype left to parse out ?
          {
            while (*c2 == ' ') c2++;       // remove all leading blanks
            cf = c2;
          }
        }
        if (cf != NULL)
        {
          if ((c3=strchr(cf,'>')) != NULL) *c3 = 0;
          strncpy(dtstr,cf,32);
          strupr(dtstr);
          if (dtstr[strlen(dtstr)-1] == 'S') dtstr[strlen(dtstr)-1] = 0; // in case "floats" instead of "float"
          datatype = ftoi(dtstr);
        }
      }
      if ((c=strchr(buf,'@')) != NULL)
      { // peel it off before we start tokenizing
        *c++ = 0; 
        tmr = atoi(c); // this will override the argument input !
      }
      if (c1 != NULL) *c1 = 0; // peel this one off too, if it's there

      strcpy(dev, strtok(buf, "["));
      strncpy(prop, strtok(NULL, "]"),80);

      if (datatype < 0 || readsize < 0)
      { // still unknown : get the default
        readtype = read_prop(dev, prop, &readsize, dtag);
        if (readtype < 0) readtype = CF_INT32; // try this ...
        if (datatype > 0) readtype = datatype;
      }
      else
      {
        readtype = datatype;
      }
      // input values
      writesize = mxGetN(prhs[0]);
      if (writesize > 0 && writetype == CF_NULL) 
      {
        writetype = readtype;
        strncpy(dtagw,dtag,TAG_NAME_SIZE);
      }
      if (mxIsChar(prhs[0]))
      {
        if (writesize == 0)
        { // no input
          writetype = CF_NULL;
        }
        else if (!isStrFormat(writetype))
        { // non-zero length string given but inconsistent format type
          writetype = CF_TEXT;
        }
      }
      break;
    default:
      mxerr(usageStr);
  }
	// the data to 'send'
  if (writesize > 0)
  {
    if (writetype == CF_STRUCT && strlen(dtagw) > 0)
    {
      err = AcquireAndRegisterStruct(dev,dtagw,writesize);
    }
    err = setMxData(prhs[0],writetype,writesize,dtagw,(BYTE *)Buffer);
    if (err != 0)
    {
      mxerr("could not set input data !");
    }
  }
  din.dArrayLength = writesize;
  din.dFormat = writetype;
  din.data.vptr = Buffer;
  strncpy(din.dTag,dtagw,TAG_NAME_SIZE);
  
  fmtsiz = GetFormatSize(readtype);
  if (readsize * fmtsiz > MAX_TINE_BUFFER) readsize = MAX_TINE_BUFFER / fmtsiz;
  if (readtype == CF_STRUCT && strlen(dtag) > 0)
  {
    err = AcquireAndRegisterStruct(dev,dtag,readsize);
  }

  strcpy((char *)field[0],"error");
  data[0] = mxCreateString("");

  dout.dArrayLength = readsize;
  dout.dFormat = readtype;
  dout.data.vptr = Buffer;
  strncpy(dout.dTag,dtag,TAG_NAME_SIZE);
  err = mex_tine_call(dev, prop, &dout, &din, acc, tmr, &dt);

  k = startindex;
  field[k] = (char *)mxCalloc(strlen(prop)+1, sizeof(char));
  strcpy((char *)field[k],prop);

  if (err) 
  {
    data[k]  = mxCreateString(Buffer);
    data[0]  = mxCreateString(Buffer);
  } 
  else 
  {
    size = readsize;
    data[k] = getMxData(readtype,readsize,dtag,(BYTE *)Buffer);
  }
  plhs[0] = mxCreateStructMatrix(1,1,k+1,&field[0]);
  strcpy((char *)field[1],"timestamp");
  strcpy((char *)field[2],"utc");
  strncpy(tsstr,getDataTimeString(dt,0),64);
  sprintf(dtstr,"%10.3f",dt);
  data[1] = mxCreateString(tsstr);
  data[2] = mxCreateString(dtstr);

  mxSetFieldByNumber(plhs[0],0,0,data[0]);
  mxSetFieldByNumber(plhs[0],0,1,data[1]);
  mxSetFieldByNumber(plhs[0],0,2,data[2]);
  mxSetFieldByNumber(plhs[0],0,k,data[k]);
err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
}
