#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"

#define	PROPMAX 1024
#define	DEVMAX 4096
#define	MAX_TINE_BUFFER 512000 //256000

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

int mexIsStockProperty(char *prop)
{
  if (!stricmp(prop,"DEVICES")) return TRUE;
  if (!stricmp(prop,"NDEVICES")) return TRUE;
  if (!stricmp(prop,"NPROPS")) return TRUE;
  if (!stricmp(prop,"NPROPERTIES")) return TRUE;
  if (!stricmp(prop,"PROPERTIES")) return TRUE;
  return FALSE;
}
int	read_prop(char *dev, char *prop, int *size)
{
  PropertyQueryStruct	svrProp;
	DTYPE dout, din;
	int cc, fmt;
  char *prp = "PROPS";
  if ((cc=getPropertyInfoFromListener(dev,prop,size,&fmt)) == 0)
  {
    return fmt;
  }

  if (mexIsStockProperty(prop)) prp = "STOCKPROPS";

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = PROPERTYQUERYSTRUCT_SIZE;
  dout.data.vptr = &svrProp;
  dout.dTag[0] = '\0';

  din.dFormat = CF_TEXT;
  din.dArrayLength = strlen(prop);
  din.data.vptr = prop;
  din.dTag[0] = '\0';
  svrProp.prpName[0] = '\0';

  if ((cc=alsnExecLink(dev,prp,&dout,&din,CA_READ,500)) != 0)
  {
    *size = 0;
    return -cc;
 	}
  if (svrProp.prpName[0] == 0) 
  {
    *size = 0;
    return -cc;
 	}
  *size = svrProp.prpSize;
  return LFMT(svrProp.prpFormat);
}
int	close_data_link(char *dev, char *prop, int type, int size)
{
	DTYPE dout;
	int cc;

  memset(&dout,0,sizeof(DTYPE));
	dout.dFormat = type;
	dout.dArrayLength = size;
  return alsnStopLink(dev,prop,&dout,NULL);
}

char Buffer[512];
#define usageStr "tine_close_link usage:\n ret=tine_close_link('/<context>/<server>/<device>[<property>]'[,'<data_type>',data_size])"
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[]) 
{
	mxArray	*data[PROPMAX];
	const char *field[PROPMAX];
	double *lp, dt = (double)time(NULL);
	int	error[PROPMAX];
	char *name, *typestr, *cbfcn=NULL, *mxerrStr=NULL;
	char *p;
	char buf[1024];
	char *c=NULL;
  char dtstr[64];
	char prp[80];
	char dev[132];
	int	len,type=CF_FLOAT,size,fmtsiz,readsize = -1,ln,datatype = -1;
	mwSize dims[2];
	int	i,j,k,err,count,n,startindex = 3,cbid=-1;
	int	lim,nmb;
  int tmr = 1000;
  int acc = CA_READ;
	mxChar *CharData;
  char tsstr[64];
  char *tmpBuffer;
    
  if (plhs == NULL || prhs == NULL) return;

  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 3: // output data size (int)
      if (!mxIsDouble(prhs[2])) mxerr(usageStr);
      if (mxGetM(prhs[2]) != 1) mxerr(usageStr);
      readsize = (int)mxGetScalar(prhs[2]);
    case 2: // input/output data type
      if (!mxIsChar(prhs[1])) mxerr(usageStr);
      if (mxGetM(prhs[1]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[1]) + 1;
      typestr = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[1],typestr,len);
      strncpy(dtstr,typestr,32);
      strupr(dtstr);
      if (dtstr[strlen(dtstr)-1] == 'S') dtstr[strlen(dtstr)-1] = 0; // in case "floats" instead of "float"
      datatype = GetFormatFromString(dtstr);
      mxFree(typestr);
    case 1: // target
      if (!mxIsChar(prhs[0])) mxerr(usageStr);
      if (mxGetM(prhs[0]) != 1) mxerr(usageStr);
      // the requested data endpoint ...
      len = mxGetN(prhs[0]) + 1;
      name = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[0],name,len);
      field[0] = (char *)mxCalloc(6, sizeof(char));
      field[1] = (char *)mxCalloc(10, sizeof(char));
      field[2] = (char *)mxCalloc(4, sizeof(char));
      memset((void *)field[0],0,sizeof(field[0]));
      memset((void *)field[1],0,sizeof(field[1]));
      memset((void *)field[2],0,sizeof(field[2]));
      if((p=strchr(name,'?')) != NULL)
      {
        strcpy((char *)field[0],"usage");
        plhs[0] = mxCreateStructMatrix(1,1,1,&field[0]);
        data[0] = mxCreateString("result = tine_close_link(''/<context>/<server>/<device>[<property>]'[,'<data_type>',data_size]");
        mxSetFieldByNumber(plhs[0],0,0,data[0]);
        mxFree(name);
        return;
 	    }
      strcpy((char *)field[0],"error");
      strcpy((char *)field[1],"timestamp");
      strcpy((char *)field[2],"utc");

      break;
    default:
      mxerr(usageStr);
  }  
    
  strncpy(buf,name,1024);
  mxFree(name);
  if((p=strchr(buf,'[')) == NULL || strchr(buf,']') == NULL)
  { // no property name specified !
    mxerr(usageStr);
  }

  *p++ = 0;
  strncpy(dev,buf,132);
  while (*p == ' ') p++;
  strncpy(prp,p,64);
  if ((p=strchr(prp,']')) != NULL) *p = 0;
  if ((len=strlen(prp)) == 0) return;
  j = 0;

  if (readsize < 0 || datatype < 0)
  {
    type = read_prop(dev, prp, &size);
  }
  if (type < 0)
  {
    err = -type;
  }
  else
  {
    if (readsize > 0) size = readsize;
    if (datatype > 0) type = datatype;
mexPrintf("closing link to %s %s %d %d\n",dev,prp,type,size);
    err = close_data_link(dev, prp, type, size);
  }
  GetLastLinkError(err,Buffer);
  strcpy((char *)field[0],"error");
  plhs[0] = mxCreateStructMatrix(1,1,1,&field[0]);
  data[0] = mxCreateString(Buffer);
  mxSetFieldByNumber(plhs[0],0,0,data[0]);

err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
}
