#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"
#if defined(WIN32)
#include <direct.h>
#include <windows.h>

static char mexWndName[32];
HANDLE hMexWnd = 0;
#define WM_ML_CALLBACK  (WM_USER+200)

#ifndef HWND_MESSAGE
#define HWND_MESSAGE    ((HWND)-3)
#endif

#elif defined(UNIX)
#include <signal.h>
#ifndef MULTITHREADED
#define MULTITHREADED
#endif
#include "threader.h"

#define WM_ML_CALLBACK  SIGUSR1

pthread_t mlabMainThrd = 0;

#endif

#define	PROPMAX 1024
#define	DEVMAX 4096
#define	MAX_TINE_BUFFER 512000 //256000

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

static int mexStartCbMsgHndlrDone = FALSE;
static char cwd[256] = "";

#define MAX_CBHANDLERS 1024
typedef struct PrpHndlrStruct
{
  int id;
  int fcnExists;
  char fcnName[128];
  mxArray *dpParams[1];
} CbHndlrTbl;
CbHndlrTbl cbHndlrTbl[MAX_CBHANDLERS];
int cbHndlrTblSize = 0;

int getCbHndlrTblEntry(int cbid,char *fcn)
{
  int i;
  for (i=0; i<cbHndlrTblSize; i++)
  {
    if (fcn != NULL && strnicmp(fcn,cbHndlrTbl[i].fcnName,128) != 0) continue;
    if (cbHndlrTbl[i].id != cbid) continue;
    return i;
  }
  return -1;
}
int addCbHndlrTblEntry(int cbid,char *fcn)
{
  char fn[512];
  struct stat sbuf;
  int i = getCbHndlrTblEntry(cbid,fcn);
  if (i == -1 && cbHndlrTblSize < MAX_CBHANDLERS)
  { // doesn't exist and there's room ... 
    i = cbHndlrTblSize++;
    cbHndlrTbl[i].id = cbid;
    if (fcn != NULL)
    { // function name given ...
      strncpy(cbHndlrTbl[i].fcnName,fcn,128);
    }
    else
    { // then try this
      strncpy(cbHndlrTbl[i].fcnName,"tine_callback",128);
    }
    sprintf(fn,"%.256s%c%.128s.m",cwd,FS_DELIMITER,cbHndlrTbl[i].fcnName);
    if (stat(fn,&sbuf) == 0) cbHndlrTbl[i].fcnExists = TRUE;
  }
  else
  {
  }
  return i;
}
int handleCbMessage(int msg,int id)
{
  switch (msg)
  {
    case WM_ML_CALLBACK:
      if (id >= 0 && id < cbHndlrTblSize) 
      {
        if (cbHndlrTbl[id].fcnExists)
        {
          cbHndlrTbl[id].dpParams[0] = mxCreateDoubleMatrix(1, 1, mxREAL);
          *((double *)mxGetPr(cbHndlrTbl[id].dpParams[0])) = (double)cbHndlrTbl[id].id;
          mexCallMATLAB(0,NULL,1,cbHndlrTbl[id].dpParams,cbHndlrTbl[id].fcnName);
          mxDestroyArray(cbHndlrTbl[id].dpParams[0]);
        }
      }
      return 1;
      break;
    default: 
      break;
  }
  return 0;
}
#if defined(WIN32)
__declspec(dllexport) LRESULT CALLBACK mexWndProc(HWND hWnd, WORD iMessage, WORD wParam, LONG lParam)
{
  int id = (int)wParam;
  switch (iMessage)
  {
    case WM_ML_CALLBACK:
      return (LRESULT)handleCbMessage(iMessage,id);
    default: 
      return DefWindowProc (hWnd, iMessage, wParam, lParam);
  }
  return 0;
}
int CreateMexWindow(void)
{
  static WNDCLASS wndclass;
  int err, pid = GetCurrentProcessId();

  if (mexStartCbMsgHndlrDone) return 0;
  sprintf(mexWndName,"mxCbWnd-%x",pid);
  mexWndName[32] = 0;

  mexStartCbMsgHndlrDone = TRUE;
  if (hMexWnd != 0 && FindWindow(mexWndName,mexWndName) != 0) 
  { /* still active */
    return 0; 
  }
  
  wndclass.style         = CS_GLOBALCLASS|CS_SAVEBITS; 
  wndclass.lpfnWndProc   = (WNDPROC) mexWndProc;
  wndclass.cbClsExtra    = 0;
  wndclass.cbWndExtra    = 0;
  wndclass.hInstance     = (HINSTANCE)HWND_MESSAGE;
  wndclass.hIcon         = LoadIcon (0, IDI_APPLICATION);
  wndclass.hCursor       = LoadCursor (0, IDC_ARROW);
  wndclass.hbrBackground = (HBRUSH)COLOR_BACKGROUND; //GetStockObject (WHITE_BRUSH);
  wndclass.lpszMenuName  = 0;
  wndclass.lpszClassName = mexWndName;

  if (!RegisterClass(&wndclass)) 
  {
    if ((err=GetLastError()) != ERROR_CLASS_ALREADY_EXISTS)
    {
      MessageBeep(-1);
      return -1;
    }
  }

  if (!(hMexWnd = CreateWindowEx (
      WS_EX_NOPARENTNOTIFY,
      mexWndName, mexWndName,               
      WS_BORDER|WS_CAPTION,         
      CW_USEDEFAULT, CW_USEDEFAULT, 
      CW_USEDEFAULT, CW_USEDEFAULT, 
      HWND_MESSAGE, 0,                 
      (HINSTANCE)HWND_MESSAGE,               
      (LPSTR)""                     
      ) ) ) 
  {
    err = GetLastError();
    MessageBeep(-1);
    MessageBeep(-1);
    return -1;
  }

  return 0;
}
void mexStartCbMsgHndlr(void)
{
  CreateMexWindow();
}
#elif defined(UNIX)
siginfo_t mexCbSigInfo;
void mexCbSigFcn(int sig, siginfo_t *inf,void *ctx)
{
  int id = mexCbSigInfo.si_int;
  handleCbMessage(SIGUSR1,id);
}
void mexStartCbMsgHndlr(void)
{
  sigset_t set;
  static struct sigaction act;
  sigemptyset(&set);
  act.sa_handler = NULL;
  act.sa_sigaction = &mexCbSigFcn;
  act.sa_mask = set;
  act.sa_flags = SA_SIGINFO;
  sigaction(SIGUSR1,&act,NULL);
  mlabMainThrd = pthread_self();
  SystemSetCycleTimer();
}
#endif

int sendCbMessage(int id)
{
  int rc = 0;
# if defined(WIN32)
  int err;
  err = (int)PostMessage((HWND)hMexWnd,WM_ML_CALLBACK,id,0);
  if (err == 0) rc = code_failure;
# elif defined(UNIX)
  mexCbSigInfo.si_int = id;
  rc = pthread_kill(mlabMainThrd,SIGUSR1);
# endif
  return rc;
}
extern "C" void _mlCbNf(int id)
{
  int rc=0;

  if (id < 0 || id >= cbHndlrTblSize) return;
  sendCbMessage(id);
}
short ftoi(char *fmt)
{
  char *c,fmtstr[64];
  short ifmt;
  strncpy(fmtstr,fmt,64);
  if ((c=strchr(fmtstr,'.')) != NULL) *c = 0;
  strupr(fmtstr);
  if ((ifmt=getSimpleFormatType(fmtstr)) != CF_NULL) return ifmt;
  return getCompoundFormatType(fmtstr);
}
#ifdef WIN32
int mexlog(char *tag,char *text,...)
{
  char str[512];
  FILE *fp = NULL;
  va_list args;
  time_t t = time(NULL);
  static char *fmod = "a+";

  sprintf(str,"%.24s[%s] ",ctime(&t),tag);

  va_start(args,text);
  _vsnprintf(&str[strlen(str)],256,text,args); str[256] = 0;
  va_end(args);

  strcat(str,"\n");
  if ((fp=fopen("l:\\log\\mex.log","a+")) == NULL) return -1;
  fwrite(str,strlen(str),1,fp);

  fclose(fp);
  return 0;
}
#endif
int read_names(const char **name, mxArray **field, char* buf, int op)
{
  PropertyQueryStruct	svrProps[PROPMAX];
  NAME64 svrTags[DEVMAX];
  char bf[80], *tname;
  DTYPE dout;
  int	cc,i,tlen;
  u_short count;

  switch (op)
  {
    case 0:
      dout.dFormat = CF_NAME32; // CF_NAME16;
      dout.dArrayLength = DEVMAX/2;
      dout.data.vptr = svrTags;
      svrTags[0].name[0] = '\0';
      dout.dTag[0] = 0;
      cc = alsnExecLink("ENS","TAGS",&dout,NULL,CA_READ,500);
      count = PROPMAX;
      tlen = 32;
      break;
    case 1:
      i = DEVMAX;
      cc = GetSystemDevices(buf,NULL, svrTags,&i);
      if (cc) break;
      count = (u_short)i;
      tlen = 64;
      /*
      dout.dFormat = CF_SHORT;
      dout.dArrayLength = 1;
      dout.data.usptr = &count;
      cc = alsnExecLink(buf,"NDEVICES",&dout,NULL,CA_READ,500);
      if(cc) break;
      if (count > DEVMAX) count = DEVMAX;
      dout.dFormat = CF_NAME16;
      dout.dArrayLength = count;
      dout.data.vptr = svrTags;
      svrTags[0].name[0] = '\0';
      dout.dTag[0] = 0;
      cc = alsnExecLink(buf,"DEVICES",&dout,NULL,CA_READ,500);
      */
      break;		
    case 2:
      dout.dFormat = CF_SHORT;
      dout.dArrayLength = 1;
      dout.data.usptr = &count;
      cc = alsnExecLink(buf,"NPROPS",&dout,NULL,CA_READ,500);
      if(cc) break;
      if (count > PROPMAX) count = PROPMAX;
      dout.dFormat = CF_STRUCT;
      dout.dArrayLength = count*PROPERTYQUERYSTRUCT_SIZE;
      dout.data.vptr = svrProps;
      svrProps[0].prpName[0] = '\0';
      cc = alsnExecLink(buf,"PROPS",&dout,NULL,CA_READ,500);
      break;
  }
  if(cc) 
  {
    GetLastLinkError(cc,buf);
    return 0;
  }
  switch (op)
  {
    case 0:
    case 1:
      for(i = 0; i < count;i++)
      {
        tname = (char *)svrTags + i*tlen;
        if(strlen(tname) == 0 ) return i;
        //if(strlen((const char *)svrTags[i].name) == 0 ) return i;
        sprintf(bf,"Device_%d",i+1);
        cc = (int)strlen(bf)+1;
        name[i] = (char *)mxCalloc(cc, sizeof(char));
        memset((void *)name[i],0,sizeof(name[i]));
        strcpy((char *)name[i],bf);
        memset((void *)bf,0,sizeof(bf));
        //if(!op)	sprintf(bf,"%s",svrTags[i].name);
        //else	sprintf(bf,"%s/%s",buf,svrTags[i].name);
        if(!op)	sprintf(bf,"%s",tname);
        else	sprintf(bf,"%s/%s",buf,tname);
        field[i] = mxCreateString(bf);
      }
      break;		
    case 2:
      for(i= 0;i < count;i++)
      {
        if(svrProps[i].prpName[0] == 0 ) return i;
        cc = (int)strlen(svrProps[i].prpName)+1;
        name[i] = (char *)mxCalloc(cc, sizeof(char));
        memset((void *)name[i],0,sizeof(name[i]));
        strcpy((char *)name[i],svrProps[i].prpName);
        field[i] = mxCreateString(svrProps[i].prpDescription);
      }
      break;
  }
  return i;
}
int mexIsStockProperty(char *prop)
{
  if (!stricmp(prop,"DEVICES")) return TRUE;
  if (!stricmp(prop,"NDEVICES")) return TRUE;
  if (!stricmp(prop,"NPROPS")) return TRUE;
  if (!stricmp(prop,"NPROPERTIES")) return TRUE;
  if (!stricmp(prop,"PROPERTIES")) return TRUE;
  return FALSE;
}
int read_prop(char *dev, char *prop, int *size, char *tag)
{ 
  PropertyQueryStruct	svrProp; // legacy
  PrpQueryStruct pqs;
  DTYPE dout, din;
  int cc, fmt, isLegacy=FALSE;
  char *prp = "PROPS";

  if ((cc=getPropertyInfoFromListenerEx(dev,prop,size,&fmt,tag)) == 0)
  {
    return fmt;
  }
  if (mexIsStockProperty(prop)) prp = "STOCKPROPS";

  memset(&din,0,sizeof(DTYPE));
  din.dFormat = CF_TEXT;
  din.dArrayLength = (int)strlen(prop);
  din.data.vptr = prop;

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  dout.data.vptr = &pqs;
  strcpy(dout.dTag,"PRPQSr4");

legacyRetry:
  if ((cc=alsnExecLink(dev,prp,&dout,&din,CA_READ,500)) != 0)
  {
    if (!isLegacy && (cc == illegal_format || cc == invalid_structure_tag))
    {
      isLegacy = TRUE;
      dout.dFormat = CF_STRUCT;
      dout.dArrayLength = PROPERTYQUERYSTRUCT_SIZE;
      dout.data.vptr = &svrProp;
      dout.dTag[0] = '\0';
      goto legacyRetry;
    }
    *size = 0;
    return -cc;
 	}

  if (isLegacy)
  {
    *size = svrProp.prpSize;
    fmt = LFMT(svrProp.prpFormat);
  }
  else
  {
    *size = pqs.prpSize;
    fmt = LFMT(pqs.prpFormat);
    if (tag != NULL) strncpy(tag,pqs.prpTag,TAG_NAME_SIZE);
  }
  return fmt;
}
int read_data_ext(char *dev, char *prop, int type, int *size, char *buf,int access_rate,double *pDataTime,char *dtag,int accessFlags)
{
  DTYPE dout;
  int cc, access=CA_READ;

  if (accessFlags > 0) access |= accessFlags;
  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = type;
  dout.dArrayLength = *size;
  dout.data.vptr = buf;
  if (dtag != NULL) strncpy(dout.dTag,dtag,TAG_NAME_SIZE);
  if ((cc=alsnExecLinkEx(dev,prop,&dout,NULL,access,access_rate,pDataTime)) != 0)
  {
    GetLastLinkError(cc,buf);
    if (pDataTime != NULL) *pDataTime = (double)time(NULL);
    return 1;
  }
  *size = dout.dArrayLength;
  return 0;
}
int read_data(char *dev,char *prop,int type,int *size,char *buf)
{
  return read_data_ext(dev,prop,type,size,buf,1000,NULL,NULL,CA_READ);
}
void name_corr(char *name)
{
  int	i, len;

  len = (int)strlen(name);
  for (i = 0; i < len; i++) 
  {
    if (name[i] == '.') 
    {
      name[i] = '_';
    } 
    else if (name[i] == '-') 
    {
      name[i] = '_';
    }
  }
}
#define getDblVal(i,b,f) (\
  (double)((f) == CF_BYTE ? (b)[i] :\
           (f) == CF_INT16 ? ((short *)(b))[i] :\
           (f) == CF_INT32 ? ((int *)(b))[i] :\
           (f) == CF_FLOAT ? ((float *)(b))[i] :\
           (f) == CF_DOUBLE ? ((double *)(b))[i] : 0))

#define get1stDblVal(i,b,f) (\
  (double)((f) == CF_INTINT ? ((INTINT *)(b))[i].i1val :\
           (f) == CF_FLTINT ? ((FLTINT *)(b))[i].fval :\
           (f) == CF_DBLDBL ? ((DBLDBL *)(b))[i].d1val :\
           (f) == CF_FLTFLT ? ((FLTFLT *)(b))[i].f1val : 0))

#define get2ndDblVal(i,b,f) (\
  (double)((f) == CF_INTINT ? ((INTINT *)(b))[i].i2val :\
           (f) == CF_FLTINT ? ((FLTINT *)(b))[i].ival :\
           (f) == CF_DBLDBL ? ((DBLDBL *)(b))[i].d2val :\
           (f) == CF_FLTFLT ? ((FLTFLT *)(b))[i].f2val : 0))

#define get1stQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i1val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].ival :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].i1val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].t1 :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].f1val : 0))

#define get2ndQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i2val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].f1val :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].i2val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].t2 :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].i1val : 0))

#define get3rdQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i3val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].f2val :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].f1val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].i1val :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].f2val : 0))

#define get4thQuadVal(i,b,f) (\
  (double)((f) == CF_INTINTINTINT ? ((DADDRESS *)(b))[i].i4val :\
           (f) == CF_INTFLTFLTFLT ? ((DFILTER *)(b))[i].f3val :\
           (f) == CF_INTINTFLTFLT ? ((FWINDOW *)(b))[i].f2val :\
           (f) == CF_TTII         ? ((DWINDOW *)(b))[i].i2val :\
           (f) == CF_FLTINTFLTINT ? ((FLTINTFLTINT *)(b))[i].i2val : 0))

#define get1stTriVal(i,b,f) (\
  (double)((f) == CF_INTINTINT ? ((INTINTINT *)(b))[i].i1val :\
           (f) == CF_FLTFLTINT ? ((FLTFLTINT *)(b))[i].f1val :\
           (f) == CF_INTFLTINT ? ((INTFLTINT *)(b))[i].i1val :\
           (f) == CF_FLTINTINT ? ((FLTINTINT *)(b))[i].fval :\
           (f) == CF_DBLDBLDBL ? ((DBLDBLDBL *)(b))[i].d1val : 0))

#define get2ndTriVal(i,b,f) (\
  (double)((f) == CF_INTINTINT ? ((INTINTINT *)(b))[i].i2val :\
           (f) == CF_FLTFLTINT ? ((FLTFLTINT *)(b))[i].f2val :\
           (f) == CF_INTFLTINT ? ((INTFLTINT *)(b))[i].fval :\
           (f) == CF_FLTINTINT ? ((FLTINTINT *)(b))[i].i1val :\
           (f) == CF_DBLDBLDBL ? ((DBLDBLDBL *)(b))[i].d2val : 0))

#define get3rdTriVal(i,b,f) (\
  (double)((f) == CF_INTINTINT ? ((INTINTINT *)(b))[i].i3val :\
           (f) == CF_FLTFLTINT ? ((FLTFLTINT *)(b))[i].ival :\
           (f) == CF_INTFLTINT ? ((INTFLTINT *)(b))[i].i2val :\
           (f) == CF_FLTINTINT ? ((FLTINTINT *)(b))[i].i2val :\
           (f) == CF_DBLDBLDBL ? ((DBLDBLDBL *)(b))[i].d3val : 0))

mxArray *getMxData(int fmt,int siz,char *tag,BYTE *buf)
{
  mxArray *da, *fdata[256];
  mxChar *mxc;
  int fmtsiz = fmtsizeof(fmt);
  int i, k, n, nsiz, nf;
  mwSize dims[2];
  const char *field[PROPMAX], *nifield[2], *nddfield[3], *sfield[256];
  double *lp;
  float *fp;
  struct structStruct *ss;
  struct structFormat *sf;

  switch(fmt)
  {	
    case CF_TEXT:
      da = mxCreateString((const char *)buf);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      dims[0] = siz;
      dims[1] = fmtsiz;
      if (siz > 1)
      {
        da = mxCreateCharArray(2,dims);
        mxc = (mxChar *)mxGetData(da);
        for (i=0; i<siz; i++)
        { /* string arrays transposed in MatLab ? */
          for (k=0; k<fmtsiz; k++) mxc[siz*k+i] = (mxChar)buf[fmtsiz*i+k];
        }
      }
      else
      {
        da = mxCreateString((const char *)buf);
      }
      break;
    case CF_STRING:
      if (siz > 0)
      {
        da = mxCreateCharMatrixFromStrings(siz,(const char **)buf);
      }
      break;
    case CF_NAME8I:
    case CF_NAME16I:
    case CF_NAME32I:
    case CF_NAME48I:
    case CF_NAME64I:
      nifield[0] = (char *)mxCalloc(5, sizeof(char));
      nifield[1] = (char *)mxCalloc(6, sizeof(char));
      strcpy((char *)nifield[0],"name");
      strcpy((char *)nifield[1],"value");
      nsiz = fmtsiz - 4;
      dims[0] = 1;
      dims[1] = nsiz;
      if (siz > 0)
      {
        da = mxCreateStructMatrix(siz,1,2,&nifield[0]);
        for (i=0; i<siz; i++)
        {
          fdata[0] = mxCreateCharArray(2,dims);
          fdata[1] = mxCreateDoubleMatrix(1,1,mxREAL);
          mxc = (mxChar *)mxGetData(fdata[0]);
          lp = (double *)mxGetPr(fdata[1]);
          for (k=0; k<nsiz; k++) mxc[k] = (mxChar)buf[fmtsiz*i+k];
          *lp = double(*(int *)&buf[fmtsiz*i+nsiz]);
          mxSetFieldByNumber(da,i,0,fdata[0]);
          mxSetFieldByNumber(da,i,1,fdata[1]);
        }
      }
      break;
    case CF_STRUCT:
      if ((ss=GetStructFromTag(tag)) != NULL && siz > 0)
      {
        fmtsiz = GetStructSize(tag);
        for (k=0, sf=ss->s; sf != NULL && k < 256; k++, sf=sf->nxt) if (sf->fmt == CF_NULL) k--;
        nf = k; // number of fields (n.b. legacy registrations might have a 'CF_NULL' in termination)
        for (n=nf-1, sf=ss->s; sf != NULL && n >= 0; n--, sf=sf->nxt)
        {
          if (sf->fmt == CF_NULL) { n++; continue; }
          nsiz = MIN(strlen(sf->field),16);
          sfield[n] = (char *)mxCalloc(nsiz+1,sizeof(char));
          strncpy((char *)sfield[n],sf->field,nsiz);
        }
        da = mxCreateStructMatrix(siz,1,nf,&sfield[0]);
        for (i=0; i<siz; i++)
        {
          for (n=nf-1, sf=ss->s; sf != NULL && n >=0; n--, sf=sf->nxt)
          { 
            nsiz = sf->off;
            switch (sf->fmt)
            {
              case CF_INT16:
              case CF_INT32:
              case CF_FLOAT:
              case CF_DOUBLE:
                fdata[n] = mxCreateDoubleMatrix(sf->siz,1,mxREAL);
                lp = (double *)mxGetPr(fdata[n]);
                for (k=0; k<sf->siz; k++) *lp++ = getDblVal(k,&buf[fmtsiz*i+nsiz],sf->fmt);
                mxSetFieldByNumber(da,i,n,fdata[n]);
                break;
              case CF_TEXT:
                dims[0] = 1;
                dims[1] = sf->siz+1;
                fdata[n] = mxCreateCharArray(2,dims);
                mxc = (mxChar *)mxGetData(fdata[n]);
                for (k=0; k<dims[1]; k++) mxc[k] = (mxChar)buf[fmtsiz*i+nsiz+k];
                mxSetFieldByNumber(da,i,n,fdata[n]);
                break;
              case CF_NULL:
                n++; 
                continue;
            }
          }
        }
      }
      break;
    case CF_USTRING:
      buf[96] = '\0';
      sprintf((char *)buf,"%d %g %g %u %s",*(int *)&buf[0],*(float *)&buf[4],*(float *)&buf[8],*(UINT32 *)&buf[12],(char *)&buf[16]);
      da = mxCreateString((const char *)buf);
      break;					
    case CF_BYTE:
    case CF_INT16:
    case CF_INT32:
    case CF_FLOAT:
    case CF_DOUBLE:
      da = mxCreateDoubleMatrix(siz,1,mxREAL);
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) *lp++ = getDblVal(i,buf,fmt);
      break;
    case CF_BOOLEAN:
      da = mxCreateDoubleMatrix(siz,1,mxREAL);
      lp = (double *)mxGetPr(da);
      for(n = 0; n < siz; n++) *lp++ = (double)(((int *)buf)[i] ? 1 : 0);
      break;
    case CF_INTINT:
    case CF_FLTINT:
    case CF_FLTFLT:
    case CF_DBLDBL:
      da = mxCreateDoubleMatrix(siz,2,mxREAL);
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) 
      {
        *lp       = get1stDblVal(i,buf,fmt);
        *(lp+siz) = get2ndDblVal(i,buf,fmt);
        lp++;
      }
      break;
    case CF_IFFF:
    case CF_FIFI:
    case CF_IIII:
    case CF_IIFF:
    case CF_TTII:
      da = mxCreateDoubleMatrix(siz,4,mxREAL),
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) 
      {
        *lp         = get1stQuadVal(i,buf,fmt);
        *(lp + siz) = get2ndQuadVal(i,buf,fmt);
        *(lp+2*siz) = get3rdQuadVal(i,buf,fmt);
        *(lp+3*siz) = get4thQuadVal(i,buf,fmt);
        lp++;
      }
      break;
    case CF_SPECTRUM:
//      da = mxCreateDoubleMatrix(siz,1,mxREAL),
//      lp = (double *)mxGetPr(da);
//      fp = (float *)&buf[96];
//      for(n = 0; n < siz; n++) *lp++ = double(fp[n]);

      sfield[0] = (char *)mxCalloc(8,sizeof(char));
      strncpy((char *)sfield[0],"comment",7);
      sfield[1] = (char *)mxCalloc(5,sizeof(char));
      strncpy((char *)sfield[1],"time",4);
      sfield[2] = (char *)mxCalloc(6,sizeof(char));
      strncpy((char *)sfield[2],"start",5);
      sfield[3] = (char *)mxCalloc(10,sizeof(char));
      strncpy((char *)sfield[3],"increment",9);
      sfield[4] = (char *)mxCalloc(7,sizeof(char));
      strncpy((char *)sfield[4],"status",6);
      sfield[5] = (char *)mxCalloc(9,sizeof(char));
      strncpy((char *)sfield[5],"spectrum",8);

      da = mxCreateStructMatrix(1,1,6,&sfield[0]);

      dims[0] = 1; dims[1] = 81;
      fdata[0] = mxCreateCharArray(2,dims);
      mxc = (mxChar *)mxGetData(fdata[0]);
      for (k=0; k<dims[1]; k++) mxc[k] = (mxChar)buf[k];
      mxSetFieldByNumber(da,0,0,fdata[0]);
      
      fdata[1] = mxCreateDoubleMatrix(1,1,mxREAL);
      lp = (double *)mxGetPr(fdata[1]);
      *lp = getDblVal(0,&buf[80],CF_INT32);
      mxSetFieldByNumber(da,0,1,fdata[1]);
      
      fdata[2] = mxCreateDoubleMatrix(1,1,mxREAL);
      lp = (double *)mxGetPr(fdata[2]);
      *lp = getDblVal(0,&buf[84],CF_FLOAT);
      mxSetFieldByNumber(da,0,2,fdata[2]);

      fdata[3] = mxCreateDoubleMatrix(1,1,mxREAL);
      lp = (double *)mxGetPr(fdata[3]);
      *lp = getDblVal(0,&buf[88],CF_FLOAT);
      mxSetFieldByNumber(da,0,3,fdata[3]);

      fdata[4] = mxCreateDoubleMatrix(1,1,mxREAL);
      lp = (double *)mxGetPr(fdata[4]);
      *lp = getDblVal(0,&buf[92],CF_INT32);
      mxSetFieldByNumber(da,0,4,fdata[4]);

      fdata[5] = mxCreateDoubleMatrix(siz,1,mxREAL),
      lp = (double *)mxGetPr(fdata[5]);
      fp = (float *)&buf[96];
      for(n = 0; n < siz; n++) *lp++ = double(fp[n]);
      mxSetFieldByNumber(da,0,5,fdata[5]);

      break;
    case CF_INTINTINT:
    case CF_DBLDBLDBL:
    case CF_FLTFLTINT:
    case CF_INTFLTINT:
    case CF_FLTINTINT:
      da = mxCreateDoubleMatrix(siz,3,mxREAL),
      lp = (double *)mxGetPr(da);
      for (i=0; i<siz; i++) 
      {
        *lp         = get1stTriVal(i,buf,fmt);
        *(lp + siz) = get2ndTriVal(i,buf,fmt);
        *(lp+2*siz) = get3rdTriVal(i,buf,fmt);
        lp++;
      }
      break;
    default:
      sprintf((char *)buf,"data type %d not handled\n",fmt);
      da  = mxCreateString((const char *)buf);
      break;
  }
  return da;
}

char *Buffer = NULL;
int mlBufferSize = MAX_TINE_BUFFER;
#ifdef ALLOW_CALLBACKS
# define usageStr "tine_read usage:\n ret=tine_read('/<context>/<server>/<device>[<property>]'[,'<data_type>',data_size,timer_interval,callback_id,'callback_fcn']);\n'ret' contains the status/error message,timestamps,and returned values"
#else
# define usageStr "tine_read usage:\n ret=tine_read('/<context>/<server>/<device>[<property>]'[,'<data_type>',data_size,timer_interval]);\n'ret' contains the status/error message,timestamps,and returned values"
#endif
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[]) 
{
  mxArray *data[PROPMAX], *fdata[256];
  const char *field[PROPMAX], *nifield[2], *nddfield[3], *sfield[256];
  double *lp, dt = (double)time(NULL);
  int error[PROPMAX];
  char *name, *typestr, *cbfcn=NULL, *mxerrStr=NULL;
  char *p;
  char buf[1024];
  char bf1[1024], *c=NULL, *c1=NULL, *c2=NULL, *c3=NULL;
  char dtstr[32],dtag[32];
  char prop[80];
  char dev[132];
  int len,type=CF_FLOAT,size,fmtsiz,nsiz,readsize = -1,ln,datatype = -1;
  int dims[2];
  int i,j,k,err,count,n,nf,startindex = 3,cbid=-1;
  int lim,nmb;
  int tmr = 1000;
  int acc = CA_READ;
  mxChar *CharData;
  char tsstr[64];
  char *tmpBuffer;
  struct structStruct *ss;
  struct structFormat *sf;
    
  if (plhs == NULL || prhs == NULL) return;
  dtag[0] = 0;
  /* Check for proper number of arguments. */
  switch (nrhs)
  {
#   ifdef ALLOW_CALLBACKS
    case 6:
      mexLock();
      if (!mxIsChar(prhs[5])) mxerr(usageStr);
      if (mxGetM(prhs[5]) != 1) mxerr(usageStr);
      len = (int)mxGetN(prhs[5]) + 1;
      cbfcn = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[5],cbfcn,len);
    case 5:
      if (mxIsDouble(prhs[4]))
      {
        cbid = (int)mxGetScalar(prhs[4]);
        if (!mexStartCbMsgHndlrDone) 
        {
          getcwd(cwd,256);
          mexStartCbMsgHndlr();
        }
      }
#   endif
    case 4: // timer interval (int)
      if (!mxIsDouble(prhs[3])) mxerr(usageStr);
      if (mxGetM(prhs[3]) != 1) mxerr(usageStr);
      tmr = (int)mxGetScalar(prhs[3]);
    case 3: // output data size (int)
      if (!mxIsDouble(prhs[2])) mxerr(usageStr);
      if (mxGetM(prhs[2]) != 1) mxerr(usageStr);
      readsize = (int)mxGetScalar(prhs[2]);
    case 2: // input/output data type
      if (!mxIsChar(prhs[1])) mxerr(usageStr);
      if (mxGetM(prhs[1]) != 1) mxerr(usageStr);
      len = (int)mxGetN(prhs[1]) + 1;
      typestr = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[1],typestr,len);
      strncpy(dtstr,typestr,32);
      if ((c=strchr(dtstr,'/')) != NULL || 
          (c=strchr(dtstr,'\\')) != NULL ||
          (c=strchr(dtstr,'.')) != NULL)
      {
        *c++ = 0;
        strncpy(dtag,c,TAG_NAME_SIZE);
      }
      strupr(dtstr);
      if (dtstr[strlen(dtstr)-1] == 'S') dtstr[strlen(dtstr)-1] = 0; // in case "floats" instead of "float"
      datatype = ftoi(dtstr);
      mxFree(typestr);
    case 1: // target
      if (!mxIsChar(prhs[0])) mxerr(usageStr);
      if (mxGetM(prhs[0]) != 1) mxerr(usageStr);
      // the requested data endpoint ...
      len = (int)mxGetN(prhs[0]) + 1;
      name = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[0],name,len);
      field[0] = (char *)mxCalloc(6, sizeof(char));
      field[1] = (char *)mxCalloc(10, sizeof(char));
      field[2] = (char *)mxCalloc(4, sizeof(char));
      memset((void *)field[0],0,sizeof(field[0]));
      memset((void *)field[1],0,sizeof(field[1]));
      memset((void *)field[2],0,sizeof(field[2]));
      if((p=strchr(name,'?')) != NULL)
      {
        strcpy((char *)field[0],"usage");
        plhs[0] = mxCreateStructMatrix(1,1,1,&field[0]);
        data[0] = mxCreateString("data = tine_read('dev/loc[prop1;prop2;....;propn]<data_size data_type>@interval');\ndev = tine_read('');\nloc = tine_read('dev/*');\nprop = tine_read('dev/loc')\nThe error field contains the number(s) of the wrong read\nprops and the corresponding props contain error message(s)!!!");
        mxSetFieldByNumber(plhs[0],0,0,data[0]);
        mxFree(name);
        return;
      }
      strcpy((char *)field[0],"error");
      strcpy((char *)field[1],"timestamp");
      strcpy((char *)field[2],"utc");

      break;
    default:
      mxerr(usageStr);
  }  
    
  if (Buffer == NULL)
  {
    if ((Buffer=(char *)malloc(MAX_TINE_BUFFER)) == NULL)
    {
      plhs[0] = mxCreateString("Out of memory !!!");
      return;
    }
    mlBufferSize = MAX_TINE_BUFFER;
  }
	
  if(len < 2 || name[0] == '[' || name[0] == '*')
  {
    k = read_names(&field[0], &data[0], bf1, 0);
    if(!k)
    {
      plhs[0] = mxCreateStructMatrix(1, 1, 1, &field[0]);
      sprintf(buf,"Names read error: %s",bf1);
      data[0] = mxCreateString(buf);
      mxSetFieldByNumber(plhs[0], 0, 0, data[0]);
      mxFree(name);
      return;
    }
    plhs[0] = mxCreateStructMatrix(1,1,k,&field[0]);
    mxSetFieldByNumber(plhs[0],0,0,data[0]);
    for(i = 1; i < k; i++)
    {
      mxSetFieldByNumber(plhs[0],0,i,data[i]);
    }
    mxFree(name);
    return;
  }	
  strcpy(buf,name);
  if((p = strchr(name,'[')) == NULL)
  { // no property name specified -> query of some sort ?
    if((p=strchr(name,'*')) != NULL && (p=strchr(name,'/')) != NULL)
    {
      //strcpy(bf1,strtok(buf,"/"));
      strncpy(bf1,buf,1024);
      if ((p=strstr(bf1,"/*")) != NULL) *p = 0;
      k = read_names(&field[0], &data[0],bf1,1);
    } 
    else 
    { // no wild card -> get the property list ?
      strcpy(bf1,name);
      k = read_names(&field[0], &data[0],bf1,2);
    }
    if (!k) 
    {
      plhs[0] = mxCreateStructMatrix(1,1,1,&field[0]);
      sprintf(buf,"Names read error: %s",bf1);
      data[0] = mxCreateString(buf);
      mxSetFieldByNumber(plhs[0],0,0,data[0]);
      mxFree(name);
      return;
    }
    plhs[0] = mxCreateStructMatrix(1,1,k,&field[0]);
    mxSetFieldByNumber(plhs[0],0,0,data[0]);
    for(i = 1; i < k; i++)
    {
      mxSetFieldByNumber(plhs[0],0,i,data[i]);
    }
    return;
  }
  if ((p = strchr(name,']')) == NULL)
  {
    plhs[0] = mxCreateStructMatrix(1,1,1,&field[0]);
    data[0] = mxCreateString("usage: tine_read('dev/loc[prop1;prop2;....;propn]<data_type @interval>'  !!!");
    mxSetFieldByNumber(plhs [0],0,0,data[0]);
    mxFree(name);
    return;
  }

  mxFree(name);
  if ((c1=strchr(buf,'<')) != NULL)
  {
    c = c1; c++;
    strncpy(bf1,c,sizeof(bf1));
    for (c2=bf1; *c2 == ' '; c2++);  // any leading blanks?
    readsize = atoi(c2);
    if ((c2=strchr(c2,' ')) != NULL) // is the datatype left to parse out ?
    {
      while (*c2 == ' ') c2++;       // remove all leading blanks
      if ((c3=strchr(c2,'>')) != NULL) *c3 = 0;
      strncpy(dtstr,c2,32);
      strupr(dtstr);
      if (dtstr[strlen(dtstr)-1] == 'S') dtstr[strlen(dtstr)-1] = 0;
      datatype = ftoi(dtstr);
    }
  }
  if ((c=strchr(buf,'@')) != NULL)
  {
    *c++ = 0; // peel it off before we start tokenizing
    tmr = atoi(c);
    if (strstr(c,"CONNECT") != NULL) acc |= CA_CONNECT;
  }
  if (c1 != NULL) *c1 = 0; // peel this one off too, if it's there
  strcpy(dev, strtok(buf, "["));
  strncpy(bf1, "", sizeof(bf1));
  strcpy(bf1, strtok(NULL, "]"));
  len = (int)strlen(bf1);

  j = 0;
  k = startindex;
  count = 0;
  for(i = 0; i < len+1; i++)
  {
    if(bf1[i] == ' ') continue;
    if(bf1[i] == ';' || bf1[i] == '\0') 
    {
      prop[j]='\0';
      j = 0;
      if (readsize < 0 || datatype < 0)
      {
        type = read_prop(dev, prop, &size, dtag);
        acc |= CA_MUTABLE; // if not given explicitly
      }
      if (type < 0)
      {
        err = -type;
        GetLastLinkError(err,Buffer);
      }
      else
      {
        if (readsize > 0) size = readsize;
        if (datatype > 0) type = datatype;
        if ((fmtsiz=fmtsizeof(type)) < 1) return;
        if (type == CF_STRUCT && strlen(dtag) > 0)
        {
          err = AcquireAndRegisterStruct(dev,dtag,size);
        }
        if (size * fmtsiz > mlBufferSize) 
        {
          if ((tmpBuffer=(char *)realloc(Buffer,size*fmtsiz)) == NULL)
          { // didn't work -> scale down
            size = mlBufferSize / fmtsiz;
          }
          else
          { // record the new settings ...
            Buffer = tmpBuffer;
            mlBufferSize = size * fmtsiz;
          }
        }
        err = read_data_ext(dev, prop, type, &size, Buffer, tmr, &dt, dtag, acc);
        if (cbid >= 0) 
        {
          n = addCbHndlrTblEntry(cbid,cbfcn);
          if (n >= 0)
          {
            sprintf(buf,"%.130s[%.64s]",dev,prop);
            tAttachNotifier(buf, _mlCbNf, n);
          }
        }
      }
      name_corr(prop);
      ln = (int)strlen(prop)+1;

      if (ln > 1) 
      {
        field[k] = (char *)mxCalloc(ln, sizeof(char));
        memset((void *)field[k],0,sizeof(field[k]));
        strcpy((char *)field[k],prop);
        memset((void *)prop,0,sizeof(prop));
      } 
      else 
      {
        k++;
        break;
      }

      if (err) 
      {
        error[count++] = k;
        data[k]  = mxCreateString(Buffer);
        data[0]  = mxCreateString(Buffer);
      }
      else
      {
        data[k] = getMxData(type,size,dtag,(BYTE *)Buffer);
      }
      k++;
    }
    else
    {
      prop[j++]=bf1[i];
    }
  }
  strcpy((char *)field[0],"error");
  plhs[0] = mxCreateStructMatrix(1,1,k,&field[0]);
  if(count)
  {
    if (count > 1)
    {
      data[0] = mxCreateDoubleMatrix(count,1,mxREAL),
      lp = (double *)mxGetPr(data[0]);
      for(i = 0; i < count; i++) *lp++ = double(error[i]);
    }
  } 
  else 
  {
    strcpy((char *)field[1],"timestamp");
    strcpy((char *)field[2],"utc");
    data[0] = mxCreateString("");
  }
  strncpy(tsstr,getDataTimeString(dt,0),64);
  sprintf(dtstr,"%10.3f",dt);
  data[1] = mxCreateString(tsstr);
  data[2] = mxCreateString(dtstr);

  mxSetFieldByNumber(plhs[0],0,0,data[0]);
  mxSetFieldByNumber(plhs[0],0,1,data[1]);
  mxSetFieldByNumber(plhs[0],0,2,data[2]);
  for(i = startindex; i < k; i++) 
  {
    mxSetFieldByNumber(plhs[0],0,i,data[i]);
  }	
err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
}
