#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"

#define	PROPMAX 10

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

#define usageStr "tine_register_device usage:\n ret=tine_register_device('device_name',device_number[,'device_redirection','device_description'])"
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[]) 
{
  mxArray	*data[PROPMAX];
  const char *field[PROPMAX];
  char *dev=NULL, *rdr=NULL, *dsc=NULL, *mxerrStr=NULL;
  char buf[128],stsbuf[64];
  double *lp;
  int	i,k,len,err,n;
  int devnr=0;
    
  if (plhs == NULL || prhs == NULL) return;
  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 4: // description
      if (!mxIsChar(prhs[3])) mxerr(usageStr);
      if (mxGetM(prhs[3]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[3]) + 1;
      dsc = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[3],dsc,len);
    case 3: // redirection
      if (!mxIsChar(prhs[2])) mxerr(usageStr);
      if (mxGetM(prhs[2]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[2]) + 1;
      rdr = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[2],rdr,len);
    case 2: // device number and name
      if (!mxIsDouble(prhs[1])) mxerr(usageStr);
      if (mxGetM(prhs[1]) != 1) mxerr(usageStr);
      devnr = (int)mxGetScalar(prhs[1]);
      if (!mxIsChar(prhs[0])) mxerr(usageStr);
      if (mxGetM(prhs[0]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[0]) + 1;
      dev = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[0],dev,len);
      break;
    default:
      mxerr(usageStr);
  }   

  err = RegisterBufferedDeviceName(dev,devnr,rdr,dsc);

  if (err != 0)
  {
    GetLastLinkError(err,stsbuf);
    sprintf(buf,"tine_register_device : <%d> %.64s",err,stsbuf);
    mexErrMsgTxt(buf);
  }
  data[0] = mxCreateDoubleMatrix(1,1,mxREAL);
  lp = (double *)mxGetPr(data[0]);
  *lp = (double)err;
  plhs[0] = data[0];
  
err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
  if (dev != NULL) mxFree(dev);
  if (rdr != NULL) mxFree(rdr);
  if (dsc != NULL) mxFree(dsc);
}
