#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"

#define	PROPMAX 10

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

#define usageStr "tine_register_fec usage:\n ret=tine_register_fec('fec_name'[,'sub_system','context','description','location','hardware','responsible',portOffset])"
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[]) 
{
  mxArray	*data[PROPMAX];
  char *fec=NULL, *sub=NULL, *ctx=NULL, *dsc=NULL, *loc=NULL, *hdw=NULL, *rsp=NULL, *mxerrStr=NULL;
  char buf[128],stsbuf[64];
  double *lp;
  int	i,k,len,err,n;
  int poff=0;
    
  if (plhs == NULL || prhs == NULL) return;
  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 8:
      if (!mxIsDouble(prhs[7])) mxerr(usageStr);
      if (mxGetM(prhs[7]) != 1) mxerr(usageStr);
      poff = (int)mxGetScalar(prhs[7]);
      if (!mxIsChar(prhs[6])) mxerr(usageStr);
      if (mxGetM(prhs[6]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[6]) + 1;
      rsp = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[6],rsp,len);
      if (!mxIsChar(prhs[5])) mxerr(usageStr);
      if (mxGetM(prhs[5]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[5]) + 1;
      hdw = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[5],hdw,len);
      if (!mxIsChar(prhs[4])) mxerr(usageStr);
      if (mxGetM(prhs[4]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[4]) + 1;
      loc = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[4],loc,len);
      if (!mxIsChar(prhs[3])) mxerr(usageStr);
      if (mxGetM(prhs[3]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[3]) + 1;
      dsc = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[3],dsc,len);
      if (!mxIsChar(prhs[2])) mxerr(usageStr);
      if (mxGetM(prhs[2]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[2]) + 1;
      ctx = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[2],ctx,len);
      if (!mxIsChar(prhs[1])) mxerr(usageStr);
      if (mxGetM(prhs[1]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[1]) + 1;
      sub = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[1],sub,len);
    case 1: // fec name
      if (!mxIsChar(prhs[0])) mxerr(usageStr);
      if (mxGetM(prhs[0]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[0]) + 1;
      fec = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[0],fec,len);
      break;
    default:
      mxerr(usageStr);
  }   

  err = RegisterFecInformation(fec,sub,ctx,dsc,loc,hdw,rsp,poff);

  if (err != 0)
  {
    GetLastLinkError(err,stsbuf);
    sprintf(buf,"tine_register_fec : <%d> %.64s",err,stsbuf);
    mexErrMsgTxt(buf);
  }
  data[0] = mxCreateDoubleMatrix(1,1,mxREAL);
  lp = (double *)mxGetPr(data[0]);
  *lp = (double)err;
  plhs[0] = data[0];
  
err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
  if (fec != NULL) mxFree(fec);
  if (sub != NULL) mxFree(sub);
  if (ctx != NULL) mxFree(ctx);
  if (dsc != NULL) mxFree(dsc);
  if (loc != NULL) mxFree(loc);
  if (hdw != NULL) mxFree(hdw);
  if (rsp != NULL) mxFree(rsp);
}
