/* tine_register_type.cpp   register a tagged structure */

#include <math.h>
#include "mex.h"
#include "tine.h"
#include "tbufsrv.h"

#define mxerr(s) { mxerrStr = (char *)s; goto err; }

#define usageStr "tine_register_type usage:\n ret=tine_register_type('name_tag',values);\n where 'name_tag' is the structure tag associated with the type and values = MatLab structure array representing the data type\n'ret' contains the status/error message"
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) 
{
  char Buffer[64];
  char *typestr, *mxerrStr=NULL;
  char dtstr[32],dtag[32];
  int	len,size=0;
  int	i,nf,fmt,err,siz,off,add;
  const char *fld;
  mxArray *sa;
 	
  if (plhs == NULL || prhs == NULL) return;
  dtag[0] = 0;

  /* Check for proper number of arguments. */
  switch (nrhs)
  {
    case 2:
      if (!mxIsChar(prhs[0])) mxerr(usageStr);
      if (mxGetM(prhs[0]) != 1) mxerr(usageStr);
      len = mxGetN(prhs[0]) + 1;
      typestr = (char *)mxCalloc(len, sizeof(char));
      mxGetString(prhs[0],typestr,len);
      strncpy(dtag,typestr,TAG_NAME_SIZE);
      mxFree(typestr);
      break;
    default:
      mxerr(usageStr);
  }  
	if (!mxIsStruct(prhs[1])) mxerr("input is not a MatLab structure !");
	size = mxGetM(prhs[1]); // set the size to what was input (num rows) !
  if ((nf=mxGetNumberOfFields(prhs[1])) == 0) mxerr("input values not a structure !");
  off = 0; add = 0;
  for (i=0; i<nf; i++)
  {
    fld = mxGetFieldNameByNumber(prhs[1],i);
    sa = mxGetField(prhs[1],0,fld);
    if (mxIsDouble(sa)) 
    {
      fmt = CF_DOUBLE;
      siz = mxGetM(sa);
      add = sizeof(double) * siz;
    }
    else if (mxIsChar(sa)) 
    {
      fmt = CF_TEXT;
      siz = mxGetN(sa);
      siz = (siz/64)*64 + 64;
      add = siz;
    }
    else
    {
      mxerr("structure field type not implemented");
    }
    AddFieldToStruct(dtag,off,siz,fmt,(char *)fld);
    off += siz*fmtsizeof(fmt);
    add = 0;
  }
  err = SealTaggedStruct(dtag,off,size);
  if (err != 0)
  {
    GetLastLinkError(err,Buffer);
    mxerr("could not register structure !");
  }

	if(err)	plhs[0] = mxCreateString(Buffer);
	else	plhs[0] = mxCreateString("");

err:
  if (mxerrStr != NULL) 
  {
    mexErrMsgTxt(mxerrStr);
  }
}