#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
//#include "tgweqm.h"

#define CMBEQM_TAG "CMBEQM"

#define PRP_POLL_INTERVAL 1001
#define PRP_STOPALL 1002
#define PRP_STARTALL 1003
#define PRP_OPTIONS 1004

#define CB_USE_LOCALTIME 0x0001
#define CB_USE_REDIRECTION 0x0002
#define CB_USE_SCHEDULING 0x0004
#define CB_IS_CLONED 0x0008
#define CB_FORWARD 0x0010
#define CB_WRITEONLY 0x0020
#define CB_STATIC 0x0040
#define CB_CONNECT 0x0080
#define CB_APPLYSTALE 0x0100
#define CB_APPLYDEFAULT 0x0200

#define MAX_CAPACITY 2048
char thisFecContext[32] = "TEST";
char thisFecServer[32];
char thisFecName[32];
char thisSubsystem[16] = "MISC";
char thisFecDescription[64];
char thisFecLocation[64] = "unknown";
char thisFecResponsible[32] = "";
int thisFecPort = 0;
int thisFecCapacity = MAX_CAPACITY; // this maximum number of servers + devices to combobulate ...
extern char *gUsageString;

// cover some non-exported routines from the lib ...
short GetElevatedMcaFmt(short fmt)
{
  switch (fmt)
  {
  case CF_INT16:
  case CF_INT32: return CF_INTINT;
  case CF_FLOAT: return CF_FLTINT;
  case CF_DOUBLE: return CF_DBLDBL;
  case CF_NAME16: return CF_NAME16I;
  case CF_NAME32: return CF_NAME32I;
  case CF_NAME64: return CF_NAME64I;
  default: break;
  }
  return fmt;
}
int IsNumFmt(short fmt)
{
  switch (LFMT(fmt))
  {
    case CF_INT16:
    case CF_INT32:
    case CF_INT64:
    case CF_FLOAT:
    case CF_DOUBLE:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
int IsIntFmt(short fmt)
{
  switch (LFMT(fmt))
  {
    case CF_INT16:
    case CF_INT32:
    case CF_INT64:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
UINT32 g_status;
void cbCmb(int id,int cc,void *ref);
void applyValueToDTYPE(DTYPE *d, float value);

void cmbeqm_bkg(void)
{
  ClearAlarm(CMBEQM_TAG,-1);

  //SetAlarm(CMBEQM_TAG,devNr,almCode,almFlag);
}
// target device
typedef struct
{
  char name[64];
  char alias[64];
  int number;
  int disabled;
} TDevs;
TDevs *tdevs = NULL;
// target property
typedef struct TPrp 
{
  char srv[64];
  char prp[64];
  char dev[64];
  char dsc[64];
  int offset;
  int ndevs;
  TDevs *devlst;
  int lid;
  int cc;
  int lsts;
  int useTgtDev;
  short *sts;
  BYTE *data;
  double ts;   // current time stamp
  int delta_t; // current update timestamp difference (this - last) in msec
  int sysstmp; // current system stamp
  int usrstmp; // current user stamp
  int delta_sys; // current update system stamp difference (this - last)
  float shift;
  float scale;
  float defaultValue; // to use on any link error
  float staleValue;
  int mono_sys; // expected monotonic delta_sys stamp (-1 => learn from initial 4 updates)
  int nto; // number of timeouts since startup
  int nmiss; // number of missed callbacks since startup
  XTCBFCNP cb;
  struct TPrp *nxt;
  struct TPrp *prv;
} TPrp;

// combobulated property
typedef struct CPrp 
{
  TPrp *tprps; // the linked list of property items
  int nprps; // length of the TPrp linked list
  int pid; // property id
  int npid; // .NAM property id
  int spid; // .SRC property id
  int qpid; // .CQS property id
  int tmr; // timer interval in ms
  int options;
  int needsScale;
  int staleCnts; // stale counts
  int staleThreshold; // officially stale count threshold
  int errorCnts; // error counts
  int defaultThreshold; // officially error count threshold
  int fmt; // export format
  int len; // acq. data length (standard: 1)
  int atyp; // acq. array type (standard: AT_CHANNEL)
  int fld; // default field index (for compound data types)
  int rpid; // reference id (for cloned entries)
  char prp[64]; // this property name
  char dsc[64]; // this property description
  DTYPE dout;
} CPrp;

CPrp *cprps = NULL;
int ncprps = 0;

typedef struct CCPrpInfo
{
  char prp[64];
  char dsc[64];
  char prp_alias[64]; // this property name
  char fmt_str[16];
  char fmt_exp[16]; // format to export
  char opt_str[128];
  float shift;
  float scale;
  float defaultValue;
  float staleValue;
  int fld; // default field index for compound formats
  int tmr;
  int pid;
} CCPrpInfo;

CSVCOLUMN csvCPrpsFileColumns[] =
{
  { "PROPERTY","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CCPrpInfo,prp),NULL,0 },
  { "PROPERTY_ALIAS","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CCPrpInfo,prp_alias),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CCPrpInfo,dsc),NULL,0 },
  { "FORMAT","",0,CF_TEXT,16,(BYTE *)OFFSETIN(CCPrpInfo,fmt_str),NULL,0 },
  { "FORMAT_EXPORT","",0,CF_TEXT,16,(BYTE *)OFFSETIN(CCPrpInfo,fmt_exp),NULL,0 },
  { "SHIFT","0.0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(CCPrpInfo,shift),NULL,0 },
  { "SCALE","1.0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(CCPrpInfo,scale),NULL,0 },
  { "FIELD_INDEX","-1",0,CF_INT32,1,(BYTE *)OFFSETIN(CCPrpInfo,fld),NULL,0 },
  { "DEFAULT_VALUE","0.0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(CCPrpInfo,defaultValue),NULL,0 },
  { "STALE_VALUE","0.0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(CCPrpInfo,staleValue),NULL,0 },
  { "INTERVAL","1000",0,CF_INT32,1,(BYTE *)OFFSETIN(CCPrpInfo,tmr),NULL,0 },
  { "OPTIONS","0",0,CF_TEXT,128,(BYTE *)OFFSETIN(CCPrpInfo,opt_str),NULL,0 },
};
CSVDB csvCPrpsFileDb =
{
  csvCPrpsFileColumns,
  sizeof(csvCPrpsFileColumns) / sizeof(csvCPrpsFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(CCPrpInfo),
  0
};

typedef struct CPrpInfo 
{
  char srv[64];
  char prp[64];
  char dsc[64];
  char prp_alias[64]; // this property name
  char dev[64];
  char dev_alias[64]; // this device name
  char fmt_str[16];
  char fmt_exp[16]; // format to export
  char opt_str[128];
  int  siz; // capacity
  int  fld; // default field index for compound formats
  int  disabled;
  int  done;
  int  prpLstSiz;
  int  tmr;
  float shift;
  float scale;
  float defaultValue;
  float staleValue;
  CCPrpInfo *prpLst;
} CPrpInfo;

CSVCOLUMN csvTPrpsFileColumns[] = 
{
  { "SERVER","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CPrpInfo,srv),NULL,0 },
  { "PROPERTY","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CPrpInfo,prp),NULL,0 },
  { "PROPERTY_ALIAS","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CPrpInfo,prp_alias),NULL,0 },
  { "DEVICE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CPrpInfo,dev),NULL,0 },
  { "DEVICE_ALIAS","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CPrpInfo,dev_alias),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CPrpInfo,dsc),NULL,0 },
  { "FORMAT","",0,CF_TEXT,16,(BYTE *)OFFSETIN(CPrpInfo,fmt_str),NULL,0 },
  { "FORMAT_EXPORT","",0,CF_TEXT,16,(BYTE *)OFFSETIN(CPrpInfo,fmt_exp),NULL,0 },
  { "CAPACITY","",0,CF_INT32,1,(BYTE *)OFFSETIN(CPrpInfo,siz),NULL,0 },
  { "INTERVAL","1000",0,CF_INT32,1,(BYTE *)OFFSETIN(CPrpInfo,tmr),NULL,0 },
  { "DISABLED","FALSE",0,CF_BOOLEAN,1,(BYTE *)OFFSETIN(CPrpInfo,disabled),NULL,0 },
  { "SHIFT","0.0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(CPrpInfo,shift),NULL,0 },
  { "SCALE","1.0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(CPrpInfo,scale),NULL,0 },
  { "FIELD_INDEX","-1",0,CF_INT32,1,(BYTE *)OFFSETIN(CPrpInfo,fld),NULL,0 },
  { "DEFAULT_VALUE","0.0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(CPrpInfo,defaultValue),NULL,0 },
  { "STALE_VALUE","0.0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(CPrpInfo,staleValue),NULL,0 },
  { "OPTIONS","0",0,CF_TEXT,128,(BYTE *)OFFSETIN(CPrpInfo,opt_str),NULL,0 },
};
CSVDB csvTPrpsFileDb =
{
  csvTPrpsFileColumns,
  sizeof(csvTPrpsFileColumns)/sizeof(csvTPrpsFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(CPrpInfo),
  0
};

CPrpInfo *cprpinfo = NULL;

int isGrouped = FALSE;
int isUseOnError = FALSE;
int gMonitorInterval = 1000;

NAME64 dbuf[MAX_CAPACITY];
int bufsiz = MAX_CAPACITY;
int getCprpMonoSysIncr(char *opts_str)
{
  char *c;
  if ((c = stristr(opts_str, "MONOTONIC")) != NULL)
  {
    int v = 1;
    c += strlen("MONOTONIC");
    if (*c++ == ':')
    {
      double dv = strtod(c, NULL);
      v = (int)(dv + 0.9);
    }
    return v;
  }
  return 0;
}
int getCprpOpts(char *opts_str, int *staleThreshold,int *defaultThreshold)
{
  int opts = 0;
  if (opts_str == NULL) return 0;
  if (!strnicmp(opts_str,"0x",2)) strtol(opts_str, NULL, 16);
  if (opts == 0)
  {
    char *p = NULL;
    char *vs;
    int v;
    strupr(opts_str);
    for (p=strtok(opts_str,"+|"); p != NULL; p=strtok(NULL, "+|"))
    {
      if (strstr(p, "LOCALTIME") != NULL) opts |= CB_USE_LOCALTIME;
      if (strstr(p, "REDIRECT") != NULL) opts |= CB_USE_REDIRECTION;
      if (strstr(p, "SCHEDULE") != NULL) opts |= CB_USE_SCHEDULING;
      if (strstr(p, "CLONE") != NULL) opts |= CB_IS_CLONED; // nominally this shouldn't show up here
      if (strstr(p, "FORWARD") != NULL) opts |= CB_FORWARD; // apply WRITE access
      if (strstr(p, "WRITEONLY") != NULL) opts |= CB_WRITEONLY; // apply WRITE access
      if (strstr(p, "STATIC") != NULL) opts |= CB_STATIC; // acquired info is static: close link after success
      if (strstr(p, "CONNECT") != NULL) opts |= CB_CONNECT; // use TCP (connected socket) communication
      if (strstr(p, "USESTALE") != NULL)
      { // apply given stale value if an update is stale
        v = 0;
        opts |= CB_APPLYSTALE; 
        if ((vs = strchr(p, '@')) != NULL)
        {
          vs++;
          v = atoi(vs);
        }
        if (staleThreshold != NULL) *staleThreshold = v;
      }
      if (strstr(p, "USEONERROR") != NULL)
      { // apply given stale value if an update is stale
        v = 0;
        opts |= CB_APPLYDEFAULT;
        if ((vs = strchr(p, '@')) != NULL)
        {
          vs++;
          v = atoi(vs);
        }
        if (defaultThreshold != NULL) *defaultThreshold = v;
      }
    }
  }
  return opts;
}
// LockLinkData(), UnlockLinkData() in EQM -> apply scale and offset there !
void cbCmb(int id,int cc,void *ref)
{
  TPrp *tp = (TPrp *)ref;
  double ts;
  int sysstmp = 0, usrstmp = 0;
  int useLocalTimestamp = FALSE;
  int scheduleData = FALSE;
  int isStale = FALSE;
  int applyDefault = FALSE;
  if (tp == NULL) return;
  if (id < 0 || id >= ncprps) return;
  if (tp->lid < 0) return;
  if (cprps[id].options & CB_USE_LOCALTIME) useLocalTimestamp = TRUE;
  if (cprps[id].options & CB_USE_SCHEDULING) scheduleData = TRUE;
  if (cc != 0)
  {
    useLocalTimestamp = TRUE;
    cprps[id].errorCnts++;
    if (cprps[id].errorCnts > cprps[id].defaultThreshold)
    {
      if (cprps[id].options & CB_APPLYDEFAULT) applyDefault = TRUE;
    }
  }
  tp->cc = cc;  // need to keep track of this status and last status ?
  tp->lsts = cc;
  if (useLocalTimestamp)
  {
    ts = MakeDataTimeStamp();
    sysstmp = MakeSystemDataStamp();
  }
  else
  {
    ts = GetCurrentDataTimeStamp(tp->lid);
    sysstmp = GetSystemDataStamp(tp->lid);
    if (sysstmp == 0) sysstmp = MakeSystemDataStamp(); // override
    usrstmp = GetDataStamp(tp->lid);
    if (sysstmp < 0) sysstmp = 0;
  }
  if (cc == link_timeout) tp->nto++;
  if (cc == 0)
  { // only keep system stamp stats on the successes
    cprps[id].errorCnts = 0;
    if (tp->ts == 0) tp->ts = ts;
    if (tp->sysstmp == 0) tp->sysstmp = sysstmp;
    tp->delta_t = (int)((ts - tp->ts)*1000.0);
    tp->delta_sys = sysstmp - tp->sysstmp;
    if ((tp->mono_sys > 0 && tp->delta_sys > tp->mono_sys) ||
      tp->delta_t > cprps[id].tmr * 2)
    {
      tp->nmiss++;
      if (cprps[id].options & CB_APPLYSTALE) isStale = TRUE;
    }
    tp->sysstmp = sysstmp;
    tp->usrstmp = usrstmp;
    tp->ts = ts;
    if (cprps[id].options & CB_STATIC) CloseLink(tp->lid);
  }
  cprps[id].staleCnts = isStale ? cprps[id].staleCnts + 1 : 0;
  if (cc == 0 || applyDefault)
  {
    if (ts > cprps[id].dout.dTimeStamp) cprps[id].dout.dTimeStamp = ts;
    if (sysstmp > cprps[id].dout.sysStamp) cprps[id].dout.sysStamp = sysstmp;
    cprps[id].dout.dStamp = usrstmp;
  }
  if (cprps[id].staleCnts > cprps[id].staleThreshold)
  {
    DTYPE d;
    d.dArrayLength = cprps[id].len * tp->ndevs;
    d.dFormat = cprps[id].dout.dFormat;
    d.data.bptr = tp->data;
    applyValueToDTYPE(&d, tp->staleValue);
  }
  if (applyDefault)
  {
    DTYPE d;
    d.dArrayLength = cprps[id].len * tp->ndevs;
    d.dFormat = cprps[id].dout.dFormat;
    d.data.bptr = tp->data;
    applyValueToDTYPE(&d, tp->defaultValue);
  }
  if (scheduleData) SystemScheduleProperty(CMBEQM_TAG, cprps[id].prp);
}
CSVCOLUMN csvTDevsFileColumns[] = 
{
  { "DEVICE_NAME","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TDevs,name),NULL,0 },
  { "DEVICE_ALIAS","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TDevs,alias),NULL,0 },
  { "DEVICE_NUMBER","",0,CF_INT32,1,(BYTE *)OFFSETIN(TDevs,number),NULL,0 },
//  { "DISABLED","",0,CF_BOOLEAN,1,(BYTE *)OFFSETIN(TDevs,disabled),NULL,0 },
};
CSVDB csvTDevsFileDb =
{
  csvTDevsFileColumns,
  sizeof(csvTDevsFileColumns)/sizeof(csvTDevsFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(TDevs),
  0
};

char *makeTgtDevName(char *srv,char *dev)
{
  static char tgt[128];
  if (srv[0] == '/')
    sprintf(tgt, "%.64s/%.64s", srv, dev);
  else
    sprintf(tgt, "/%.32s/%.32s/%.64s", thisFecContext, srv, dev);
  return tgt;
}

#define punt(t,c) { tineDebug=1; feclog("%s : %.32s",t,cc2str(c)); exit(1); }
int assignDevListToPrp(TPrp *prp, CPrpInfo *pinf)
{ // read file, acq list, or take as is
  TDevs *devlst = NULL;
  int i, cc = 0, cap = 0, isWldCrd = FALSE, isDevGrp = FALSE;
  char cpath[256], fn[64], tgt[132], *ctx = thisFecContext, *srv, *dev, *c;
  char tgtdev[DEVICE_NAME_SIZE+4], *tqdev = "#0";
  if (prp == NULL || pinf == NULL) return -argument_list_error;
  strncpy(tgt, prp->srv, 64); tgt[64] = 0;
  if (tgt[0] == '/') tokenizeFullName(tgt, &ctx, &srv, &dev); else srv = prp->srv;
  sprintf(cpath, "%s%c%.32s%c%.32s%c%.64s%c", 
    SystemGetStartupDirectory(), FS_DELIMITER, ctx, FS_DELIMITER, srv, FS_DELIMITER, prp->prp, FS_DELIMITER);
  fn[0] = 0;
  strncpy(tgtdev, pinf->dev, DEVICE_NAME_SIZE);
  isWldCrd = strchr(tgtdev, '*') != NULL;
  if (!strncmp(tgtdev, "DEVGRP", 6)) isWldCrd = isDevGrp = TRUE;
  if ((c=strchr(tgtdev, '[')) != NULL)
  {
    *c++ = 0;
    cap = atoi(c);
    isWldCrd = isDevGrp = TRUE;
    tqdev = tgtdev;
  }
  if (isWldCrd || isDevGrp || strlen(tgtdev) == 0)
  { // acquire names from server
    bufsiz = 1024;
    if (isWldCrd && strlen(tgtdev) > 1 && pinf->siz > 1)
    { // wildcard distinguished device list ...
      strncpy(prp->dev, tgtdev, 64);
      prp->useTgtDev = TRUE;
    }
    if ((cc=GetSystemDevices(makeTgtDevName(prp->srv, tqdev), prp->prp, dbuf, &bufsiz)) == 0)
    {
      int k, nlen, elen = 0, contains=TRUE, matches;
      char strc[72], *s, *beginTok = NULL, *midTok = NULL, *endTok = NULL;
      char strpp[72], *pre = NULL, *post = NULL;
      if (strchr(pinf->dev_alias, '*') != NULL && strlen(pinf->dev_alias) > 1)
      { // device alias decorations: any other dev_alias will revert to using the acquired dev list !
        strncpy(strpp, pinf->dev_alias, 64); strpp[64] = 0; s = strpp;
        if (*s == '*')
        { // no beginToken -> only a post fix
          while (*s == '*') s++;  // just in case ...
          if ((c = strchr(s, '*')) != NULL) *c = 0;  // just in case ...
          post = s;
        }
        else
        { // start with a pre fix
          pre = strpp;
          if ((s = strchr(strpp, '*')) != NULL)
          { // and there's a post fix ...
            *s++ = 0;
            while (*s == '*') s++; // just in case ...
            if ((c = strchr(s, '*')) != NULL) *c = 0; // just in case ...
            post = s;
          }
        }
        if (pre != NULL && strlen(pre) == 0) pre = NULL;
        if (post != NULL && strlen(post) == 0) post = NULL;
      }
      if (strlen(tgtdev) > 1  && !isDevGrp)
      { // targeted devices ...
        strncpy(strc, tgtdev, 64); strc[64] = 0; s = strc;
        if (*s == '-')
        { // starts with a minus: skip those devices that match ...
          contains = FALSE;
          s++;
        }
        if (*s == '*')
        { // no beginToken
          while (*s == '*') s++;
          if ((c = strchr(s, '*')) != NULL)
          { /* there's a midToken ("*abc*") */
            *c = 0; midTok = s;
          }
          else
          { /* there's an end token */
            endTok = s;
          }
        }
        else
        {
          beginTok = s;
          if ((c = strchr(beginTok, '*')) != NULL)
          { /* found begin Token end */
            *c++ = 0;
            s = c; /* re-position (could still be an end token */
            while (*s == '*') s++; /* is it "abc*xyz" ? */
            if (strlen(s) > 0) endTok = s;
          }
        }
        if (endTok != NULL) elen = (int)strlen(endTok);
      }
      devlst = (TDevs *)calloc(bufsiz, sizeof(TDevs));
      if (devlst == NULL) punt("assignDevListToPrp", out_of_local_memory);
      for (i = 0, k = 0; i < bufsiz; i++)
      {
        nlen = MIN((int)strlen(dbuf[i].name),64);
        matches = TRUE;
        if (beginTok != NULL && strnicmp(dbuf[i].name, beginTok, strlen(beginTok))) matches = FALSE;
        if (midTok != NULL && strstr(dbuf[i].name, midTok) == NULL) matches = FALSE;
        if (endTok != NULL && strnicmp(&dbuf[i].name[nlen - elen], endTok, elen)) matches = FALSE;
        if (matches && !contains) continue;
        if (!matches && contains) continue;
        strncpy(devlst[k].name, dbuf[i].name, 64);
        // now prepare the alias ...
        if (pre == NULL && post == NULL)
        { // no pre or post fix instructions: take it as is
          strncpy(devlst[k].alias, dbuf[i].name, 64);
        }
        else
        {
          char *dn = dbuf[i].name;
          char *ppre = pre == NULL ? "" : pre;
          char *ppost = post == NULL ? "" : post;
          if (pre != NULL && *pre == '-')
          { // instruction to remove the prefix from the targeted name
            int plen = (int)strlen(pre) - 1;
            if (!strnicmp(dn, &pre[1], plen)) dn += plen;
            ppre = "";
          }
          if (post != NULL && *post == '-')
          { // instruction to remove the prefix from the targeted name
            int plen = (int)strlen(post) - 1;
            if ((c = strstr(dn, &post[1])) != NULL) *c = 0;
            ppost = "";
          }
          snprintf(devlst[k].alias, 64, "%s%.64s%s", ppre, dn, ppost);
        }
        devlst[k].number = k;
        k++;
        if (cap > 0 && k >= cap) break;
      }
      prp->ndevs = k;
      prp->devlst = devlst;
      // write to cache file
      MakeDirectoryPath(cpath, TRUE);
#     ifndef USE_CSVWRITEFILE
      {
        char fn[256];
        FILE *fp = NULL;
        sprintf(fn, "%.132sdevices.csv", cpath);
        if ((fp = fopen(fn, "w")) != NULL)
        {
          fprintf(fp, "DEVICE_NAME, DEVICE_ALIAS, DEVICE_NUMBER\n");
          for (i = 0; i < prp->ndevs; i++)
            fprintf(fp, "%.64s, %.64s, %d\n", devlst[i].name, devlst[i].alias, i);
          fclose(fp);
        }
      }
#     else
      csvTDevsFileDb.siz = prp->ndevs;
      csvTDevsFileDb.tgt = (BYTE *)prp->devlst;
      csvTDevsFileDb.tgt_flgs = CSVDB_INITIALIZED | CSVDB_PREALLOC;
      csvStdInit(&csvTDevsFileDb);
      csvWriteFile(cpath, "devices.csv", &csvTDevsFileDb);
#     endif
      ccerr(0); // we're done !
    }
    else
    { // error contacting server (read from cache file)!
      strcpy(fn, "devices.csv");
    }
  }
  if (strstr(tgtdev, ".csv") != NULL)
  { // get device list explicitly from file
    if (!strcmp(tgtdev, ".csv")) strcpy(fn, "devices.csv"); else strncpy(fn, tgtdev, 64);
  }
  if (strlen(fn) > 0)
  {
rdDevListFile:
    csvTDevsFileDb.siz = 0; csvTDevsFileDb.tgt = NULL;
    cc = csvReadFile(cpath, fn, &csvTDevsFileDb, (void **)&prp->devlst);
    if (cc == 0)
    {
      prp->ndevs = csvTDevsFileDb.siz;
    }
    else
    {
      feclog("could not read devices file %.64s at location %.256s : %.32s", fn, cpath, cc2str(cc));
      if (strlen(cpath) > 0)
      {
        feclog("try working directory");
        cpath[0] = 0;
        goto rdDevListFile;
      }
    }
    ccerr(cc);
  }
  // device name and alias given directly in config file ...
  devlst = (TDevs *)calloc(1, sizeof(TDevs));
  if (devlst == NULL) punt("assignDevListToPrp", out_of_local_memory);
  strncpy(devlst[0].name, tgtdev, 64);
  if (strlen(pinf->dev_alias) == 0) strncpy(pinf->dev_alias, tgtdev, 64);
  strncpy(devlst[0].alias, pinf->dev_alias, 64);
  devlst[0].disabled = pinf->disabled;  // flag as disabled if declared so here ...
  devlst[0].number = 1;
  prp->devlst = devlst;
  prp->ndevs = 1;
err:
  return cc ? -cc : prp->ndevs;
}
void applyValueToDTYPE(DTYPE *d, float value)
{
  if (d == NULL) return;
  switch (d->dFormat)
  {
    case CF_IMAGE:
    { // nothing to do with the value, but get the frame buffer right!
      DIMAGE *img = (DIMAGE *)d->data.vptr;
      img->frameBuffer = d->data.bptr + sizeof(hdrIMAGE) + 2 * sizeof(size_t);
      img->frameBufferSize = d->dArrayLength;
      break;
    }
    case CF_SPECTRUM:
    {
      USPECTRUM *sp = (USPECTRUM *)d->data.vptr;
      sp->status = (int)value;
      sp->s_inc = value;
      sp->s_start = value;
      sp->tm = (int)value;
      break;
    }
    case CF_FLTINT:
    {
      FLTINT *v = (FLTINT *)d->data.vptr;
      v->fval = value;
      v->ival = (int)value;
      break;
    }
    case CF_INTINT:
    {
      INTINT *v = (INTINT *)d->data.vptr;
      v->i1val = (int)value;
      v->i2val = (int)value;
      break;
    }
    case CF_FLTFLT:
    {
      FLTFLT *v = (FLTFLT *)d->data.vptr;
      v->f1val = value;
      v->f2val = value;
      break;
    }
    case CF_DBLDBL:
    {
      DBLDBL *v = (DBLDBL *)d->data.vptr;
      v->d1val = value;
      v->d2val = (int)value;
      break;
    }
    case CF_INTINTINTINT:
    {
      DADDRESS *v = (DADDRESS *)d->data.vptr;
      v->i1val = (int)value;
      v->i2val = (int)value;
      v->i3val = (int)value;
      v->i4val = (int)value;
      break;
    }
    default:
      if (IsNumFmt(d->dFormat))
      { // this'll work for the 1st value ...
        PutValuesFromFloatEx(d, &value, 1, 0);
        if (d->dArrayLength > 1)
        {
          int i;
          int fsiz = GetFormatSize(d->dFormat);
          for (i = 1; i < (int)d->dArrayLength; i++)
          {
            memcpy(&d->data.bptr[i*fsiz], &d->data.bptr[0], fsiz);
          }
        }
      }
      break;
  }
}
char *getTgtDevice(CPrp *cprp, TPrp *tprp)
{
  if (tprp->devlst[0].name[0] != '$')
  {
    if (tprp->useTgtDev) return tprp->dev;
    return tprp->devlst[0].name;
  }
  if (!strnicmp(tprp->devlst[0].name, "$PRP", 4)) return cprp->prp;
  if (!strnicmp(tprp->devlst[0].name, "$PROP", 5)) return cprp->prp;
  if (!strnicmp(tprp->devlst[0].name, "$DSC", 4)) return cprp->dsc;
  if (!strnicmp(tprp->devlst[0].name, "$DESC", 5)) return cprp->dsc;
  return tprp->devlst[0].name;
}

int nTotalLinks = 0;
#define LINKTABLE_SIZE 2048
int gLinkTableSize = LINKTABLE_SIZE;
#define CMBCFG_FILENAME "combobulate.csv"
char cmbCfgFile[64] = CMBCFG_FILENAME;
int isSrvClone(char *srv)
{
  if (!stricmp(srv, "this")) return TRUE;
  if (!stricmp(srv, "me")) return TRUE;
  if (!stricmp(srv, "self")) return TRUE;
  return FALSE;
}
int isMcaType(int arrayType)
{
  arrayType &= ~AT_DOUBLE;
  switch (arrayType)
  {
    case AT_SPECTRUM:
    case AT_IMAGE:
      return FALSE;
    default :
      return TRUE;
  }
}
void cmbeqm_ini(void)
{
  char prp[80] = "", aprp[80] = "";
  char wd[128];
  int i, j, k=-1, n, nt, p = 0, np = -1, cc = 0, m, efmt, pid;
  DTYPE dout,din,dmca;
  int keep_scanning = FALSE, use_local = FALSE, mode = CM_TIMER, acc = CA_READ;
  int isText = FALSE, isImage = FALSE, isSpectrum = FALSE;
  int use_alias_devices = FALSE, has_devices_from_target = FALSE, tgt_composite = FALSE;
  int fsiz, dsiz, ndevs, offset;
  TPrp *tprp, *nprp;
  
  char *tdevnam = "";

  DTYPEZERO(dout); DTYPEZERO(dmca);
  nTotalLinks = 0;
//  if (isUseOnError) mode |= CM_USEONERROR;

  snprintf(wd, 128, "%s%c", SystemGetStartupDirectory(), FS_DELIMITER);
  if ((cc=csvReadFile(wd, cmbCfgFile,&csvTPrpsFileDb,(void **)&cprpinfo)) != 0)
  {
    tineDebug = 1;
    feclog("cannot read gateway properties file %s! : %.32s",cmbCfgFile,cc2str(cc));
    exit(1);
  }
  n = csvTPrpsFileDb.siz; nt = 0;
  // quick check for property files ...
  for (i = 0; i < n; i++)
  {
    if (strstr(cprpinfo[i].prp, ".csv") != NULL && strlen(cprpinfo[i].prp) > 4)
    { // properties given in a list (and not as is)
      if (strlen(cprpinfo[i].prp_alias) == 0) strncpy(cprpinfo[i].prp_alias, cprpinfo[i].prp, 64);
      if (strnicmp(cprpinfo[i].prp, prp, 64) == 0)
      { // same file (and already read in) -> copy the reference !
        cprpinfo[i].prpLst = (CCPrpInfo *)csvCPrpsFileDb.tgt; 
        cprpinfo[i].prpLstSiz = csvCPrpsFileDb.siz;
        nt += cprpinfo[i].prpLstSiz;
        continue;
      }
      strncpy(prp, cprpinfo[i].prp, 64);
      csvCPrpsFileDb.siz = 0; csvCPrpsFileDb.tgt = NULL;
      if ((cc=csvReadFile(wd, prp, &csvCPrpsFileDb, (void **)&cprpinfo[i].prpLst)) != 0)
      {
        feclog("cannot read gateway properties list file %s%.64s ! : %.32s", wd, prp, cc2str(cc));
        exit(1);
      }
      cprpinfo[i].prpLstSiz = csvCPrpsFileDb.siz;
      for (j = 0; j < csvCPrpsFileDb.siz; j++)
      { // check for empty important entries ...
        if (strlen(cprpinfo[i].prpLst[j].prp_alias) == 0) 
          strncpy(cprpinfo[i].prpLst[j].prp_alias, cprpinfo[i].prpLst[j].prp, 64);
        if (strlen(cprpinfo[i].prpLst[j].fmt_exp) == 0)
          strncpy(cprpinfo[i].prpLst[j].fmt_exp, cprpinfo[i].prpLst[j].fmt_str, 16);
      }
    }
    else
    { // single property given as is ...
      cprpinfo[i].prpLst = (CCPrpInfo *)SystemCalloc(1, sizeof(CCPrpInfo));
      if (cprpinfo[i].prpLst == NULL)
      {
        feclog("cannot allocate memory for %.64s !", cprpinfo[i].prp);
        exit(1);
      }
      strncpy(cprpinfo[i].prpLst[0].prp, cprpinfo[i].prp, 64);
      if (strlen(cprpinfo[i].prp_alias) == 0) strncpy(cprpinfo[i].prp_alias, cprpinfo[i].prp, 64);
      strncpy(cprpinfo[i].prpLst[0].prp_alias, cprpinfo[i].prp_alias, 64);
      strncpy(cprpinfo[i].prpLst[0].dsc, cprpinfo[i].dsc, 64);
      strncpy(cprpinfo[i].prpLst[0].fmt_str, cprpinfo[i].fmt_str, 16);
      if (strlen(cprpinfo[i].fmt_exp) == 0) 
        strncpy(cprpinfo[i].fmt_exp, cprpinfo[i].fmt_str, 16);
      strncpy(cprpinfo[i].prpLst[0].fmt_exp, cprpinfo[i].fmt_exp, 16);
      cprpinfo[i].prpLst[0].defaultValue = cprpinfo[i].defaultValue;
      cprpinfo[i].prpLst[0].staleValue = cprpinfo[i].staleValue;
      if (cprpinfo[i].scale == 0) cprpinfo[i].scale = 1.0; // double check !
      cprpinfo[i].prpLst[0].scale = cprpinfo[i].scale;
      cprpinfo[i].prpLst[0].shift = cprpinfo[i].shift;
      cprpinfo[i].prpLst[0].fld = cprpinfo[i].fld;
      cprpinfo[i].prpLst[0].tmr = cprpinfo[i].tmr;
      strncpy(cprpinfo[i].prpLst[0].opt_str,cprpinfo[i].opt_str,128);
      cprpinfo[i].prpLstSiz = 1;
      if (strlen(cprpinfo[i].dev_alias) == 0)
        strncpy(cprpinfo[i].dev_alias, cprpinfo[i].dev, 64);
    }
    nt += cprpinfo[i].prpLstSiz; // augment the worst case total number ...
  }
  cprps = (CPrp *)calloc(nt,sizeof(CPrp)); // worst case total number of properties
  if (cprps == NULL) punt("cmbeqm_ini",out_of_local_memory);
  // line up entries per 'property alias' (the combobulated export property)
  while (TRUE)
  {
    keep_scanning = FALSE;
    for (i = p; i < n; i++)
    {
      if (cprpinfo[i].done) continue;
      keep_scanning = TRUE;
      if (strnicmp(cprpinfo[i].prp_alias, cprpinfo[p].prp_alias, 64))
      { // find the next alias property to check for
        if (np < 0) np = i; // peg the index
        if (isSrvClone(cprpinfo[i].srv))
        { // special cloning case ...
          int cp;
          for (cp = 0; cp < k+1; cp++)
          {
            if (strnicmp(cprps[cp].prp, cprpinfo[i].prp, 64) == 0)
            { // target property is an already formed alias (export) property
              k++;
              cprps[k].pid = cp;
              cprps[k].dout = cprps[cp].dout;
              cprps[k].nprps = cprps[cp].nprps;
              cprps[k].tprps = cprps[cp].tprps;
              cprps[k].len = cprps[cp].len;
              cprps[k].atyp = cprps[cp].atyp;
              strncpy(cprps[k].prp, cprpinfo[i].prp_alias, 64);
              strncpy(cprps[k].dsc, cprpinfo[i].prpLst[0].dsc, 64);
              cprps[k].fmt = GetFormatFromString(cprpinfo[i].prpLst[0].fmt_exp);
              cprps[k].fld = cprpinfo[i].prpLst[0].fld;
              cprps[k].options = CB_IS_CLONED;
              cprps[k].rpid = cp;
              cprpinfo[i].done = TRUE;
              if (np == i) np = -1;
            }
          }
        }
      }
      else
      { // latched another to target (or the first one)
        int seglen; // segment length for this target (number of elements)
        cprpinfo[i].done = TRUE;
        for (j = 0; j < cprpinfo[p].prpLstSiz; j++)
        { // work thru the property list (often a list of 1)
          if (i == p)
          { // the first one!
            k++; // k = last property, so: k + 1 and loop over range
            cprps[k].pid = k;
            cprpinfo[i].prpLst[j].pid = k;
            strncpy(cprps[k].prp, cprpinfo[i].prpLst[j].prp_alias, 64);
            DTYPEZERO(cprps[k].dout);
            cprps[k].dout.dFormat = GetFormatFromString(cprpinfo[i].prpLst[j].fmt_str);
            cprps[k].atyp = strchr(cprpinfo[i].prpLst[j].fmt_str,'.') != NULL ? 
              GetArrayTypeFromString(cprpinfo[i].prpLst[j].fmt_str) : AT_CHANNEL;
            if (cprps[k].dout.dFormat == CF_IMAGE) cprps[k].atyp = AT_IMAGE;
            isText = cprps[k].dout.dFormat == CF_TEXT ? TRUE : FALSE;
            isImage = cprps[k].dout.dFormat == CF_IMAGE ? TRUE : FALSE;
            isSpectrum = cprps[k].dout.dFormat == CF_SPECTRUM ? TRUE : FALSE;
            switch (cprps[k].atyp)
            {
              default:
                cprps[k].atyp = AT_CHANNEL; // the normal default situation
                cprps[k].len = 1;
                if (isText)
                {
                  char *c = strchr(cprpinfo[i].prpLst[j].fmt_str, '.');
                  if (c != NULL) cprps[k].len = atoi(++c);
                }
                break;
              case AT_SPECTRUM:
                if ((cprps[k].options & CB_USE_REDIRECTION) == 0) cprps[k].atyp |= AT_DOUBLE;
              case AT_IMAGE:
                cprps[k].len = cprpinfo[i].siz;
                break;
            }
            strncpy(cprps[k].dsc, cprpinfo[i].prpLst[j].dsc, 64);
            cprps[k].fmt = GetFormatFromString(cprpinfo[i].prpLst[j].fmt_exp);
            cprps[k].fld = cprpinfo[i].prpLst[j].fld;
            if (isText && cprps[k].fmt == CF_TEXT) cprps[k].atyp = AT_UNKNOWN;
          }
          else
          { // additional device entries with the same property (properties)
            k = cprpinfo[p].prpLst[j].pid;
            isText = cprps[k].dout.dFormat == CF_TEXT ? TRUE : FALSE;
            isImage = cprps[k].dout.dFormat == CF_IMAGE ? TRUE : FALSE;
            isSpectrum = cprps[k].dout.dFormat == CF_SPECTRUM ? TRUE : FALSE;
          }
          tprp = (TPrp *)calloc(1, sizeof(TPrp));
          if (tprp == NULL) punt("cmbeqm_ini", out_of_local_memory);
          strncpy(tprp->srv, cprpinfo[i].srv, 64);
          strncpy(tprp->prp, cprpinfo[i].prpLst[j].prp, 64);
          strncpy(tprp->dsc, cprpinfo[i].prpLst[j].dsc, 64);
          tprp->shift = cprpinfo[i].prpLst[j].shift;
          tprp->scale = cprpinfo[i].prpLst[j].scale;
          tprp->defaultValue = cprpinfo[i].prpLst[j].defaultValue;
          tprp->staleValue = cprpinfo[i].prpLst[j].staleValue;
          tprp->mono_sys = getCprpMonoSysIncr(cprpinfo[i].prpLst[j].opt_str);
          tprp->ndevs = assignDevListToPrp(tprp, &cprpinfo[i]); // read file, acq list, or take as is
          if (cprpinfo[i].siz > 1 && cprpinfo[i].siz < tprp->ndevs) 
            tprp->ndevs = cprpinfo[i].siz;
          tprp->nxt = cprps[k].tprps;
          if (cprps[k].tprps != NULL) cprps[k].tprps->prv = tprp;
          cprps[k].tprps = tprp;
          seglen = cprps[k].len * tprp->ndevs; // normal segment length
          if (isText) seglen = cprps[k].atyp == AT_CHANNEL ? tprp->ndevs : cprps[k].len;
          if (isImage)
          { // massage the number of elements
            seglen += IMAGE_HDR_SIZE + 2 * sizeof(size_t);
          }
          if (isSpectrum)
          { // massage the number of elements 
            seglen += SPECTRUM_HDR_SIZE / sizeof(float);
          }
          cprps[k].dout.dArrayLength += seglen;
          cprps[k].tmr = cprpinfo[i].prpLst[j].tmr;
          cprps[k].options |= getCprpOpts(cprpinfo[i].prpLst[j].opt_str,&cprps[k].staleThreshold,&cprps[k].defaultThreshold);
          if (tprp->defaultValue != 0) cprps[k].options |= CB_APPLYDEFAULT;
          if (isUseOnError) cprps[k].options |= CB_APPLYDEFAULT;
          cprps[k].nprps++; // number of accessed target properties (links)
        }
      }
    }
    if (np < 0) break;
    if (!keep_scanning) break;
    p = np; np = -1; // reset things for the next scan
  }
  ncprps = k + 1;
  for (i=0; i<ncprps; i++)
  {
    ndevs = 0; tprp = NULL;
    if (cprps[i].options & (CB_USE_REDIRECTION|CB_IS_CLONED|CB_WRITEONLY))
    { // cut to the chase if we're redirecting ...
      for (nprp = cprps[i].tprps; nprp != NULL; nprp = nprp->nxt) ndevs++;
      if (cprps[i].options & CB_IS_CLONED)
      { // pass the data allocation pointer along just in case ...
        int cp = cprps[i].pid;
        cprps[i].dout.data.bptr = cprps[cp].dout.data.bptr;
        if (cprps[i].fld != cprps[cp].fld)
        { // different field: need a distinct property id
          cprps[i].pid = i;
        }
      }
      goto register_combobulator_property;
    }
    m = mode;
    if (cprps[i].options & CB_CONNECT) m |= CM_CONNECT;
    fsiz = fmtsizeof(cprps[i].dout.dFormat);
    //hsiz = getFormatHeaderSize(cprps[i].dout.dFormat);
    // assume header finagled into the effective num elements
    isText = cprps[i].dout.dFormat == CF_TEXT ? TRUE : FALSE;
    isImage = cprps[i].dout.dFormat == CF_IMAGE ? TRUE : FALSE;
    isSpectrum = cprps[i].dout.dFormat == CF_SPECTRUM ? TRUE : FALSE;
    if (isText) fsiz *= cprps[i].len;
    // n.b. formats with headers will not be combobulating an existing MCA !
    dsiz = cprps[i].dout.dArrayLength * fsiz;
    cprps[i].dout.data.bptr = (BYTE *)calloc(dsiz,1); // single block with all memory for this property
    if (cprps[i].dout.data.bptr == NULL)
    {
      feclog("cannot allocate %d elements of size %d", cprps[i].dout.dArrayLength, fsiz);
      punt("cmbeqm_ini", out_of_local_memory);
    }
    for (nprp = cprps[i].tprps; nprp != NULL; nprp = nprp->nxt) tprp = nprp; // find the end
    while (tprp != NULL)
    { // use segments of DTYPE in the tprps list ...
      offset = isText ? cprps[i].len * ndevs : cprps[i].len * ndevs * fsiz;
      if (isImage) offset = (cprps[i].len + IMAGE_HDR_SIZE + 2 * sizeof(size_t)) * ndevs;
      if (isSpectrum) offset = (cprps[i].len + SPECTRUM_HDR_SIZE / sizeof(float)) * ndevs;
      tprp->data = cprps[i].dout.data.bptr + offset;
      dout.dArrayLength = cprps[i].len * tprp->ndevs;
      dout.dFormat = cprps[i].dout.dFormat;
      dout.data.bptr = tprp->data;
      applyValueToDTYPE(&dout, tprp->defaultValue);

      if (tprp->devlst != NULL && !tprp->devlst[0].disabled)
      {
        tdevnam = getTgtDevice(&cprps[i], tprp);
        tprp->lid = AttachLinkEx2(makeTgtDevName(tprp->srv, tdevnam), tprp->prp, &dout, NULL, CA_READ, cprps[i].tmr, cbCmb, m, i, tprp);
      }
      else
      {
        tprp->lsts = not_running;
      }
      tprp->offset = ndevs;
      ndevs += tprp->ndevs;
      if (tprp->scale != 1 || tprp->shift != 0) cprps[i].needsScale = TRUE;
      tprp = tprp->prv; // backwards thru the list => forward ordered
      nTotalLinks++;
      if (nTotalLinks > gLinkTableSize - 100)
      { // realloc ...
        gLinkTableSize += 1024;
        SetConnectionTableCapacity(gLinkTableSize);
      }
    }
register_combobulator_property:
    efmt = GetElevatedMcaFmt(cprps[i].fmt);
    dmca.dArrayLength = cprps[i].dout.dArrayLength;
    if (!isMcaType(cprps[i].atyp))
    { /* SPECTRUM or IMAGE types ? */
      dmca.dArrayLength = cprps[i].len;
      efmt = cprps[i].fmt;
    }
    pid = (cprps[i].options & CB_IS_CLONED) ? cprps[i].pid : i;
    acc = (cprps[i].options & (CB_FORWARD | CB_WRITEONLY)) ? CA_READ | CA_WRITE : CA_READ;
    if (cprps[i].fmt != efmt)
    {
      dmca.dFormat = efmt;
      cprps[i].pid = RegisterPropertyInformation(CMBEQM_TAG, cprps[i].prp, &dmca, NULL, CA_READ, AT_CHANNEL, cprps[i].dout.dArrayLength, cprps[i].dsc, pid, NULL);
    }
    dmca.dFormat = cprps[i].fmt;
    DTYPEZERO(din);
    if (acc & CA_WRITE)
    {
      din.dArrayLength = 1;
      din.dFormat = cprps[i].fmt;
    }
    cprps[i].pid = RegisterPropertyInformation(CMBEQM_TAG, cprps[i].prp, &dmca, &din, acc, cprps[i].atyp, dmca.dArrayLength, cprps[i].dsc, pid, NULL);
    sprintf(prp, "%.60s.NAM", cprps[i].prp);
    dmca.dFormat = CF_NAME64;
    dmca.dArrayLength = !isMcaType(cprps[i].atyp) ? ndevs : cprps[i].dout.dArrayLength;
    cprps[i].npid = RegisterPropertyInformation(CMBEQM_TAG, prp, &dmca, NULL, CA_READ, AT_CHANNEL, dmca.dArrayLength, "channel names", ncprps+ pid, NULL);
    sprintf(aprp, "%.56s.SRC.NAM", cprps[i].prp);
    RegisterPropertyAlias(CMBEQM_TAG, prp, aprp);
    sprintf(aprp, "%.56s.CQS.NAM", cprps[i].prp);
    RegisterPropertyAlias(CMBEQM_TAG, prp, aprp);
    sprintf(prp, "%.60s.SRC", cprps[i].prp);
    cprps[i].spid = RegisterPropertyInformation(CMBEQM_TAG, prp, &dmca, NULL, CA_READ, AT_CHANNEL, dmca.dArrayLength, "channel sources", 2*ncprps + pid, NULL);
    if ((cprps[i].options & CB_USE_REDIRECTION) == 0)
    {
      dmca.dFormat = CF_NAME64;
      sprintf(prp, "%.60s.CQS", cprps[i].prp);
      cprps[i].qpid = RegisterPropertyInformation(CMBEQM_TAG, prp, &dmca, NULL, CA_READ, AT_CHANNEL, dmca.dArrayLength, "channel connection quality stats", 3 * ncprps + pid, NULL);
      dmca.dFormat = CF_INT32;
      RegisterPropertyInformation(CMBEQM_TAG, prp, &dmca, NULL, CA_READ, AT_CHANNEL, dmca.dArrayLength, "channel connection quality stats", 3 * ncprps + pid, NULL);
      dmca.dFormat = CF_INTINT;
      RegisterPropertyInformation(CMBEQM_TAG, prp, &dmca, NULL, CA_READ, AT_CHANNEL, dmca.dArrayLength, "channel connection quality stats", 3 * ncprps + pid, NULL);
      dmca.dFormat = CF_INTINTINTINT;
      RegisterPropertyInformation(CMBEQM_TAG, prp, &dmca, NULL, CA_READ, AT_CHANNEL, dmca.dArrayLength, "channel connection quality stats", 3 * ncprps + pid, NULL);
    }
  }
  // and the combobulator server properties :
  dout.dArrayLength = 1;
  dout.dFormat = CF_INT32;
  RegisterPropertyInformation(CMBEQM_TAG,"Monitor.PollingInterval",&dout,&dout,CA_READ|CA_WRITE,AT_SCALAR,1,"[500:10000 ms]monitor polling interval",PRP_POLL_INTERVAL,NULL);
  RegisterPropertyInformation(CMBEQM_TAG,"Monitor.StopAll",NULL,NULL,CA_WRITE,AT_SCALAR,0,"Stop all monitoring",PRP_STOPALL,NULL);
  RegisterPropertyInformation(CMBEQM_TAG,"Monitor.StartAll",NULL,NULL,CA_WRITE,AT_SCALAR,0,"Stop all monitoring",PRP_STARTALL,NULL);
  dout.dArrayLength = ncprps;
  RegisterPropertyInformation(CMBEQM_TAG, "Monitor.Options", &dout, NULL, CA_READ, AT_CHANNEL, 0, "configured monitor options", PRP_OPTIONS, NULL);
  dout.dFormat = CF_NAME64;
  RegisterPropertyInformation(CMBEQM_TAG, "Monitor.Options", &dout, NULL, CA_READ, AT_CHANNEL, 0, "configured monitor options", PRP_OPTIONS, NULL);
  for (i = 0; i < ncprps; i++) strncpy(dbuf[i].name, cprps[i].prp, 64);
  AssignDeviceListToProperty(CMBEQM_TAG, "Monitor.PollingInterval", dbuf, ncprps);
  AssignDeviceListToProperty(CMBEQM_TAG, "Monitor.StopAll", dbuf, ncprps);
  AssignDeviceListToProperty(CMBEQM_TAG, "Monitor.StartAll", dbuf, ncprps);
  AssignDeviceListToProperty(CMBEQM_TAG, "Monitor.Options", dbuf, ncprps);
  // register all known devices as a device (let the .NAM sort it out)
  for (i=0; i<ncprps; i++)
  { 
    RegisterDeviceName(CMBEQM_TAG, cprps[i].prp,i);
  }
  ndevs = ncprps;
  for (i = 0; i < ncprps; i++)
  {
    tprp = NULL;
    for (nprp=cprps[i].tprps; nprp != NULL; nprp = nprp->nxt) tprp = nprp; // find the end
    while (tprp != NULL)
    {
      for (n = 0; n < tprp->ndevs; n++)
      {
        if (GetDeviceNumber(CMBEQM_TAG, tprp->devlst[n].alias) >= 0) continue; // already known
        RegisterDeviceName(CMBEQM_TAG, tprp->devlst[n].alias, ndevs++);
      }
      tprp = tprp->prv;
    }
  }
  AssignNumDevices(CMBEQM_TAG, ndevs);
  SetServerCycleIdle(FALSE);
}

void start_links(int pid)
{
  int i,cc = 0,mode=CM_TIMER, m;
  int fsiz, hsiz, ndevs, offset;
  DTYPE dout,din;
  TPrp *tprp;
  char *tdevnam = "";
//  if (isUseOnError) mode |= CM_USEONERROR;
  DTYPEZERO(dout); DTYPEZERO(din);

  for (i = 0; i < ncprps; i++)
  {
    if (pid >= 0 && i != pid) continue;
    if (cprps[i].options & (CB_USE_REDIRECTION|CB_IS_CLONED|CB_WRITEONLY)) continue;
    m = mode;
    if (cprps[i].options & CB_CONNECT) m |= CM_CONNECT;
    fsiz = fmtsizeof(cprps[i].dout.dFormat);
    hsiz = getFormatHeaderSize(cprps[i].dout.dFormat);
    fsiz += hsiz;
    cprps[i].dout.data.bptr = (BYTE *)calloc(cprps[i].dout.dArrayLength, fsiz);
    ndevs = 0;
    for (tprp = cprps[i].tprps; tprp != NULL; tprp = tprp->nxt)
    { // use segments of DTYPE in the tprps list ...
      offset = cprps[i].len * fsiz;
      if (cprps[i].dout.dFormat == CF_IMAGE) offset += IMAGE_HDR_SIZE + 2 * sizeof(size_t);
      if (cprps[i].dout.dFormat == CF_SPECTRUM) offset += SPECTRUM_HDR_SIZE / sizeof(float);
      offset *= ndevs;
      if (hsiz > 0) tprp->ndevs = 1; // guarantee this !
      tprp->data = cprps[i].dout.data.bptr + offset;
      dout.dArrayLength = cprps[i].len * tprp->ndevs;
      dout.dFormat = cprps[i].dout.dFormat;
      dout.data.bptr = tprp->data;
      applyValueToDTYPE(&dout, tprp->defaultValue);
      if (tprp->devlst != NULL && !tprp->devlst[0].disabled)
      {
        tdevnam = getTgtDevice(&cprps[i], tprp);
        tprp->lid = AttachLinkEx2(makeTgtDevName(tprp->srv, tdevnam), tprp->prp, &dout, NULL, CA_READ, cprps[i].tmr, cbCmb, m, i, tprp);
      }
      else
      {
        tprp->lsts = not_running;
      }
      tprp->offset = ndevs;
      ndevs += tprp->ndevs;
    }
  }
  return;
}
void stop_links(int pid)
{
  int i;
  TPrp *tprp;
  for (i = 0; i < ncprps; i++)
  {
    if (pid >= 0 && i != pid) continue;
    for (tprp = cprps[i].tprps; tprp != NULL; tprp = tprp->nxt)
    {
      if (tprp->devlst != NULL && !tprp->devlst[0].disabled)
      {
        CloseLink(tprp->lid);
      }
    }
  }
}
void cmbeqm_exi(void)
{
  stop_links(-1);
}
int pushImageBytes(DIMAGE *imgDst, DIMAGE *imgSrc, UINT32 *img_size)
{
  if (*img_size > imgSrc->frameBufferSize) *img_size = imgSrc->frameBufferSize;
  memcpy(imgDst, imgSrc, sizeof(hdrIMAGE));
  if (imgSrc->frameHeader.appendedFrameSize < (int)*img_size)
    *img_size = imgSrc->frameHeader.appendedFrameSize;
  imgDst->frameBuffer = imgSrc->frameBuffer;
  imgDst->frameBufferSize = imgSrc->frameBufferSize;
  return 0;
}
char *getOptionsString(int opt)
{
  static char opts[512];
  int len;
  memset(opts, 0, 512);
  if (opt & CB_USE_LOCALTIME) strcat(opts, "LOCALTIME|");
  if (opt & CB_USE_REDIRECTION) strcat(opts, "REDIRECT|");
  if (opt & CB_USE_SCHEDULING) strcat(opts, "SCHEDULE|");
  if (opt & CB_IS_CLONED) strcat(opts, "CLONED|");
  if (opt & CB_FORWARD) strcat(opts, "FORWARD|");
  if (opt & CB_WRITEONLY) strcat(opts, "WRITEONLY|");
  if (opt & CB_STATIC) strcat(opts, "STATIC|");
  if (opt & CB_CONNECT) strcat(opts, "CONNECT|");
  if (opt & CB_APPLYSTALE) strcat(opts, "USE_STALE|");
  if (opt & CB_APPLYDEFAULT) strcat(opts, "USE_DEFAULT|");
  len = (int)strlen(opts);
  if (len > 0) opts[len - 1] = 0;
  return opts;
}
int isAllowedStringFormat(short fmt)
{
  switch (LFMT(fmt))
  {
  case CF_TEXT:
  case CF_NAME8:
  case CF_NAME16:
  case CF_NAME32:
  case CF_NAME48:
  case CF_NAME64:
    return TRUE;
  default:
    break;
  }
  return FALSE;
}
int cmbeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int devnr,prpid,ndevs,i,cc;

  prpid = GetPropertyId(CMBEQM_TAG,devProperty);
  if (prpid < 0) return illegal_property;

  switch (prpid)
  {
    case PRP_POLL_INTERVAL:
      cc = illegal_read_write;
      if ((i = GetPropertyId(CMBEQM_TAG, devName)) < 0) return illegal_device;
      if (access & CA_WRITE)
      {
        int tmr;
        if (din->dArrayLength != 1) return dimension_error;
        if ((cc=GetValuesAsLong(din,&tmr,1)) != 0) return cc;
        cprps[i].tmr = tmr;
        stop_links(i);
        start_links(i);
      }
      if (dout->dArrayLength > 0)
        cc = PutValuesFromLong(dout,&cprps[i].tmr,1);
      return cc;
    case PRP_STOPALL:
      if ((access & CA_WRITE) == 0) return illegal_read_write;
      if ((i = GetPropertyId(CMBEQM_TAG, devName)) < 0) return illegal_device;
      stop_links(i);
      return 0;
    case PRP_STARTALL:
      if ((access & CA_WRITE) == 0) return illegal_read_write;
      if ((i = GetPropertyId(CMBEQM_TAG, devName)) < 0) return illegal_device;
      start_links(i);
      return 0;
    case PRP_OPTIONS:
    {
      int pid = GetPropertyId(CMBEQM_TAG, devName);
      int fsiz = GetFormatSize(dout->dFormat);
      if (pid < 0) return illegal_device;
      if ((access & CA_WRITE)) return illegal_read_write;
      if (dout->dArrayLength < 1) return dimension_error;
      if (isAllowedStringFormat(dout->dFormat))
      {
        if (dout->dFormat == CF_TEXT)
        {
          char *opts = getOptionsString(cprps[pid].options);
          int len = MIN((int)strlen(opts),(int)dout->dArrayLength);
          dout->dArrayLength = len;
          strncpy(dout->data.cptr, opts, len);
          return 0;
        }
        if ((int)dout->dArrayLength > ncprps - pid) dout->dArrayLength = (UINT32)(ncprps-pid);
        for (i = pid; i < (int)dout->dArrayLength; i++)
        {
          strncpy(dout->data.cptr + fsiz * i, getOptionsString(cprps[i].options), fsiz);
        }
        return 0;
      }
      return PutValuesFromAny(dout, &cprps[0].options, CF_INT32, 1, sizeof(CPrp), ncprps, pid);
    }
    default:
      if (prpid < ncprps)
      { // specific property
        int rpid;
        int fsiz = fmtsizeof(cprps[prpid].dout.dFormat);
        int hsiz = getFormatHeaderSize(cprps[prpid].dout.dFormat);
        int efmt = GetElevatedMcaFmt(cprps[prpid].dout.dFormat);
        int needsScale = cprps[prpid].needsScale;
        int needsElevation = cprps[prpid].dout.dFormat != efmt && dout->dFormat == efmt;
        int isSpectrumType = cprps[prpid].atyp & AT_SPECTRUM;
        if (isSpectrumType && needsElevation)
        {
          return illegal_format;
        }
        if (cprps[prpid].options & CB_USE_REDIRECTION)
        { // redirect to the correct element ...
          TPrp *tprp;
          for (tprp = cprps[prpid].tprps; tprp != NULL; tprp = tprp->nxt)
          { // straight-up search (as we don't do this very often ...)
            for (i = 0; i < tprp->devlst->number; i++)
            {
              if (!strnicmp(tprp->devlst[i].alias, devName, 64))
              { // found it!
                char tgt[80], *ctx, *srv;
                strncpy(tgt, tprp->srv, 64); tgt[64] = 0;
                if (tgt[0] == '/')
                { // target server leads with context info
                  ctx = &tgt[1];
                  srv = strchr(ctx, '/');
                  if (srv == NULL) return configuration_error;
                  *srv++ = 0;
                }
                else
                {
                  ctx = gDeviceContext;
                  srv = tgt;
                }
                SetRedirectionParameters(CMBEQM_TAG, ctx, srv, tprp->devlst[i].name, tprp->prp);
                return server_redirection;
              }
            }
          }
          return configuration_error; // if we land here then something is not configured correctly!
        }
        if (cprps[prpid].options & (CB_FORWARD | CB_WRITEONLY))
        { // special property options given ...
          if (access & CA_WRITE)
          { // forward if WRITE access is present
            TPrp *tp = NULL;
            TDevs *dp;
            for (tp = cprps[prpid].tprps; tp != NULL; tp = tp->nxt)
            {
              if ((dp=tp->devlst) == NULL) break;
              for (i = 0; i < tp->ndevs; i++)
              if (!strncmp(dp[i].alias, devName, 64))
              {
                char tgt[132];
                sprintf(tgt, "%.64s/%.64s", tp->srv, dp[i].name);
                return ExecLinkEx(tgt, tp->prp, dout, din, access, 800);
              }
            }
            return illegal_device;
          }
          // it's just a READ so fall thru, unless ...
          if (cprps[prpid].options & CB_WRITEONLY) return illegal_read_write;
        }
        if (!IsNumFmt(cprps[prpid].dout.dFormat)) needsScale = needsElevation = FALSE;
        LockLinkData();
        fsiz += hsiz;
        rpid = (cprps[prpid].options & CB_IS_CLONED) ? cprps[prpid].rpid : prpid;
        if (cprps[prpid].staleCnts <= cprps[prpid].staleThreshold)
        { // not marked stale: use the learned data stamps
          dout->dTimeStamp = cprps[rpid].dout.dTimeStamp;
          dout->sysStamp = cprps[rpid].dout.sysStamp;
          dout->dStamp = cprps[rpid].dout.dStamp;
        }
        devnr = GetDeviceNumberEx(CMBEQM_TAG, devName, devProperty);
        if (devnr < 0) ccerr(illegal_device);
        ndevs = cprps[prpid].dout.dArrayLength;
        if (dout->dArrayLength > (UINT32)ndevs) dout->dArrayLength = (UINT32)ndevs;
        if (isSpectrumType) ndevs /= cprps[prpid].len; // use the 'real number' below ...
        if (needsScale || needsElevation)
        { // must be a number type
          SINT16 sval, *svals = NULL;
          SINT32 ival, *ivals = NULL;
          SINT64 lval, *lvals = NULL;
          float fval, *fvals = NULL;
          double dval, *dvals = NULL;
          BYTE *bvals = NULL;
          int n, k = 0;
          TPrp *tp = NULL, *np;
          int isTrace = cprps[prpid].atyp == AT_SPECTRUM ? TRUE : FALSE;
          int numElems = 0;
          for (np = cprps[prpid].tprps; np != NULL; np = np->nxt) tp = np; // find the end
          cc = 0;
          if (!needsElevation)
          { // need to allocate a collection buffer
            if ((bvals = (BYTE *)calloc(dout->dArrayLength, fmtsizeof(dout->dFormat))) == NULL)
              ccerr(out_of_server_memory);
            svals = (SINT16 *)bvals;
            ivals = (SINT32 *)bvals;
            lvals = (SINT64 *)bvals;
            fvals = (float *)bvals;
            dvals = (double *)bvals;
          }      
          while (tp != NULL && k < (int)dout->dArrayLength && cc == 0)
          {
            numElems = isTrace ? cprps[prpid].len : tp->ndevs;
            for (n = 0; n < numElems && k < (int)dout->dArrayLength; n++, k++)
            {
              switch (cprps[prpid].dout.dFormat)
              {
                case CF_INT16:
                  sval = cprps[prpid].dout.data.sptr[k];
                  if (tp->scale < 1)
                  {
                    fval = (float)sval;
                    fval *= tp->scale;
                    sval = (SINT16)fval;
                  }
                  else
                  {
                    sval *= (SINT16)tp->scale;
                  }
                  sval += (SINT16)tp->shift;
                  if (needsElevation)
                  {
                    ((INTINT *)dout->data.vptr)[k].i1val = (int)sval;
                    ((INTINT *)dout->data.vptr)[k].i2val = tp->lsts;
                  }
                  else
                  {
                    //dout->data.sptr[k] = sval;
                    svals[k] = sval;
                  }
                  break;
                case CF_INT32:
                  ival = cprps[prpid].dout.data.lptr[k];
                  if (tp->scale < 1)
                  {
                    dval = (double)ival;
                    dval *= tp->scale;
                    ival = (SINT32)dval;
                  }
                  else
                  {
                    ival *= (SINT32)tp->scale;
                  }
                  ival += (SINT32)tp->shift;
                  if (needsElevation)
                  {
                    ((INTINT *)dout->data.vptr)[k].i1val = ival;
                    ((INTINT *)dout->data.vptr)[k].i2val = tp->lsts;
                  }
                  else
                  {
                    //dout->data.lptr[k] = ival;
                    ivals[k] = ival;
                  }
                  break;
                case CF_INT64:
                  lval = cprps[prpid].dout.data.llptr[k];
                  if (tp->scale < 1)
                  {
                    dval = (double)ival;
                    dval *= tp->scale;
                    lval = (SINT64)dval;
                  }
                  else
                  {
                    lval *= (SINT64)tp->scale;
                  }
                  lval += (SINT64)tp->shift;
                  //dout->data.llptr[k] = lval;
                  lvals[k] = lval;
                  break;
                case CF_FLOAT:
                  fval = cprps[prpid].dout.data.fptr[k];
                  fval *= tp->scale;
                  fval += tp->shift;
                  if (needsElevation)
                  {
                    ((FLTINT *)dout->data.vptr)[k].fval = fval;
                    ((FLTINT *)dout->data.vptr)[k].ival = tp->lsts;
                  }
                  else
                  {
                    //dout->data.fptr[k] = fval;
                    fvals[k] = fval;
                  }
                  break;
                case CF_DOUBLE:
                  dval = cprps[prpid].dout.data.dptr[k];
                  dval *= (double)tp->scale;
                  dval += (double)tp->shift;
                  if (needsElevation)
                  {
                    ((DBLDBL *)dout->data.vptr)[k].d1val = dval;
                    ((DBLDBL *)dout->data.vptr)[k].d2val = (double)tp->lsts;
                  }
                  else
                  {
                    //dout->data.dptr[k] = dval;
                    dvals[k] = dval;
                  }
                  break;
                default:
                  cc = illegal_format;
                  break;
              }
            }
            tp = tp->prv;
          }
          if (bvals != NULL)
          { // needed a collection buffer!
            PutValuesFromAny(dout, bvals, cprps[prpid].dout.dFormat, 1, fsiz, ndevs, devnr);
            free(bvals);
          }
        }
        else
        { // normal case and a one-liner ...
          int fldIdx = cprps[prpid].fld, fldOff = 0, fldFmt = CF_NULL;
          cc = -1;
          switch (cprps[prpid].dout.dFormat)
          {
            case CF_USTRING:
              switch (dout->dFormat)
              {
                case CF_USTRING: break;
                case CF_NAME16:
                case CF_NAME32:
                case CF_NAME64:
                case CF_STRING:
                  cc = PutValuesFromString(dout, cprps[prpid].dout.data.cptr, cprps[prpid].dout.dFormat, ndevs, devnr);
                  break;
                case CF_FLOAT:
                case CF_DOUBLE:
                  fldOff = 4; fldFmt = CF_FLOAT;
                  cc = PutValuesFromAny(dout, cprps[prpid].dout.data.bptr + fldOff, fldFmt, 1, fsiz, ndevs, devnr);
                  break;
                default:
                  cc = illegal_format;
                  break;
              }
              break;
            case CF_TEXT:
            {
              int n, off = 0, nlen, slen, flen, cplen;
              flen = nlen = GetFormatSize(dout->dFormat);
              slen = cprps[prpid].len;
              switch (dout->dFormat)
              {
                case CF_USTRING:
                case CF_NAME16:
                case CF_NAME32:
                case CF_NAME64:
                  if (dout->dFormat == CF_USTRING)
                  {
                    nlen = 80;
                    off = 16;
                  }
                  cplen = MIN(nlen, slen);
                  for (n = 0, i = devnr; i < (int)cprps[0].dout.dArrayLength && n < (int)dout->dArrayLength; i++, n++)
                  {
                    strncpy(dout->data.cptr + off + n*flen, cprps[prpid].dout.data.cptr + i*slen, cplen);
                  }
                  cc = 0;
                  break;
                case CF_TEXT:
                  if (dout->dArrayLength < (UINT32)slen) ccerr(dimension_error);
                  strncpy(dout->data.cptr, cprps[prpid].dout.data.cptr + devnr*slen, slen);
                  break;
                default:
                  cc = illegal_format;
                  break;
              }
              break;
            }
            case CF_IMAGE:
            {
              int dsiz = cprps[prpid].len + IMAGE_HDR_SIZE + 2 * sizeof(size_t);
              cc = pushImageBytes((DIMAGE *)(dout->data.vptr), (DIMAGE *)(cprps[prpid].dout.data.bptr+devnr*dsiz), &dout->dArrayLength);
              break;
            }
            case CF_SPECTRUM:
              if (din->dArrayLength == 1)
              {
                if ((cc = GetValuesAsLong(din, &fldIdx, 1)) != 0) ccerr(cc);
              }
              if (fldIdx < 0) fldIdx = 3;
              if (fldIdx == 0) fldFmt = CF_TEXT;
              else if (fldIdx == 1 || fldIdx == 4) fldFmt = CF_INT32;
              else fldFmt = CF_FLOAT;
              if (fldIdx > 0) fldOff = 76 + fldIdx * 4;
              cc = PutValuesFromAny(dout, cprps[prpid].dout.data.bptr + fldOff, fldFmt, 1, fsiz, ndevs, devnr);
              break;
            default: 
              if (din->dArrayLength == 1)
              {
                if ((cc = GetValuesAsLong(din, &fldIdx, 1)) != 0) ccerr(cc);
              }
              if (fldIdx >= 0)
              {
                if ((cc = GetFieldInFormat(cprps[prpid].dout.dFormat, NULL, NULL, fldIdx, &fldOff, &fldFmt)) != 0) ccerr(cc);
                cc = PutValuesFromAny(dout, cprps[prpid].dout.data.bptr + fldOff, fldFmt, 1, fsiz, ndevs, devnr);
              }
              break;
          }
          if (cc < 0)
          { // not handled in a special case above ...
            int off = cprps[prpid].len * devnr;
            int num = cprps[prpid].len * ndevs;
            if (!isMcaType(cprps[prpid].atyp))
            { // refers to a specific device (i.e. not a combobulated MCA) ...
              TPrp *tp;
              int found = FALSE;
              for (tp = cprps[prpid].tprps; tp != NULL && !found; tp = tp->nxt)
              {
                for (i = 0; i < tp->ndevs; i++)
                {
                  if (!strnicmp(tp->devlst[i].alias, devName, 64))
                  {
                    dout->sysStamp = tp->sysstmp;
                    dout->dStamp = tp->usrstmp;
                    dout->dTimeStamp = tp->ts;
                    if (!HAS_DATA(tp->lsts) && cprps[prpid].len == dout->dArrayLength)
                    { // if a caller just wants a specific spectrum array and it's in an error state
                      ccerr(tp->lsts);
                    }
                    found = TRUE;
                    break;
                  }
                }
              }
            }
            cc = PutValuesFromAny(dout, cprps[prpid].dout.data.vptr, cprps[prpid].dout.dFormat, 1, fsiz, num, off);
          }
        }
err:
        UnlockLinkData();
        return cc;
      }
      { // a meta property request
        int tpid, n, k=0, isNameRequest=FALSE, isSrcRequest=FALSE, isCqsRequest=FALSE;
        char *c, tprp[72], mprp[16];
        TPrp *tp = NULL, *np;
        NAME64 *dst = (NAME64 *)dout->data.vptr;
        strncpy(tprp, devProperty, 64);
        if ((c=strrchr(tprp, '.')) == NULL) return illegal_property;
        *c++ = 0;
        strncpy(mprp, c, 4); mprp[4] = 0;
        if (!strnicmp(mprp,"NAM",3)) isNameRequest = TRUE;
        else if (!strnicmp(mprp, "SRC",3)) isSrcRequest = TRUE;
        else if (!strnicmp(mprp, "CQS",4)) isCqsRequest = TRUE;
        if ((tpid=GetPropertyId(CMBEQM_TAG, tprp)) < 0) return illegal_property;
        if (dout->dFormat != CF_NAME64 && !isCqsRequest) return illegal_format;
        for (np = cprps[tpid].tprps; np != NULL; np = np->nxt) tp = np; // find the end
        while (tp != NULL && k < (int)dout->dArrayLength)
        {
          for (n = 0; n < tp->ndevs && k < (int)dout->dArrayLength; n++)
          {
            if (isNameRequest)
            {
              strncpy(dst[k++].name, tp->devlst[n].alias, DEVICE_NAME_SIZE);
            }
            else if (isSrcRequest)
            {
              snprintf(dst[k++].name, 63, "%.64s[%.64s]", makeTgtDevName(tp->srv, tp->devlst[n].name), tp->prp);
            }
            else if (isCqsRequest)
            {
              switch (dout->dFormat)
              {
                case CF_INTINTINTINT:
                  ((DADDRESS *)dout->data.vptr)[k].i1val = tp->nto;
                  ((DADDRESS *)dout->data.vptr)[k].i2val = tp->nmiss;
                  ((DADDRESS *)dout->data.vptr)[k].i3val = tp->delta_t;
                  ((DADDRESS *)dout->data.vptr)[k].i4val = tp->delta_sys;
                  break;
                case CF_INTINT:
                  ((INTINT *)dout->data.vptr)[k].i1val = tp->nto;
                  ((INTINT *)dout->data.vptr)[k].i2val = tp->nmiss;
                  break;
                case CF_INT32:
                  dout->data.lptr[k] = tp->nto;
                  break;
                case CF_NAME64:
                  snprintf(dst[k].name, 63, "timeouts: %d, missed: %d, dts: %d, dn: %d", tp->nto, tp->nmiss, tp->delta_t, tp->delta_sys);
                  break;
                default:
                  return illegal_format;
              }
              k++;
            }
          }
          tp = tp->prv;
        }
        if (k < (int)dout->dArrayLength) dout->dArrayLength = (UINT32)k;
        return 0;
      }
   }
   return 0;
}
char *getAbbreviatedFecName(char *srv, int maxlen)
{
  static char fec[FEC_NAME_SIZE + 8];
  int i, k, slen, sflen = 0;
  char *suf = NULL, tmp[EXPORT_NAME_SIZE + 8];
  strncpy(tmp, srv, EXPORT_NAME_SIZE); tmp[EXPORT_NAME_SIZE] = 0;
  slen = (int)strlen(tmp);
  for (i = slen - 1; i > 0 && isdigit(tmp[i]); i--);
  suf = &tmp[i + 1];
  if ((sflen = (int)strlen(suf)) > 5)
  { /* allow up to a 5-digit numerical name */
    suf = NULL;
    sflen = 0;
  }
  slen -= sflen;
  memset(fec, 0, FEC_NAME_SIZE + 8);
  for (i = 0, k = 0; i<slen && k<maxlen; i++)
  {
    if (tmp[i] == 0) break;
    if (tmp[i] == '*') continue;
    if (tmp[i] == '.') continue;
    if (tmp[i] == '_') continue;
    if (tmp[i] == '-') continue;
    fec[k++] = tmp[i];
  }
  if (sflen > 0)
  {
    if ((int)strlen(fec) > maxlen - sflen) fec[maxlen - sflen - 1] = 0;
    strncat(fec, suf, sflen);
  }
  return fec;
}
char *suggestFecName(char *ctx, char *srv, int maxlen)
{
  static char fec[FEC_NAME_SIZE + 8];
  int len;

  memset(fec, 0, FEC_NAME_SIZE + 8);
  if (srv == NULL) return fec;
  if (ctx != NULL)
  {
    char *c;
    if ((c = strstr(ctx, ".SIM")) != NULL) *c++ = 0;
    else if ((c = strstr(ctx, "_SIM")) != NULL) *c++ = 0;
    else if ((c = strstr(ctx, ".TEST")) != NULL) *c++ = 0;
    else if ((c = strstr(ctx, "_TEST")) != NULL) *c++ = 0;
    strncpy(fec, ctx, 2);
    if (c != NULL) strncat(fec, c, 1);
  }
  if (maxlen > FEC_NAME_SIZE) maxlen = FEC_NAME_SIZE;
  len = maxlen - (int)strlen(fec);
  strncat(fec, getAbbreviatedFecName(srv, len), len);
  strupr(fec);
  if ((int)strlen(fec) < (maxlen - 2) && strstr(fec, "FEC") == NULL) strcat(fec, "FEC");
  return fec;
}
void PreSystemInit(void)
{
  char *usr = getenv("USERNAME");
  if (strlen(thisFecServer) == 0) goto err;
  if (usr == NULL) usr = "UNKNOWN";
  if (strlen(thisFecName) == 0)
  {
    strncpy(thisFecName,suggestFecName(thisFecContext, thisFecServer, 16),16);
  }
  if (strlen(thisFecResponsible) == 0) strncpy(thisFecResponsible, usr, 32);
  if (thisFecPort == 0) thisFecPort = GetPortOffset(thisFecName);
  FeclogDepth = 1000;
  RegisterFecInformation(thisFecName, thisSubsystem, thisFecContext, thisFecDescription, thisFecLocation, "none", thisFecResponsible, (UINT16)thisFecPort);
  SetConnectionTableCapacity(gLinkTableSize);
  return;
err:
  printf("%s\n", gUsageString);
  exit(0);
}
void PostSystemInit(void)
{
  RegisterEquipmentModule(thisFecServer, CMBEQM_TAG, thisFecCapacity, cmbeqm, cmbeqm_ini, cmbeqm_bkg, 200, cmbeqm_exi);

}
