#ifndef _ARCEQM_H_
#define _ARCEQM_H_

#ifndef PM_EXPORT
# define PM_EXPORT
#endif

#ifndef ulong
# define ulong UINT32
#endif

extern char ArchiveDirectoryPath[128];
#define TRCVIEW_FILE "triggers.csv"
#ifdef WIN32
#define CONFIGS_PATH "..\\CONFIGS"
#else
#define CONFIGS_PATH "../CONFIGS"
#endif
typedef struct
{
  char ctx[32]; 
  char srv[32];
  char prp[64];
  char dev[64]; 
} TrgChannel;

typedef struct
{
  char fileName[64];
  char desc[64];
  TrgChannel *channels;
  int nChannels;
} TrgComponent;

typedef struct
{
  char ctx[32];
  char trg[64];
  char desc[64];
  char file[64];
  TrgComponent *components;
  int nComponents;
} TrgViewFile;

typedef struct
{
  short numProperties;
  short numDevices;
  long  eventCode;
  char  comment[64];
} AIB;  /* Archive Info Block */
#define AIBStructSize (2*2+4+64)
typedef struct
{
  float SamplingRate;
  long  numPreTriggerValues;
  long  TimeOffset;
  long  LowCount;
  long  HighCount;
  char  Units[16];
  double LowEGU;
  double HighEGU;
  short  numDevices;
  short  reserved;
} PIB;  /* Property Info Block */
#define PIBStructSize (5*4+16+2*8+2*2)

typedef struct
{
   char channelID[32]; /* Kanalidentifikator                         */
   UINT32 eventcode;   /* Ereignisbewerter                           */
   UINT16 arraysize;   /* Anzahl der Daten                           */
   UINT16 datatype;    /* Datentyp                                   */
   float samplingrate; /* Abtastfrequenz                             */
   UINT32 pretrigger;  /* Anzahl der Daten vor dem Triggerzeitpunkt  */
   UINT32 timeoffset;  /* Zeitoffset                                 */
   UINT32 lowcount;    /* kleinster Mewert in digits                */
   UINT32 highcount;   /* grter Mewert in digits                  */
   char egu[16];       /* physikalische Einheit                      */
   float egulow;       /* kleinster Mewert in physikalischer Einheit*/
   float eguhigh;      /* grter Mewert in physikalischer Einheit  */
   UINT32 datalength;  /* supercedes arraysize, when arraysize = -1  */
   char desc[4][64];   /* 4x Kommentarzeile                          */
} DHS;  /* Data Header Structure */
#define DHSStuctSize (32 + 4 + 2*2 + 4 + 4*4 + 16 + 4*2 + 4 + 4*64)

typedef struct
{
   UINT32 evnCode;
   UINT32 evnNr;
   UINT32 datSiz;
   UINT16 datFmt;
   UINT16 aryTyp;
   UINT32 preTrg;
   UINT32 reserved;
   float shiftX;
   float scaleX;
   float shiftY;
   float scaleY;
   float rangeMinX; 
   float rangeMaxX;
   float scaleMinY;
   float scaleMaxY;
   char unitsX[16];
   char unitsY[16];
   char desc[256];
   char fcnTag[64];
} EVNDHDR;  /* Event Data Header Structure (modern) */
#define EVNDHDR_SIZE (14*4 + 2*16 + 256 + 64) /* 408 bytes */

typedef struct
{
  char  deviceServerName[16];
  char  deviceProperty[16];
  char  deviceName[16];
  short dataFormat;
  short dataDimension;
} DEVSPEC5; /* Device Specifier */

typedef struct
{
  ulong   eventNumber;
  UINT32  starttime;
  UINT32  stoptime;
  DEVSPEC5 devSpecifier;
} ADRB5; /* Archived Data Request Block */

#define ADRBStructSize (3*4 + 3*16 + 2*2)

typedef struct
{
  char  deviceContext[CONTEXT_NAME_SHORTSIZE];
  char  deviceServerName[EXPORT_NAME_SHORTSIZE];
  char  deviceProperty[PROPERTY_NAME_SHORTSIZE];
  char  deviceName[32];
  short dataFormat;
  short reserved;
  ulong dataIndex;
  ulong dataDimension;
} DEVSPECX; /* Extended Device Specifier */

typedef struct
{
  ulong    eventNumber;
  UINT32   starttime;
  UINT32   stoptime;
  DEVSPECX devSpecifier;
} ADRBX; /* Extended Archived Data Request Block */

#define ADRBXStructSize (3*4 + 2*16 + 2*32 + 2*2 + 2*4)

typedef struct
{
  char  deviceContext[CONTEXT_NAME_SIZE];
  char  deviceServerName[EXPORT_NAME_SIZE];
  char  deviceProperty[PROPERTY_NAME_SIZE];
  char  deviceName[DEVICE_NAME_SIZE];
  char  dataTag[TAG_NAME_SIZE];
  short dataFormat;
  short reserved;
  UINT32 dataIndex;
  UINT32 dataDimension;
} DEVSPEC; /* Device Specifier Release 4 */

typedef struct
{
  UINT32  eventNumber;
  UINT32  starttime;
  UINT32  stoptime;
  DEVSPEC devSpecifier;
} ADRB; /* Extended Archived Data Request Block */

#define MAKECODE(id,cd) (((id)<<16)|(cd))
#define BITSinBYTE 8
#define EVENT_IGNORE 0xffffffff

#define ARCEQM_TAG "ARCEQM"
#define ARCEQM_EXPORT "EVENTS"
#define ARCEQM_VERSION 1

#endif /* _ARCEQM_H_ */

