#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "tine.h"
#if defined(UNIX)
#include "dirent.h"
#elif defined(WIN32)
#include "direct.h"
#endif
#include "archive.h"

int adjustWorkAreaSize(UINT32 newSizeInBytes); /* from tine lib (requires a static lib on windows) */

#define HSTEQM_NAME "HSTEQM"
#define HSTEXPORT_NAME "HISTORY"

#define MAX_NUM_CONFIGS 500
#define ONEDAY 3600*24

char *archErrLst[] = { 
  "no stored data in the time range given",
  "given time range invalid",
};
int archErrLstSize = sizeof(archErrLst)/sizeof(archErrLst[0]);

extern TINE_EXPORT int gScanForNameLists;
extern int gAppendProperyAliasNamesToQueries;
int totalNrRecords, nrecords;
char cbuff[256];

#define MAX_ADBFILES 20
char adbFileName[64] = "";
char adbfile[MAX_ADBFILES][64];
int nadbfiles = 0;
int archdebug = 0,archPending = 0;
int recordsCapacity = 0;
int g_isTerminating = 0;
int AppendProblemsLog(char *s) { return 0;}

char thisFecName[FEC_NAME_SIZE] = "";
char thisContext[CONTEXT_NAME_SIZE] = "";
char *thisSubSystem = "SER";
char thisFecDescription[FEC_DESC_SIZE] = "";
char thisLocation[FEC_LOCATION_SIZE] = "";
char thisServerName[EXPORT_NAME_SIZE] = HSTEXPORT_NAME;
char thisFailoverMasterName[EXPORT_NAME_SIZE] = "";
char thisFailoverSlaveMasterName[EXPORT_NAME_SIZE] = "";

UINT16 thisPortOffset = 0;
int startupInputsGiven = FALSE;

#define DATABLOB_SIZE 65536 * 4
#define MAXNAMES (DATABLOB_SIZE/64)
#define MAXNAMEI 4000
NAME64I name64Iblob[MAXNAMEI];
NAME64 name64blob[MAXNAMES];
INTINT *iiblob = (INTINT *)name64Iblob;
BYTE *gDataBuffer = NULL;
int gDataBufferCapacity = 0;
NAME64DBLDBL *nddblob = (NAME64DBLDBL *)name64blob;
#define MAXN64DD (DATABLOB_SIZE/80)

int getNumLines(FILE *fp);
int getdata(char *fname, int ntotal, int offset);
int opendata(int mon, int day, int hour);
int getEntryFromRectyp(int rec);
int getArchivedData(char *keyword,int index,int fieldindex,time_t starttime,
          time_t stoptime,int arraysize,BYTE *data,int datafmt,char *datatag,int sample,
          float upr, float lwr, double *ts);
int getArchivedDataTrace(char *keyword,time_t tgttime,DTYPE *dout,int index,int fieldindex,int sample);
void hstinit(void);
void hstexit(void);
int hsteqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access);
int OpenDataFile(DAQRecord *daqr,int *hasAlternate);
time_t getArchiveStartTimeFromDepth(char *keyword,int depth);
int getItemSize(int fmt,char *tag);
int getDeviceAndSubDeviceIndicies(char *keyword,char *devName,int *subindex,time_t starttime,time_t endtime);
void rprtFileErr(const char *src,const char *txt,const char *filename);
double assignKeyValue(const char *key,char **kvset,int nkv);

typedef struct
{
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  char units[UNITS_SIZE];
  char desc[PROPERTY_DESC_SIZE];
  int idx;
  int sidx;
  float scale;
  float offset;
  float max;
  float min;
  int plot;
} HstViewSet;

#define MCA_SINGLE_ENTRY  0
#define MCA_APPEND_ENTRY  1
#define MCA_REVERSE_ENTRY 2

typedef struct
{
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  char tag[TAG_NAME_SIZE];
  char desc[PROPERTY_DESC_SIZE];
  char shell[64];
  char alias[PROPERTY_NAME_SIZE];
  char partner[PROPERTY_NAME_SIZE];
  char reserved[128];
  int append;
} McaSet;

#define HSTGRPS_FILE "groups.csv"
#define HSTVIEW_FILE "configs.csv"
#define MCAVIEW_FILE "configs.csv"
#define TRACEVIEW_FILE "configs.csv"
#define TRACES_FILE "traces.csv"

typedef struct
{
  char fileName[64];
  char desc[64];
} HstViewFile;
HstViewFile *hstViewFiles = NULL;
int gNumHstViewFiles = 0;
HstViewFile *hstGrpsFiles = NULL;
int gNumHstGrpsFiles = 0;
HstViewFile **hstGrpsViewFiles = NULL;
int *gNumHstGrpsViewFiles = NULL;

HstViewFile *mcaViewFiles = NULL;
int gNumMcaViewFiles = 0;

HstViewFile *traceViewFiles = NULL;
int gNumTraceViewFiles = 0;

CSVCOLUMN csvHstViewFileColumns[] =
{
  { "FILENAME","",0,CF_TEXT,64,(BYTE *)OFFSETIN(HstViewFile,fileName),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,64,(BYTE *)OFFSETIN(HstViewFile,desc),NULL,0 },
};
CSVDB csvHstViewFileDb =
{
  csvHstViewFileColumns,
  sizeof(csvHstViewFileColumns)/sizeof(csvHstViewFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(HstViewFile)
};
CSVDB csvHstGrpsFileDb =
{
  csvHstViewFileColumns,
  sizeof(csvHstViewFileColumns)/sizeof(csvHstViewFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(HstViewFile)
};
CSVDB csvTraceViewFileDb =
{
  csvHstViewFileColumns,
  sizeof(csvHstViewFileColumns)/sizeof(csvHstViewFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(HstViewFile)
};
CSVDB csvMcaViewFileDb =
{
  csvHstViewFileColumns,
  sizeof(csvHstViewFileColumns)/sizeof(csvHstViewFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(HstViewFile)
};
/* borrowed from arctools.c: */
FILE *createAndOpen(char *fname,char *attr)
{
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  char fn[128];
  struct _finddata_t fp;
# endif
  char *c,tmpname[128],dirname[128];
  static char strdel[2] = { FS_DELIMITER , 0};

  strncpy(tmpname,fname,128); dirname[0] = 0;
  for (c=strtok(tmpname,strdel); c != NULL; c=strtok(NULL,strdel))
  {
    strcat(dirname,c);
    if (!strncmp(dirname,fname,128)) /* okay the directory should be there */
    {
      return fopen(fname,attr);
    }
    strcat(dirname,strdel);
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL) 
    {
      closedir(dp);
    }
    else
    {
      if (mkdir(dirname,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH)) return NULL;
    }
#   elif defined(WIN32)
    sprintf(fn,"%.120s*.*",dirname);
    if (_findfirst(fn, &fp) == -1)
    {
      if (mkdir(dirname)) return NULL;
    }
#   endif
  }
  return NULL;
}
int isValidFileName(char *fn)
{
  int i,len;
  if (fn == NULL || strlen(fn) == 0) return FALSE;
  if (fn[0] == '.') return FALSE;
  len = (int)strlen(fn);
  for (i=0; i<len; i++)
  {
    if (!isalnum(fn[i]) && fn[i] != '.') return FALSE;
  }
  return TRUE;
}
int readViewerFiles(void)
{
  int cc=0;
  char fn[128];
  if (hstViewFiles != NULL) 
  {
    free(hstViewFiles);
    hstViewFiles = NULL;
  }
  sprintf(fn,"%.96sCONFIGS%c",ArchiveDBpath,FS_DELIMITER);
  if ((cc=csvReadFile(fn,HSTVIEW_FILE,&csvHstViewFileDb,(void **)&hstViewFiles)) != 0) ccerr(cc);
  gNumHstViewFiles = csvHstViewFileDb.siz;
err:
  return cc;
}
int writeViewerFiles(char *subdir,char *name,HstViewFile *viewFiles,int nViewFiles)
{
  int cc=0, i;
  char fn[128];
  FILE *fp=NULL;
  sprintf(fn,"%.96s%s%c%s",ArchiveDBpath,subdir,FS_DELIMITER,name);
  if ((fp=createAndOpen(fn,"wt")) == NULL) ccerr(file_error);
  fprintf(fp,"FileName, Description\n");
  for (i=0; i<nViewFiles; i++)
  {
    fprintf(fp,"%.64s, %.64s\n",viewFiles[i].fileName,viewFiles[i].desc);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int readGroupsFiles(void)
{
  int i, cc=0;
  char fn[128];
  CSVDB db;
  if (hstGrpsFiles != NULL) 
  {
    free(hstGrpsFiles);
    hstGrpsFiles = NULL;
  }
  sprintf(fn,"%.96sCONFIGS%c",ArchiveDBpath,FS_DELIMITER);
  if ((cc=csvReadFile(fn,HSTGRPS_FILE,&csvHstGrpsFileDb,(void **)&hstGrpsFiles)) != 0) ccerr(cc);
  gNumHstGrpsFiles = csvHstGrpsFileDb.siz;
  if (gNumHstGrpsFiles <= 0) ccerr(database_not_loaded);
  if ((hstGrpsViewFiles=(HstViewFile **)SystemCalloc(gNumHstGrpsFiles,sizeof(HstViewFile))) == NULL)
    ccerr(out_of_server_memory);
  if ((gNumHstGrpsViewFiles=(int *)SystemCalloc(gNumHstGrpsFiles,sizeof(int))) == NULL)
    ccerr(out_of_server_memory);
  for (i=0; i<gNumHstGrpsFiles; i++)
  {
    sprintf(fn,"%.96sCONFIGS%c%s%c",ArchiveDBpath,FS_DELIMITER,hstGrpsFiles[i].fileName,FS_DELIMITER);
    db.col = csvHstViewFileColumns;
    db.ncols = sizeof(csvHstViewFileColumns)/sizeof(csvHstViewFileColumns[0]);
    db.siz = 0;
    db.ini = csvStdInit;
    db.itr = csvStdIterator;
    db.exi = csvStdExit;
    db.cnt = 0;
    db.tgt = NULL;
    db.tgt_siz = sizeof(HstViewFile);
    if (hstGrpsViewFiles[i] != NULL)
    {
      free(hstGrpsViewFiles[i]);
      hstGrpsViewFiles[i] = NULL;
    }
    if ((cc=csvReadFile(fn,HSTVIEW_FILE,&db,(void **)&hstGrpsViewFiles[i])) != 0) ccerr(cc);
    gNumHstGrpsViewFiles[i] = db.siz;
  }
err:
  return cc;
}
int readTraceViewerFiles(void)
{
  int cc=0;
  char fn[128];
  if (traceViewFiles != NULL)
  {
    free(traceViewFiles);
    traceViewFiles = NULL;
  }
  sprintf(fn,"%.96sTRACE%c",ArchiveDBpath,FS_DELIMITER);
  if ((cc=csvReadFile(fn,TRACEVIEW_FILE,&csvTraceViewFileDb,(void **)&traceViewFiles)) != 0) ccerr(cc);
  gNumTraceViewFiles = csvTraceViewFileDb.siz;
err:
  return cc;
}
int getViewFileListIndex(char *filename,int n,HstViewFile *hst)
{
  int i;

  for (i=0; i<n; i++)
  {
    if (!strncmp(hst[i].desc,filename,64)) return i;
  }
  return -1;
}
int appendViewerFile(char *desc,char *filename,char *subdir,int *n,CSVDB *pdb,HstViewFile **phst)
{
  int i,k,cc=0;
  char fn[256],cfgfn[72];
  FILE *fp = NULL;
  HstViewFile *hst = NULL;
  if (!isValidFileName(filename)) ccerr(invalid_name);
  sprintf(fn,"%.96s%.16s%c%.16s",ArchiveDBpath,subdir,FS_DELIMITER,filename);
  if (*n == 0)
  { /* doesn't exist yet : write the header */
    if ((fp=createAndOpen(fn,"wt")) == NULL) ccerr(file_error);
    fprintf(fp,"FileName, Description\n");
  }
  else
  {
    if ((fp=fopen(fn,"at")) == NULL) ccerr(file_error);
  }
  for (i=0,k=0; i<64; i++)
  { /* make a new file name */
    if (desc[i] == 0) break;
    if (!isalnum(desc[i])) continue;
    cfgfn[k++] = tolower(desc[i]);
  }
  cfgfn[k] = 0;
  strcat(cfgfn,".csv");
  fprintf(fp,"%.64s, %.64s\n",cfgfn,desc);
err:
  if (fp != NULL) fclose(fp);
  if (cc != 0) return -cc;
  sprintf(fn,"%.96s%.16s%c",ArchiveDBpath,subdir,FS_DELIMITER);
  if ((cc=csvReadFile(fn,filename,pdb,(void **)&hst)) != 0) ccerr(cc);
  if (*phst != NULL) free(*phst);
  *phst = hst;
  *n = pdb->siz;
  return *n;
}
int readMcaFiles(void)
{
  int cc=0;
  char fn[128];
  if (mcaViewFiles != NULL)
  {
    free(mcaViewFiles);
    mcaViewFiles = NULL;
  }
  sprintf(fn,"%.96sMCA%c",ArchiveDBpath,FS_DELIMITER);
  if ((cc=csvReadFile(fn,MCAVIEW_FILE,&csvMcaViewFileDb,(void **)&mcaViewFiles)) != 0) ccerr(cc);
  gNumMcaViewFiles = csvMcaViewFileDb.siz;

err:
  return cc;
}
HstViewFile *getViewerFile(char *fileName)
{
  int i;
  for (i=0; i<gNumHstViewFiles; i++)
  {
    if (strnicmp(hstViewFiles[i].fileName,fileName,64)) continue;
    return &hstViewFiles[i];
  }
  return NULL;
}
HstViewFile *getTraceViewerFile(char *fileName)
{
  int i;
  for (i=0; i<gNumTraceViewFiles; i++)
  {
    if (strnicmp(traceViewFiles[i].fileName,fileName,64)) continue;
    return &traceViewFiles[i];
  }
  return NULL;
}
HstViewFile *getMcaFile(char *fileName)
{
  int i;
  for (i=0; i<gNumMcaViewFiles; i++)
  {
    if (strnicmp(mcaViewFiles[i].fileName,fileName,64)) continue;
    return &mcaViewFiles[i];
  }
  return NULL;
}
int readViewSetFile(char *fileName,char *subdir,char *txtbuf,int *buflen)
{
  FILE *fp=NULL;
  char fn[128];
  int cc = 0, siz;

  if (fileName == NULL || subdir == NULL || txtbuf == NULL || buflen == NULL) ccerr(argument_list_error);
  if (*buflen < 1) ccerr(dimension_error);
  sprintf(fn,"%.96s%s%c%s",ArchiveDBpath,subdir,FS_DELIMITER,fileName);
  if (strnicmp(&fn[strlen(fn)-4],".csv",4)) strcat(fn,".csv");
  if ((fp=fopen(fn,"rt")) == NULL) 
  {
    rprtFileErr("readViewSetFile","fopen",fn);
    ccerr(no_such_file);
  }
  siz = (int)fread(txtbuf,1,*buflen,fp);
  *buflen = siz;
err:
  if (fp != NULL) fclose(fp);
  if (cc)
  {
    feclog("error reading config file %s : %s",fileName,cc2str(cc));
  }
  return cc;
}
char *gtypes[] = { "LIN", "LOG", "SLOG", "TD", "BIT", "TEXT", "ABS", "HEX" };
int ngtypes = sizeof(gtypes)/sizeof(gtypes[0]);
char *disptypes[] = {
  "POLYLINE",
  "BARLINE",
  "HISTOGRAM",
  "DOTS",
  "AHISTOGRAM",
  "RECT",
  "CIRCLE",
  "TXTBOX",
  "POLYDOTS",
  "BARDOTS",
  "HISTDOTS",
  "AHISTDOTS",
  "BARHIST",
  "BARHISTDOTS",
  "POLYXXX",
  "ANNOTATE",
  "STATUS"
};
int ndisptypes = sizeof(disptypes)/sizeof(disptypes[0]);

int getDisplayType(char *str)
{
  int i;
  if (str == NULL) return 0;
  for (i=0; i<ndisptypes; i++)
  {
    if (!stricmp(disptypes[i],str)) return i;
  }
  return 0;
}

int commitViewSet(char *fileName,char *subdir,HstViewSet *hstset,int setsiz)
{
  FILE *fp=NULL;
  char fn[128],plotstr[64];
  int i = 0, cc = 0;
  int ptype,dtype;

  if (fileName == NULL || hstset == NULL || setsiz < 1) ccerr(argument_list_error);
  sprintf(fn,"%.96s%s%c%s",ArchiveDBpath,subdir,FS_DELIMITER,fileName);
  if (strnicmp(&fn[strlen(fn)-4],".csv",4)) strcat(fn,".csv");
  if ((fp=createAndOpen(fn,"wt")) == NULL) ccerr(file_error);
  fprintf(fp,"Context, Server, Property, Device, Index, SubIndex, Scale, Offset, Min, Max, Plot, Units, Description\n");
  for (i=0; i<setsiz; i++)
  {
    ptype = (hstset[i].plot & 0xffff);
    dtype = (hstset[i].plot>>16);
    if (ptype < 0 || ptype >= ngtypes) ptype = 0;
    if (dtype < 0 || dtype >= ndisptypes) dtype = 0;
    sprintf(plotstr,"%.8s|%.12s",gtypes[ptype],disptypes[dtype]);
    fprintf(fp,"%s, %s, %s, %s, %d, %d, %g, %g, %g, %g, %s, %s, %s\n",
      hstset[i].ctx,hstset[i].srv,hstset[i].prp,hstset[i].dev,hstset[i].idx,hstset[i].sidx,hstset[i].scale,hstset[i].offset,
      hstset[i].min,hstset[i].max,plotstr,hstset[i].units,hstset[i].desc);
  }

err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int commitMcaSet(char *fileName,McaSet *mcaset,int setsiz)
{
  FILE *fp=NULL;
  char scratch[256],appnd[64],fn[256];
  int i = 0, cc = 0;

  if (fileName == NULL || mcaset == NULL || setsiz < 1) ccerr(argument_list_error);
  sprintf(fn,"%.96sMCA%c%.72s",ArchiveDBpath,FS_DELIMITER,fileName);
  if (strnicmp(&fn[strlen(fn)-4],".csv",4)) strcat(fn,".csv");
  if ((fp=createAndOpen(fn,"wt")) == NULL) ccerr(file_error);
  fprintf(fp,"Context, Server, Property, Tag, Description, Append, Shell, Alias, Partner\n");
  for (i=0; i<setsiz; i++)
  {
    memset(scratch,0,256);
    strncat(scratch,mcaset[i].srv,EXPORT_NAME_SIZE);
    if (strlen(mcaset[i].dev) > 0) sprintf(&scratch[strlen(scratch)],"/%.64s",mcaset[i].dev);
    strcpy(appnd,mcaset[i].append ? "TRUE" : "FALSE");
    if (mcaset[i].append & MCA_REVERSE_ENTRY) strcat(appnd,"|REVERSE");

    fprintf(fp,"%.32s, %.128s, %.64s, %.16s, %.64s, %.64s, %.64s, %.64s, %.64s\n",
      mcaset[i].ctx,scratch,mcaset[i].prp,mcaset[i].tag,mcaset[i].desc,appnd,mcaset[i].shell,mcaset[i].alias,mcaset[i].partner);
  }

err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int populateViewSet(char *fileName,char *subdir,HstViewSet *hstset,int *setsiz)
{
  FILE *fp=NULL;
  char str[256],scratch[256],fn[128],ctxStr[CONTEXT_NAME_SIZE+8],devStr[DEVICE_NAME_SIZE+8];
  char *c,*ctx,*srv,*dev;
  int n = 0, done = 0,cc = 0;
  int ctx_col=-1,srv_col=-1,dev_col=-1,nam_col=-1,idx_col=-1,sub_col=-1,sca_col=-1,off_col=-1;
  int min_col=-1,max_col=-1,plt_col=-1,uni_col=-1,hlp_col=-1;

  if (fileName == NULL || hstset == NULL || setsiz == NULL) ccerr(argument_list_error);
  if (*setsiz < 1) ccerr(dimension_error);
  sprintf(fn,"%.96s%s%c%s",ArchiveDBpath,subdir,FS_DELIMITER,fileName);
  if (strnicmp(&fn[strlen(fn)-4],".csv",4)) strcat(fn,".csv");
  if ((fp=fopen(fn,"rt")) == NULL) 
  {
    rprtFileErr("populateViewSet","fopen",fn);
    ccerr(no_such_file);
  }
  while (csvGetLine(str,255,fp) != NULL )
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      ctx_col = findcol(str,"CONTEXT");
      dev_col = findcol(str,"DEVICE");
      if ((srv_col=findcol(str,"SERVER")) < 0) ccerr(no_such_line);
      if ((nam_col=findcol(str,"NAME")) < 0 && 
          (nam_col=findcol(str,"PROPERTY")) < 0) ccerr(no_such_line);
      idx_col = findcol(str,"INDEX");
      sub_col = findcol(str,"SUBINDEX");
      sca_col = findcol(str,"SCALE");
      off_col = findcol(str,"OFFSET");
      min_col = findcol(str,"MIN");
      max_col = findcol(str,"MAX");
      if ((plt_col=findcol(str,"LIN/LOG")) < 0) plt_col = findcol(str,"PLOT");
      uni_col = findcol(str,"UNITS");
      if ((hlp_col=findcol(str,"DESCRIPTION")) < 0) hlp_col = findcol(str,"HELP");
      done = 1;
      continue;
    }
    if (n >= *setsiz) break;
    memset(&hstset[n],0,sizeof(HstViewSet));
    strncpy(scratch,colptr(srv_col,str),256);
    /* allow '/<context>/<server>/<device>' */
    if (scratch[0] == '/')
    {
      ctx = &scratch[1];
      if ((c=strchr(ctx,'/')) == NULL) continue;
      *c++ = 0;
      srv = c;
    }
    else
    {
      ctx = NULL;
      srv = scratch;
    }
    if ((c=strchr(srv,'/')) != NULL)
    {
      *c++ = 0;
      dev = c;
    }
    else
    {
      dev = NULL;
    }
    if (ctx_col >= 0)
    { /* context column takes precedence */
      strncpy(ctxStr,colptr(ctx_col,str),CONTEXT_NAME_SIZE);
      ctxStr[CONTEXT_NAME_SIZE] = 0;
      ctx = ctxStr;
    }
    if (ctx == NULL) ctx = gDeviceContext;
    strncpy(hstset[n].ctx,ctx,CONTEXT_NAME_SIZE);
    strncpy(hstset[n].srv,srv,EXPORT_NAME_SIZE);
    if (idx_col >= 0) 
    {
      strncpy(scratch,colptr(idx_col,str),16);
      hstset[n].idx = atoi(scratch);
    }
    if (dev_col >= 0)
    { /* device column takes precedence */
      strncpy(devStr,colptr(dev_col,str),DEVICE_NAME_SIZE);
      devStr[DEVICE_NAME_SIZE] = 0;
      dev = devStr;
    }
    if (dev == NULL || strlen(dev) == 0) 
    { /* not given : fix as the index number */
      sprintf(devStr,"#%d",hstset[n].idx);
      dev = devStr; 
    }
    strncpy(hstset[n].dev,dev,DEVICE_NAME_SIZE);
    strncpy(hstset[n].prp,colptr(nam_col,str),PROPERTY_NAME_SIZE);
    if (sub_col >= 0) 
    {
      strncpy(scratch,colptr(sub_col,str),16);
      hstset[n].sidx = atoi(scratch);
    }
    if (sca_col >= 0) 
    {
      strncpy(scratch,colptr(sca_col,str),32);
      hstset[n].scale = (float)strtod(scratch,NULL);
    }
    if (off_col >= 0) 
    {
      strncpy(scratch,colptr(off_col,str),32);
      hstset[n].offset = (float)strtod(scratch,NULL);
    }
    if (min_col >= 0) 
    {
      strncpy(scratch,colptr(min_col,str),32);
      hstset[n].min = (float)strtod(scratch,NULL);
    }
    if (max_col >= 0) 
    {
      strncpy(scratch,colptr(max_col,str),32);
      hstset[n].max = (float)strtod(scratch,NULL);
    }
    if (plt_col >= 0) 
    {
      strncpy(scratch,colptr(plt_col,str),32);
      if ((c=strchr(scratch,'|')) != NULL)
      {
        *c++ = 0;
        hstset[n].plot = (getDisplayType(c) << 16);
      }
      else
      {
        hstset[n].plot = 0;
      }
      hstset[n].plot |= getGraphType(scratch);
    }
    if (uni_col >= 0) 
    {
      strncpy(hstset[n].units,colptr(uni_col,str),UNITS_SIZE);
    }
    if (hlp_col >= 0) 
    {
      strncpy(hstset[n].desc,colptr(hlp_col,str),PROPERTY_DESC_SIZE);
    }
    n++;
  }
  *setsiz = n;
err:
  if (fp != NULL) fclose(fp);
  if (cc)
  {
    feclog("error reading config file %s : %s",fileName,cc2str(cc));
  }
  return cc;
}
void rprtFileErr(const char *src,const char *txt,const char *filename)
{
  char msgbuf[256], errbuf[256] = "";
  static int errcnt = 0;
  static int last_errno = 0;
# if defined(WIN32)
  LPVOID lpMsgBuf;
  DWORD lastErr = GetLastError();
  if (ServerExitCondition) return;
  if (FormatMessage( 
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,lastErr,
        MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US), 
        (LPTSTR) &lpMsgBuf,0,NULL) > 0)
  {
    snprintf(errbuf,256,"%s : %s",txt,(char *)lpMsgBuf);
    LocalFree( lpMsgBuf );
  }
  else
  { 
    snprintf(errbuf,256,"%s : %d",txt,lastErr);
  }
  last_errno = (int)lastErr;
# elif defined(UNIX)
  if (ServerExitCondition) return;
  last_errno = errno;
  snprintf(errbuf,256,"%s : %s",txt,strerror(errno));
# endif
  if (errno == ENOENT) return;
  snprintf(msgbuf,256,"in %.16s %.16s file %.32s: %.256s",src,txt,filename,errbuf);
  feclogEx(TINE_LOGLEVEL_ERROR,msgbuf);
  if (++errcnt > 10)
  {
    feclogEx(TINE_LOGLEVEL_ERROR,"too many file errors (%d) : die and live another day",errcnt);
    exit(1);
  }
}
int populateMcaSet(char *fileName,McaSet *mcaset,int *setsiz)
{
  FILE *fp=NULL;
  char str[256],scratch[256],fn[128];
  char *c,*srv,*dev;
  int n = 0, done = 0,cc = 0;
  int ctx_col=-1,srv_col=-1,prp_col=-1,tag_col=-1,dsc_col=-1;
  int alias_col=-1,partner_col=-1,append_col=-1,shell_col=-1;

  if (fileName == NULL || mcaset == NULL || setsiz == NULL) ccerr(argument_list_error);
  if (*setsiz < 1) ccerr(dimension_error);
  sprintf(fn,"%.96sMCA%c%s",ArchiveDBpath,FS_DELIMITER,fileName);
  if (strnicmp(&fn[strlen(fn)-4],".csv",4)) strcat(fn,".csv");
  if ((fp=fopen(fn,"rt")) == NULL) 
  {
    rprtFileErr("populateMcaSet","fopen",fn);
    ccerr(no_such_file);
  }
  while (csvGetLine(str,255,fp) != NULL )
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      ctx_col = findcol(str,"CONTEXT");
      if ((srv_col=findcol(str,"SERVER")) < 0) ccerr(no_such_line);
      if ((prp_col=findcol(str,"PROPERTY")) < 0) ccerr(no_such_line);
      tag_col = findcol(str,"TAG");
      dsc_col = findcol(str,"DESCRIPTION");
      alias_col = findcol(str,"ALIAS");
      partner_col = findcol(str,"PARTNER");
      if ((shell_col=findcol(str,"SHELL")) < 0) shell_col = findcol(str,"HIGH");
      append_col = findcol(str,"APPEND");
      done = 1;
      continue;
    }
    if (n >= *setsiz) break;
    memset(&mcaset[n],0,sizeof(McaSet));
    strncpy(scratch,colptr(srv_col,str),256);
    srv = scratch;
    if ((c=strchr(srv,'/')) != NULL)
    {
      *c++ = 0;
      dev = c;
    }
    else
    {
      dev = NULL;
    }
    if (ctx_col >= 0) 
    {
      strncpy(mcaset[n].ctx,colptr(ctx_col,str),CONTEXT_NAME_SIZE);
    }
    strncpy(mcaset[n].srv,srv,EXPORT_NAME_SIZE);
    if (dev != NULL) strncpy(mcaset[n].dev,dev,DEVICE_NAME_SIZE);
    strncpy(mcaset[n].prp,colptr(prp_col,str),PROPERTY_NAME_SIZE);
    if (tag_col >= 0) 
    {
      strncpy(mcaset[n].tag,colptr(tag_col,str),TAG_NAME_SIZE);
    }
    if (dsc_col >= 0) 
    {
      strncpy(mcaset[n].desc,colptr(dsc_col,str),PROPERTY_DESC_SIZE);
    }
    if (shell_col >= 0) 
    {
      strncpy(mcaset[n].shell,colptr(shell_col,str),64);
    }
    if (alias_col >= 0) 
    {
      strncpy(mcaset[n].alias,colptr(alias_col,str),64);
    }
    if (partner_col >= 0) 
    {
      strncpy(mcaset[n].partner,colptr(partner_col,str),64);
    }
    if (append_col >= 0) 
    {
      strncpy(scratch,colptr(append_col,str),64);
      strupr(scratch);
      if (strstr(scratch,"TRUE") != NULL || strstr(scratch,"YES") != NULL) mcaset[n].append = MCA_APPEND_ENTRY;
      if (strstr(scratch,"REVERSE") != NULL) mcaset[n].append = MCA_APPEND_ENTRY|MCA_REVERSE_ENTRY;
    }
    n++;
  }
  *setsiz = n;
err:
  if (fp != NULL) fclose(fp);
  if (cc)
  {
    feclog("error reading config file %s : %s",fileName,cc2str(cc));
  }
  return cc;
}

DAQRecord *daqTbl;
int nxtags;

#define addNumberRecords(f) {\
  sprintf(fn,"%s%s",ArchiveDBpath,f);\
  if ((fp = fopen(fn,"rt")) == NULL) break;\
  totalNrRecords += getNumLines(fp);\
  fclose(fp); }

#define cctxt(txt) { feclog("HST: %s",txt); ccerr(database_not_loaded); }
#define quit(i) { feclog("HST: %s",erlst[i]); exit(1); }

time_t tvalid = 0;
int getDatabaseInfo(void)
{
  FILE *fp=NULL;
  char str[256],scratch[170];
  int done = 0,cc = 0;
  int database_col,valid_col=-1;

  if ((fp=fopen("database.csv","rt")) == NULL && 
      (fp=fopen("startup.csv","rt")) == NULL)
  {
    if (strlen(adbFileName) == 0)
    {
      if (strlen(thisContext) > 0)
      { /* if context known : append it to default name */
        sprintf(adbFileName,"data%.32s.csv",thisContext);
        strlwr(adbFileName);
      }
      else
      {
        strncpy(adbFileName,DEFAULT_ADB_FILENAME,64);
      }
    }
    feclog("did not find data base file: using database %s",adbFileName);
    addADBtoList(adbFileName);
    return 0;
  }
  while (csvGetLine(str,255,fp) != NULL )
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((database_col=findcol(str,"DATABASE")) < 0) ccerr(no_such_line);
      valid_col = findcol(str,"LASTVALID");
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(database_col,str),170);
    addADBtoList(scratch);
    if (valid_col >= 0)
    {
      tvalid = parseTimeString(colptr(valid_col,str));
    }
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int addADBtoList(char *fn)
{
  int i;

  for (i=0; i<nadbfiles && strncmp(adbfile[i],fn,64); i++);
  if (i < nadbfiles) return 0;
  if (nadbfiles == MAX_ADBFILES) return out_of_local_memory;
  strncpy(adbfile[i],fn,64);
  nadbfiles++;
  return 0;
}

extern int gNumVolatileDaqTblEntries;
NAME64 *volatileKeys = NULL;
int gNumVolatileKeys = 0;
int readVolatileKeys(void)
{
  int i, k, n, cc=0;
  if (volatileKeys != NULL)
  {
    free(volatileKeys);
    volatileKeys = NULL;
  }
  if (gNumVolatileDaqTblEntries == 0) 
  { /* nothing volatile found (oh, what luck!) */
    return 0;
  }
  if ((volatileKeys=(NAME64 *)calloc(gNumVolatileDaqTblEntries,sizeof(NAME64))) == NULL) 
    ccerr(out_of_server_memory);
  for (i=0,n=0; i<totalNrRecords && n<gNumVolatileDaqTblEntries; i++)
  {
    if (!daqTbl[i].hasVolatileMca || daqTbl[i].nparam == 0) continue;
    if (daqTbl[i].nparam > 1) feclog("ambiguous volatile table entry: %s maps to %d parameters (assume all are volatile)!",
      daqTbl[i].prp, daqTbl[i].nparam);
    for (k=0; k<daqTbl[i].nparam; k++)
    {
      strncpy(volatileKeys[n].name,daqTbl[i].daqParam[k].param,64);
      n++;
    }
  }
  gNumVolatileKeys = n;
err:
  if (cc != 0) feclog("readVolatileKeys: %s",cc2str(cc));
  return cc;
}
//TODO: hash the volatile keys list !
int isVolatileKeyword(char *keyword)
{ /* keyword is never NULL ! */
  int i, p;
  char *c=strrchr(keyword,'.');
  if (c != NULL && stricmp(c,".NAM") == 0) 
  { /* a .NAM keyword is never volatile ! */
    return FALSE;
  }
  if ((i=getEntryFromKeyword(keyword,NULL,&p)) < 0) return FALSE;
  if (daqTbl[i].hasVolatileMca) return TRUE;
  return FALSE;
}
int dumpVolatileEntries(int a0,int a1,int a2,int a3)
{
  int i, k, n=0;
  DAQParameter *daqParam;
  ttyoutput("the following record keywords have been marked as volatile:");
  for (i=0; i<nrecords; i++)
  {
    if (!daqTbl[i].active) continue;
    if (!daqTbl[i].hasVolatileMca) continue;
    daqParam = daqTbl[i].daqParam;
    for (k=0; k<daqTbl[i].nparam; k++)
    {
      ttyoutput("\t%.64s",daqParam[k].param);
      n++;
    }
  }
  return n;
}
#define ADMINS_FILE "admins.csv"
int nadmins = 0;
NAME16 *admins = NULL;
int isAdministrator(char *nam)
{
  int i;
  if (nadmins == 0) return -1;
  for (i=0; i<nadmins && strnicmp(nam,admins[i].name,USERNAME_SIZE); i++);
  return nadmins - i;
}
int getAdministrators(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int usr_col=-1;
  char str[256],hdr[256];
 
  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (admins != NULL)
  {
    free(admins);
    nadmins = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((admins=(NAME16 *)calloc(n,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
  nadmins = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; csvGetLine(str,255,fp) != NULL && i<nadmins;)
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((usr_col=findcol(hdr,"USERNAME")) < 0 &&
          (usr_col=findcol(hdr,"USERS")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(admins[i].name,colptr(usr_col,str),USERNAME_SIZE);
    i++;
  }
  feclog("%d administrators found",nadmins);
  if (nadmins == 0) feclog("EVERYONE can administor");
  else for (i=0; i<n; i++) if (strlen(admins[i].name) > 0) feclog("%.16s can administor",admins[i].name);
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
void PreSystemInit(void)
{
  SetWorkAreaSize(MAX_TRANSPORTSIZE*16);
  SetPacketMTU(6400);
  gScanForNameLists = FALSE;
  gRetardSingleContractRemoval = FALSE;
  SystemPresetMemory = TRUE;
  gSyncTolerance = 1.0;

  if (startupInputsGiven)
  {
    int p;
    if (strlen(thisContext) == 0)
    {
      strcpy(thisContext,"TEST");
    }
    if (strlen(thisFecName) == 0)
    {
      strncpy(thisFecName,thisContext,10); thisFecName[10] = 0;
      strncat(thisFecName,"ARCFEC",6);
    }
    if (thisPortOffset == 0)
    {
      p = GetPortOffset(thisFecName);
      if (p >= 0) thisPortOffset = p;
    }
    sprintf(thisFecDescription,"%.32s central archive server",thisContext);
    RegisterFecInformation(thisFecName,thisSubSystem,thisContext,thisFecDescription,thisLocation,"none","administrator",thisPortOffset);
  }

}
void PostSystemInit(void)
{
  int cc;
  DTYPE dout;

  signal(SIGFPE,SIG_IGN); /* catch floating point domain errors */
  if (strlen(thisContext) == 0)
  {
    strncpy(thisContext,gDeviceContext,CONTEXT_NAME_SIZE);
  }
  if ((cc=RegisterEquipmentModule(thisServerName,HSTEQM_NAME,1,hsteqm,hstinit,NULL,0,hstexit)) != 0)
  {
    dbglog("Can't register %.6s : %s\n",HSTEQM_NAME,cc2str(cc));
  }
  else
  {
    dout.dFormat = CF_STRUCT; dout.dArrayLength = 1024; dout.dTag[0] = 0;
    RegisterPropertyInformation(HSTEQM_NAME,"FETCH",
         &dout,NULL,CA_READ,AT_UNKNOWN,0,"general data retrieval (deprectated)",-1,NULL);
  }
  if (strlen(thisFailoverMasterName) > 0)
  {
    if (strlen(thisFailoverSlaveMasterName) > 0)
    { /* we're a failover slave ... */
      SetFailoverSlave(HSTEQM_NAME,thisFailoverMasterName,thisFailoverSlaveMasterName);
    }
    else
    { /* we're a failover master ... */
      SetFailoverMaster(HSTEQM_NAME,thisFailoverMasterName);
    }
  }
  getAdministrators(ADMINS_FILE);
  SetFailoverThreshold(60);
}
int needsRestart = FALSE;
void RestartServer(void)
{
  static int restartCountdown = 5;
  if (restartCountdown-- > 0) return;
  FlushContractTable();
  SetServerCycleIdle(TRUE);
  FreeEQPModules();
  ServerExitCondition = 0;
  getRegisteredExports(EXPORT_FILE,NULL);
  PostSystemInit();
  SetServerCycleIdle(FALSE);
  needsRestart = FALSE;
  restartCountdown = 5;
}
void RegisterStructs(void)
{
  static int done = FALSE;
  if (done) return;
  done = TRUE;
  if (addFieldToStruct("ARS",OFFSETIN(ARS,starttime),1,CF_INT32,"starttime")) quit(out_of_local_memory);
  if (addFieldToStruct("ARS",OFFSETIN(ARS,index),1,CF_INT16,"index")) quit(out_of_local_memory);
  if (addFieldToStruct("ARS",OFFSETIN(ARS,minutes),1,CF_INT16,"minutes")) quit(out_of_local_memory);
  if (addFieldToStruct("ARS",OFFSETIN(ARS,keyword),16,CF_TEXT,"keyword")) quit(out_of_local_memory);
  if (sealTaggedStruct("ARS",sizeof(ARS),1)) quit(out_of_local_memory);

  if (addFieldToStruct("HstCfg",OFFSETIN(HstViewSet,ctx),CONTEXT_NAME_SIZE,CF_TEXT,"context")) quit(out_of_local_memory);
  if (addFieldToStruct("HstCfg",OFFSETIN(HstViewSet,srv),EXPORT_NAME_SIZE,CF_TEXT,"server")) quit(out_of_local_memory);
  if (addFieldToStruct("HstCfg",OFFSETIN(HstViewSet,dev),DEVICE_NAME_SIZE,CF_TEXT,"device")) quit(out_of_local_memory);
  if (addFieldToStruct("HstCfg",OFFSETIN(HstViewSet,prp),PROPERTY_NAME_SIZE,CF_TEXT,"property")) quit(out_of_local_memory);
  if (addFieldToStruct("HstCfg",OFFSETIN(HstViewSet,units),UNITS_SIZE,CF_TEXT,"units")) quit(out_of_local_memory);
  if (addFieldToStruct("HstCfg",OFFSETIN(HstViewSet,desc),PROPERTY_DESC_SIZE,CF_TEXT,"description")) quit(out_of_local_memory);
  if (addFieldToStruct("HstCfg",OFFSETIN(HstViewSet,idx),1,CF_INT32,"index")) quit(out_of_local_memory);
  if (addFieldToStruct("HstCfg",OFFSETIN(HstViewSet,sidx),1,CF_INT32,"subIndex")) quit(out_of_local_memory);
  if (addFieldToStruct("HstCfg",OFFSETIN(HstViewSet,scale),1,CF_FLOAT,"scale")) quit(out_of_local_memory);
  if (addFieldToStruct("HstCfg",OFFSETIN(HstViewSet,offset),1,CF_FLOAT,"offset")) quit(out_of_local_memory);
  if (addFieldToStruct("HstCfg",OFFSETIN(HstViewSet,max),1,CF_FLOAT,"max")) quit(out_of_local_memory);
  if (addFieldToStruct("HstCfg",OFFSETIN(HstViewSet,min),1,CF_FLOAT,"min")) quit(out_of_local_memory);
  if (addFieldToStruct("HstCfg",OFFSETIN(HstViewSet,plot),1,CF_INT32,"plot")) quit(out_of_local_memory);
  /* terminate the structure definition like this! */
  if (sealTaggedStruct("HstCfg",sizeof(HstViewSet),MAX_NUM_CONFIGS)) quit(out_of_local_memory);

  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,ctx),CONTEXT_NAME_SIZE,CF_TEXT,"context")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,srv),EXPORT_NAME_SIZE,CF_TEXT,"server")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,dev),DEVICE_NAME_SIZE,CF_TEXT,"device")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,prp),PROPERTY_NAME_SIZE,CF_TEXT,"property")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,tag),TAG_NAME_SIZE,CF_TEXT,"tag")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,desc),PROPERTY_DESC_SIZE,CF_TEXT,"description")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,shell),64,CF_TEXT,"shell")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,alias),PROPERTY_NAME_SIZE,CF_TEXT,"alias")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,partner),PROPERTY_NAME_SIZE,CF_TEXT,"partner")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,reserved),128,CF_TEXT,"reserved")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,append),1,CF_INT32,"append")) quit(out_of_local_memory);
  if (sealTaggedStruct("McaCfg",sizeof(McaSet),MAX_NUM_CONFIGS)) quit(out_of_local_memory);

  if (addFieldToStruct("HstView",OFFSETIN(HstViewFile,fileName),64,CF_TEXT,"fileName")) quit(out_of_local_memory);
  if (addFieldToStruct("HstView",OFFSETIN(HstViewFile,desc),64,CF_TEXT,"desc")) quit(out_of_local_memory);
  if (sealTaggedStruct("HstView",sizeof(HstViewFile),MAX_NUM_CONFIGS)) quit(out_of_local_memory);

  RegisterUserCommand("ArchiveDebug",NULL,&archdebug,NULL,CA_READ+CA_WRITE);
}
extern CSVDB csvExportsInfoDb;
int populateXIS(char *fecname,char *eqmname,DBLNKLIST *fecList,XIS **xis);
XIS *additionalKeywords = NULL;
int nAdditionalKeywords = 0;
int getAdditionalKeywords(void)
{ /* re-scans the exports.csv file (if there) for aliases or redirections */
  int i, sz, cc=0;

  if (additionalKeywords != NULL) return 0;
  if (isUsingXMLFecDatabase)
  {
    sz = populateXIS(gFecName,HSTEQM_NAME,gFecInfoList,&additionalKeywords);
    if (sz < 0) ccerr(-sz);
  }
  else
  {
    if ((cc=csvReadFile(fecDbPath,"exports.csv",&csvExportsInfoDb,(void **)&additionalKeywords)) != 0) ccerr(cc);
    sz = csvExportsInfoDb.siz;
  }
  nAdditionalKeywords = sz;
  for (i=0; i<nAdditionalKeywords; i++)
  {
    checkDaqSubSystem(additionalKeywords[i].SubSystem);
  }
  feclog("getAdditionalKeywords : added %d additional keywords",sz);
  SetExportedSubSystem(HSTEQM_NAME,"SER");
err:
  if (cc != 0 && cc != no_such_file) feclog("getAdditionalKeywords : %.32s",cc2str(cc));
  return cc;
}
int getRedirectedIndex(char *prp)
{
  int i;
  for (i=0; i<nAdditionalKeywords; i++)
  {
    if (strcmp(additionalKeywords[i].EqmName,HSTEQM_NAME)) continue;
    if (!strnicmp(prp,additionalKeywords[i].PrpName,PROPERTY_NAME_SIZE)) 
    {
      if (strlen(additionalKeywords[i].redirection) > 0) return i;
    }
  }
  return -1;
}
char primaryDbFile[64];
time_t dbLoadTime = 0;
void daqInit(void)
{
  int id,i,n,np,off=0,cc=0,rlen=0,nprps=0;
  static char fn[80],prp[PROPERTY_NAME_SIZE+8],desc[256];
  DAQParameter *daqParam;
  int p, rec;
  UINT32 cap;
  FILE *fp;
  DTYPE dout, din;

  srvlog(ALF,"daqInit","--- initializing ---");

  if (gDataBuffer == NULL)
  {
    gDataBuffer = (BYTE *)malloc(DATABLOB_SIZE);
    if (gDataBuffer == NULL) ccerr(out_of_server_memory);
    gDataBufferCapacity = DATABLOB_SIZE;
    srvlog(ALF,"daqInit","allocated %d bytes for reading trace data",gDataBufferCapacity);
  }
  gAppendProperyAliasNamesToQueries = FALSE;
  gAllocDataBuffers = FALSE;
  nadbfiles = 0;

  getFilterTable();
  getDatabaseInfo();
  RegisterStructs();
  readViewerFiles();
  readGroupsFiles();
  readMcaFiles();
  readTraceViewerFiles();
  totalNrRecords = 0;
  nextFreeRecord = -1;
  for (i=0; i<nadbfiles; i++)
  {
    addNumberRecords(adbfile[i]);
  }
  recordsCapacity = totalNrRecords + RECORD_CAPACITY_BUFFER;
  if ((daqTbl=(DAQRecord *)calloc(recordsCapacity,sizeof(DAQRecord))) == NULL) ccerr(out_of_server_memory);
  if ((prpAlias=(PRP_ALIAS *)calloc(recordsCapacity,sizeof(PRP_ALIAS))) == NULL) ccerr(out_of_server_memory);
  if ((subSystems=(SUB_SYS *)calloc(recordsCapacity,sizeof(SUB_SYS))) == NULL) ccerr(out_of_server_memory);

  registerDaqStructs(recordsCapacity*2);

  for (i=0; i<nadbfiles; i++)
  {
    if (!(n=loadDaqDb(adbfile[i],off))) break;
    inputDataLoad(n,nrecords); /* load input data for dynamic configuration changes */
    if (!(np=loadDaqParamDb(n,off))) cctxt("No Parameters loaded!");
    off += n;
  }
  ArchiveDBfile = primaryDbFile;
  strncpy(ArchiveDBfile,adbfile[0],64); primaryDbFile[63] = 0;
  nrecords = totalNrRecords;
  fillinArrayTypes(nrecords);
  addExtSrcInfo();
  for (i=0,ntags=0; i<totalNrRecords; i++)       
  { /* register properties */
    if (daqTbl[i].option&NEVER) continue; /* it's not being archived */
    if (!daqTbl[i].active) continue;      /* legacy entry, don't count it! */
    if (daqTbl[i].daqParam == NULL)
    {
      daqTbl[i].active = FALSE;
      feclog("record %d has no valid keyword parameters!",daqTbl[i].rec);
      continue;
    }
    ntags += daqTbl[i].nparam;
    for (n=0; n<daqTbl[i].nparam; n++)
    {
      if (daqTbl[i].daqParam[n].arrayType & AT_SPECTRUM)
      {
        sprintf(desc,"[%g:%g %.16s!%.16s][%g:%g %.16s]%.64s",daqTbl[i].daqParam[n].minimum,
                daqTbl[i].daqParam[n].maximum,daqTbl[i].daqParam[n].units,daqTbl[i].daqParam[n].plot,
                daqTbl[i].daqParam[n].xminimum,daqTbl[i].daqParam[n].xmaximum,daqTbl[i].daqParam[n].xunits,
                daqTbl[i].daqParam[n].desc);
      }
      else
      {
        sprintf(desc,"[%g:%g %.16s!%.16s]%.64s",daqTbl[i].daqParam[n].minimum,
                daqTbl[i].daqParam[n].maximum,daqTbl[i].daqParam[n].units,
                daqTbl[i].daqParam[n].plot,daqTbl[i].daqParam[n].desc);
      }
      dout.dFormat = daqTbl[i].daqParam[n].fmt;
      dout.dArrayLength = daqTbl[i].daqParam[n].size;
      if (daqTbl[i].fmt == CF_STRUCT) strncpy(dout.dTag,daqTbl[i].tag,TAG_NAME_SIZE); else dout.dTag[0] = 0;
      /* n.b. scaleOffset as the row length is a kluge to be removed at some later date (only applies to .ENUM) */
      rlen = (daqTbl[i].daqParam[n].arrayType & AT_DOUBLE) ? daqTbl[i].daqParam[n].rowsize : daqTbl[i].daqParam[n].size;
      if ((id=RegisterPropertyInformation(HSTEQM_NAME,daqTbl[i].daqParam[n].param,
           &dout,NULL,CA_READ,daqTbl[i].daqParam[n].arrayType,(UINT16)rlen,desc,-1,NULL)) < 0)
      {
        cc = -id;
        feclog("Couldn't register property %.64s: %.32s",daqTbl[i].daqParam[n].param,cc2str(cc));
      }
      nprps++;
      if (!strnicmp(daqTbl[i].ctx,thisContext,CONTEXT_NAME_SIZE) && 
           daqTbl[i].nparam == 1 && (daqTbl[i].option & ONCE) == 0)
      { /* this is a one to one acq property and hst parameter and not a names acq */
        if (strlen(daqTbl[i].srv) + strlen(daqTbl[i].prp) < PROPERTY_NAME_SIZE)
        { /* the lengths will fit as an alias */
#         if defined(WIN32)
          _snprintf(prp,PROPERTY_NAME_SIZE,"%s.%s",daqTbl[i].srv,daqTbl[i].prp);
#         elif defined(UNIX)
          snprintf(prp,PROPERTY_NAME_SIZE,"%s.%s",daqTbl[i].srv,daqTbl[i].prp);
#         else
          sprintf(prp,"%.23s.%40s",daqTbl[i].srv,daqTbl[i].prp);
#         endif
          if (strnicmp(prp,daqTbl[i].daqParam[n].param,PROPERTY_NAME_SIZE) != 0)
          { /* and this isn't already the property name ! */
            RegisterPropertyAlias(HSTEQM_NAME,daqTbl[i].daqParam[n].param,prp);
          }
        }
      }
    }
  }
  nxtags = 0;
  for (n=0; n<nPrpAlias; n++)
  { /* .alias should be an existing record ! => .name is fictive alias name */
    if ((i=getArchivedEntryFromKeyword(prpAlias[n].alias,&rec,&p)) < 0) continue;
    daqParam = daqTbl[i].daqParam;
    dout.dFormat = daqParam[p].fmt;
    dout.dArrayLength = daqParam[p].size;
    if (daqTbl[i].fmt == CF_STRUCT) strncpy(dout.dTag,daqTbl[i].tag,TAG_NAME_SIZE); else dout.dTag[0] = 0;
    rlen = (daqParam[p].arrayType & AT_DOUBLE) ? (int)daqParam[p].scaleOffset : daqParam[p].size;
    sprintf(desc,"[%g:%g %.16s!%.16s]%.64s",daqParam[p].minimum,
            daqParam[p].maximum,daqParam[p].units,
            daqParam[p].plot,daqParam[p].desc);
    if ((id=RegisterPropertyInformation(HSTEQM_NAME,prpAlias[n].name,
         &dout,NULL,CA_READ,daqParam[p].arrayType,(UINT16)rlen,desc,-1,NULL)) < 0)
    {
      cc = -id;
      feclog("Couldn't register property %.64s: %.32s",prpAlias[n].name,cc2str(cc));
    }
    nprps++;
    nxtags++;
  }

  if (nprps > 256) gIgnoreMetaPropertyQueries = TRUE;
  getAdditionalKeywords();

  n = ntags + nAdditionalKeywords;
  RegisterProperty(HSTEQM_NAME,"NSUBSYSTEMS",1,CF_INT32,CA_READ,"number of archive subsystems");
  RegisterProperty(HSTEQM_NAME,"SUBSYSTEMS",nSubSystems,CF_NAME16,CA_READ,"archive subsystems");
  RegisterProperty(HSTEQM_NAME,"NTAGS",1,CF_INT32,CA_READ,"number of keywords");
  RegisterProperty(HSTEQM_NAME,"NKEYWORDS",1,CF_INT32,CA_READ,"number of keywords");
  RegisterProperty(HSTEQM_NAME,"NKEYWORDS.VOLATILE",1,CF_INT32,CA_READ,"number of volatile keywords");
  RegisterProperty(HSTEQM_NAME,"TAGS",n,CF_NAME32,CA_READ,"archive keywords");
  RegisterProperty(HSTEQM_NAME,"KEYWORDS",n,CF_NAME64,CA_READ,"archive keywords");
  RegisterProperty(HSTEQM_NAME,"KEYWORDS.VOLATILE",gNumVolatileKeys,CF_NAME64,CA_READ,"volatile archive keywords");
  RegisterProperty(HSTEQM_NAME,"KEYWORDS.PRIMARY",n,CF_NAME64,CA_READ,"primary archive keywords");
  RegisterProperty(HSTEQM_NAME,"RECORDSIZE",1,CF_INT32,CA_READ,"record keyword data length at timestamp");
  RegisterProperty(HSTEQM_NAME,"KEYWORD.SIZE",1,CF_INT32,CA_READ,"keyword data length at timestamp");
  RegisterProperty(HSTEQM_NAME,"RECORDSOURCE",3,CF_NAME32,CA_READ,"source of record");
  RegisterProperty(HSTEQM_NAME,"RECORD.SOURCE",3,CF_NAME32,CA_READ,"source of record");

  RegisterProperty(HSTEQM_NAME,"DB.LASTLOADED",1,CF_INT32,CA_READ,"most recent archive database load time (UTC)");
  RegisterProperty(HSTEQM_NAME,"DB.LASTLOADED",32,CF_TEXT,CA_READ,"most recent archive database load time (string)");

  RegisterProperty(HSTEQM_NAME,"FILTERTABLE",32,CF_NAME64,CA_READ,"read filter table");
  dout.dFormat = CF_STRUCT; dout.dArrayLength = 32; strncpy(dout.dTag,DAQFltrTag,TAG_NAME_SIZE);
  RegisterPropertyInformation(HSTEQM_NAME,"FILTERTABLE",&dout,&dout,CA_READ|CA_WRITE,AT_UNKNOWN,0,"read/update filter table",-1,NULL);

  /* database access properties */
  DTYPEZERO(din);
  dout.dFormat = CF_STRUCT; dout.dArrayLength = 1000; strncpy(dout.dTag,DAQRecTag,TAG_NAME_SIZE);
  RegisterPropertyInformation(HSTEQM_NAME,"RECORD.INFO",&dout,NULL,CA_READ,AT_UNKNOWN,0,"Record information database",-1,NULL);
  dout.dFormat = CF_TEXT; dout.dArrayLength = 256; dout.dTag[0] = 0;
  RegisterPropertyInformation(HSTEQM_NAME,"RECORD.ERR",&dout,NULL,CA_READ,AT_UNKNOWN,0,"Last DBM Record Configuration error",-1,NULL);
  dout.dFormat = CF_TEXT; dout.dArrayLength = 1024; dout.dTag[0] = 0;
  RegisterPropertyInformation(HSTEQM_NAME,"RECORD.CMT",&dout,&dout,CA_READ|CA_WRITE,AT_UNKNOWN,0,"set/get annotate record (device='prop@ts')",-1,NULL);
  dout.dFormat = CF_INT32; dout.dArrayLength = 100; dout.dTag[0] = 0;
  din.dFormat = CF_INT32; din.dArrayLength = 2;
  RegisterPropertyInformation(HSTEQM_NAME,"RECORD.CMTS",&dout,&din,CA_READ,AT_UNKNOWN,0,"get record annotation time list (device='prop')",-1,NULL);
  dout.dFormat = CF_INTINT; dout.dArrayLength = 100; dout.dTag[0] = 0;
  RegisterPropertyInformation(HSTEQM_NAME,"RECORD.CMTS",&dout,&din,CA_READ,AT_UNKNOWN,0,"get record annotation time and active list (device='prop')",-1,NULL);
  dout.dFormat = CF_INT32; dout.dArrayLength = 100; dout.dTag[0] = 0;
  RegisterPropertyInformation(HSTEQM_NAME,"RECORDS.WITH.CMTS",&dout,&din,CA_READ,AT_UNKNOWN,0,"get records with annotations in time range",-1,NULL);
  dout.dFormat = CF_STRUCT; dout.dArrayLength = 100; strncpy(dout.dTag,DAQPrmTag,TAG_NAME_SIZE);
  RegisterPropertyInformation(HSTEQM_NAME,"KEYWORD.INFO",&dout,NULL,CA_READ,AT_UNKNOWN,0,"Record keyword information database",-1,NULL);
  dout.dFormat = CF_INT32; dout.dArrayLength = 100; memset(dout.dTag,0,TAG_NAME_SIZE);
  RegisterPropertyInformation(HSTEQM_NAME,"KEYWORD.INPT",&dout,NULL,CA_READ,AT_UNKNOWN,0,"Record keyword input",-1,NULL);
  dout.dFormat = CF_NAME64DBLDBL; dout.dArrayLength = 100; 
  RegisterPropertyInformation(HSTEQM_NAME, "KEYWORD.PRIORS", &dout, &dout, CA_READ|CA_WRITE, AT_UNKNOWN, 0, "Record keyword prior information", -1, NULL);
  dout.dFormat = CF_TEXT; dout.dArrayLength = 256;
  RegisterPropertyInformation(HSTEQM_NAME,"KEYWORD.SOURCE",&dout,NULL,CA_READ,AT_UNKNOWN,0,"Record keyword source",-1,NULL);
  dout.dFormat = CF_NAME64; dout.dArrayLength = 256; 
  RegisterPropertyInformation(HSTEQM_NAME,"KEYWORD.SOURCES",&dout,NULL,CA_READ,AT_UNKNOWN,0,"Record keyword sources",-1,NULL);
  dout.dFormat = CF_DBLDBL; dout.dArrayLength = 1; memset(dout.dTag,0,TAG_NAME_SIZE);
  RegisterPropertyInformation(HSTEQM_NAME,"KEYWORD.TOLERANCE",&dout,NULL,CA_READ,AT_UNKNOWN,0,"Record keyword tolerance",-1,NULL);
  RegisterPropertyInformation(HSTEQM_NAME,"DB.RELOAD",NULL,NULL,CA_WRITE,AT_UNKNOWN,0,"reload the database",-1,NULL);
  RegisterPropertyInformation(HSTEQM_NAME,"RESTART",NULL,NULL,CA_WRITE,AT_UNKNOWN,0,"restart the server's equipment modules",-1,NULL);

  din.dFormat = CF_TEXT; din.dArrayLength = 64;
  dout.dFormat = CF_INT32; dout.dArrayLength = 100; memset(dout.dTag,0,TAG_NAME_SIZE);
  RegisterPropertyInformation(HSTEQM_NAME,"RECORD.SEARCH",&dout,&din,CA_READ,AT_UNKNOWN,0,"return records matching keyword search string",-1,NULL);
  din.dFormat = CF_TEXT; din.dArrayLength = 64;
  dout.dFormat = CF_STRUCT; dout.dArrayLength = 100; strncpy(dout.dTag,DAQRecTag,TAG_NAME_SIZE);
  RegisterPropertyInformation(HSTEQM_NAME,"RECORD.SEARCH",&dout,&din,CA_READ,AT_UNKNOWN,0,"return records matching keyword search string",-1,NULL);

  /* just for a test */
  RegisterPropertyInformation(HSTEQM_NAME,"DB.DUMP",NULL,NULL,CA_WRITE,AT_UNKNOWN,0,"back out the running database",-1,NULL);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = MAX_NUM_CONFIGS;
  strcpy(dout.dTag,"HstCfg");
  RegisterPropertyInformation(HSTEQM_NAME,"CONFIGURATION",
           &dout,NULL,CA_READ,AT_UNKNOWN,0,"registered configuration set",-1,NULL);
  RegisterProperty(HSTEQM_NAME,"CONFIGURATION.NAM",100,CF_NAME32,CA_READ,"configuration names");
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 100;
  strcpy(dout.dTag,"HstView");
  RegisterPropertyInformation(HSTEQM_NAME,"CONFIGURATION.NAM",
           &dout,NULL,CA_READ,AT_UNKNOWN,0,"configuration names",-1,NULL);

  if (gNumHstGrpsFiles > 0)
  {
    dout.dFormat = CF_STRUCT;
    dout.dArrayLength = MAX_NUM_CONFIGS;
    strcpy(dout.dTag,"HstCfg");
    for (n=0; n<gNumHstGrpsFiles; n++)
    {
      sprintf(prp,"CONFIGURATION.%s",hstGrpsFiles[n].fileName);
      sprintf(desc,"%.48s configuration set",hstGrpsFiles[n].desc);
      RegisterPropertyInformation(HSTEQM_NAME,prp,&dout,NULL,CA_READ,AT_UNKNOWN,0,desc,-1,NULL);
      sprintf(prp,"CONFIGURATION.%s.NAM",hstGrpsFiles[n].fileName);
      RegisterProperty(HSTEQM_NAME,prp,MAX_NUM_CONFIGS,CF_NAME32,CA_READ,"configuration names");
    }
  }

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = MAX_NUM_CONFIGS;
  strcpy(dout.dTag,"McaCfg");
  RegisterPropertyInformation(HSTEQM_NAME,"CONFIGURATION.MCA",
     &dout,NULL,CA_READ,AT_UNKNOWN,0,"registered MCA configuration set",-1,NULL);
  RegisterProperty(HSTEQM_NAME,"CONFIGURATION.MCA.NAM",MAX_NUM_CONFIGS,CF_NAME32,CA_READ,"configuration names");

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = MAX_NUM_CONFIGS;
  strcpy(dout.dTag,"HstCfg");
  RegisterPropertyInformation(HSTEQM_NAME,"CONFIGURATION.TRACE",
     &dout,NULL,CA_READ,AT_UNKNOWN,0,"registered TRACE configuration set",-1,NULL);
  RegisterProperty(HSTEQM_NAME,"CONFIGURATION.TRACE.NAM",MAX_NUM_CONFIGS,CF_NAME32,CA_READ,"configuration names");

  SetExportedSubSystem(HSTEQM_NAME,"HIST");

  readVolatileKeys();

  cap = recordsCapacity * sizeof(DAQRecInfo);
  if (cap > srvWorkAreaSize) adjustWorkAreaSize(cap);
err:
  if (cc)
  {
    feclog("HST initialization error : %s",cc2str(cc));
  }
  else 
  {
    feclog("HST: Data Access server initialized");
    dbLoadTime = time(NULL);
  }
  return;
}
int daqReload(int argc,char *argv[])
{
  argc = argc; argv = argv;
  feclog("reload DAQ database");
  /* flushStructCache(); */
  freeDaqDb();
  if (additionalKeywords != NULL) free(additionalKeywords);
  additionalKeywords = NULL; nAdditionalKeywords = 0;
  if (keywordList != NULL) free(keywordList), keywordList = NULL;
  nkeywords = 0;
  /* we should be back to square one now */
  daqInit();
  return 0;
}
void hstinit(void)
{
  gAppendProperyAliasNamesToQueries = FALSE;
  RegisterDeviceName(HSTEQM_NAME,"keyword",0);
  RegisterUserCommand("FilterTable",dumpFilterTable,NULL,NULL,CA_READ);
  RegisterUserCommand("ArchiveRecord",dumpArchiveRecord,NULL,NULL,CA_READ);
  RegisterUserCommand("VolatileKeys",dumpVolatileEntries,NULL,NULL,CA_READ);
  RegisterUserFunction("reload",daqReload);
  daqInit();
  return;
}
void hstexit(void)
{
  feclog("exit called: free DAQ database");
  freeDaqDb();
  if (additionalKeywords != NULL) free(additionalKeywords);
  additionalKeywords = NULL; nAdditionalKeywords = 0;
}
int fileExists(int i,time_t *mtime)
{
  int hFile, found = 0;
  struct stat sbuf;
  daqTbl[i].hasLegacyStorage = FALSE;
  if ((hFile=OpenDataFile(&daqTbl[i],NULL)) != -1) 
  {
    found = -1;
    fstat(hFile,&sbuf);
    *mtime = sbuf.st_mtime;
    close(hFile);
  }
  return found;
}
char *months[] = {"Jan","Feb","Mar","Apr","May","Jun",
                  "Jul","Aug","Sep","Oct","Nov","Dec"};
int getmonth(char *m)
{
  int i;
  for (i=0; i<12; i++) if (!stricmp(m,months[i])) return i+1;
  return 0;
}

int getEntryFromRectyp(int rec)
{
  int i;

  for (i=0; i<totalNrRecords; i++)
    if (daqTbl[i].rec == rec) return i;
  return -1;
}
int getLegacyEntryFromKeyword(char *keyword,int *rec,int *param)
{
  DAQParameter *daqParam;
  int i,j;

  for (i=0; i<totalNrRecords; i++)
  {
    if (daqTbl[i].active && !(daqTbl[i].option&NEVER)) continue;
    daqParam = daqTbl[i].daqParam;
    for (j=0; j<daqTbl[i].nparam; j++)
    {
      if (!stricmp(daqParam[j].param,keyword))
      {
        *rec = daqTbl[i].rec;
        *param  = j;
        return i;
      }
    }
  }
  return -1;
}
int getArchivedEntryFromKeywordAndTime(char *keyword,int *rec,int *param,time_t tgt)
{
  DAQParameter *daqParam;
  int i,j,yy,mm,best=-1,keyExists=FALSE,activeId=-no_data_in_interval;
  time_t mtime = 0;
  struct tm *lt=localtime((time_t *)&tgt);

  yy = lt->tm_year; mm = lt->tm_mon + 1;
  for (i=0; i<totalNrRecords; i++)
  {
    if (daqTbl[i].option&NEVER) continue;
    daqParam = daqTbl[i].daqParam;
    for (j=0; j<daqTbl[i].nparam; j++)
    {
      if (!stricmp(daqParam[j].param,keyword))
      {
        keyExists = TRUE;
        if (daqTbl[i].active) activeId = i;
        *rec = daqTbl[i].rec;
        *param  = j;
        getDataFilename(i,yy,mm);
        if (fileExists(i,&mtime)) 
        {
          if (mtime > tgt) return i;
          if (daqTbl[i].active) return i;
          daqTbl[i].lastOnDisk = (time_t)getLastArchivedTimeStamp(&daqTbl[i]);
          best = i;
        }
      } 
    }
  }
  if (best >= 0) return best;
  return keyExists ? activeId : -illegal_property;
}

int getIndexFromDeviceName(char *devName,char *devProperty)
{
  DAQParameter *daqParam;
  DTYPE ddin,ddout;
  int cc, i, siz, n, p, rec;
  char namProperty[PROPERTY_NAME_SIZE+4] = "";

  if ((n=getArchivedEntryFromKeyword(devProperty,&rec,&p)) < 0) return n;
  daqParam = daqTbl[n].daqParam;
  if ((siz=daqParam[p].size) > MAXNAMES) siz = MAXNAMES; 
  ddout.dArrayLength = siz;
  ddout.dFormat = CF_NAME64;
  ddout.data.vptr = name64blob;
  ddin.dArrayLength = 0;
  ddin.dFormat = CF_NULL;
  ddin.data.vptr = NULL;
  strncpy(namProperty,devProperty,PROPERTY_NAME_SIZE-4);
  strcat(namProperty,".NAM");
  if ((cc=hsteqm("#0",namProperty,&ddout,&ddin,CA_READ)) != 0) return -cc;
  for (i=0; i<siz && strnicmp(devName,((NAME64 *)ddout.data.vptr)[i].name,64); i++);
  if (i == siz) return -illegal_equipment_number;
  return i; 
}
int getIndexFromDeviceNameEx(char *devName,char *devProperty,int starttime,int stoptime)
{
  DAQParameter *daqParam;
  DTYPE ddout,ddin;
  int cc, i, siz, n, p, nn, np, rec;
  int startstopArray[2];
  char namProperty[PROPERTY_NAME_SIZE+4] = "";

  i = GetDeviceNumberEx(HSTEQM_NAME,devName,devProperty); /* the current registered/cached results */
  if ((n=getArchivedEntryFromKeyword(devProperty,&rec,&p)) < 0) return n;
  if (strnicmp(&devProperty[strlen(devProperty)-4],".NAM",4) == 0) return 0; /* a .NAM property */
  daqParam = daqTbl[n].daqParam;
  if ((siz=daqParam[p].size) > MAXNAMES) siz = MAXNAMES; 
  if (siz == 1) return 0;
  strncpy(namProperty,devProperty,PROPERTY_NAME_SIZE-4);
  strcat(namProperty,".NAM");
  nn = getArchivedEntryFromKeyword(namProperty,&rec,&np);
  if (nn >= 0 && i == 0)
  { /* has associated names: trap device = 'keyword' */
    if (!stricmp(devName,"keyword")) i = -1;
  }
  if (i >= 0) return i; /* go with the cached results */
  ddout.dArrayLength = siz;
  ddout.dFormat = CF_NAME64;
  ddout.data.vptr = name64blob;
  ddin.dArrayLength = 2;
  ddin.dFormat = CF_INT32;
  if (starttime == 0)
  {
    startstopArray[1] = (int)(tvalid > 0 ? tvalid : time(NULL));
    startstopArray[0] = startstopArray[1]-3600;
  }
  else
  {
    startstopArray[0] = starttime;
    startstopArray[1] = stoptime;
  }
  ddin.data.lptr = (SINT32 *)startstopArray;
  if ((cc=hsteqm("#0",namProperty,&ddout,&ddin,CA_READ)) != 0) return -cc;
  for (i=0; i<siz && strnicmp(devName,((NAME64 *)ddout.data.vptr)[i].name,64); i++);
  if (i == siz) return -illegal_equipment_number;
  return i; 
}
INTINT *getIndiciesFromVolatileDeviceName(char *devName,char *devProperty,int starttime,int stoptime)
{
  DAQParameter *daqParam;
  DTYPE ddout,ddin;
  int cc, i, j, k, siz, n, p, rec;
  int startstopArray[2];
  char namProperty[PROPERTY_NAME_SIZE+4] = "";
  INTINT ii[1002];

  if ((n=getArchivedEntryFromKeyword(devProperty,&rec,&p)) < 0) return NULL;
  if (strnicmp(&devProperty[strlen(devProperty)-4],".NAM",4) == 0) return NULL; /* a .NAM property */
  daqParam = daqTbl[n].daqParam;

  strncpy(namProperty,devProperty,PROPERTY_NAME_SIZE-4);
  strcat(namProperty,".NAM");

  ddin.dArrayLength = 2;
  ddin.dFormat = CF_INT32;
  if (starttime == 0)
  {
    startstopArray[1] = (int)(tvalid > 0 ? tvalid : time(NULL));
    startstopArray[0] = startstopArray[1]-3600;
  }
  else
  {
    startstopArray[0] = starttime;
    startstopArray[1] = stoptime;
  }
  ddin.data.lptr = (SINT32 *)startstopArray;
  ddout.dArrayLength = MAXNAMEI;
  ddout.dFormat = CF_NAME64I;
  ddout.data.vptr = name64Iblob;

  if ((cc=hsteqm("#0",namProperty,&ddout,&ddin,CA_READ)) != 0) return NULL;
  ii[0].i2val = starttime;
  for (i=0, k=1; i<(int)ddout.dArrayLength && k<1000; i++)
  {
    if (name64Iblob[i].ival > ii[k-1].i2val) ii[k++].i2val = name64Iblob[i].ival;
  }
  if ((siz=daqParam[p].size) > MAXNAMES) siz = MAXNAMES; 
  for (i=0; i<k; i++)
  {
    ddout.dArrayLength = siz;
    ddout.dFormat = CF_NAME64;
    ddout.data.vptr = name64blob;
    startstopArray[1] = startstopArray[0] = ii[i].i2val+1;
    if ((cc=hsteqm("#0",namProperty,&ddout,&ddin,CA_READ)) != 0) 
    {
      ii[i].i1val = -cc;
      continue;
    }
    for (j=0; j<siz && strnicmp(devName,((NAME64 *)ddout.data.vptr)[j].name,64); j++);
    ii[i].i1val = (j < siz) ? j : -illegal_device_number;
  }
  ii[k].i1val = -1; ii[k].i2val = 0;
  n = k;
  iiblob[0] = ii[0];
  for (i=0,k=0; i<n; i++)
  {
    /* if (ii[i].i1val < 0) continue; */
    if (i > 0 && ii[i].i1val == ii[i-1].i1val) continue;
    iiblob[k++] = ii[i];
  }
  if (k < MAXNAMEI) iiblob[k].i2val = iiblob[k].i1val = 0;
  
  return iiblob;
}

NAME64DBLDBL *getIndiciesFromVolatileDeviceNameEx(char *devName,char *keyword,int starttime,int stoptime)
{
  DAQParameter *daqParam;
  DTYPE ddout,ddin;
  int cc, i, j, k, siz, n, p, rec;
  int startstopArray[2];
  char namProperty[PROPERTY_NAME_SIZE+4] = "";
  NAME64DBLDBL ndd[1002], *pndd;
  PriorReference *pr;
  int pi, npi = 0, pend, dv = -1;

  if ((n=getArchivedEntryFromKeyword(keyword,&rec,&p)) < 0) return NULL;
  if (strnicmp(&keyword[strlen(keyword)-4],".NAM",4) == 0) return NULL; /* a .NAM property */
  daqParam = daqTbl[n].daqParam;
  if (devName[0] == '#') dv = atoi(&devName[1]);
  if ((pr=daqParam->pr) != NULL)
  { // prior references are available ...
    for (k=0; k<daqParam->npriors; k++)
    { // usually pr = NULL !
      if (starttime > pr[k].end) continue; // forget about it
      for (pi=0; pi<pr[k].ndevs; pi++) if (!strnicmp(devName,pr[k].aliaslst[pi].name,64) || dv == pi) break;
      if (pi == pr[k].ndevs) continue;
      pend = MIN((int)pr[k].end,stoptime);
      pndd = getIndiciesFromVolatileDeviceNameEx(pr[k].devlst[pi].name,pr[k].keyword,starttime,pend);
      if (pndd != NULL)
      {
        for (i=0; i<1000 && pndd[i].d2val > 0; i++) ndd[i] = pndd[i];
        npi = i;
      }
      if (pend < stoptime)
      { // keep going ...
        starttime = pend;
        break;
      }
      // we're done ...
      strncpy(ndd[i].name, ndd[i - 1].name, 64);
      ndd[i].d1val = ndd[i - 1].d1val;
      ndd[i].d2val = stoptime;
      i++;
      ndd[i].d1val = -1; ndd[i].d2val = 0;
      n = i;
      goto fin;
    }
  }
  strncpy(namProperty,keyword,PROPERTY_NAME_SIZE-4);
  strcat(namProperty,".NAM");

  ddin.dArrayLength = 2;
  ddin.dFormat = CF_INT32;
  if (starttime == 0)
  {
    startstopArray[1] = (int)(tvalid > 0 ? tvalid : time(NULL));
    startstopArray[0] = startstopArray[1]-3600*24;
  }
  else
  {
    startstopArray[0] = starttime;
    startstopArray[1] = stoptime;
  }
  ddin.data.lptr = (SINT32 *)startstopArray;
  ddout.dArrayLength = MAXNAMEI;
  ddout.dFormat = CF_NAME64I;
  ddout.data.vptr = name64Iblob;
  // get a trend of the 1st name so that we have  a list of timestamps
  if ((cc=hsteqm("#0",namProperty,&ddout,&ddin,CA_READ)) != 0) 
  { // no .NAM given => default to index = 1
    strcpy(nddblob[0].name,keyword);
    nddblob[0].d1val = 0; nddblob[0].d2val = 0;
    nddblob[1].d1val = -1; nddblob[1].d2val = 0;
    return nddblob;
  }
  if (npi > 0 && name64Iblob[0].ival < stoptime) 
  { // fix the starttime at the 1st observed name entry
    starttime = name64Iblob[0].ival;
  }
  ndd[npi].d2val = starttime;
  strncpy(ndd[npi].name,keyword,64);
  for (i=0, k=npi+1; i<(int)ddout.dArrayLength && k<1000; i++)
  {
    if (name64Iblob[i].ival > ndd[k-1].d2val) ndd[k++].d2val = name64Iblob[i].ival;
  }
  if ((siz=daqParam[p].size) > MAXNAMES) siz = MAXNAMES; 
  for (i=npi; i<k; i++)
  { // now get the snapshots at each timestamp and look for differences ...
    ddout.dArrayLength = siz;
    ddout.dFormat = CF_NAME64;
    ddout.data.vptr = name64blob;
    startstopArray[1] = startstopArray[0] = (int)ndd[i].d2val+1;
    if ((cc=hsteqm("#0",namProperty,&ddout,&ddin,CA_READ)) != 0) 
    {
      ndd[i].d1val = -cc;
      continue;
    }
    for (j=0; j<siz && strnicmp(devName,((NAME64 *)ddout.data.vptr)[j].name,64); j++) if (j == dv) break;
    ndd[i].d1val = (j < siz) ? j : -illegal_device_number;
    strncpy(ndd[i].name,keyword,64);
  }
  ndd[k].d1val = -1; ndd[k].d2val = 0;
  n = k;
fin:
  nddblob[0] = ndd[0];
  for (i=npi,k=npi; i<n; i++)
  { // now run thru the list and remove double entries ...
    if (i > 0 && ndd[i].d1val == ndd[i-1].d1val) continue;
    nddblob[k++] = ndd[i];
  }
  if (k < MAXN64DD) nddblob[k].d2val = nddblob[k].d1val = 0;
  return nddblob;
}

int srchcmp(char *subsys,char *desc,char *tok)
{
  char str[96],*c=str,*ss,substr[96],keystr[96];
  int i,found=FALSE,orflg=FALSE,andflg=FALSE;
  strncpy(str,tok,80); str[80] = 0;
  while (c != NULL)
  {
    ss = c; found = FALSE;
    if ((c=strchr(ss,',')) != NULL || (c=strchr(ss,'|')) != NULL)
    {
      *c++ = 0; orflg = TRUE;
    }   
    if ((c=strchr(ss,'+')) != NULL || (c=strchr(ss,'&')) != NULL)
    {
      *c++ = 0; andflg = TRUE;
    }
    for (i=0; i<48 && ss[i] == ' '; i++);   
    strncpy(substr,&ss[i],48); strupr(substr);
    if ((ss=strchr(substr,' ')) != NULL) *ss = 0; 
    strcpy(keystr,subsys);
    strcat(keystr,desc); 
    strupr(keystr);
    if (strstr(keystr,substr) != NULL) found = TRUE;
    if (andflg && !found) return FALSE;
    if (!andflg && found) return TRUE;
  }
  return found ? TRUE : FALSE;
}
void assignNrOfPoints(BYTE *data,int fmt,int npoints)
{
  if (data == NULL) return;
  switch (fmt)
  {
    case CF_FLTINT:
      ((FLTINT *)data)[0].ival = (SINT32)(npoints);
      ((FLTINT *)data)[0].fval = (float)(npoints);
      break;
    case CF_FLTINTINT:
      ((FLTINTINT *)data)[0].i1val = (SINT32)(npoints);
      ((FLTINTINT *)data)[0].i2val = (SINT32)(npoints);
      ((FLTINTINT *)data)[0].fval = (float)(npoints);
      break;
    case CF_INTINT:
      ((INTINT *)data)[0].i1val = (SINT32)(npoints);
      ((INTINT *)data)[0].i2val = (SINT32)(npoints);
      break;
    case CF_DBLDBL:
      ((DBLDBL *)data)[0].d1val = (double)(npoints);
      ((DBLDBL *)data)[0].d2val = (double)(npoints);
      break;
    case CF_NAME16I:
      ((NAME16I *)data)[0].ival = (SINT32)(npoints);
      break;
    case CF_NAME32I:
      ((NAME32I *)data)[0].ival = (SINT32)(npoints);
      break;
    case CF_NAME48I:
      ((NAME48I *)data)[0].ival = (SINT32)(npoints);
      break;
    case CF_NAME64I:
      ((NAME64I *)data)[0].ival = (SINT32)(npoints);
      break;
    case CF_INT32:
      *(SINT32 *)data = (SINT32)(npoints);
      break;
    case CF_FLOAT:
      *(float *)data = (float)(npoints);
      break;
    case CF_INTFLTINT:
      ((INTFLTINT *)data)[0].i1val = (SINT32)(npoints);
      ((INTFLTINT *)data)[0].i2val = (SINT32)(npoints);
      ((INTFLTINT *)data)[0].fval = (float)(npoints);
      break;
    case CF_DOUBLE:
      *(double *)data = (double)(npoints);
      break;
  }
}
int prepSearchRecord(DTYPE *dout,int offset,int idx)
{
  switch (dout->dFormat)
  {
    case CF_STRUCT: 
      if (strcmp(dout->dTag,DAQRecTag) == 0)
      { 
        DAQRecInfo *dst = (DAQRecInfo *)dout->data.vptr;
        dst[offset].active = daqTbl[idx].active;
        strncpy(dst[offset].ctx,daqTbl[idx].ctx,CONTEXT_NAME_SIZE);
        strncpy(dst[offset].srv,daqTbl[idx].srv,EXPORT_NAME_SIZE);
        strncpy(dst[offset].dev,daqTbl[idx].dev,DEVICE_NAME_SIZE);
        strncpy(dst[offset].prp,daqTbl[idx].prp,PROPERTY_NAME_SIZE);
        dst[offset].fmt = daqTbl[idx].fmt;
        if (dst[offset].fmt == CF_STRUCT)
          strncpy(dst[offset].tag,daqTbl[idx].tag,TAG_NAME_SIZE);
        else
          strncpy(dst[offset].tag,daqTbl[idx].fn_inp,80);
        strncpy(dst[offset].fn_key,daqTbl[idx].fn_key,80);
        dst[offset].hbt = daqTbl[idx].hbt;
        dst[offset].len = daqTbl[idx].len;
        dst[offset].infmt = daqTbl[idx].infmt;
        dst[offset].inlen = daqTbl[idx].inlen;
        dst[offset].opt = daqTbl[idx].option;
        dst[offset].rec = daqTbl[idx].rec;
        dst[offset].tmr = daqTbl[idx].tmr;
        break;
      }
      return invalid_structure_tag;
    case CF_INT32:
      dout->data.lptr[offset] = daqTbl[idx].rec;
      break;
    case CF_INT16:
      dout->data.sptr[offset] = daqTbl[idx].rec;
      break;
    default:
      return illegal_format;
  }
  return 0;
}
int strfound(char *s,char *tok)
{
  char beginToken[64], endToken[64], midToken[64];
  int blen,elen,mlen,dlen;
  if (!strcmp(tok,"*")) return TRUE; /* always matches */
  if ((dlen=(int)strlen(s)) == 0) return FALSE;
  if (strchr(tok,'*') == NULL) return (strstr(s,tok) != NULL); /* not a wildcard compare */
  getWildcardTokens((char *)tok,beginToken,endToken,midToken);
  blen = (int)strlen(beginToken); if (blen > 64) blen = 64;
  elen = (int)strlen(endToken); if (elen > 64) elen = 64;
  mlen = (int)strlen(midToken); if (mlen > 64) mlen = 64;
  if (blen > 0)
  { /* check against begin token */
    if (strncmp(s,beginToken,blen)) return FALSE;
  }
  if (elen > 0)
  { /* check against end token */
    if (strncmp(&s[dlen-elen],endToken,elen)) return FALSE;
  }
  if (mlen > 0)
  { /* check against mid token */
    if (strstr(s,midToken) == NULL) return FALSE;
  }
  return TRUE;
}
int findRecords(char *search,DTYPE *dout)
{
  int i,j,n=0;
  char keyupr[256], searchupr[96];
  DAQParameter *daqParam;
  DAQRecInfo *dst = (DAQRecInfo *)dout->data.vptr;
  strncpy(searchupr,search,PROPERTY_NAME_SIZE);
  searchupr[PROPERTY_NAME_SIZE] = 0;
  strupr(searchupr);
  for (i=0; i<nrecords; i++)
  {
    sprintf(keyupr,"/%.32s/%32s/%.64s[%.64s]",daqTbl[i].ctx,daqTbl[i].srv,daqTbl[i].dev,daqTbl[i].prp);
    strupr(keyupr);
    if (strstr(keyupr,searchupr) != NULL)
    {
      if (prepSearchRecord(dout,n,i)) return illegal_format;
      if (n++ >= (int)dout->dArrayLength) return 0;
      continue;
    }
    daqParam = daqTbl[i].daqParam;
    for (j=0; j<daqTbl[i].nparam; j++)
    {
      strncpy(keyupr,daqParam[j].param,PROPERTY_NAME_SIZE);
      keyupr[PROPERTY_NAME_SIZE] = 0;
      strupr(keyupr);
      if (strstr(keyupr,searchupr) != NULL)
      {
        if (prepSearchRecord(dout,n,i)) return illegal_format;
        if (n++ >= (int)dout->dArrayLength) return 0;
        break;
      }
    }
  }
  dout->dArrayLength = (UINT32)n;
  return 0;
}
int getRegisteredKeyword(char *dst,char *src)
{
  int i;
  strncpy(dst,src,PROPERTY_NAME_SIZE);
  for (i=0; i<nPrpAlias; i++)
  {
    if (!strnicmp(prpAlias[i].name,src,PROPERTY_NAME_SIZE))
    {
      strncpy(dst,prpAlias[i].alias,PROPERTY_NAME_SIZE);
      return i;
    }
  }
  return 0;
}
#define isStsRequest(f) ((f) == CF_NAME32DBLDBL)
int hsteqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int fmt,fmtsize,i,j,n,r,len, ival, flen, slen, soff, ioff, depth;
  int index=0,subindex=0,fieldindex=0,sample=0,startpos=0,hasInput=FALSE;
  int recsreadFlag=0,inputgiven=0,ncallers=1;
  float lwrlimit=0,uprlimit=0, fvals[2];
  time_t tnow=time(NULL),starttime,stoptime;
  double dts=0;
  char *c,*c1,fn[136],pn[96],keyword[PROPERTY_NAME_SIZE+4],sstr[256],usr[USERNAME_SIZE+4];
  int cc,rec,p,idxGrps,isNameQuery,isSnapshot,isVolatile,isRecentRange = FALSE;
  SINT32 tgt;
  UINT32 ivals[2];
  int nitpairs,startt,stopt,nread, dlen, doff;
  NAME64DBLDBL *itpair, itthis;
  BYTE *dptr;
  CSVDB db;
  NAME16 caller;

  if (tvalid > 0) 
  {
    tnow = tvalid;
    if ((i=getArchivedEntryFromKeywordAndTime(devProperty,&r,&n,tnow)) >= 0) 
    { /* but if this one's still active, reset tnow to current time! */
      if (daqTbl[i].active > 0) tnow = time(NULL);
    }
  }
  starttime = tnow; 
  stoptime = tnow;

  if (strlen(devName) == 0) devName = "#0";
  if (access & CA_WRITE)
  {
    GetCallerInfo(HSTEQM_NAME,&caller,NULL,NULL,NULL,&ncallers);
    strncpy(usr,caller.name,16); strupr(usr);
  }
  if (!stricmp(devProperty,"NTAGS") || !stricmp(devProperty,"NKEYWORDS"))
  {
    len = ntags + nAdditionalKeywords; 
    return PutValuesFromLong(dout,(SINT32 *)&len,1);
  }
  if (!stricmp(devProperty,"NKEYWORDS.VOLATILE"))
  {
    return PutValuesFromLong(dout,&gNumVolatileKeys,1);
  }
  if (!stricmp(devProperty,"TAGS") || !stricmp(devProperty,"KEYWORDS"))
  {
    sstr[0] = 0;
    if (din->dArrayLength > 0)
    {
      switch (din->dFormat)
      {
        case CF_NAME64:
        case CF_NAME32:
        case CF_NAME16:
          slen = fmtsizeof(din->dFormat);
          break;
        case CF_TEXT:
          if ((slen=din->dArrayLength) > SUBSYSTEM_NAME_SIZE) slen = SUBSYSTEM_NAME_SIZE;
          break;
        default: 
          return illegal_format;
      }
      sstr[slen] = 0;
      strncpy(sstr,din->data.cptr,slen);
    }
    flen = fmtsizeof(dout->dFormat);
    switch (dout->dFormat)
    {
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        /* just a names list */
        slen = flen;
        soff = 0;
        ioff = -1;
        break;
      case CF_NAME8I:
      case CF_NAME16I:
      case CF_NAME32I:
      case CF_NAME48I:
      case CF_NAME64I:
        /* a names list with default format */
        slen = flen - 4;
        soff = 0;
        ioff = slen;
        break;
      case CF_USTRING:
        /* a names list with default format */
        slen = 64; /* properties are not longer than 64 chars (USTRING has 80) */
        soff = 16;
        ioff = 0;
        break;
      default: return illegal_format;
    }

    for (i=0,n=0; i<totalNrRecords; i++)
    {
      if (!daqTbl[i].active) continue;
      if (daqTbl[i].option&NEVER) continue; /* it's not being archived */
      if (dout->dArrayLength < (UINT32)n) break;
      for (j=0; j<daqTbl[i].nparam; j++)
      {
        if (dout->dArrayLength < (UINT32)n) break;
        if (strlen(sstr) > 0 && !srchcmp(daqTbl[i].daqParam[j].subsys,daqTbl[i].daqParam[j].desc,sstr)) continue;
        strncpy((char *)(dout->data.bptr+n*flen),daqTbl[i].daqParam[j].param,slen);
        if (ioff >= 0) 
        { /* send the default format with the call */
          ival = daqTbl[i].daqParam[j].fmt;
          memcpy(dout->data.bptr+n*flen + ioff,&ival,4);
        }
        n++;
      }
    }
    for (i=0; i<nAdditionalKeywords; i++)
    {
      if (strcmp(additionalKeywords[i].EqmName,HSTEQM_NAME)) continue;
      if (dout->dArrayLength < (UINT32)n) break;
      if (strlen(sstr) > 0 && !srchcmp(additionalKeywords[i].SubSystem,additionalKeywords[i].description,sstr)) continue;
      strncpy((char *)(dout->data.bptr+n*flen),additionalKeywords[i].PrpName,slen);
      if (ioff >= 0) 
      { /* send the default format with the call */
        ival =  ftoi(additionalKeywords[i].fmtOut);
        memcpy(dout->data.bptr+n*flen + ioff,&ival,4);
      }
      n++;
    }
    dout->dArrayLength = n;
    return 0;
  }
  if (!stricmp(devProperty,"NSUBSYSTEMS"))
  {
    return PutValuesFromLong(dout,(SINT32 *)&nSubSystems,1);
  }
  if (!stricmp(devProperty,"SUBSYSTEMS"))
  {
    switch (dout->dFormat)
    {
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        len = fmtsizeof(dout->dFormat);
        break;
      default: 
        return illegal_format;
    }
    for (i=0; i<(int)dout->dArrayLength && i<nSubSystems; i++)
    {
      strncpy((char *)(dout->data.bptr+i*len),subSystems[i].name,len);
    }
    if (i < (int)dout->dArrayLength) dout->dArrayLength = (UINT32)i;
    return 0;
  }
  if (!stricmp(devProperty,"KEYWORDS.VOLATILE"))
  {
    if (dout->dArrayLength > (UINT32) gNumVolatileKeys) dout->dArrayLength = gNumVolatileKeys; 
    if (gNumVolatileKeys == 0) return 0;
    return PutValuesFromString(dout,(char *)volatileKeys,CF_NAME64,gNumVolatileKeys,0);
  }
  if (!stricmp(devProperty,"KEYWORDS.PRIMARY"))
  {
    if (dout->dFormat != CF_NAME64) return illegal_format;
    return getArchivedKeywords((NAME64 *)dout->data.vptr,&dout->dArrayLength);
  }
  cc = 0;
  if (!stricmp(devProperty,"RECORD.INFO"))  
  { /* DAQRecInfo database */
    if (access & CA_WRITE)
    {
      DAQRecInfo *dqr = (DAQRecInfo *)din->data.vptr;
      if (!isAdministrator(usr)) 
      {
        sprintf(daqUpdateErrStr,"access denied: user %.16s is not an administrator",usr);
        SetEqmCompletion(HSTEQM_NAME,daqUpdateErrStr);
        return access_denied;
      }
      if (din->dArrayLength == 0) return illegal_read_write;
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,DAQRecTag) != 0) return invalid_structure_tag;
      cc = addDaqRecInfoElements(dqr,din->dArrayLength);
      if (cc != 0) return cc;
    }
    else if (din->dArrayLength > 0)
    { /* not a WRITE but it does have input specifying particular record(s) */
      hasInput = TRUE;
      cc = GetValuesAsStringEx(din,keyword,CF_TEXT,PROPERTY_NAME_SIZE,0);
      if (cc != 0) return cc;
    }
    if (dout->dArrayLength > 0)
    {
      DAQRecInfo *dqr = (DAQRecInfo *)dout->data.vptr;
      if (dout->dArrayLength > (UINT32)totalNrRecords) dout->dArrayLength = totalNrRecords;
      if (dout->dArrayLength == 0) return 0;
      if (dout->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(dout->dTag,DAQRecTag) != 0) return illegal_format;
      if (hasInput) cc = getDaqRecInfoEx(dqr,&dout->dArrayLength,keyword);
      else          cc = getDaqRecInfo(dqr,&dout->dArrayLength);
    }
    return cc;
  }
  if (!stricmp(devProperty,"RECORD.ERR"))
  { /* last record update error (from db manager) */
    if (access & CA_WRITE) return illegal_read_write;
    if (dout->dArrayLength < 1) return dimension_error;
    return PutValuesFromString(dout,daqUpdateErrStr,CF_TEXT,256,0);
  }
  if (!stricmp(devProperty,"RECORD.CMT"))
  { /* annotate the record */
    int isAa=FALSE, liveAr=FALSE, isRm=FALSE;
    time_t fillto = 0;
    strncpy(pn,devName,96);
    if ((c=strchr(pn,'@')) == NULL) return invalid_parameter;
    *c++ = 0;
    if ((c1=strchr(c,'[')) != NULL)
    {
      *c1++ = 0;
      if (!strnicmp(c1,"KEEPALIVE",9)) isAa = TRUE;
    }
    if (!stricmp(c,"NOW"))
    {
      starttime = tnow;
      liveAr = TRUE;
    }
    else
    {
      starttime = strtol(c,NULL,0);
    }
    if ((i=getAliasTableIndex(pn)) >= 0)
    {
      strncpy(pn,gAliasTable[i].name,PROPERTY_NAME_SIZE);
    }
    if ((i=getArchivedEntryFromKeywordAndTime(pn,&rec,&p,starttime)) < 0) 
    {
      if (stricmp(pn,"State") != 0) return -i;
      // let a 'context annotation' slide thru if there is no "State" keyword ...
      rec = 0xffff;
    }
    if (access & CA_WRITE)
    {
      switch (din->dFormat)
      {
        case CF_TEXT:
          len = din->dArrayLength;
          c = din->data.cptr;
          break;
        case CF_NULL:
          len = 0;
          c = NULL;
          break;
        default:
          return illegal_format;
      }
      cc = annotateRecord(rec,starttime,c,len);
      if (cc) return cc;
      if (rec == 0xffff) goto AnnotationRead;
      if (c == NULL || strlen(c) == 0) 
      { /* signal to REMOVE an annotation */
        isRm = TRUE;
        isAa = FALSE;
      }
      applyActiveAnnotation(rec,starttime,!isAa);
      srvlog(ALF,"dispatch","annotate record %.64s -> time: %.24s (%d) active: %s, remove: %s",
        daqTbl[i].daqParam[0].param,ctime(&starttime),starttime,btoa(isAa),btoa(isRm));
    }
AnnotationRead:
    if (dout->dArrayLength > 0)
    {
      if (dout->dFormat != CF_TEXT) return illegal_format;
      return getRecordAnnotation(rec,starttime,dout->data.cptr,&dout->dArrayLength);
    }
    return 0;
  }
  if (!stricmp(devProperty,"RECORD.CMTS"))
  { /* get record annotations */
    strncpy(pn,devName,96);
    if ((c=strchr(pn,'@')) != NULL) *c = 0;
    if ((i=getArchivedEntryFromKeywordAndTime(pn,&rec,&p,starttime)) < 0) return -i;
    starttime = stoptime - 86400;
    if (din != NULL && din->dArrayLength > 0)
    {
      if ((cc=GetValuesAsLong(din,ivals,2)) != 0) return cc;
      starttime = ivals[0];
      if (din->dArrayLength > 1) stoptime = ivals[1];
    }
    return getRecordAnnotations(rec,starttime,stoptime,dout);
  }
  if (!stricmp(devProperty,"RECORDS.WITH.CMTS"))
  { /* get records with annotations in time range*/
    starttime = stoptime - 86400;
    if (din != NULL && din->dArrayLength > 0)
    {
      if ((cc=GetValuesAsLong(din,ivals,2)) != 0) return cc;
      starttime = ivals[0];
      if (din->dArrayLength > 1) stoptime = ivals[1];
    }
    return getRecordsWithAnnotations(starttime,stoptime,dout);
  }
  if (!stricmp(devProperty,"KEYWORD.INFO"))  
  { /* DAQRecInfo database */
    if (access & CA_WRITE)
    {
      if (!isAdministrator(usr)) 
      {
        sprintf(daqUpdateErrStr,"access denied: user %.16s is not an administrator",usr);
        SetEqmCompletion(HSTEQM_NAME,daqUpdateErrStr);
        return access_denied;
      }
      if (din->dArrayLength == 0) return illegal_read_write;
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,DAQPrmTag) != 0) return illegal_format;
      cc = addDaqPrmInfoElem(devName,(DAQPrmInfo *)din->data.vptr,din->dArrayLength);
      if (cc != 0) return cc;
    }
#   ifdef READ_LOADEDDB
    if (getEntryFromKeywordFile(devName,&n) < 0)
    { /* must then be entered as #<rec number> */
      n = GetDeviceNumber(HSTEQM_NAME,devName);
    }
#   endif
    if (dout->dArrayLength > 0)
    {
      if (dout->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(dout->dTag,DAQPrmTag) != 0) return illegal_format;
      n = GetDeviceNumber(HSTEQM_NAME,devName);
      if (n < 0)
      { /* not a simple '#<recno>' */
        c = devName; /* devName = keyword file name ? */
        getRegisteredKeyword(keyword,devName);
        if ((i=getArchivedEntryFromKeyword(keyword,&r,&j)) >= 0)
        { /* devName is a keyword property */
          c = daqTbl[i].fn_key;
        }
        cc = readDaqPrmFile(ArchiveDBpath,c,(DAQPrmInfo *)dout->data.vptr,(int *)&dout->dArrayLength);
      }
      else
      {
        cc = getDaqPrmInfo(n,(DAQPrmInfo *)dout->data.vptr,(int *)&dout->dArrayLength);
      }
    }
    return cc;
  }
  if (!stricmp(devProperty,"KEYWORD.INPT"))  
  { /* DAQRecInfo database */
    if (access & CA_WRITE)
    {
      if (!isAdministrator(usr)) 
      {
        sprintf(daqUpdateErrStr,"access denied: user %.16s is not an administrator",usr);
        SetEqmCompletion(HSTEQM_NAME,daqUpdateErrStr);
        return access_denied;
      }
      if (din->dArrayLength == 0) return illegal_read_write;
      cc = addDaqInpt(devName,din->data.bptr,(int)din->dFormat,(int)din->dArrayLength);
      if (cc != 0) return cc;
    }
    if (getEntryFromKeywordFile(devName,&n) < 0)
    { /* must then be entered as #<rec number> */
      n = GetDeviceNumber(HSTEQM_NAME,devName);
    }
    cc = getDaqPrmInpt(n,dout);
    return cc;
  }
  if (!stricmp(devProperty, "KEYWORD.PRIORS"))
  { /* prior info if any ... */
    getRegisteredKeyword(keyword, devName);
    if ((i=getArchivedEntryFromKeyword(keyword, &r, &j)) < 0) return invalid_keyword;
    if (access & CA_WRITE)
    {
      if (!isAdministrator(usr))
      {
        sprintf(daqUpdateErrStr, "access denied: user %.16s is not an administrator", usr);
        SetEqmCompletion(HSTEQM_NAME, daqUpdateErrStr);
        return access_denied;
      }
      if (din->dArrayLength == 0) return illegal_read_write;
      if (din->dFormat != CF_NAME64DBLDBL) return illegal_format;
      cc = addDaqPrmPriors(&daqTbl[i].daqParam[j], (NAME64DBLDBL *)din->data.vptr, din->dArrayLength);
      if (cc != 0) return cc;
    }
    if (dout->dArrayLength > 0)
    {
      if (dout->dFormat != CF_NAME64DBLDBL) return illegal_format;
      getRegisteredKeyword(keyword, devName);
      if ((i=getArchivedEntryFromKeyword(keyword, &r, &j)) >= 0)
      {
        int k, np = daqTbl[i].daqParam[j].npriors;
        PriorReference *pr = daqTbl[i].daqParam[j].pr;
        NAME64DBLDBL *dst = (NAME64DBLDBL *)dout->data.vptr;
        dout->dArrayLength = np;
        if (np == 0 || pr == NULL) return 0;
        for (k = 0; k < np; k++)
        {
          strncpy(dst[k].name, pr[k].keyword, 64);
          dst[k].d1val = (double)pr[k].begin;
          dst[k].d2val = (double)pr[k].end;
        }
        return 0;
      }
    }
    return invalid_keyword;
  }
  if (!stricmp(devProperty,"KEYWORD.SOURCES"))
  { /* Keyword source (db source or learned source) */
    int k, fecid, srvid, isDoocsSrc=FALSE;
    DTYPE d;
    DTYPEZERO(d);
    if (!isSimpleStringFormat(dout->dFormat) || dout->dFormat == CF_TEXT) return illegal_format;
    len = fmtsizeof(dout->dFormat);
    for (i=0, k=0; i<nrecords && k<(int)dout->dArrayLength/2; i++)
    {
      if (daqTbl[i].active <= 0) continue;
      if (daqTbl[i].daqParam == NULL || daqTbl[i].nparam == 0) continue;
      if ((c=strstr(daqTbl[i].daqParam[0].param,".NAM")) != NULL && !strcmp(c,".NAM")) continue;
      if ((c=strstr(daqTbl[i].daqParam[0].param,".SRCADDR")) != NULL && !strcmp(c,".SRCADDR")) continue;
      fecid = locateExport(daqTbl[i].srv,daqTbl[i].ctx,daqTbl[i].dev,&srvid);
      if (fecid < 0) continue;
      isDoocsSrc = !strncmp(SrvTbl[srvid].EqmName,DOOCSEQM_TAG,EQM_NAME_SHORTSIZE);
      if ((n=daqTbl[i].daqParam[0].srcIndex) > 0)
      {
        n--; /* re-shuffle index */
        d.dArrayLength = 1;
        d.dFormat = dout->dFormat;
        d.data.bptr = &dout->data.bptr[(k+1)*len];
        if (getArchivedDataTrace(daqTbl[n].daqParam[0].param,starttime,&d,0,0,0) != 0) continue;
      }
      else if (isDoocsSrc) 
      {
        char subs[16];
        getSubSysDecoration(daqTbl[i].daqParam[0].subsys,subs);
        snprintf(&dout->data.cptr[(k+1)*len],len,"/%.32s%s/%.32s/%.64s[%.64s]",
          daqTbl[i].ctx,subs,daqTbl[i].srv,daqTbl[i].dev,daqTbl[i].prp);
      }
      else
      {
        snprintf(&dout->data.cptr[(k+1)*len],len,"/%.32s/%.32s/%.64s[%.64s]",
          daqTbl[i].ctx,daqTbl[i].srv,daqTbl[i].dev,daqTbl[i].prp);
      }
      strncpy(&dout->data.cptr[k*len],daqTbl[i].daqParam[0].param,len);
      k += 2;
    }
    dout->dArrayLength = (UINT32)k*2;
    return 0;
  }
  if (!stricmp(devProperty,"KEYWORD.SOURCE"))  
  { /* Keyword source (db source or learned source) */
    int fecid, srvid, isDoocsSrc=FALSE;
    strncpy(keyword,devName,PROPERTY_NAME_SIZE);
    i = getArchivedEntryFromKeyword(keyword,&r,&j);
    if (i < 0) 
    {
      if ((i=getRedirectedIndex(keyword)) >= 0)
      {
        char *c, *ctx, *srv, *prp;
        strncpy(sstr,additionalKeywords[i].redirection,DEVICE_REDIR_SIZE);
        ctx = sstr;
        if (ctx[0] == '/')
        {
          ctx++;
          if ((srv=strchr(ctx,'/')) == NULL) return non_existent;
          *srv++ = 0;
          if ((prp=strchr(srv,'[')) == NULL) return non_existent;
          *prp++ = 0;
          if ((c=strchr(prp,']')) != NULL) *c = 0;
          if ((c=strchr(srv,'/')) != NULL) *c = 0;
        }
        SetRedirectionParameters(HSTEQM_NAME,ctx,srv,prp,devProperty);
        return server_redirection;
      }
      return non_existent;
    }
    fecid = locateExport(daqTbl[i].srv,daqTbl[i].ctx,daqTbl[i].dev,&srvid);
    if (fecid < 0) return fecid;
    if (!strncmp(SrvTbl[srvid].EqmName,DOOCSEQM_TAG,EQM_NAME_SHORTSIZE)) isDoocsSrc = TRUE;
    if ((n=daqTbl[i].daqParam[j].srcIndex) > 0)
    {
      n--; /* re-shuffle index */
      cc = getArchivedDataTrace(daqTbl[n].daqParam[0].param,starttime,dout,0,0,0);
    }
    else if (isDoocsSrc) 
    {
      DTYPE d;
      int k;
      char subs[16];
      DUSTRING *us=NULL;
      if (!isStringFormat(dout->dFormat) || dout->dFormat == CF_TEXT) return illegal_format;
      if (strlen(keyword) > 60) return invalid_keyword;
      if (daqTbl[i].daqParam[j].arrayType & AT_CHANNEL)
      {
        if (strlen(daqTbl[i].daqParam[j].assoc) > 0) 
        {
          strncpy(keyword,daqTbl[i].daqParam[j].assoc,PROPERTY_NAME_SIZE);
        }
        else
        {
          strcat(keyword,".NAM");
        }
        subs[0] = 0;
        if ((n=getArchivedEntryFromKeyword(keyword,&r,&j)) < 0) return -n;
        DTYPEZERO(d); d.dArrayLength = dout->dArrayLength; d.dFormat = CF_USTRING;
        d.data.vptr = (DUSTRING *)malloc(d.dArrayLength * sizeof(DUSTRING));
        if (d.data.vptr == NULL) return out_of_local_memory;
        cc = getArchivedDataTrace(daqTbl[n].daqParam[0].param,starttime,&d,0,0,0);
        us = (DUSTRING *)d.data.vptr; slen = fmtsizeof(dout->dFormat);
      }
      getSubSysDecoration(daqTbl[i].daqParam[j].subsys,subs);
      if (us == NULL)
      {
        if (dout->dFormat != CF_TEXT) dout->dArrayLength = 1;
        sprintf(sstr,"/%.32s%s/%.32s/%.64s[%.64s]",
          daqTbl[i].ctx,subs,daqTbl[i].srv,daqTbl[i].dev,daqTbl[i].prp);
        return PutValuesFromString(dout,sstr,CF_TEXT, (int)strlen(sstr),0);
      }
      for (k=0; k<(int)dout->dArrayLength && k<daqTbl[i].daqParam[j].size; k++)
      {
        if (strlen(us[k].str) == 0) break;
        sprintf(sstr,"/%.32s%s/%.32s/%.64s[%.64s]",
          daqTbl[i].ctx,subs,daqTbl[i].srv,us[k].str,daqTbl[i].prp);
        strncpy(us[k].str,sstr,80);
      }
      cc = PutValuesFromString(dout,(char *)us,CF_USTRING,k,0);
      dout->dArrayLength = (UINT32)k;
      free(us);
      return cc;
    }
    else
    {
      sprintf(sstr,"/%.32s/%.32s/%.64s[%.64s]",
        daqTbl[i].ctx,daqTbl[i].srv,daqTbl[i].dev,daqTbl[i].prp);
      cc = PutValuesFromString(dout,sstr,CF_TEXT, (int)strlen(sstr),0);
      if (dout->dFormat != CF_TEXT) dout->dArrayLength = 1;
    }
    return cc;
  }
  if (!stricmp(devProperty,"KEYWORD.TOLERANCE"))  
  { /* keyword tolerance (for given interval) */
    starttime = stoptime - 86400;
    if (din != NULL && din->dArrayLength > 0)
    { /* a time range specified */
      if ((cc=GetValuesAsLong(din,ivals,2)) != 0) return cc;
      starttime = ivals[0];
      if (din->dArrayLength > 1) stoptime = ivals[1];
      /* TODO: tie into tolerance history */
    }
    if (dout->dArrayLength < 1) return dimension_error;
    strncpy(keyword,devName,PROPERTY_NAME_SIZE);
    i = getArchivedEntryFromKeyword(keyword,&r,&j);
    if (i < 0) 
    {
      if ((i=getRedirectedIndex(keyword)) >= 0)
      {
        char *c, *ctx, *srv, *prp;
        strncpy(sstr,additionalKeywords[i].redirection,DEVICE_REDIR_SIZE);
        ctx = sstr;
        if (ctx[0] == '/')
        {
          ctx++;
          if ((srv=strchr(ctx,'/')) == NULL) return non_existent;
          *srv++ = 0;
          if ((prp=strchr(srv,'[')) == NULL) return non_existent;
          *prp++ = 0;
          if ((c=strchr(prp,']')) != NULL) *c = 0;
          if ((c=strchr(srv,'/')) != NULL) *c = 0;
        }
        SetRedirectionParameters(HSTEQM_NAME,ctx,srv,prp,devProperty);
        return server_redirection;
      }
      return non_existent;
    }
    fvals[0] = daqTbl[i].daqParam[j].aTolerance;
    fvals[1] = daqTbl[i].daqParam[j].pTolerance;
    switch (dout->dFormat)
    {
      case CF_DBLDBL:
        ((DBLDBL *)dout->data.dptr)[0].d1val = (double)fvals[0];
        ((DBLDBL *)dout->data.dptr)[0].d2val = (double)fvals[1];
        dout->dArrayLength = 1;
        return 0;
      case CF_FLTFLT:
        memcpy(dout->data.fptr,fvals,2*sizeof(float));
        dout->dArrayLength = 1;
        return 0;
      default:
        if (dout->dArrayLength > 2) dout->dArrayLength = 2;
        return PutValuesFromFloat(dout,fvals,2);
    }
    return 0;
  }
  if (!stricmp(devProperty,"DB.LASTLOADED"))
  {
    if ((access & CA_WRITE)) return illegal_read_write;
    if (isNumberFormat(dout->dFormat))
    {
      UINT32 t = (UINT32)dbLoadTime;
      return PutValuesFromLong(dout,&t,1);
    }
    if (isStringFormat(dout->dFormat))
    {
      char dbLoadTimeStr[32];
      sprintf(dbLoadTimeStr,"%.24s",ctime(&dbLoadTime));
      return PutValuesFromString(dout,dbLoadTimeStr,CF_TEXT,32,0);
    }
    return illegal_format;
  }
  if (!stricmp(devProperty,"DB.RELOAD"))  
  { /* dump database */
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (!isAdministrator(usr)) 
    {
      sprintf(daqUpdateErrStr,"access denied: user %.16s is not an administrator",usr);
      SetEqmCompletion(HSTEQM_NAME,daqUpdateErrStr);
      return access_denied;
    }
    return daqReload(0,NULL);
  }
  if (!stricmp(devProperty,"DB.DUMP"))  
  { /* dump database */
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (!isAdministrator(usr)) 
    {
      sprintf(daqUpdateErrStr,"access denied: user %.16s is not an administrator",usr);
      SetEqmCompletion(HSTEQM_NAME,daqUpdateErrStr);
      return access_denied;
    }
    sprintf(sstr,".%s",delStr);
    return dumpDaqDb(sstr,"dbtest.csv",TRUE);
  }
  if (!stricmp(devProperty,"RESTART"))  
  { /* dump database */
    if ((access & CA_WRITE) == 0) return illegal_read_write;
    if (!isAdministrator(usr)) 
    {
      sprintf(daqUpdateErrStr,"access denied: user %.16s is not an administrator",usr);
      SetEqmCompletion(HSTEQM_NAME,daqUpdateErrStr);
      return access_denied;
    }
    needsRestart = TRUE;
    return 0;
  }
  if (!stricmp(devProperty,"FILTERTABLE"))
  {
    if (access & CA_WRITE)
    {
      if (!isAdministrator(usr)) 
      {
        sprintf(daqUpdateErrStr,"access denied: user %.16s is not an administrator",usr);
        SetEqmCompletion(HSTEQM_NAME,daqUpdateErrStr);
        return access_denied;
      }
      if (din->dArrayLength < 1) return dimension_error;
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,DAQFltrTag)) return invalid_structure_tag;
      cc = addDaqFilters((DAQFilter *)din->data.vptr,din->dArrayLength);
      if (cc != 0) return cc;
    }
    flen = fmtsizeof(dout->dFormat);
    switch (dout->dFormat)
    {
      case CF_NULL:
        dout->dArrayLength = 0;
        return 0;
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        /* just a names list */
        slen = flen;
        soff = 0;
        ioff = -1;
        break;
      case CF_NAME8I:
      case CF_NAME16I:
      case CF_NAME32I:
      case CF_NAME48I:
      case CF_NAME64I:
        /* a names list with current validity status */
        slen = flen - 4;
        soff = 0;
        ioff = slen;
        break;
      case CF_STRUCT:
        if (strcmp(dout->dTag,DAQFltrTag)) return invalid_structure_tag;
        return getDaqFltrTbl((DAQFilter *)dout->data.vptr,(int *)&dout->dArrayLength);
      default: 
        return illegal_format;
    }
    if (dout->dArrayLength > (UINT32)nDaqFilterTblEntries) dout->dArrayLength = nDaqFilterTblEntries;
    for (i=0,n=0; i<nDaqFilterTblEntries && n<(int)dout->dArrayLength; i++)
    {
      strncpy((char *)(dout->data.bptr+n*flen),daqFilterTbl[i].tag,slen);
      if (ioff >= 0) 
      { /* send the current valid state the call */
        ival = daqFilterTbl[i].valid;
        memcpy(dout->data.bptr+n*flen + ioff,&ival,4);
      }
      n++;
    }
    return 0;
  }
 
  if (!stricmp(devProperty,"CONFIGURATION.MCA"))
  {
    if (din->dArrayLength > 0)
    {
      if (!(access & CA_WRITE)) return illegal_read_write;
      if (!isAdministrator(usr)) 
      {
        sprintf(daqUpdateErrStr,"access denied: user %.16s is not an administrator",usr);
        SetEqmCompletion(HSTEQM_NAME,daqUpdateErrStr);
        return access_denied;
      }
      switch (din->dFormat)
      {
        case CF_STRUCT:
          if (strcmp(din->dTag,"McaCfg")) return invalid_structure_tag;
          if ((i=GetDeviceNumberEx(HSTEQM_NAME,devName,devProperty)) < 0 &&
              (i=getViewFileListIndex(devName,gNumMcaViewFiles,mcaViewFiles)) < 0)
          {
            if ((i=appendViewerFile(devName,MCAVIEW_FILE,"MCA",&gNumMcaViewFiles,&csvMcaViewFileDb,&mcaViewFiles)) < 0) return -i;
            i--; /* if we successfully appended, then we're at the end of the list */
          }
          cc = commitMcaSet(mcaViewFiles[i].fileName,(McaSet *)din->data.vptr,din->dArrayLength);
          if (cc != 0) return cc;
          break;
        default:
          return illegal_format;
      }
    }
    if (dout->dArrayLength > 0)
    {
      if (gNumMcaViewFiles == 0)
      { /* there aren't any ! */
        dout->dArrayLength = (UINT32)0;
        return 0;
      }
      if ((i=GetDeviceNumberEx(HSTEQM_NAME,devName,devProperty)) < 0) return illegal_device;
      switch (dout->dFormat)
      {
        case CF_TEXT:
          return readViewSetFile(mcaViewFiles[i].fileName,"MCA",dout->data.cptr,(int *)&dout->dArrayLength);
        case CF_STRUCT:
          if (strcmp(dout->dTag,"McaCfg")) return invalid_structure_tag;
          return populateMcaSet(mcaViewFiles[i].fileName,(McaSet *)dout->data.vptr,(int *)&dout->dArrayLength);
        default:
          return illegal_format;
      }
    }
    return 0;
  }
  if (!stricmp(devProperty,"CONFIGURATION.MCA.NAM"))
  {
    if (din->dArrayLength > 0 )
    { 
      if (!(access & CA_WRITE)) return illegal_read_write;
      if (!isAdministrator(usr)) 
      {
        sprintf(daqUpdateErrStr,"access denied: user %.16s is not an administrator",usr);
        SetEqmCompletion(HSTEQM_NAME,daqUpdateErrStr);
        return access_denied;
      }
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,"HstView")) return invalid_structure_tag;
      cc = writeViewerFiles("MCA",MCAVIEW_FILE,(HstViewFile *)din->data.vptr,din->dArrayLength);
      if (cc != 0) return cc;
      if ((cc=readMcaFiles()) != 0) return cc;
    }
    if (dout->dArrayLength == 0) return 0;
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        len = fmtsizeof(dout->dFormat);
        break;
      case CF_STRUCT:
        if (strcmp(dout->dTag,"HstView")) return invalid_structure_tag;
        len = 0;
        break;
      default: 
        return illegal_format;
    }
    for (i=0; i<(int)dout->dArrayLength && i<gNumMcaViewFiles; i++)
    {
      if (len > 0) 
      {
        strncpy((char *)(dout->data.bptr+i*len),mcaViewFiles[i].desc,len);
      }
      else 
      {
        ((HstViewFile *)dout->data.vptr)[i] = mcaViewFiles[i];
      }
    }
    if (i < (int)dout->dArrayLength) dout->dArrayLength = (UINT32)i;
    return 0;
  }  
  if (!stricmp(devProperty,"CONFIGURATION.TRACE"))
  { /* devName is the file description ! */
    if (din->dArrayLength > 0)
    {
      if (!(access & CA_WRITE)) return illegal_read_write;
      if (!isAdministrator(usr)) 
      {
        sprintf(daqUpdateErrStr,"access denied: user %.16s is not an administrator",usr);
        SetEqmCompletion(HSTEQM_NAME,daqUpdateErrStr);
        return access_denied;
      }
      switch (din->dFormat)
      {
        case CF_STRUCT:
          if (strcmp(din->dTag,"HstCfg")) return invalid_structure_tag;
          if ((i=GetDeviceNumberEx(HSTEQM_NAME,devName,devProperty)) < 0 &&
              (i=getViewFileListIndex(devName,gNumTraceViewFiles,traceViewFiles)) < 0)
          {
            if ((i=appendViewerFile(devName,TRACEVIEW_FILE,"TRACE",&gNumTraceViewFiles,&csvTraceViewFileDb,&traceViewFiles)) < 0) return -i;
            i--; /* if we successfully appended, then we're at the end of the list */
          }
          cc = commitViewSet(traceViewFiles[i].fileName,"TRACE",(HstViewSet *)din->data.vptr,din->dArrayLength);
          if (cc != 0) return cc;
          break;
        default:
          return illegal_format;
      }
    }
    if (dout->dArrayLength > 0)
    {
      if (gNumTraceViewFiles == 0)
      { /* there aren't any ! */
        dout->dArrayLength = (UINT32)0;
        return 0;
      }
      if ((i=GetDeviceNumberEx(HSTEQM_NAME,devName,devProperty)) < 0) return illegal_device;
      switch (dout->dFormat)
      {
        case CF_TEXT:
          return readViewSetFile(traceViewFiles[i].fileName,"TRACE",dout->data.cptr,(int *)&dout->dArrayLength);
        case CF_STRUCT:
          if (strcmp(dout->dTag,"HstCfg")) return invalid_structure_tag;
          return populateViewSet(traceViewFiles[i].fileName,"TRACE",(HstViewSet *)dout->data.vptr,(int *)&dout->dArrayLength);
        default:
          return illegal_format;
      }
    }
    return 0;
  }
  if (!stricmp(devProperty,"CONFIGURATION.TRACE.NAM"))
  {
    if (din->dArrayLength > 0 )
    { 
      if (!(access & CA_WRITE)) return illegal_read_write;
      if (!isAdministrator(usr)) 
      {
        sprintf(daqUpdateErrStr,"access denied: user %.16s is not an administrator",usr);
        SetEqmCompletion(HSTEQM_NAME,daqUpdateErrStr);
        return access_denied;
      }
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,"HstView")) return invalid_structure_tag;
      cc = writeViewerFiles("TRACE",TRACEVIEW_FILE,(HstViewFile *)din->data.vptr,din->dArrayLength);
      if (cc != 0) return cc;
      if ((cc=readTraceViewerFiles()) != 0) return cc;
    }
    if (dout->dArrayLength == 0) return 0;
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        len = fmtsizeof(dout->dFormat);
        break;
      case CF_STRUCT:
        if (strcmp(dout->dTag,"HstView")) return invalid_structure_tag;
        len = 0;
        break;
      default: 
        return illegal_format;
    }
    for (i=0; i<(int)dout->dArrayLength && i<gNumTraceViewFiles; i++)
    {
      if (len > 0) 
      {
        strncpy((char *)(dout->data.bptr+i*len),traceViewFiles[i].desc,len);
      }
      else 
      {
        ((HstViewFile *)dout->data.vptr)[i] = traceViewFiles[i];
      }
    }
    if (i < (int)dout->dArrayLength) dout->dArrayLength = (UINT32)i;
    return 0;
  }  
  if (!strnicmp(devProperty,"CONFIGURATION",13))
  {
    idxGrps = -1; isNameQuery = FALSE;
    if ((c=strchr(devProperty,'.')) != NULL)
    { /* has an extension */
      strncpy(sstr,c,48); /* copy into substring */
      if ((c=strstr(sstr,".NAM")) != NULL)
      { /* is a names query */
        isNameQuery = TRUE;
        *c = 0; /* peal the .NAM off */
      }
      if (strlen(sstr) > 0)
      { /* there's still an extension */
        for (i=0; i<gNumHstGrpsFiles; i++)
        { /* move past the '.' for comparison */
          if (!stricmp(&sstr[1],hstGrpsFiles[i].fileName)) break;
        }
        if (i == gNumHstGrpsFiles) return illegal_property;
        idxGrps = i;
      }
    }
    if (!isNameQuery)
    { /* caller wants the configurations */
      if (din->dArrayLength > 0)
      { /* input coming */
        if (!(access & CA_WRITE)) return illegal_read_write;
        if (!isAdministrator(usr)) 
        {
          sprintf(daqUpdateErrStr,"access denied: user %.16s is not an administrator",usr);
          SetEqmCompletion(HSTEQM_NAME,daqUpdateErrStr);
          return access_denied;
        }
        switch (din->dFormat)
        {
          case CF_STRUCT:
            if (strcmp(din->dTag,"HstCfg")) return invalid_structure_tag;
            if ((i=GetDeviceNumberEx(HSTEQM_NAME,devName,devProperty)) < 0 &&
                (i=getViewFileListIndex(devName,gNumHstViewFiles,hstViewFiles)) < 0)
            {
              if (idxGrps < 0)
              {
                if ((i=appendViewerFile(devName,HSTVIEW_FILE,"CONFIGS",&gNumHstViewFiles,&csvHstViewFileDb,&hstViewFiles)) < 0) return -i;
              }
              else
              {
                sprintf(pn,"CONFIGS%c%.64s",FS_DELIMITER,hstGrpsFiles[idxGrps].fileName);
                db.col = csvHstViewFileColumns;
                db.ncols = sizeof(csvHstViewFileColumns)/sizeof(csvHstViewFileColumns[0]);
                db.siz = 0;
                db.ini = csvStdInit;
                db.itr = csvStdIterator;
                db.exi = csvStdExit;
                db.cnt = 0;
                db.tgt = NULL;
                db.tgt_siz = sizeof(HstViewFile);
                if ((i=appendViewerFile(devName,HSTVIEW_FILE,pn,&gNumHstGrpsViewFiles[idxGrps],&db,&hstGrpsViewFiles[idxGrps])) < 0) return -i;
              }
              i--; /* if we successfully appended, then we're at the end of the list */
            }
            if (idxGrps < 0)
            { 
              strncpy(fn,hstViewFiles[i].fileName,128);
            }
            else
            {
              sprintf(fn,"%.64s%c%.64s",hstGrpsFiles[idxGrps].fileName,FS_DELIMITER,hstGrpsViewFiles[idxGrps][i].fileName);
            }
            cc = commitViewSet(fn,"CONFIGS",(HstViewSet *)din->data.vptr,din->dArrayLength);
            if (cc != 0) return cc;
            break;
          default:
            return illegal_format;
        }
      }      
      if (dout->dArrayLength > 0)
      {
        if ((i=GetDeviceNumberEx(HSTEQM_NAME,devName,devProperty)) < 0) return illegal_device;
        if (idxGrps < 0)
        { /* normal archive configs case */
          strncpy(fn,hstViewFiles[i].fileName,128);
        }
        else
        {
          sprintf(fn,"%.64s%c%.64s",hstGrpsFiles[idxGrps].fileName,FS_DELIMITER,hstGrpsViewFiles[idxGrps][i].fileName);
        }
        if (gNumHstViewFiles == 0)
        { /* there aren't any ! */
          dout->dArrayLength = (UINT32)0;
          return 0;
        }
        switch (dout->dFormat)
        {
          case CF_TEXT:
            return readViewSetFile(fn,"CONFIGS",dout->data.cptr,(int *)&dout->dArrayLength);
          case CF_STRUCT:
            if (strcmp(dout->dTag,"HstCfg")) return invalid_structure_tag;
            return populateViewSet(fn,"CONFIGS",(HstViewSet *)dout->data.vptr,(int *)&dout->dArrayLength);
          default:
            return illegal_format;
        }
      }
    }
    else
    { /* is a names query ! (.NAM) -> get/set 'device' list */
      if (din->dArrayLength > 0 )
      { 
        if (!(access & CA_WRITE)) return illegal_read_write;
        if (!isAdministrator(usr)) 
        {
          sprintf(daqUpdateErrStr,"access denied: user %.16s is not an administrator",usr);
          SetEqmCompletion(HSTEQM_NAME,daqUpdateErrStr);
          return access_denied;
        }
        if (din->dFormat != CF_STRUCT) return illegal_format;
        if (strcmp(din->dTag,"HstView")) return invalid_structure_tag;
        if (idxGrps < 0)
        {
          cc = writeViewerFiles("CONFIGS",HSTVIEW_FILE,(HstViewFile *)din->data.vptr,din->dArrayLength);
          if (cc != 0) return cc;
          if ((cc=readViewerFiles()) != 0) return cc;
        }
        else
        {
          sprintf(fn,"%.64s%c%.64s","CONFIGS",FS_DELIMITER,hstGrpsFiles[idxGrps].fileName);
          cc = writeViewerFiles(fn,HSTVIEW_FILE,(HstViewFile *)din->data.vptr,din->dArrayLength);
          if (cc != 0) return cc;
          if ((cc=readGroupsFiles()) != 0) return cc;
        }
      }
      if (dout->dArrayLength == 0) return 0;
      /* caller wants the 'device names' list */
      switch (dout->dFormat)
      {
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME64:
          len = fmtsizeof(dout->dFormat);
          break;
        case CF_STRUCT:
          if (strcmp(dout->dTag,"HstView")) return invalid_structure_tag;
          len = 0;
          break;
        default: 
          return illegal_format;
      }
      if (idxGrps < 0)
      {
        for (i=0; i<(int)dout->dArrayLength && i<gNumHstViewFiles; i++)
        {
          if (len > 0) 
          {
            strncpy((char *)(dout->data.bptr+i*len),hstViewFiles[i].desc,len);
          }
          else 
          {
            ((HstViewFile *)dout->data.vptr)[i] = hstViewFiles[i];
          }
        }
      }
      else
      {
        for (i=0; i<(int)dout->dArrayLength && i<gNumHstGrpsViewFiles[idxGrps]; i++)
        {
          if (len > 0) 
          {
            strncpy((char *)(dout->data.bptr+i*len),hstGrpsViewFiles[idxGrps][i].desc,len);
          }
          else 
          {
            ((HstViewFile *)dout->data.vptr)[i] = hstGrpsViewFiles[idxGrps][i];
          }
        }
      }
      if (i < (int)dout->dArrayLength) dout->dArrayLength = (UINT32)i;
      return 0;
    }
    return 0;
  }
  if (!stricmp(devProperty,"RECORD.SEARCH"))
  { /* return array of record indices matching search criteria */
    NAME64 searchname;
    char *searchstr = searchname.name;
    if (din->dArrayLength == 0) return invalid_data;
    if ((cc=GetValuesAsStringEx(din,searchstr,CF_NAME64,1,0)) != 0) return cc;
    return findRecords(searchstr,dout);
  }
  if (!stricmp(devProperty,"RECORDSOURCE") || !stricmp(devProperty,"RECORD.SOURCE"))
  { /* see also KEYWORD.SOURCE above */
    if (din->dArrayLength == 0) return argument_list_error;
    tgt = (int)time(NULL);
    switch(din->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        len = fmtsizeof(din->dFormat);
        break;
      case CF_NAME16I:
      case CF_NAME32I:
      case CF_NAME64I:
        len = fmtsizeof(din->dFormat) - 4;
        tgt = *(SINT32 *)(din->data.cptr+len);
        break;
      case CF_TEXT:
        len = din->dArrayLength;
        break;
      default: 
        return illegal_format;
    }
    if (len > PROPERTY_NAME_SIZE) len = PROPERTY_NAME_SIZE;
    strncpy(keyword,din->data.cptr,len); keyword[len] = 0;
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        break;
      default:
        return illegal_format;
    }
    len = fmtsizeof(dout->dFormat);
    if ((i=getArchivedEntryFromKeywordAndTime(keyword,&rec,&p,tgt)) < 0)
      return -i;
    switch (dout->dArrayLength)
    {
      case 3:
        strncpy(dout->data.cptr+2*len,daqTbl[i].dev,len);
      case 2:
        strncpy(dout->data.cptr+1*len,daqTbl[i].prp,len);
      case 1:
        strncpy(dout->data.cptr,daqTbl[i].srv,len);
        break;
      default:
        return dimension_error;
    }
    return 0;
  }
  if (!stricmp(devProperty,"RECORDSIZE") || !stricmp(devProperty,"KEYWORD.SIZE"))
  {
    tgt = (int)time(NULL);
    switch (din->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        len = fmtsizeof(din->dFormat);
        strncpy(keyword,din->data.cptr,len); keyword[len] = 0;
        break;
      case CF_NAME16I:
      case CF_NAME32I:
      case CF_NAME64I:
        len = fmtsizeof(din->dFormat) - 4;
        tgt = *(SINT32 *)(din->data.cptr+len);
        strncpy(keyword,din->data.cptr,len); keyword[len] = 0;
        break;
      case CF_TEXT:
        len = din->dArrayLength;
        if (len > PROPERTY_NAME_SIZE) len = PROPERTY_NAME_SIZE;
        strncpy(keyword,din->data.cptr,len); keyword[len] = 0;
        break;
      default: 
        if ((cc=getValuesAsLong(din,&tgt,1)) != 0) return cc;
        strncpy(keyword,devName,PROPERTY_NAME_SIZE);
    }
    if ((i=getArchivedEntryFromKeywordAndTime(keyword,&rec,&p,tgt)) < 0) ccerr(-i);
    len = daqTbl[i].daqParam[p].size;
    return PutValuesFromLong(dout,(SINT32 *)&len,1);
  }
  /* generally true below: */
  if (dout->dArrayLength < 1) return illegal_read_write;
  fmt = dout->dFormat == CF_II      ? CF_INT32 :
        dout->dFormat == CF_FI      ? CF_FLOAT :
        dout->dFormat == CF_TDS     ? CF_TDS :
        dout->dFormat == CF_FII     ? CF_FII :
        dout->dFormat == CF_DBLTIME ? CF_DOUBLE :
        dout->dFormat == CF_DBLDBL  ? CF_DOUBLE :
        dout->dFormat == CF_DBLDBLDBL  ? CF_DBLDBLDBL :
        dout->dFormat == CF_NAME32DBLDBL  ? CF_NAME32DBLDBL :
        dout->dFormat == CF_NAME64DBLDBL  ? CF_NAME64DBLDBL :
        dout->dFormat == CF_NAME8I  ? CF_NAME8 :
        dout->dFormat == CF_NAME16I ? CF_NAME16 :
        dout->dFormat == CF_NAME32I ? CF_NAME32 :
        dout->dFormat == CF_NAME48I ? CF_NAME48 :
        dout->dFormat == CF_NAME64DBL ? CF_NAME64DBL :
        dout->dFormat == CF_NAME64TIME ? CF_NAME64TIME :
        dout->dFormat == CF_NAME64I ? CF_NAME64 : -1;
  fmtsize = fmtsizeof(fmt);
  
  /* else devProperty is the Keyword */
  if (fmt == -1 && dout->dFormat == CF_HISTORY)
  {
    fmt = CF_HISTORY;
  }
  isSnapshot = (fmt == -1) ? TRUE : FALSE;
  getRegisteredKeyword(keyword,devProperty);
  index = -1;
  if (din != NULL && din->dArrayLength > 0)
  {
    switch (din->dFormat)
    {
      case CF_DOUBLE:
        switch (din->dArrayLength)
        {
          default:
          case 5: fieldindex = (int)din->data.dptr[4];
          case 4: sample    = (int)din->data.dptr[3];
          case 3: index = (int)din->data.dptr[2];
          case 2: stoptime  = (int)din->data.dptr[1];
                  inputgiven = TRUE;
          case 1: starttime = (int)din->data.dptr[0];
          case 0: break;
        }
        break;
      case CF_INT32:
        switch (din->dArrayLength)
        {
          default:
          case 5: fieldindex = din->data.lptr[4];
          case 4: sample    = din->data.lptr[3];
          case 3: index = din->data.lptr[2];
          case 2: stoptime  = din->data.lptr[1];
                  inputgiven = TRUE;
          case 1: starttime = din->data.lptr[0];
          case 0: break;
        }
        break;
      case CF_CHAR:
      case CF_STRING:
      case CF_KEYVALUE:
        { /* n.b. sizeIn => size in bytes/chars */
          int nkv = 0;
          char *kvb[10], *sbuf, **kv;
          if (din->dFormat == CF_CHAR)
          { /* inpt = single string (array of characters: tokenize it ... */
            char *tok = "\n,";
            if ((sbuf=alloca(din->dArrayLength + 1)) == NULL) ccerr(out_of_local_memory);
            strncpy(sbuf,din->data.cptr,din->dArrayLength);
            for (c = strtok(sbuf, tok); c != NULL; c = strtok(NULL, tok))
            {
              if (strlen(c) == 0) continue;
              kvb[nkv++] = c;
            }
            kv = kvb;
          }
          else
          { /* Stock/Meta properties don't call prepAdjustableArrayInputPreCall, prepAdjustableArrayInputPostCall */
            nkv = din->dArrayLength;
            kv = din->data.strptr;
          }
          if (kv == NULL) ccerr(invalid_data);
          starttime = (time_t)assignKeyValue("starttime", kv, nkv);
          stoptime = (time_t)assignKeyValue("stoptime", kv, nkv);
          index = (int)assignKeyValue("index", kv, nkv);
          sample = (int)assignKeyValue("sample", kv, nkv);
          fieldindex = (int)assignKeyValue("fieldindex", kv, nkv);
          inputgiven = TRUE;
        }
        break;
      case CF_FWINDOW:
        starttime = ((FWINDOW *)din->data.vptr)[0].i1val;
        stoptime  = ((FWINDOW *)din->data.vptr)[0].i2val;
        lwrlimit  = ((FWINDOW *)din->data.vptr)[0].f1val;
        uprlimit  = ((FWINDOW *)din->data.vptr)[0].f2val;
        if (lwrlimit < uprlimit) sample    = -1;
        break;
      case CF_WINDOW:
        starttime = ((DWINDOW *)din->data.vptr)[0].t1;
        stoptime  = ((DWINDOW *)din->data.vptr)[0].t2;
        sample    = ((DWINDOW *)din->data.vptr)[0].i2val;
        break;
      default:
        return illegal_format;
    }
  }
  else
  {
    recsreadFlag = 0;
    depth = isSnapshot ? 1 : dout->dArrayLength;
    starttime = getArchiveStartTimeFromDepth(keyword,depth);
    isRecentRange = TRUE;
  }
  if (stoptime > tnow) stoptime = tnow;

  /* start and stop both known at this point */
  isVolatile = isVolatileKeyword(keyword);
  if (dout->dArrayLength == 1 && inputgiven)
  { /* assign num points in interval to 1st element */
    recsreadFlag = 1;
  }
  if (recsreadFlag && isSnapshot && stoptime > starttime)
  { /* is NOT a snapshot request after all, just a number query */
    fmt = dout->dFormat;
    isSnapshot = FALSE;
  }
  if (isStsRequest(fmt) || isSnapshot)
  {
    if ((isVolatile && !isRecentRange) || strlen(devName) == 0) devName = "#0"; // nominally irrelevant;
  }
  if (!isVolatile)
  { /* not volatile */
    if (index <= 0 && /* not assigned with input from the call */
        (index=getIndexFromDeviceNameEx(devName,keyword, (int)starttime, (int)stoptime)) < 0) /* check cached or archived lists */
    { /* then try whatever is current OR possible sub-device */
      if (index == -illegal_equipment_number &&
          (index=getDeviceAndSubDeviceIndicies(keyword,devName,&subindex,starttime,stoptime)) < 0)
      {
        return -index;
      }
      if (index < 0)
      {
        if ((index=getIndexFromDeviceName(devName,keyword)) < 0) return illegal_device;
      }
    }
    itthis.d1val = (double)index;
    itthis.d2val = (double)starttime;
    strncpy(itthis.name,keyword,64);
    nitpairs = 1;
    itpair = &itthis;
  }
  else
  { /* is volatile -> scan for possible indices */
    int devfound = FALSE;
    time_t vstarttime = starttime;
    if (stoptime - vstarttime < ONEDAY) vstarttime = stoptime - ONEDAY;
    itpair = getIndiciesFromVolatileDeviceNameEx(devName,keyword, (int)vstarttime, (int)stoptime);
    if (itpair == NULL) ccerr(no_data_in_interval);
    /* if (itpair[0].i1val < 0) itpair[0].i1val = 1; */
    if (starttime > itpair[0].d2val) itpair[0].d2val = (double)starttime; 
    for (i=0; i<1000; i++) 
    {
      if (itpair[i].d2val == 0) break;
      if (itpair[i].d1val >= 0) devfound = TRUE;
    }
    if (!devfound) return illegal_device;
    nitpairs = i;
  }

  if (!isSnapshot) 
  { /* assume request for pair-wise data + timestamp */
    int ilimit = 0;
    char *kwd;
    nread = 0;
    dptr = dout->data.bptr; doff = 0;
    for (i=0; i<nitpairs; i++)
    {
      if ((index=(int)itpair[i].d1val) < 0) continue;
      startt = (int)itpair[i].d2val;
      stopt = (i < nitpairs-1) ? (int)itpair[i+1].d2val - 1 : (int)stoptime;
      dlen = dout->dArrayLength-recsreadFlag;
      if (dout->dArrayLength > 1 && stoptime > starttime) 
      {
        dlen -= nread;
        ilimit = (int)(dout->dArrayLength * (float)(stopt - startt)/(float)(stoptime - starttime));
        if (dlen > ilimit) dlen = ilimit;
      }
      if (i == nitpairs - 1 && startt > stopt) stopt = startt;
      kwd = strlen(itpair[i].name) > 0 ? itpair[i].name : keyword;
      n = getArchivedData(kwd,index,fieldindex,startt,stopt,dlen,&dptr[doff],
                fmt,dout->dTag,sample,uprlimit,lwrlimit,&dts);
      if (n < 0) continue;
      nread += n;
      doff += n * getItemSize(fmt,dout->dTag);
    }
    if (nread == 0 && n < 0) ccerr(-n);
    if (recsreadFlag) assignNrOfPoints(dout->data.bptr,dout->dFormat,nread);
    if (nread < (int)dout->dArrayLength) dout->dArrayLength = nread + recsreadFlag;
    if (dts > 100) dout->dTimeStamp = dts;
    return 0;
  }
  /* if we're here, then the caller wants a trace at a timestamp : */
  cc = getArchivedDataTrace(keyword,starttime,dout,index,fieldindex,sample);
err:
  if (cc >= 512) 
  {
    if (cc < 512 + archErrLstSize) SetEqmCompletion(HSTEQM_NAME,archErrLst[cc-512]);
  }
  return cc;
}
#define NULLVALUE 0.05
int isTransition(BYTE *d1, BYTE *d0, int fmt)
{
  switch (fmt)
  {
    case CF_FLOAT:
      if (*(float *)d1 > NULLVALUE || *(float *)d0 <= NULLVALUE) return 0;
      return -1;
    case CF_DOUBLE:
      if (*(double *)d1 > NULLVALUE || *(double *)d0 <= NULLVALUE) return 0;
      return -1;
    default: return 0;
  }
}
static void checkPoiFile(DAQRecord *daqr,int p,int *percentPoi,int *isPoi)
{
  char filenm[256],fn[32];
  struct stat sbuf;
  int sizPoi=0, sizFile=0, pci=0;
  strncpy(fn,daqr->fname,16);
  /* routine only used locally -> no NULL check */
  *percentPoi = 0; *isPoi = FALSE;
  if (strnicmp(fn,"pi",2) != 0 && strnicmp(fn,"st",2) != 0)
  { /* not looking at a points-of-interest file */
    if (!daqr->daqParam[p].keepsPointsOfInterest) return;
#   ifdef WIN32
    sprintf(filenm,"%s\\%04d\\%02d\\%s",ArchiveDatapath,daqr->archYear,daqr->archMonth,fn);
#   else
    sprintf(filenm,"%s/%04d/%02d/%s",ArchiveDatapath,daqr->archYear,daqr->archMonth,fn);
#   endif
    if (stat(filenm,&sbuf) == 0) sizFile = sbuf.st_size;
    if (sizFile == 0) return;
    fn[0] = 'p'; fn[1] = 'i';
#   ifdef WIN32
    sprintf(filenm,"%s\\%04d\\%02d\\%s",ArchiveDatapath,daqr->archYear,daqr->archMonth,fn);
#   else
    sprintf(filenm,"%s/%04d/%02d/%s",ArchiveDatapath,daqr->archYear,daqr->archMonth,fn);
#   endif
    if (stat(filenm,&sbuf) == 0) 
    {
      sizPoi = sbuf.st_size;
      /* if (sizPoi > daqr->reclen * 1000) return; */ /* ridiculously large poi file */
      if (sizPoi == 0 || sizPoi > (sizFile/10)) return; /* ridiculously large poi file */
      pci = (sizPoi*100)/sizFile; /* estimate percent interesting */
      if (pci <= 0) pci = 1;
      *percentPoi = pci;
    }
  }
  else
  {
    *isPoi = TRUE;
  }
}
#define EIGHTHOURS (3600 * 8)
double lastReadDoubleValue = 0.0;
long loadData(DAQRecord *daqr,int p, int index, int fieldindex, double starttime,double endtime,
              void *datbuf,int datfmt,char *dattag,int bufsize,int sample,
              float upr,float lwr, double *lasttime,int useAltFile)
{
  int i,n,pidx,nrecs,raster = 1,useMinimalStorage,hasSts,cc = 0;
  int hFile = -1,hPoolFile = -1;
  int pcntPointsOfInterest=0,isPoiFile=FALSE,oneAndDone=FALSE,stringsMap=FALSE,scanAlternate=FALSE;
  int recsize, hdrsiz, foffset=0, ffmt=CF_NULL;
  int hasAlt = useAltFile, hasAlt2 = FALSE, isStsReq=FALSE;
  long pstart,pstop,pnow,ppos,p0,number = 0,npoi,cnt = 0,poiDir=0,lastPoiDir=0;
  size_t poolPos, poolLen, plen = 0;
  DTYPE d;
  int fmtsize,dsize,fsize,rsize,hsize,poisize,datlen,cfmt,noff = 0,bsiz,numoff = 0;
  SINT32 timestamp, sysstamp = 0, usrstamp = 0;
  float fval,fvalnull;
  double dtimestamp = 0, timenull, datatime, thisReadDoubleValue;
  double rtsstart, rtsstop, *prtsstart, *prtsstop;
  BYTE hdr[HISTORY_HDR_SIZE], *data;
  DAQRecord daqpoi, daqsts;
  DAQParameter prmsts;
  BYTE *nxtptr=NULL;

  if (bufsize < 0) return 0;
  if (endtime < starttime) return 0; /* don't even look ! */
  if (daqr->daqParam[p].fmt == CF_USTRING && isStringFormat((short)datfmt)) stringsMap = TRUE;
  if (!stringsMap && 
      !isSimpleFormat(daqr->daqParam[p].fmt) && 
      datfmt != CF_HISTORY && 
      GetFormatHeaderSize(daqr->daqParam[p].fmt) == 0)
  {
    char dev[132] = "";
    sprintf(dev,"/%.32s/%.32s",daqr->ctx,daqr->srv);
    cc = getFieldInFormat(daqr->daqParam[p].fmt,dev,daqr->tag,fieldindex,&foffset,&ffmt);
    if (cc != 0) return -cc;
  }
  if (isStsRequest(datfmt))
  { /* looking for possible non-zero record status information */
    memcpy(&daqsts,daqr,sizeof(DAQRecord));
    daqsts.fname[0] = 's'; daqsts.fname[1] = 't';
    daqsts.useMinimalStorage = TRUE;
    daqsts.fmt = CF_NAME32I;
    daqsts.reclen = fmtsizeof(CF_NAME32I);
    memcpy(&prmsts,daqr->daqParam,sizeof(DAQParameter));
    prmsts.arrayType = AT_SINGLE;
    prmsts.fmt = CF_NAME32I;
    prmsts.size = 1;
    daqsts.daqParam = &prmsts;
    daqr = &daqsts;
    index = 0;
    p = 0;
    isStsReq = TRUE;
  }
  else
  { /* normal data request */
    checkPoiFile(daqr,p,&pcntPointsOfInterest,&isPoiFile);
  }
  if (pcntPointsOfInterest)
  {
    memcpy(&daqpoi,daqr,sizeof(DAQRecord));
    daqpoi.fname[0] = 'p'; daqpoi.fname[1] = 'i';
    daqpoi.useMinimalStorage = FALSE;
  }
  daqr->hasLegacyStorage = FALSE;
  prtsstart = &rtsstart; prtsstop = &rtsstop;
retryLocateRecord:
  if ((pstart=locateRecord(daqr,starttime,prtsstart,0,&hasAlt)) < 0) 
  {
    if (pstart == -file_error) cc = file_error;
    goto err;
  }
  hasAlt2 = useAltFile;
  if ((pstop=locateRecord(daqr,endtime,prtsstop,0,&hasAlt2)) < 0)
  {
    if (pstop == -file_error) cc = file_error;
    goto err;
  }
  hasAlt = hasAlt | hasAlt2;
  if (pstart <= 0 && pstop <= 0 && 
      !daqr->hasLegacyStorage && !daqr->useMinimalStorage)
  { /* maybe a transition month ? */
    if (isPoiFile) return 0;
    daqr->hasLegacyStorage = TRUE;
    scanAlternate = FALSE;
    goto retryLocateRecord;
  }
  else 
  {
    scanAlternate = useAltFile ? FALSE : hasAlt;
  }
  if (isStsReq) scanAlternate = FALSE;
  if (scanAlternate)
  {
    if (rtsstart > starttime + EIGHTHOURS && rtsstart < endtime)
    { /* scan the alternate file first */
      bsiz = (int)(((double)bufsize)*(rtsstart-starttime)/(endtime-starttime));
      if (bsiz <= 0) bsiz = bufsize/2;
      n = loadData(daqr,p,index,fieldindex,starttime,rtsstart-1,datbuf,datfmt,dattag,bsiz,sample,upr,lwr,lasttime,TRUE);
      if (n > 0) numoff = n;
      scanAlternate = FALSE;
    }
    else if (rtsstop < endtime - EIGHTHOURS)
    {
      bsiz = (int)(((double)bufsize)*(endtime-rtsstop)/(endtime-starttime));
      if (bsiz <= 0 || bsiz >= bufsize) bsiz = bufsize/2;
      bufsize -= bsiz;
    }
  }
  useMinimalStorage = daqr->useMinimalStorage|daqr->hasLegacyStorage;
  if (useAltFile) useMinimalStorage = !daqr->useMinimalStorage;
  hasSts = daqr->hasSts;
  hdrsiz = getRecordPrefixLength(useMinimalStorage);
  recsize = daqr->reclen + hdrsiz;

  nrecs = (pstop-pstart)/recsize;
  if (isPoiFile && nrecs == 0) return 0;
  if (bufsize == 0) 
  {
    if (lasttime != NULL) *lasttime = (double)daqr->lastOnDisk;
    return nrecs;
  }
  if (sample > 0) 
  { /* requested raster provided by caller */
    raster = sample;
  }
  else 
  {
    if (bufsize < nrecs) raster += nrecs/(bufsize+1);
  }
  if (raster == 1) pcntPointsOfInterest = 0;
  if (isPoiFile && raster > 10) 
  {
    oneAndDone = TRUE;
  }
  if (pcntPointsOfInterest) 
  { /* try to give it some more room for important data */
    poisize = (nrecs * pcntPointsOfInterest)/100;
    if (pcntPointsOfInterest == 1 || poisize <= 0) 
    { /* there are points of interest but too small a number to deal with with integer arithmetic */
      poisize = (bufsize * 5)/100;
    }
    if (poisize < bufsize/4)
    { /* tolerate up to 25% of the caller's buffer designated 'interesting' */
      if (raster > 25) raster *= 2; else raster = 1 + nrecs/(bufsize-2*poisize);
    }
    else
    { /* too much is 'interesting' -> turn it off! */
      pcntPointsOfInterest = 0;
    }
  }
  hasAlt = useAltFile;
  if ((hFile=OpenDataFile(daqr,&hasAlt)) == -1) ccerr(file_error);
  if (daqr->hasDataPool)
  { /* complex data type : open companion file as well */
    char fn[132];
#   ifdef WIN32
    sprintf(fn,"%s\\%04d\\%02d\\td%.14s",
      ArchiveDatapath,daqr->archYear,daqr->archMonth,&daqr->fname[2]);
#   else
    sprintf(fn,"%s/%04d/%02d/td%.14s",
      ArchiveDatapath,daqr->archYear,daqr->archMonth,&daqr->fname[2]);
#   endif
    if ((hPoolFile=open(fn,O_RDONLY)) == -1) ccerr(file_error);
    /* if (datfmt != CF_HISTORY) ccerr(illegal_format); */
    pidx = index;
    index = 0;
  }

  /*startscan: */
  number = numoff; cnt = 0; datatime = 0;
  pnow = lseek(hFile,pstart,SEEK_SET);

  if ((rsize=getItemSize(datfmt,dattag)) < 0) ccerr(-rsize);
  if (datfmt == CF_HISTORY) cfmt = ftoi(dattag);
  d.dFormat = daqr->daqParam[p].fmt;
  d.dArrayLength = 1;
  fmtsize = daqr->daqParam[p].fmt == CF_STRUCT ? GetStructSize(daqr->tag) : fmtsizeof(daqr->daqParam[p].fmt);
  hsize = GetFormatHeaderSize(daqr->daqParam[p].fmt);
  dsize = daqr->daqParam[p].size * fmtsize + hsize;
  if (archdebug) dbglog("HSTEQM: rastering at %d records",raster);
  while(read(hFile,hdr,hdrsiz) == hdrsiz)   
  { /* read through file */
    cnt++;
    if (index < 0 || index >= daqr->daqParam[p].size) goto err;
    if (!useMinimalStorage) 
    { 
      dtimestamp = *(double *)hdr;
      timestamp = (SINT32)dtimestamp;
      sysstamp = *(SINT32 *)&hdr[8];
      usrstamp = *(SINT32 *)&hdr[12];
    }
    else 
    { 
      timestamp = *(SINT32 *)hdr;
      dtimestamp = (double)timestamp; 
    }
    if (pcntPointsOfInterest && datatime > 0 && dtimestamp > datatime-2)
    { /* n.b.  poi files do NOT use minimal storage => full timestamps */
      if ((poisize=bufsize-number-1) > 2) poisize = 2;
      npoi = loadData(&daqpoi,p,index,fieldindex,datatime+0.500,dtimestamp-0.500,
               (BYTE *)datbuf + rsize * number,
               datfmt,dattag,poisize,sample,upr,lwr,NULL,FALSE);
      if (npoi >= bufsize - number) goto err;
      if (npoi > 0) number += npoi;
    }
    if (dtimestamp > endtime) goto err;
    if (daqr->reclen > gDataBufferCapacity)
    {
      int nsiz = daqr->reclen + 65536;
      BYTE *dp = (BYTE *)realloc(gDataBuffer,nsiz);
      if (dp == NULL)
      {
        srvlog(ALF,"loadData","could not realloc trace data buffer to %d bytes",nsiz);
      }
      else
      {
        gDataBuffer = dp;
        gDataBufferCapacity = nsiz;
        srvlog(ALF,"loadData","reallocated trace data buffer to %d bytes",nsiz);
      }
    }
    if (read(hFile,gDataBuffer,daqr->reclen) != (int)daqr->reclen) goto err;
    if (dtimestamp < starttime) goto nxtrec;

    datatime = dtimestamp;
    /* point to the proper location for the parameter 'p' and array 'index' requested */
    d.data.bptr = gDataBuffer + daqr->daqParam[p].offset + hsize + index*fmtsize;
    /* if a field index (for complex data types) is given, locate that spot as well ... */
    if (ffmt != CF_NULL)
    {
      d.data.bptr += foffset;
      d.dFormat = ffmt;
    }
    if (daqr->hasDataPool)
    { /* read in the real data and re-assign the data pointer */
      int rlen = DATABLOB_SIZE;
      plen = sizeof(size_t);
      memcpy(&poolPos,gDataBuffer,plen);
      memcpy(&poolLen,&gDataBuffer[plen],plen);
      lseek(hPoolFile,(long)poolPos,SEEK_SET);
      if (poolLen < (size_t)rlen) rlen = (int)poolLen;
      if (read(hPoolFile,gDataBuffer,rlen) != rlen) goto err;
      if (isStringFormat((short)daqr->fmt))
      { /* find the position ofthe pidx string */
        memcpy(&plen,gDataBuffer+pidx*sizeof(plen),sizeof(plen));
        d.data.cptr = (char *)gDataBuffer + daqr->len * sizeof(size_t);
        d.data.cptr += plen;
        poolLen = strlen(d.data.cptr) + 1;
      }
      if (daqr->fmt == CF_IMAGE)
      { /* reduce trend calls to just the frame number */
        DIMAGE *img = (DIMAGE *)gDataBuffer;
        d.dFormat = CF_INT32;
        d.dArrayLength = 1;
        d.data.lptr = &img->frameHeader.frameNumber;
      }
    }
    switch (datfmt)
    {
      case CF_HISTORY:
        data = nxtptr == NULL ? (BYTE *)datbuf + rsize*number : nxtptr;
        fsize = fmtsize;
        if (daqr->hasDataPool)
        {
          fsize = (int)poolLen + HISTORY_HDR_SIZE; /* make this work out in case switch below */
          nxtptr = (BYTE *)datbuf + poolLen + HISTORY_HDR_SIZE;
        }
        *(double *)data = dtimestamp;
        *(SINT32 *)(&data[8]) = sysstamp;
        *(SINT32 *)(&data[12]) = usrstamp;
        /* reset the fmt and data ptr */
        data += HISTORY_HDR_SIZE;
        switch (cfmt)
        {
          case CF_INT32:
            getValuesAsLong(&d,(SINT32 *)data,1);
            thisReadDoubleValue = (double)(*((SINT32 *)data));
            break;
          case CF_INT64:
            getValuesAsDLong(&d,(SINT64 *)data,1);
            thisReadDoubleValue = (double)(*((SINT64 *)data));
            break;
          case CF_FLOAT:
            getValuesAsFloat(&d,(float *)data,1);
            thisReadDoubleValue = (double)(*((float *)data));
            break;
          case CF_DOUBLE:
            getValuesAsDouble(&d,(double *)data,1);
            thisReadDoubleValue = *((double *)data);
            break;
          case CF_IMAGE:
            { /*
              DIMAGE *img = (DIMAGE *)d.data.vptr;
              memcpy(img,data,IMAGE_HDR_SIZE);
              */
              return not_implemented;
            }
            break;
          case CF_SPECTRUM:
            return not_implemented;
          case CF_STRUCT:
            if (strncmp(d.dTag,daqr->tag,TAG_NAME_SIZE)) return invalid_structure_tag;
          default:
            if (cfmt == d.dFormat)
            { /* requested fmt = stored format ! */
              memcpy(data,d.data.vptr,fsize);
              break;
            }
            ccerr(illegal_format);
        }
        break;
      case CF_INT32: 
        ((INTINT *)datbuf)[number].i2val = timestamp;
        getValuesAsLong(&d,&((INTINT *)datbuf)[number].i1val,1);
        thisReadDoubleValue = (double)((INTINT *)datbuf)[number].i1val;
        if (sample < 0 && (d.data.lptr[0] < lwr || d.data.lptr[0] > upr))
        {
          ((INTINT *)datbuf)[number].i1val = 0; /* flag the boundary */
          if (number > 0 && ((INTINT *)datbuf)[number-1].i1val == 0) continue;
        }
        break;
      case CF_FLOAT: 
        if (number > 1 && timestamp == (int)((FLTINT *)datbuf)[number-1].ival)
        {
          number--;
          break;
        }
        ((FLTINT *)datbuf)[number].ival = timestamp;
        getValuesAsFloat(&d,&((FLTINT *)datbuf)[number].fval,1);
        thisReadDoubleValue = (double)((FLTINT *)datbuf)[number].fval;
        if (raster > 1 && number > 1 && (daqr->option & FAST))
        { /* if we're skipping thru, try not to miss important jumps ... */
          fval = ((FLTINT *)datbuf)[number].fval; timenull = 0;
          for (i=0; i<raster; i++)
          {
            if (!isTransition((BYTE *)&fval,(BYTE *)&((FLTINT *)datbuf)[number-1].fval,datfmt)) break;
            if (i == raster-1) break; /* stop right here */
            if (i > 0) { fvalnull = fval; timenull = dtimestamp; }
            ppos = pnow - recsize * (1+i);
            if ((p0=lseek(hFile,ppos,SEEK_SET)) < 0) goto err;
            if (read(hFile,hdr,hdrsiz) != hdrsiz) goto err;
            if (!useMinimalStorage) 
            {
              dtimestamp = *(double *)hdr;
              timestamp = (SINT32)dtimestamp; 
            }
            else 
            { 
              timestamp = *(SINT32 *)hdr; dtimestamp = (double)timestamp; 
            }
            if (read(hFile,gDataBuffer,daqr->reclen) != (int)daqr->reclen) goto err;
            getValuesAsFloat(&d,&fval,1);
          }
          if (timenull > 0)
          {
            ((FLTINT *)datbuf)[number].ival = (SINT32)timenull;
            ((FLTINT *)datbuf)[number].fval = fvalnull;
          }
          if (lseek(hFile,pnow+recsize,SEEK_SET) < 0) goto err; /* back to where we were */
        }
        if (sample < 0 &&
            (d.data.fptr[0] < lwr || d.data.fptr[0] > upr))
        {
          ((FLTINT *)datbuf)[number].ival = 0; /* flag the boundary */
          if (number > 0 && ((FLTINT *)datbuf)[number-1].ival == 0) continue;
        }
        break;
      case CF_TDS: /* special case where the timestamp is already part of the format */
        {
          int sts = 0;
          ((INTFLTINT *)datbuf)[number].i1val = timestamp;
          getValuesAsFloat(&d,&((INTFLTINT *)datbuf)[number].fval,1);
          thisReadDoubleValue = (double)((INTFLTINT *)datbuf)[number].fval;
          if (sample < 0 && 
              (d.data.fptr[0] < lwr || d.data.fptr[0] > upr))
          {
            ((INTFLTINT *)datbuf)[number].i1val = 0; /* flag the boundary */
            if (number > 0 && ((INTFLTINT *)datbuf)[number-1].i1val == 0) continue;
          }
          switch (d.dFormat)
          {
            case CF_FLTINT:
            case CF_INTINT:
            case CF_NAME16I:
            case CF_NAME32I:
            case CF_NAME64I:
              fsize = fmtsizeof(d.dFormat);
              sts = *(int *)(&d.data.bptr[fsize-4]);
              break;
            case CF_DBLDBL:
              sts = (int)d.data.dptr[1];
              break;
          }
          sts &= 0x0fff;
          ((INTFLTINT *)datbuf)[number].i2val = sts;
        }
        break;
      case CF_FII: /* special case where we add 2 stamps */
        ((FLTINTINT *)datbuf)[number].i1val = timestamp;
        ((FLTINTINT *)datbuf)[number].i2val = sysstamp;
        getValuesAsFloat(&d,&((FLTINTINT *)datbuf)[number].fval,1);
        thisReadDoubleValue = (double)((FLTINTINT *)datbuf)[number].fval;
        if (sample < 0 && 
            (d.data.fptr[0] < lwr || d.data.fptr[0] > upr))
        {
          ((FLTINTINT *)datbuf)[number].i1val = 0; /* flag the boundary */
          if (number > 0 && ((FLTINTINT *)datbuf)[number-1].i1val == 0) continue;
        }
        break;
      case CF_DOUBLE: 
        ((DBLDBL *)datbuf)[number].d2val = dtimestamp;
        getValuesAsDouble(&d,&((DBLDBL *)datbuf)[number].d1val,1);
        thisReadDoubleValue = (double)((DBLDBL *)datbuf)[number].d1val;
        if (sample < 0 && 
            (d.data.dptr[0] < lwr || d.data.dptr[0] > upr))
        {
          ((DBLDBL *)datbuf)[number].d2val = 0; /* flag the boundary */
          if (number > 0 && ((DBLDBL *)datbuf)[number-1].d1val == 0) continue;
        }
        break;
      case CF_DBLDBLDBL: /* special case where we add 2 stamps */
        ((DBLDBLDBL *)datbuf)[number].d2val = dtimestamp;
        ((DBLDBLDBL *)datbuf)[number].d3val = sysstamp;
        getValuesAsDouble(&d,&((DBLDBLDBL *)datbuf)[number].d1val,1);
        thisReadDoubleValue = (double)((DBLDBLDBL *)datbuf)[number].d1val;
        if (sample < 0 && 
            (d.data.fptr[0] < lwr || d.data.fptr[0] > upr))
        {
          ((DBLDBLDBL *)datbuf)[number].d2val = 0; /* flag the boundary */
          if (number > 0 && ((DBLDBLDBL *)datbuf)[number-1].d1val == 0) continue;
        }
        break;
      case CF_NAME32DBLDBL: /* status retrieval */
        strncpy(((NAME32DBLDBL *)datbuf)[number].name,((NAME32I *)d.data.vptr)->name,32);
        ((NAME32DBLDBL *)datbuf)[number].d1val = (double)((NAME32I *)d.data.vptr)->ival;
        ((NAME32DBLDBL *)datbuf)[number].d2val = timestamp;
        break;
      case CF_NAME64DBLDBL: /* special case where we add 2 stamps */
        switch (d.dFormat)
        {
          case CF_TEXT:
          case CF_NAME8:
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME48:
          case CF_NAME64: break;
          case CF_USTRING: 
            noff = 16;
            break;
          default: ccerr(illegal_format);
        }
        datlen = dsize;
        if (datlen > 64) datlen = 64;
        strncpy(((NAME64DBLDBL *)datbuf)[number].name,d.data.cptr+noff,datlen);
        ((NAME64DBLDBL *)datbuf)[number].d1val = dtimestamp;
        ((NAME64DBLDBL *)datbuf)[number].d2val = sysstamp;
        break;
      case CF_NAME16: 
        switch (d.dFormat)
        {
          case CF_TEXT:
          case CF_NAME8:
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME48:
          case CF_NAME64: break;
          case CF_USTRING: 
            noff = 16;
            break;
          default: ccerr(illegal_format);
        }
        datlen = fmtsizeof(datfmt);
        if (dsize > datlen) dsize = datlen;
        strncpy(((NAME16I *)datbuf)[number].name,d.data.cptr+noff,dsize);
        ((NAME16I *)datbuf)[number].ival = timestamp;
        break;
      case CF_NAME32: 
        switch (d.dFormat)
        {
          case CF_TEXT:
          case CF_NAME8:
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME48:
          case CF_NAME64: break;
          case CF_USTRING: 
            noff = 16;
            break;
          default: ccerr(illegal_format);
        }
        datlen = fmtsizeof(datfmt);
        if (dsize > datlen) dsize = datlen;
        strncpy(((NAME32I *)datbuf)[number].name,d.data.cptr+noff,dsize);
        ((NAME32I *)datbuf)[number].ival = timestamp;
        break;
      case CF_NAME48: 
        switch (d.dFormat)
        {
          case CF_TEXT:
          case CF_NAME8:
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME48:
          case CF_NAME64: break;
          case CF_USTRING: 
            noff = 16;
            break;
          default: ccerr(illegal_format);
        }
        datlen = fmtsizeof(datfmt);
        if (dsize > datlen) dsize = datlen;
        strncpy(((NAME48I *)datbuf)[number].name,d.data.cptr+noff,dsize);
        ((NAME48I *)datbuf)[number].ival = timestamp;
        break;
      case CF_NAME64:
        switch (d.dFormat)
        {
          case CF_TEXT:
          case CF_NAME8:
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME48:
          case CF_NAME64: 
            break;
          case CF_USTRING: 
            noff = 16;
            break;
          default: ccerr(illegal_format);
        }
        datlen = fmtsizeof(datfmt);
        if (dsize > datlen) dsize = datlen;
        strncpy(((NAME64I *)datbuf)[number].name,d.data.cptr+noff,dsize);
        ((NAME64I *)datbuf)[number].ival = timestamp;
        break;
      case CF_NAME64TIME:
      case CF_NAME64DBL:
        switch (d.dFormat)
        {
          case CF_TEXT:
          case CF_NAME8:
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME48:
          case CF_NAME64: break;
          case CF_USTRING: 
            noff = 16;
            break;
          default: ccerr(illegal_format);
        }
        datlen = dsize;
        if (datlen > 64) datlen = 64;
        strncpy(((NAME64DBL *)datbuf)[number].name,d.data.cptr+noff,datlen);
        ((NAME64DBL *)datbuf)[number].dval = dtimestamp;
        break;
      default: goto err;
    }
    if (isPoiFile)
    { /* don't add unnecessary points of interest */
      poiDir = (thisReadDoubleValue - lastReadDoubleValue) > 0 ? 1 : -1;
      /* if (bufsize == 2 && poiDir == lastPoiDir) goto err; */ /* just add the single point !*/
      lastPoiDir = poiDir;
    }
    lastReadDoubleValue = thisReadDoubleValue;
    if (++number >= bufsize) goto err;
    if (number == 1 && oneAndDone) goto err;
nxtrec:
    if (raster > 1) /* move along through file */
    {
      if ((pnow=lseek(hFile,recsize*(raster-1),SEEK_CUR)) < 0) goto err;
    }
    else pnow += recsize;
  }
err:
  if (hFile != -1) close(hFile);
  if (hPoolFile != -1) close(hPoolFile);
  if (archdebug) dbglog("HSTEQM: scanned %ld records",cnt);
  if (lasttime != NULL) *lasttime = dtimestamp;

  if (scanAlternate && cc == 0)
  {
    if (rtsstop < endtime - EIGHTHOURS)
    { /* scan the alternate file last */
      BYTE *dp = (BYTE *)datbuf + rsize * number;
      n = loadData(daqr,p,index,fieldindex,rtsstop+1,endtime,dp,datfmt,dattag,bsiz,sample,upr,lwr,lasttime,TRUE);
      if (n > 0) numoff = n;
    }
    scanAlternate = FALSE;
  }
  return cc > 0 ? -cc : number + noff;
}
time_t getArchiveStartTimeFromDepth(char *keyword,int depth)
{
  time_t endtime, stime;
  struct tm *lt;
  static int rec,p;
  int hFile=-1,useMinimalStorage,cc=0,recsPerDay,n,yr,mo;
  int recsize, hdrsiz, fsiz;

  endtime = tvalid > 0 ? tvalid : time(NULL);
  stime = endtime-depth;
  if ((n=getArchivedEntryFromKeywordAndTime(keyword,&rec,&p,endtime)) < 0) 
    ccerr(-n);
  if (depth == 1) return stime;
  useMinimalStorage = daqTbl[n].useMinimalStorage;
  hdrsiz = getRecordPrefixLength(useMinimalStorage);
  recsize = daqTbl[n].reclen + hdrsiz;

  /* see how big last month's file was ... */
  lt = localtime((time_t *)&endtime);
  yr = lt->tm_year; mo = lt->tm_mon;
  if (mo == 0) yr--;
  mo = (mo - 1 + 12)%12;  
  getDataFilename(n,yr,mo+1);
  if ((hFile=OpenDataFile(&daqTbl[n],NULL)) == -1) ccerr(no_such_file);
  if ((fsiz=lseek(hFile,0,SEEK_END)) <= 0) ccerr(file_error);
  recsPerDay = fsiz / (recsize * 30);  /* approximately ... */
  if (recsPerDay < 48) recsPerDay = 48; /* be reasonable ... */
  stime = endtime - (depth * 86400)/recsPerDay;
err:
  if (hFile != -1) close(hFile);
  if (archdebug)
  {
    dbglog("getArchiveStartTimeFromDepth %d pts, recs/day %d, file size %d : %s",
      depth,recsPerDay,fsiz,cc2str(cc));
  }
  return stime;
}
int getItemSize(int fmt,char *tag)
{
  int siz = fmtsizeof(fmt);
  switch (fmt)
  {
    case CF_DBLDBLDBL:
      break;
    case CF_DOUBLE: /* add 1 double */
    case CF_TDS:    /* add 2 int32s */
    case CF_FII:    /* add 2 int32s */
      siz += 8;
      break;
    case CF_HISTORY:
      {
        int cfmt = ftoi(tag);
        int fsize = getCarriedFormatSize(tag);
        if (fsize < 1) return -illegal_format;
        siz = HISTORY_HDR_SIZE + fsize;
      }
      break;
    case CF_NAME64DBL:
    case CF_NAME64TIME:
      break;
    default:
      siz += sizeof(UINT32); /* add 1 int32 */
      break;
  }
  return siz;
}
int getDeviceAndSubDeviceIndicies(char *keyword,char *devName,int *subindex,time_t starttime,time_t endtime)
{
  int idx, sidx=-1, rec, i, p, n, rlen=0, slen, len=0;
  DAQParameter *daqParam;
  DTYPE dout;
  char *sub, dev[128], prp[76], *dptr;
  if ((n=getArchivedEntryFromKeywordAndTime(keyword,&rec,&p,starttime)) < 0) return n;
  daqParam = daqTbl[n].daqParam;
  strncpy(dev,devName,128);
  if ((sub=strchr(dev,'/')) != NULL)
  {
    *sub++ = 0;
    len = (int)strlen(sub);
    rlen = daqParam->rowsize;
  }
  if ((idx=getIndexFromDeviceNameEx(dev,keyword, (int)starttime, (int)endtime)) < 0) return idx;
  if (daqParam->arrayType & AT_DOUBLE && len > 0)
  { /* makes sense to look for a companion entry */
    char key[76];
    strncpy(prp,keyword,56);
    strcat(prp,".NAM.NAM");
    getRegisteredKeyword(key,prp);
    if ((n=getArchivedEntryFromKeywordAndTime(key,&rec,&p,starttime)) >= 0)
    { /* as expected: a companion .NAM.NAM entry */
      daqParam = daqTbl[n].daqParam;
      slen = fmtsizeof(daqParam->fmt);
      if (daqTbl[n].data == NULL)
      { /* read from disk (not yet cached) */
        daqTbl[n].data = calloc(daqTbl[n].len,fmtsizeof(CF_NAME64));
        if (daqTbl[n].data == NULL) return -out_of_server_memory;
        DTYPEZERO(dout);
        dout.dArrayLength = daqTbl[n].len;
        dout.dFormat = CF_NAME64;
        dout.data.cptr = daqTbl[n].data;
        getArchivedDataTrace(key,time(NULL),&dout,0,0,0);
      }
      dptr = daqTbl[n].data + daqParam->offset;
      for (i=0; i<daqParam->size; i++)
      { /* run thru the list */
        if (strnicmp(sub,dptr + i*slen,len) == 0) 
        { /* got'm */
          sidx = i;
          idx = idx * rlen + sidx;
          break;
        }
      }
    }
  }
  if (subindex != NULL) *subindex = sidx;
  return idx;
}
int getArchivedData(char *keyword,int index,int fieldindex,time_t starttime,
          time_t endtime,int arraysize,BYTE *data,int datafmt,char *datatag,int sample,
          float upr,float lwr,double *ts)
{
  DAQParameter *daqParam;
  int rec, p, recs, rece, reci, pst, pet, pit;
  long ps;
  int recsread=0,ptr=0,recsfound=0,recsize;
  int asize=0,incsize=0,daystotal,daysinmonth,n,ns,ne, ni;
  int finalPass=FALSE,nquery=FALSE,chkTransition=FALSE,notTransition=FALSE;
  struct tm st,et,it;
  time_t itime, ts_ulimit=endtime, ts_llimit=starttime;
  double ldts=0;
  time_t call_start = time(NULL), tt;

  if (archdebug)
  {
    dbglog("HSTEQM: searching for %s <%d> %d entries beginning %.24s",
            keyword,index,arraysize,ctime(&starttime));
  }

  if (endtime < starttime) return -invalid_range; /* don't even look ! */
  if ((ne=getArchivedEntryFromKeywordAndTime(keyword,&rece,&pet,endtime)) < 0) 
    return ne;
  if ((ns=getArchivedEntryFromKeywordAndTime(keyword,&recs,&pst,starttime)) < 0) ns = ne;
  n = ns; rec = recs; p = pst;
  if (ns != ne) chkTransition = TRUE;
tof:
  daqParam = daqTbl[n].daqParam;
  if (index >= daqParam[p].size) 
  {
    if (chkTransition)
    {
      p = pet;
      n = ne;
      chkTransition = FALSE;
      goto tof;
    }
    goto bof;
  }
  if (index >= daqParam[p].size || index < 0) return -invalid_index;

  if (arraysize < 0) return -buffer_too_small;
  if (arraysize == 0) nquery = TRUE;
  if ((recsize=getItemSize(datafmt,datatag)) < 0) return recsize;
  memcpy(&et,localtime(&endtime), sizeof(struct tm));
  memcpy(&st,localtime(&starttime),sizeof(struct tm));
  if (et.tm_year == st.tm_year && et.tm_mon == st.tm_mon)
  { /* this month contains the transition (if any) */
    asize = arraysize;
    if (chkTransition) asize /= 2;
    finalPass = TRUE;
  }
  else
  {
    asize = sample > 0 ? arraysize - 1 : 0;
  }
  daystotal = (int)(endtime-starttime)/(3600*24);
  if ((endtime-starttime)%(3600*24) > 0) daystotal++;
  if (asize == 0 && daystotal == 0) daystotal++;
  if (daystotal == 1 && st.tm_mday != et.tm_mday) daystotal++;
  if (chkTransition) daystotal++;

  while (recsread < arraysize || nquery)
  {  
    if (asize < arraysize || nquery)
    {
      it = st;
      if (it.tm_year == et.tm_year && it.tm_mon == et.tm_mon)
      {
        asize = chkTransition ? (arraysize-recsread)/2 : arraysize;
        finalPass = TRUE;
        ts_ulimit = endtime;
        notTransition = FALSE;
      }
      else
      {
        it.tm_mon = (it.tm_mon+1)%12; if (it.tm_mon == 0) it.tm_year++;
        it.tm_mday = 1; it.tm_hour = 0; it.tm_min = 0; it.tm_sec = 0;
        itime = mktime(&it);
        daysinmonth = (int)MAX((itime-starttime)/(3600*24),1);
        if (!nquery && !sample)
        { /* not a number query and no specific raster given */
          incsize = MAX((arraysize * daysinmonth)/daystotal,1);
          asize += incsize;
          if (asize >= arraysize) finalPass = TRUE; 
          notTransition = ((ni=getArchivedEntryFromKeywordAndTime(keyword,&reci,&pit,itime)) == ns) ? TRUE : FALSE;
        }
        ts_ulimit = itime;
      }
    }
    if (asize > arraysize) asize = arraysize;
    getDataFilename(n,st.tm_year,st.tm_mon+1);
    ps = loadData(&daqTbl[n],p,index,fieldindex,(double)starttime,(double)endtime,&data[ptr],datafmt,datatag,asize-recsread,sample,upr,lwr,&ldts,FALSE);
    if (ps < 0) return ps;
    if (nquery) recsfound += ps; else recsread += ps;
    ptr = recsread * recsize;
    if (chkTransition && !notTransition && ldts > 1 && (time_t)ldts < ts_ulimit)
    { /* between defunct and active record */
      n = ne; rec = rece; p = pet;
      starttime = (time_t)ldts;
      chkTransition = FALSE;
      daqParam = daqTbl[n].daqParam;
      if (index > daqParam[p].size) goto bof;
      if (index > daqParam[p].size || index < 0) return -invalid_index;
    }
    else
    {
      if (finalPass) break;
      starttime = itime;
    }
    memcpy(&st,localtime(&starttime),sizeof(struct tm));
  }
bof:
  tt = time(NULL) - call_start;
  if (tt > 2) 
    feclogEx(TINE_LOGLEVEL_WARN,"getArchivedData %.64s (range %d days) took %d seconds to complete",
      keyword,(endtime-ts_llimit)/86400,tt);
  if (archdebug) dbglog("HSTEQM: returning %d records in %d seconds",recsread,tt);
  if (ts != NULL) *ts = ldts;
  return nquery ? recsfound : recsread;
}
int getArchivedDataTrace(char *keyword,time_t tgttime,DTYPE *dout,int index,int fieldindex,int sample)
{
  DAQParameter *daqParam;
  static int rec,p;
  int n, cc = 0;

  if (archdebug)
  {
    dbglog("HSTEQM: searching for %s trace %d entries at %.24s",keyword,dout->dArrayLength,ctime(&tgttime));
  }

  if ((n=getArchivedEntryFromKeywordAndTime(keyword,&rec,&p,tgttime)) < 0) return -n;
  daqParam = daqTbl[n].daqParam;
  cc = loadTraceData(gDataBuffer,n,p,tgttime,dout,index,fieldindex,sample);
  return cc;
}

